open Ast0
open AstUtils
open BasicAst
open Datatypes
open LiftSubst
open List0
open PeanoNat
open Specif
open String0
open Typing0
open UnivSubst0
open Universes0
open Config0
open Monad_utils
open UGraph0
open Utils

type __ = Obj.t
let __ = let rec f _ = Obj.repr f in Obj.repr f

module RedFlags =
 struct
  type t = { beta : bool; iota : bool; zeta : bool; delta : bool;
             fix_ : bool; cofix_ : bool }

  (** val beta : t -> bool **)

  let beta x = x.beta

  (** val iota : t -> bool **)

  let iota x = x.iota

  (** val zeta : t -> bool **)

  let zeta x = x.zeta

  (** val delta : t -> bool **)

  let delta x = x.delta

  (** val fix_ : t -> bool **)

  let fix_ x = x.fix_

  (** val cofix_ : t -> bool **)

  let cofix_ x = x.cofix_

  (** val default : t **)

  let default =
    { beta = true; iota = true; zeta = true; delta = true; fix_ = true;
      cofix_ = true }
 end

(** val zip : (term * term list) -> term **)

let zip t0 =
  mkApps (fst t0) (snd t0)

(** val reduce_stack :
    RedFlags.t -> global_env -> context -> nat -> term -> term list ->
    (term * term list) option **)

let rec reduce_stack flags _UU03a3_ _UU0393_ n t0 stack =
  match n with
  | O -> None
  | S n0 ->
    (match t0 with
     | Coq_tRel c ->
       if flags.RedFlags.zeta
       then bind (Obj.magic option_monad) (nth_error (Obj.magic _UU0393_) c)
              (fun d ->
              match d.decl_body with
              | Some b ->
                reduce_stack flags _UU03a3_ _UU0393_ n0 (lift (S c) O b) stack
              | None -> ret (Obj.magic option_monad) (t0, stack))
       else ret (Obj.magic option_monad) (t0, stack)
     | Coq_tCast (c, _, _) -> reduce_stack flags _UU03a3_ _UU0393_ n0 c stack
     | Coq_tLambda (_, _, b) ->
       if flags.RedFlags.beta
       then (match stack with
             | [] -> ret (Obj.magic option_monad) (t0, stack)
             | a :: args' ->
               reduce_stack flags _UU03a3_ _UU0393_ n0 (subst1 a O b) args')
       else ret (Obj.magic option_monad) (t0, stack)
     | Coq_tLetIn (_, b, _, c) ->
       if flags.RedFlags.zeta
       then reduce_stack flags _UU03a3_ _UU0393_ n0 (subst1 b O c) stack
       else ret (Obj.magic option_monad) (t0, stack)
     | Coq_tApp (f, args) ->
       reduce_stack flags _UU03a3_ _UU0393_ n0 f (app args stack)
     | Coq_tConst (c, u) ->
       if flags.RedFlags.delta
       then (match lookup_env _UU03a3_ c with
             | Some g ->
               (match g with
                | ConstantDecl (_, c0) ->
                  let { cst_type = _; cst_body = cst_body0; cst_universes =
                    _ } = c0
                  in
                  (match cst_body0 with
                   | Some body ->
                     let body' = subst_instance_constr u body in
                     reduce_stack flags _UU03a3_ _UU0393_ n0 body' stack
                   | None -> ret (Obj.magic option_monad) (t0, stack))
                | InductiveDecl (_, _) ->
                  ret (Obj.magic option_monad) (t0, stack))
             | None -> ret (Obj.magic option_monad) (t0, stack))
       else ret (Obj.magic option_monad) (t0, stack)
     | Coq_tCase (ind_and_nbparams, p, c, brs) ->
       let (ind, par) = ind_and_nbparams in
       if flags.RedFlags.iota
       then bind (Obj.magic option_monad)
              (reduce_stack flags _UU03a3_ _UU0393_ n0 c []) (fun c' ->
              let (y, args) = c' in
              (match y with
               | Coq_tConstruct (_, c0, _) ->
                 reduce_stack flags _UU03a3_ _UU0393_ n0
                   (iota_red par c0 args brs) stack
               | _ ->
                 ret (Obj.magic option_monad) ((Coq_tCase ((ind, par), p,
                   (zip c'), brs)), stack)))
       else ret (Obj.magic option_monad) (t0, stack)
     | Coq_tFix (mfix, idx) ->
       if flags.RedFlags.fix_
       then bind (Obj.magic option_monad) (Obj.magic unfold_fix mfix idx)
              (fun nf ->
              let (narg, fn) = nf in
              (match nth_error stack narg with
               | Some c ->
                 bind (Obj.magic option_monad)
                   (reduce_stack flags _UU03a3_ _UU0393_ n0 c []) (fun c' ->
                   match fst c' with
                   | Coq_tConstruct (_, _, _) ->
                     reduce_stack flags _UU03a3_ _UU0393_ n0 fn stack
                   | _ -> ret (Obj.magic option_monad) (t0, stack))
               | None -> ret (Obj.magic option_monad) (t0, stack)))
       else ret (Obj.magic option_monad) (t0, stack)
     | _ -> ret (Obj.magic option_monad) (t0, stack))

(** val reduce_stack_term :
    RedFlags.t -> global_env -> context -> nat -> term -> term option **)

let reduce_stack_term flags _UU03a3_ _UU0393_ n c =
  bind (Obj.magic option_monad)
    (Obj.magic reduce_stack flags _UU03a3_ _UU0393_ n c []) (fun res ->
    ret (Obj.magic option_monad) (zip res))

(** val fix_decls : term mfixpoint -> context_decl list **)

let fix_decls l =
  let rec aux acc = function
  | [] -> acc
  | d :: ds0 -> aux ((vass d.dname d.dtype) :: acc) ds0
  in aux [] l

(** val map_constr_with_binders :
    (context -> term -> term) -> context -> term -> term **)

let map_constr_with_binders f _UU0393_ t0 = match t0 with
| Coq_tEvar (ev, args) -> Coq_tEvar (ev, (map (f _UU0393_) args))
| Coq_tCast (c, kind, t1) -> Coq_tCast ((f _UU0393_ c), kind, (f _UU0393_ t1))
| Coq_tProd (na, a, b) ->
  let a' = f _UU0393_ a in
  Coq_tProd (na, a', (f (snoc _UU0393_ (vass na a')) b))
| Coq_tLambda (na, t1, m) -> Coq_tLambda (na, (f _UU0393_ t1), (f _UU0393_ m))
| Coq_tLetIn (na, b, t1, c) ->
  let b' = f _UU0393_ b in
  let t' = f _UU0393_ t1 in
  Coq_tLetIn (na, b', t', (f (snoc _UU0393_ (vdef na b' t')) c))
| Coq_tApp (u, v) -> Coq_tApp ((f _UU0393_ u), (map (f _UU0393_) v))
| Coq_tCase (ind, p, c, brs) ->
  let brs' = map (on_snd (f _UU0393_)) brs in
  Coq_tCase (ind, (f _UU0393_ p), (f _UU0393_ c), brs')
| Coq_tProj (p, c) -> Coq_tProj (p, (f _UU0393_ c))
| Coq_tFix (mfix, idx) ->
  let _UU0393_' = app (fix_decls mfix) _UU0393_ in
  let mfix' = map (map_def (f _UU0393_) (f _UU0393_')) mfix in
  Coq_tFix (mfix', idx)
| Coq_tCoFix (mfix, k) ->
  let _UU0393_' = app (fix_decls mfix) _UU0393_ in
  let mfix' = map (map_def (f _UU0393_) (f _UU0393_')) mfix in
  Coq_tCoFix (mfix', k)
| _ -> t0

(** val reduce_opt :
    RedFlags.t -> global_env -> context -> nat -> term -> term option **)

let rec reduce_opt flags _UU03a3_ _UU0393_ n c =
  match n with
  | O -> None
  | S n0 ->
    (match reduce_stack_term flags _UU03a3_ _UU0393_ n0 c with
     | Some c' ->
       Some
         (map_constr_with_binders (fun _UU0393_0 t0 ->
           match reduce_opt flags _UU03a3_ _UU0393_0 n0 t0 with
           | Some t1 -> t1
           | None -> t0) _UU0393_ c')
     | None -> None)

(** val isConstruct : term -> bool **)

let isConstruct = function
| Coq_tApp (f, _) ->
  (match f with
   | Coq_tConstruct (_, _, _) -> true
   | _ -> false)
| Coq_tConstruct (_, _, _) -> true
| _ -> false

(** val isCoFix : term -> bool **)

let isCoFix = function
| Coq_tCoFix (_, _) -> true
| _ -> false

type conv_pb =
| Conv
| Cumul

(** val conv_pb_rect : 'a1 -> 'a1 -> conv_pb -> 'a1 **)

let conv_pb_rect f f0 = function
| Conv -> f
| Cumul -> f0

(** val conv_pb_rec : 'a1 -> 'a1 -> conv_pb -> 'a1 **)

let conv_pb_rec f f0 = function
| Conv -> f
| Cumul -> f0

(** val eq_term : checker_flags -> universes_graph -> term -> term -> bool **)

let rec eq_term h _UU03c6_ t0 u =
  match t0 with
  | Coq_tRel n -> (match u with
                   | Coq_tRel n' -> eq_nat n n'
                   | _ -> false)
  | Coq_tVar id ->
    (match u with
     | Coq_tVar id' -> eq_string id id'
     | _ -> false)
  | Coq_tEvar (ev, args) ->
    (match u with
     | Coq_tEvar (ev', args') ->
       (&&) (eq_evar ev ev') (forallb2 (eq_term h _UU03c6_) args args')
     | _ -> false)
  | Coq_tSort s ->
    (match u with
     | Coq_tSort s' -> try_eqb_universe h _UU03c6_ s s'
     | _ -> false)
  | Coq_tCast (f, _, t1) ->
    (match u with
     | Coq_tCast (f', _, t') ->
       (&&) (eq_term h _UU03c6_ f f') (eq_term h _UU03c6_ t1 t')
     | _ -> false)
  | Coq_tProd (_, b, t1) ->
    (match u with
     | Coq_tProd (_, b', t') ->
       (&&) (eq_term h _UU03c6_ b b') (eq_term h _UU03c6_ t1 t')
     | _ -> false)
  | Coq_tLambda (_, b, t1) ->
    (match u with
     | Coq_tLambda (_, b', t') ->
       (&&) (eq_term h _UU03c6_ b b') (eq_term h _UU03c6_ t1 t')
     | _ -> false)
  | Coq_tLetIn (_, b, t1, c) ->
    (match u with
     | Coq_tLetIn (_, b', t', c') ->
       (&&) ((&&) (eq_term h _UU03c6_ b b') (eq_term h _UU03c6_ t1 t'))
         (eq_term h _UU03c6_ c c')
     | _ -> false)
  | Coq_tApp (f, args) ->
    (match u with
     | Coq_tApp (f', args') ->
       (&&) (eq_term h _UU03c6_ f f')
         (forallb2 (eq_term h _UU03c6_) args args')
     | _ -> false)
  | Coq_tConst (c, u0) ->
    (match u with
     | Coq_tConst (c', u') ->
       (&&) (eq_constant c c') (eqb_univ_instance h _UU03c6_ u0 u')
     | _ -> false)
  | Coq_tInd (i, u0) ->
    (match u with
     | Coq_tInd (i', u') ->
       (&&) (eq_ind i i') (eqb_univ_instance h _UU03c6_ u0 u')
     | _ -> false)
  | Coq_tConstruct (i, k, u0) ->
    (match u with
     | Coq_tConstruct (i', k', u') ->
       (&&) ((&&) (eq_ind i i') (eq_nat k k'))
         (eqb_univ_instance h _UU03c6_ u0 u')
     | _ -> false)
  | Coq_tCase (ind_and_nbparams, p, c, brs) ->
    let (ind, par) = ind_and_nbparams in
    (match u with
     | Coq_tCase (ind_and_nbparams0, p', c', brs') ->
       let (ind', par') = ind_and_nbparams0 in
       (&&)
         ((&&)
           ((&&) ((&&) (eq_ind ind ind') (eq_nat par par'))
             (eq_term h _UU03c6_ p p')) (eq_term h _UU03c6_ c c'))
         (forallb2 (fun pat pat0 ->
           let (_, b) = pat in let (_, b') = pat0 in eq_term h _UU03c6_ b b')
           brs brs')
     | _ -> false)
  | Coq_tProj (p, c) ->
    (match u with
     | Coq_tProj (p', c') ->
       (&&) (eq_projection p p') (eq_term h _UU03c6_ c c')
     | _ -> false)
  | Coq_tFix (mfix, idx) ->
    (match u with
     | Coq_tFix (mfix', idx') ->
       (&&)
         (forallb2 (fun x y ->
           (&&) (eq_term h _UU03c6_ x.dtype y.dtype)
             (eq_term h _UU03c6_ x.dbody y.dbody)) mfix mfix')
         (eq_nat idx idx')
     | _ -> false)
  | Coq_tCoFix (mfix, idx) ->
    (match u with
     | Coq_tCoFix (mfix', idx') ->
       (&&)
         (forallb2 (fun x y ->
           (&&) (eq_term h _UU03c6_ x.dtype y.dtype)
             (eq_term h _UU03c6_ x.dbody y.dbody)) mfix mfix')
         (Nat.eqb idx idx')
     | _ -> false)

(** val leq_term :
    checker_flags -> universes_graph -> term -> term -> bool **)

let rec leq_term h _UU03c6_ t0 u =
  match t0 with
  | Coq_tRel n -> (match u with
                   | Coq_tRel n' -> eq_nat n n'
                   | _ -> false)
  | Coq_tVar id ->
    (match u with
     | Coq_tVar id' -> eq_string id id'
     | _ -> false)
  | Coq_tEvar (ev, args) ->
    (match u with
     | Coq_tEvar (ev', args') ->
       (&&) (eq_nat ev ev') (forallb2 (eq_term h _UU03c6_) args args')
     | _ -> false)
  | Coq_tSort s ->
    (match u with
     | Coq_tSort s' -> try_leqb_universe h _UU03c6_ s s'
     | _ -> false)
  | Coq_tCast (f, _, t1) ->
    (match u with
     | Coq_tCast (f', _, t') ->
       (&&) (eq_term h _UU03c6_ f f') (eq_term h _UU03c6_ t1 t')
     | _ -> false)
  | Coq_tProd (_, b, t1) ->
    (match u with
     | Coq_tProd (_, b', t') ->
       (&&) (eq_term h _UU03c6_ b b') (leq_term h _UU03c6_ t1 t')
     | _ -> false)
  | Coq_tLambda (_, b, t1) ->
    (match u with
     | Coq_tLambda (_, b', t') ->
       (&&) (eq_term h _UU03c6_ b b') (eq_term h _UU03c6_ t1 t')
     | _ -> false)
  | Coq_tLetIn (_, b, t1, c) ->
    (match u with
     | Coq_tLetIn (_, b', t', c') ->
       (&&) ((&&) (eq_term h _UU03c6_ b b') (eq_term h _UU03c6_ t1 t'))
         (leq_term h _UU03c6_ c c')
     | _ -> false)
  | Coq_tApp (f, args) ->
    (match u with
     | Coq_tApp (f', args') ->
       (&&) (eq_term h _UU03c6_ f f')
         (forallb2 (eq_term h _UU03c6_) args args')
     | _ -> false)
  | Coq_tConst (c, u0) ->
    (match u with
     | Coq_tConst (c', u') ->
       (&&) (eq_constant c c') (eqb_univ_instance h _UU03c6_ u0 u')
     | _ -> false)
  | Coq_tInd (i, u0) ->
    (match u with
     | Coq_tInd (i', u') ->
       (&&) (eq_ind i i') (eqb_univ_instance h _UU03c6_ u0 u')
     | _ -> false)
  | Coq_tConstruct (i, k, u0) ->
    (match u with
     | Coq_tConstruct (i', k', u') ->
       (&&) ((&&) (eq_ind i i') (eq_nat k k'))
         (eqb_univ_instance h _UU03c6_ u0 u')
     | _ -> false)
  | Coq_tCase (ind_and_nbparams, p, c, brs) ->
    let (ind, par) = ind_and_nbparams in
    (match u with
     | Coq_tCase (ind_and_nbparams0, p', c', brs') ->
       let (ind', par') = ind_and_nbparams0 in
       (&&)
         ((&&)
           ((&&) ((&&) (eq_ind ind ind') (eq_nat par par'))
             (eq_term h _UU03c6_ p p')) (eq_term h _UU03c6_ c c'))
         (forallb2 (fun pat pat0 ->
           let (_, b) = pat in let (_, b') = pat0 in eq_term h _UU03c6_ b b')
           brs brs')
     | _ -> false)
  | Coq_tProj (p, c) ->
    (match u with
     | Coq_tProj (p', c') ->
       (&&) (eq_projection p p') (eq_term h _UU03c6_ c c')
     | _ -> false)
  | Coq_tFix (mfix, idx) ->
    (match u with
     | Coq_tFix (mfix', idx') ->
       (&&)
         (forallb2 (fun x y ->
           (&&) (eq_term h _UU03c6_ x.dtype y.dtype)
             (eq_term h _UU03c6_ x.dbody y.dbody)) mfix mfix')
         (eq_nat idx idx')
     | _ -> false)
  | Coq_tCoFix (mfix, idx) ->
    (match u with
     | Coq_tCoFix (mfix', idx') ->
       (&&)
         (forallb2 (fun x y ->
           (&&) (eq_term h _UU03c6_ x.dtype y.dtype)
             (eq_term h _UU03c6_ x.dbody y.dbody)) mfix mfix')
         (eq_nat idx idx')
     | _ -> false)

(** val nodelta_flags : RedFlags.t **)

let nodelta_flags =
  { RedFlags.beta = true; RedFlags.iota = true; RedFlags.zeta = true;
    RedFlags.delta = false; RedFlags.fix_ = true; RedFlags.cofix_ = true }

(** val unfold_one_fix :
    global_env -> nat -> context -> term mfixpoint -> nat -> term list ->
    term option **)

let unfold_one_fix _UU03a3_ n _UU0393_ mfix idx l =
  bind (Obj.magic option_monad) (Obj.magic unfold_fix mfix idx) (fun unf ->
    let (arg, fn) = unf in
    bind (Obj.magic option_monad) (nth_error l arg) (fun c ->
      bind (Obj.magic option_monad)
        (Obj.magic reduce_stack RedFlags.default _UU03a3_ _UU0393_ n c [])
        (fun cred ->
        let (cred0, _) = cred in
        if negb (isConstruct cred0) then None else Some fn)))

(** val unfold_one_case :
    global_env -> nat -> context -> term -> term option **)

let unfold_one_case _UU03a3_ n _UU0393_ c =
  bind (Obj.magic option_monad)
    (reduce_stack_term RedFlags.default _UU03a3_ _UU0393_ n c) (fun cred ->
    if negb ((||) (isConstruct cred) (isCoFix cred)) then None else Some cred)

(** val reducible_head :
    global_env -> nat -> context -> term -> term list -> term option **)

let reducible_head _UU03a3_ n _UU0393_ c l =
  match c with
  | Coq_tConst (c0, _) ->
    (match lookup_env _UU03a3_ c0 with
     | Some g ->
       (match g with
        | ConstantDecl (_, c1) -> c1.cst_body
        | InductiveDecl (_, _) -> None)
     | None -> None)
  | Coq_tCase (ind', p', c', brs) ->
    (match unfold_one_case _UU03a3_ n _UU0393_ c' with
     | Some c'0 -> Some (Coq_tCase (ind', p', c'0, brs))
     | None -> None)
  | Coq_tProj (p, c0) ->
    (match unfold_one_case _UU03a3_ n _UU0393_ c0 with
     | Some c' -> Some (Coq_tProj (p, c'))
     | None -> None)
  | Coq_tFix (mfix, idx) -> unfold_one_fix _UU03a3_ n _UU0393_ mfix idx l
  | _ -> None

(** val lookup_env : global_env -> ident -> global_decl option **)

let lookup_env =
  lookup_env

(** val opt_bool_to_bool : bool option -> bool **)

let opt_bool_to_bool = function
| Some b -> b
| None -> false

(** val isconv :
    checker_flags -> global_env -> universes_graph -> nat -> conv_pb ->
    context -> term -> term list -> term -> term list -> bool option **)

let isconv h _UU03a3_ g =
  let rec isconv0 n leq _UU0393_ t1 l1 t2 l2 =
    match n with
    | O -> None
    | S n0 ->
      bind (Obj.magic option_monad)
        (Obj.magic reduce_stack nodelta_flags _UU03a3_ _UU0393_ n0 t1 l1)
        (fun red1 ->
        bind (Obj.magic option_monad)
          (Obj.magic reduce_stack nodelta_flags _UU03a3_ _UU0393_ n0 t2 l2)
          (fun red2 ->
          let (t3, l3) = red1 in
          let (t4, l4) = red2 in isconv_prog0 n0 leq _UU0393_ t3 l3 t4 l4))
  and isconv_prog0 n leq _UU0393_ t1 l1 t2 l2 =
    match n with
    | O -> None
    | S n0 ->
      let isconv_stacks = fun l3 l4 ->
        ret option_monad
          (forallb2 (fun x y ->
            opt_bool_to_bool (isconv0 n0 Conv _UU0393_ x [] y [])) l3 l4)
      in
      let on_cond = fun b ->
        if b then isconv_stacks l1 l2 else ret option_monad false
      in
      let fallback = fun _ ->
        match reducible_head _UU03a3_ n0 _UU0393_ t1 l1 with
        | Some t3 ->
          bind (Obj.magic option_monad)
            (Obj.magic reduce_stack nodelta_flags _UU03a3_ _UU0393_ n0 t3 l1)
            (fun redt ->
            let (t4, l3) = redt in isconv_prog0 n0 leq _UU0393_ t4 l3 t2 l2)
        | None ->
          (match reducible_head _UU03a3_ n0 _UU0393_ t2 l2 with
           | Some t3 ->
             bind (Obj.magic option_monad)
               (Obj.magic reduce_stack nodelta_flags _UU03a3_ _UU0393_ n0 t3
                 l2) (fun redt ->
               let (t4, l3) = redt in isconv_prog0 n0 leq _UU0393_ t1 l1 t4 l3)
           | None ->
             Obj.magic on_cond
               (match leq with
                | Conv -> eq_term h g t1 t2
                | Cumul -> leq_term h g t1 t2))
      in
      (match t1 with
       | Coq_tCast (_, _, _) ->
         (match t2 with
          | Coq_tCast (_, _, _) -> None
          | _ -> fallback ())
       | Coq_tProd (na, b, t0) ->
         (match t2 with
          | Coq_tProd (_, b', t') ->
            bind (Obj.magic option_monad)
              (isconv0 n0 Conv _UU0393_ b [] b' []) (fun cnv ->
              if cnv
              then isconv0 n0 leq (snoc _UU0393_ (vass na b)) t0 [] t' []
              else ret (Obj.magic option_monad) false)
          | _ -> fallback ())
       | Coq_tLambda (na, b, t0) ->
         (match t2 with
          | Coq_tLambda (_, b', t') ->
            bind (Obj.magic option_monad)
              (isconv0 n0 Conv _UU0393_ b [] b' []) (fun cnv ->
              if cnv
              then isconv0 n0 Conv (snoc _UU0393_ (vass na b)) t0 [] t' []
              else ret (Obj.magic option_monad) false)
          | _ -> fallback ())
       | Coq_tApp (_, _) ->
         (match t2 with
          | Coq_tApp (_, _) -> None
          | _ -> fallback ())
       | Coq_tConst (c, _) ->
         (match t2 with
          | Coq_tConst (c', _) ->
            if eq_constant c c'
            then bind (Obj.magic option_monad)
                   (Obj.magic isconv_stacks l1 l2) (fun b ->
                   if b
                   then ret (Obj.magic option_monad) true
                   else (match lookup_env _UU03a3_ c with
                         | Some g0 ->
                           (match g0 with
                            | ConstantDecl (_, c0) ->
                              let { cst_type = _; cst_body = cst_body0;
                                cst_universes = _ } = c0
                              in
                              (match cst_body0 with
                               | Some body ->
                                 isconv0 n0 leq _UU0393_ body l1 body l2
                               | None -> ret (Obj.magic option_monad) false)
                            | InductiveDecl (_, _) ->
                              ret (Obj.magic option_monad) false)
                         | None -> ret (Obj.magic option_monad) false))
            else (match lookup_env _UU03a3_ c' with
                  | Some g0 ->
                    (match g0 with
                     | ConstantDecl (_, c0) ->
                       let { cst_type = _; cst_body = cst_body0;
                         cst_universes = _ } = c0
                       in
                       (match cst_body0 with
                        | Some body -> isconv0 n0 leq _UU0393_ t1 l1 body l2
                        | None ->
                          (match lookup_env _UU03a3_ c with
                           | Some g1 ->
                             (match g1 with
                              | ConstantDecl (_, c1) ->
                                let { cst_type = _; cst_body = cst_body1;
                                  cst_universes = _ } = c1
                                in
                                (match cst_body1 with
                                 | Some body ->
                                   isconv0 n0 leq _UU0393_ body l1 t2 l2
                                 | None -> ret (Obj.magic option_monad) false)
                              | InductiveDecl (_, _) ->
                                ret (Obj.magic option_monad) false)
                           | None -> ret (Obj.magic option_monad) false))
                     | InductiveDecl (_, _) ->
                       (match lookup_env _UU03a3_ c with
                        | Some g1 ->
                          (match g1 with
                           | ConstantDecl (_, c0) ->
                             let { cst_type = _; cst_body = cst_body0;
                               cst_universes = _ } = c0
                             in
                             (match cst_body0 with
                              | Some body ->
                                isconv0 n0 leq _UU0393_ body l1 t2 l2
                              | None -> ret (Obj.magic option_monad) false)
                           | InductiveDecl (_, _) ->
                             ret (Obj.magic option_monad) false)
                        | None -> ret (Obj.magic option_monad) false))
                  | None ->
                    (match lookup_env _UU03a3_ c with
                     | Some g0 ->
                       (match g0 with
                        | ConstantDecl (_, c0) ->
                          let { cst_type = _; cst_body = cst_body0;
                            cst_universes = _ } = c0
                          in
                          (match cst_body0 with
                           | Some body ->
                             isconv0 n0 leq _UU0393_ body l1 t2 l2
                           | None -> ret (Obj.magic option_monad) false)
                        | InductiveDecl (_, _) ->
                          ret (Obj.magic option_monad) false)
                     | None -> ret (Obj.magic option_monad) false))
          | _ -> fallback ())
       | Coq_tCase (ind_and_nbparams, p, c, brs) ->
         (match t2 with
          | Coq_tCase (_, p', c', brs') ->
            if (&&) ((&&) (eq_term h g p p') (eq_term h g c c'))
                 (forallb2 (fun pat pat0 ->
                   let (_, b) = pat in let (_, b') = pat0 in eq_term h g b b')
                   brs brs')
            then ret (Obj.magic option_monad) true
            else bind (Obj.magic option_monad)
                   (Obj.magic reduce_stack_term RedFlags.default _UU03a3_
                     _UU0393_ n0 c) (fun cred ->
                   bind (Obj.magic option_monad)
                     (Obj.magic reduce_stack_term RedFlags.default _UU03a3_
                       _UU0393_ n0 c') (fun c'red ->
                     if (&&) (eq_term h g cred c) (eq_term h g c'red c')
                     then ret (Obj.magic option_monad) true
                     else isconv0 n0 leq _UU0393_ (Coq_tCase
                            (ind_and_nbparams, p, cred, brs)) l1 (Coq_tCase
                            (ind_and_nbparams, p, c'red, brs')) l2))
          | _ -> fallback ())
       | Coq_tProj (p, c) ->
         (match t2 with
          | Coq_tProj (p', c') ->
            Obj.magic on_cond ((&&) (eq_projection p p') (eq_term h g c c'))
          | _ -> fallback ())
       | Coq_tFix (mfix, idx) ->
         (match t2 with
          | Coq_tFix (mfix', idx') ->
            if (&&) (eq_term h g t1 t2)
                 (opt_bool_to_bool (Obj.magic isconv_stacks l1 l2))
            then ret (Obj.magic option_monad) true
            else (match unfold_one_fix _UU03a3_ n0 _UU0393_ mfix idx l1 with
                  | Some t3 ->
                    bind (Obj.magic option_monad)
                      (Obj.magic reduce_stack nodelta_flags _UU03a3_ _UU0393_
                        n0 t3 l1) (fun redt ->
                      let (t4, l3) = redt in
                      isconv_prog0 n0 leq _UU0393_ t4 l3 t2 l2)
                  | None ->
                    (match unfold_one_fix _UU03a3_ n0 _UU0393_ mfix' idx' l2 with
                     | Some t3 ->
                       bind (Obj.magic option_monad)
                         (Obj.magic reduce_stack nodelta_flags _UU03a3_
                           _UU0393_ n0 t3 l2) (fun redt ->
                         let (t4, l3) = redt in
                         isconv_prog0 n0 leq _UU0393_ t1 l1 t4 l3)
                     | None -> ret (Obj.magic option_monad) false))
          | _ -> fallback ())
       | Coq_tCoFix (_, _) ->
         (match t2 with
          | Coq_tCoFix (_, _) -> Obj.magic on_cond (eq_term h g t1 t2)
          | _ -> fallback ())
       | _ -> fallback ())
  in isconv0

(** val isconv_prog :
    checker_flags -> global_env -> universes_graph -> nat -> conv_pb ->
    context -> term -> term list -> term -> term list -> bool option **)

let isconv_prog h _UU03a3_ g =
  let rec isconv0 n leq _UU0393_ t1 l1 t2 l2 =
    match n with
    | O -> None
    | S n0 ->
      bind (Obj.magic option_monad)
        (Obj.magic reduce_stack nodelta_flags _UU03a3_ _UU0393_ n0 t1 l1)
        (fun red1 ->
        bind (Obj.magic option_monad)
          (Obj.magic reduce_stack nodelta_flags _UU03a3_ _UU0393_ n0 t2 l2)
          (fun red2 ->
          let (t3, l3) = red1 in
          let (t4, l4) = red2 in isconv_prog0 n0 leq _UU0393_ t3 l3 t4 l4))
  and isconv_prog0 n leq _UU0393_ t1 l1 t2 l2 =
    match n with
    | O -> None
    | S n0 ->
      let isconv_stacks = fun l3 l4 ->
        ret option_monad
          (forallb2 (fun x y ->
            opt_bool_to_bool (isconv0 n0 Conv _UU0393_ x [] y [])) l3 l4)
      in
      let on_cond = fun b ->
        if b then isconv_stacks l1 l2 else ret option_monad false
      in
      let fallback = fun _ ->
        match reducible_head _UU03a3_ n0 _UU0393_ t1 l1 with
        | Some t3 ->
          bind (Obj.magic option_monad)
            (Obj.magic reduce_stack nodelta_flags _UU03a3_ _UU0393_ n0 t3 l1)
            (fun redt ->
            let (t4, l3) = redt in isconv_prog0 n0 leq _UU0393_ t4 l3 t2 l2)
        | None ->
          (match reducible_head _UU03a3_ n0 _UU0393_ t2 l2 with
           | Some t3 ->
             bind (Obj.magic option_monad)
               (Obj.magic reduce_stack nodelta_flags _UU03a3_ _UU0393_ n0 t3
                 l2) (fun redt ->
               let (t4, l3) = redt in isconv_prog0 n0 leq _UU0393_ t1 l1 t4 l3)
           | None ->
             Obj.magic on_cond
               (match leq with
                | Conv -> eq_term h g t1 t2
                | Cumul -> leq_term h g t1 t2))
      in
      (match t1 with
       | Coq_tCast (_, _, _) ->
         (match t2 with
          | Coq_tCast (_, _, _) -> None
          | _ -> fallback ())
       | Coq_tProd (na, b, t0) ->
         (match t2 with
          | Coq_tProd (_, b', t') ->
            bind (Obj.magic option_monad)
              (isconv0 n0 Conv _UU0393_ b [] b' []) (fun cnv ->
              if cnv
              then isconv0 n0 leq (snoc _UU0393_ (vass na b)) t0 [] t' []
              else ret (Obj.magic option_monad) false)
          | _ -> fallback ())
       | Coq_tLambda (na, b, t0) ->
         (match t2 with
          | Coq_tLambda (_, b', t') ->
            bind (Obj.magic option_monad)
              (isconv0 n0 Conv _UU0393_ b [] b' []) (fun cnv ->
              if cnv
              then isconv0 n0 Conv (snoc _UU0393_ (vass na b)) t0 [] t' []
              else ret (Obj.magic option_monad) false)
          | _ -> fallback ())
       | Coq_tApp (_, _) ->
         (match t2 with
          | Coq_tApp (_, _) -> None
          | _ -> fallback ())
       | Coq_tConst (c, _) ->
         (match t2 with
          | Coq_tConst (c', _) ->
            if eq_constant c c'
            then bind (Obj.magic option_monad)
                   (Obj.magic isconv_stacks l1 l2) (fun b ->
                   if b
                   then ret (Obj.magic option_monad) true
                   else (match lookup_env _UU03a3_ c with
                         | Some g0 ->
                           (match g0 with
                            | ConstantDecl (_, c0) ->
                              let { cst_type = _; cst_body = cst_body0;
                                cst_universes = _ } = c0
                              in
                              (match cst_body0 with
                               | Some body ->
                                 isconv0 n0 leq _UU0393_ body l1 body l2
                               | None -> ret (Obj.magic option_monad) false)
                            | InductiveDecl (_, _) ->
                              ret (Obj.magic option_monad) false)
                         | None -> ret (Obj.magic option_monad) false))
            else (match lookup_env _UU03a3_ c' with
                  | Some g0 ->
                    (match g0 with
                     | ConstantDecl (_, c0) ->
                       let { cst_type = _; cst_body = cst_body0;
                         cst_universes = _ } = c0
                       in
                       (match cst_body0 with
                        | Some body -> isconv0 n0 leq _UU0393_ t1 l1 body l2
                        | None ->
                          (match lookup_env _UU03a3_ c with
                           | Some g1 ->
                             (match g1 with
                              | ConstantDecl (_, c1) ->
                                let { cst_type = _; cst_body = cst_body1;
                                  cst_universes = _ } = c1
                                in
                                (match cst_body1 with
                                 | Some body ->
                                   isconv0 n0 leq _UU0393_ body l1 t2 l2
                                 | None -> ret (Obj.magic option_monad) false)
                              | InductiveDecl (_, _) ->
                                ret (Obj.magic option_monad) false)
                           | None -> ret (Obj.magic option_monad) false))
                     | InductiveDecl (_, _) ->
                       (match lookup_env _UU03a3_ c with
                        | Some g1 ->
                          (match g1 with
                           | ConstantDecl (_, c0) ->
                             let { cst_type = _; cst_body = cst_body0;
                               cst_universes = _ } = c0
                             in
                             (match cst_body0 with
                              | Some body ->
                                isconv0 n0 leq _UU0393_ body l1 t2 l2
                              | None -> ret (Obj.magic option_monad) false)
                           | InductiveDecl (_, _) ->
                             ret (Obj.magic option_monad) false)
                        | None -> ret (Obj.magic option_monad) false))
                  | None ->
                    (match lookup_env _UU03a3_ c with
                     | Some g0 ->
                       (match g0 with
                        | ConstantDecl (_, c0) ->
                          let { cst_type = _; cst_body = cst_body0;
                            cst_universes = _ } = c0
                          in
                          (match cst_body0 with
                           | Some body ->
                             isconv0 n0 leq _UU0393_ body l1 t2 l2
                           | None -> ret (Obj.magic option_monad) false)
                        | InductiveDecl (_, _) ->
                          ret (Obj.magic option_monad) false)
                     | None -> ret (Obj.magic option_monad) false))
          | _ -> fallback ())
       | Coq_tCase (ind_and_nbparams, p, c, brs) ->
         (match t2 with
          | Coq_tCase (_, p', c', brs') ->
            if (&&) ((&&) (eq_term h g p p') (eq_term h g c c'))
                 (forallb2 (fun pat pat0 ->
                   let (_, b) = pat in let (_, b') = pat0 in eq_term h g b b')
                   brs brs')
            then ret (Obj.magic option_monad) true
            else bind (Obj.magic option_monad)
                   (Obj.magic reduce_stack_term RedFlags.default _UU03a3_
                     _UU0393_ n0 c) (fun cred ->
                   bind (Obj.magic option_monad)
                     (Obj.magic reduce_stack_term RedFlags.default _UU03a3_
                       _UU0393_ n0 c') (fun c'red ->
                     if (&&) (eq_term h g cred c) (eq_term h g c'red c')
                     then ret (Obj.magic option_monad) true
                     else isconv0 n0 leq _UU0393_ (Coq_tCase
                            (ind_and_nbparams, p, cred, brs)) l1 (Coq_tCase
                            (ind_and_nbparams, p, c'red, brs')) l2))
          | _ -> fallback ())
       | Coq_tProj (p, c) ->
         (match t2 with
          | Coq_tProj (p', c') ->
            Obj.magic on_cond ((&&) (eq_projection p p') (eq_term h g c c'))
          | _ -> fallback ())
       | Coq_tFix (mfix, idx) ->
         (match t2 with
          | Coq_tFix (mfix', idx') ->
            if (&&) (eq_term h g t1 t2)
                 (opt_bool_to_bool (Obj.magic isconv_stacks l1 l2))
            then ret (Obj.magic option_monad) true
            else (match unfold_one_fix _UU03a3_ n0 _UU0393_ mfix idx l1 with
                  | Some t3 ->
                    bind (Obj.magic option_monad)
                      (Obj.magic reduce_stack nodelta_flags _UU03a3_ _UU0393_
                        n0 t3 l1) (fun redt ->
                      let (t4, l3) = redt in
                      isconv_prog0 n0 leq _UU0393_ t4 l3 t2 l2)
                  | None ->
                    (match unfold_one_fix _UU03a3_ n0 _UU0393_ mfix' idx' l2 with
                     | Some t3 ->
                       bind (Obj.magic option_monad)
                         (Obj.magic reduce_stack nodelta_flags _UU03a3_
                           _UU0393_ n0 t3 l2) (fun redt ->
                         let (t4, l3) = redt in
                         isconv_prog0 n0 leq _UU0393_ t1 l1 t4 l3)
                     | None -> ret (Obj.magic option_monad) false))
          | _ -> fallback ())
       | Coq_tCoFix (_, _) ->
         (match t2 with
          | Coq_tCoFix (_, _) -> Obj.magic on_cond (eq_term h g t1 t2)
          | _ -> fallback ())
       | _ -> fallback ())
  in isconv_prog0

(** val try_reduce : global_env -> context -> nat -> term -> term **)

let try_reduce _UU03a3_ _UU0393_ n t0 =
  match reduce_opt RedFlags.default _UU03a3_ _UU0393_ n t0 with
  | Some t' -> t'
  | None -> t0

type type_error =
| UnboundRel of nat
| UnboundVar of char list
| UnboundMeta of nat
| UnboundEvar of nat
| UndeclaredConstant of char list
| UndeclaredInductive of inductive
| UndeclaredConstructor of inductive * nat
| NotConvertible of context * term * term * term * term
| NotASort of term
| NotAProduct of term * term
| NotAnInductive of term
| IllFormedFix of term mfixpoint * nat
| UnsatisfiedConstraints of ConstraintSet.t
| UnsatisfiableConstraints of ConstraintSet.t
| NotEnoughFuel of nat

(** val type_error_rect :
    (nat -> 'a1) -> (char list -> 'a1) -> (nat -> 'a1) -> (nat -> 'a1) ->
    (char list -> 'a1) -> (inductive -> 'a1) -> (inductive -> nat -> 'a1) ->
    (context -> term -> term -> term -> term -> 'a1) -> (term -> 'a1) ->
    (term -> term -> 'a1) -> (term -> 'a1) -> (term mfixpoint -> nat -> 'a1)
    -> (ConstraintSet.t -> 'a1) -> (ConstraintSet.t -> 'a1) -> (nat -> 'a1)
    -> type_error -> 'a1 **)

let type_error_rect f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13 = function
| UnboundRel x -> f x
| UnboundVar x -> f0 x
| UnboundMeta x -> f1 x
| UnboundEvar x -> f2 x
| UndeclaredConstant x -> f3 x
| UndeclaredInductive x -> f4 x
| UndeclaredConstructor (x, x0) -> f5 x x0
| NotConvertible (x, x0, x1, x2, x3) -> f6 x x0 x1 x2 x3
| NotASort x -> f7 x
| NotAProduct (x, x0) -> f8 x x0
| NotAnInductive x -> f9 x
| IllFormedFix (x, x0) -> f10 x x0
| UnsatisfiedConstraints x -> f11 x
| UnsatisfiableConstraints x -> f12 x
| NotEnoughFuel x -> f13 x

(** val type_error_rec :
    (nat -> 'a1) -> (char list -> 'a1) -> (nat -> 'a1) -> (nat -> 'a1) ->
    (char list -> 'a1) -> (inductive -> 'a1) -> (inductive -> nat -> 'a1) ->
    (context -> term -> term -> term -> term -> 'a1) -> (term -> 'a1) ->
    (term -> term -> 'a1) -> (term -> 'a1) -> (term mfixpoint -> nat -> 'a1)
    -> (ConstraintSet.t -> 'a1) -> (ConstraintSet.t -> 'a1) -> (nat -> 'a1)
    -> type_error -> 'a1 **)

let type_error_rec f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13 = function
| UnboundRel x -> f x
| UnboundVar x -> f0 x
| UnboundMeta x -> f1 x
| UnboundEvar x -> f2 x
| UndeclaredConstant x -> f3 x
| UndeclaredInductive x -> f4 x
| UndeclaredConstructor (x, x0) -> f5 x x0
| NotConvertible (x, x0, x1, x2, x3) -> f6 x x0 x1 x2 x3
| NotASort x -> f7 x
| NotAProduct (x, x0) -> f8 x x0
| NotAnInductive x -> f9 x
| IllFormedFix (x, x0) -> f10 x x0
| UnsatisfiedConstraints x -> f11 x
| UnsatisfiableConstraints x -> f12 x
| NotEnoughFuel x -> f13 x

(** val string_of_type_error : type_error -> char list **)

let string_of_type_error = function
| UnboundRel n ->
  append
    ('U'::('n'::('b'::('o'::('o'::('u'::('n'::('d'::(' '::('r'::('e'::('l'::(' '::[])))))))))))))
    (string_of_nat n)
| UnboundVar id ->
  append
    ('U'::('n'::('b'::('o'::('u'::('n'::('d'::(' '::('v'::('a'::('r'::(' '::[]))))))))))))
    id
| UnboundMeta m ->
  append
    ('U'::('n'::('b'::('o'::('u'::('n'::('d'::(' '::('m'::('e'::('t'::('a'::(' '::[])))))))))))))
    (string_of_nat m)
| UnboundEvar ev ->
  append
    ('U'::('n'::('b'::('o'::('u'::('n'::('d'::(' '::('e'::('v'::('a'::('r'::(' '::[])))))))))))))
    (string_of_nat ev)
| UndeclaredConstant c ->
  append
    ('U'::('n'::('d'::('e'::('c'::('l'::('a'::('r'::('e'::('d'::(' '::('c'::('o'::('n'::('s'::('t'::('a'::('n'::('t'::(' '::[]))))))))))))))))))))
    c
| UndeclaredInductive c ->
  append
    ('U'::('n'::('d'::('e'::('c'::('l'::('a'::('r'::('e'::('d'::(' '::('i'::('n'::('d'::('u'::('c'::('t'::('i'::('v'::('e'::(' '::[])))))))))))))))))))))
    c.inductive_mind
| UndeclaredConstructor (c, _) ->
  append
    ('U'::('n'::('d'::('e'::('c'::('l'::('a'::('r'::('e'::('d'::(' '::('i'::('n'::('d'::('u'::('c'::('t'::('i'::('v'::('e'::(' '::[])))))))))))))))))))))
    c.inductive_mind
| NotConvertible (_, t0, u, t', u') ->
  append
    ('T'::('e'::('r'::('m'::('s'::(' '::('a'::('r'::('e'::(' '::('n'::('o'::('t'::(' '::('c'::('o'::('n'::('v'::('e'::('r'::('t'::('i'::('b'::('l'::('e'::(':'::(' '::[])))))))))))))))))))))))))))
    (append (string_of_term t0)
      (append (' '::[])
        (append (string_of_term u)
          (append
            (' '::('a'::('f'::('t'::('e'::('r'::(' '::('r'::('e'::('d'::('u'::('c'::('t'::('i'::('o'::('n'::(':'::(' '::[]))))))))))))))))))
            (append (string_of_term t')
              (append (' '::[]) (string_of_term u')))))))
| NotASort _ ->
  'N'::('o'::('t'::(' '::('a'::(' '::('s'::('o'::('r'::('t'::[])))))))))
| NotAProduct (_, _) ->
  'N'::('o'::('t'::(' '::('a'::(' '::('p'::('r'::('o'::('d'::('u'::('c'::('t'::[]))))))))))))
| NotAnInductive _ ->
  'N'::('o'::('t'::(' '::('a'::('n'::(' '::('i'::('n'::('d'::('u'::('c'::('t'::('i'::('v'::('e'::[])))))))))))))))
| IllFormedFix (_, _) ->
  'I'::('l'::('l'::('-'::('f'::('o'::('r'::('m'::('e'::('d'::(' '::('r'::('e'::('c'::('u'::('r'::('s'::('i'::('v'::('e'::(' '::('d'::('e'::('f'::('i'::('n'::('i'::('t'::('i'::('o'::('n'::[]))))))))))))))))))))))))))))))
| UnsatisfiedConstraints _ ->
  'U'::('n'::('s'::('a'::('t'::('i'::('s'::('f'::('i'::('e'::('d'::(' '::('c'::('o'::('n'::('s'::('t'::('r'::('a'::('i'::('n'::('t'::('s'::[]))))))))))))))))))))))
| UnsatisfiableConstraints _ ->
  'U'::('n'::('s'::('a'::('t'::('i'::('s'::('f'::('i'::('a'::('b'::('l'::('e'::(' '::('c'::('o'::('n'::('s'::('t'::('r'::('a'::('i'::('n'::('t'::('s'::[]))))))))))))))))))))))))
| NotEnoughFuel _ ->
  'N'::('o'::('t'::(' '::('e'::('n'::('o'::('u'::('g'::('h'::(' '::('f'::('u'::('e'::('l'::[]))))))))))))))

type 'a typing_result =
| Checked of 'a
| TypeError of type_error

(** val typing_result_rect :
    ('a1 -> 'a2) -> (type_error -> 'a2) -> 'a1 typing_result -> 'a2 **)

let typing_result_rect f f0 = function
| Checked x -> f x
| TypeError x -> f0 x

(** val typing_result_rec :
    ('a1 -> 'a2) -> (type_error -> 'a2) -> 'a1 typing_result -> 'a2 **)

let typing_result_rec f f0 = function
| Checked x -> f x
| TypeError x -> f0 x

(** val typing_monad : __ typing_result coq_Monad **)

let typing_monad =
  { ret = (fun _ a -> Checked a); bind = (fun _ _ m f ->
    match m with
    | Checked a -> f a
    | TypeError t0 -> TypeError t0) }

(** val monad_exc : (type_error, __ typing_result) coq_MonadExc **)

let monad_exc =
  { raise = (fun _ e -> TypeError e); catch = (fun _ m f ->
    match m with
    | Checked _ -> m
    | TypeError t0 -> f t0) }

(** val check_conv_gen :
    checker_flags -> coq_Fuel -> conv_pb -> global_env -> universes_graph ->
    context -> term -> term -> unit typing_result **)

let check_conv_gen h f conv_pb0 _UU03a3_ g _UU0393_ t0 u =
  match isconv h _UU03a3_ g (fuel f) conv_pb0 _UU0393_ t0 [] u [] with
  | Some b ->
    if b
    then ret (Obj.magic typing_monad) ()
    else raise (Obj.magic monad_exc) (NotConvertible (_UU0393_, t0, u, t0, u))
  | None -> raise (Obj.magic monad_exc) (NotEnoughFuel (fuel f))

(** val check_conv_leq :
    checker_flags -> coq_Fuel -> global_env -> universes_graph -> context ->
    term -> term -> unit typing_result **)

let check_conv_leq h f =
  check_conv_gen h f Cumul

(** val check_conv :
    checker_flags -> coq_Fuel -> global_env -> universes_graph -> context ->
    term -> term -> unit typing_result **)

let check_conv h f =
  check_conv_gen h f Conv

(** val conv_spec :
    checker_flags -> coq_Fuel -> global_env_ext -> Coq_wGraph.t -> context ->
    term -> term -> (conv, __) isEquiv **)

let conv_spec _ _ _ _ _ _ _ =
  todo
    ('C'::('h'::('e'::('c'::('k'::('e'::('r'::('.'::('c'::('o'::('n'::('v'::('_'::('s'::('p'::('e'::('c'::[])))))))))))))))))

(** val cumul_spec :
    checker_flags -> coq_Fuel -> global_env_ext -> Coq_wGraph.t -> context ->
    term -> term -> (cumul, __) isEquiv **)

let cumul_spec _ _ _ _ _ _ _ =
  todo
    ('C'::('h'::('e'::('c'::('k'::('e'::('r'::('.'::('c'::('u'::('m'::('u'::('l'::('_'::('s'::('p'::('e'::('c'::[]))))))))))))))))))

(** val reduce_cumul :
    checker_flags -> global_env_ext -> context -> nat -> term -> cumul **)

let reduce_cumul _ _ _ _ _ =
  todo
    ('C'::('h'::('e'::('c'::('k'::('e'::('r'::('.'::('r'::('e'::('d'::('u'::('c'::('e'::('_'::('c'::('u'::('m'::('u'::('l'::[]))))))))))))))))))))

(** val hnf_stack :
    coq_Fuel -> global_env -> context -> term -> (term * term list)
    typing_result **)

let hnf_stack f _UU03a3_ _UU0393_ t0 =
  match reduce_stack RedFlags.default _UU03a3_ _UU0393_ (fuel f) t0 [] with
  | Some t' -> ret (Obj.magic typing_monad) t'
  | None -> raise (Obj.magic monad_exc) (NotEnoughFuel (fuel f))

(** val reduce :
    coq_Fuel -> global_env -> context -> term -> term typing_result **)

let reduce f _UU03a3_ _UU0393_ t0 =
  match reduce_opt RedFlags.default _UU03a3_ _UU0393_ (fuel f) t0 with
  | Some t' -> ret (Obj.magic typing_monad) t'
  | None -> raise (Obj.magic monad_exc) (NotEnoughFuel (fuel f))

(** val reduce_to_sort :
    coq_Fuel -> global_env -> context -> term -> universe typing_result **)

let reduce_to_sort f _UU03a3_ _UU0393_ t0 = match t0 with
| Coq_tRel _ ->
  bind (Obj.magic typing_monad) (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0)
    (fun t' ->
    let (y, y0) = t' in
    (match y with
     | Coq_tSort s ->
       (match y0 with
        | [] -> ret (Obj.magic typing_monad) s
        | _ :: _ -> raise (Obj.magic monad_exc) (NotASort t0))
     | _ -> raise (Obj.magic monad_exc) (NotASort t0)))
| Coq_tVar _ ->
  bind (Obj.magic typing_monad) (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0)
    (fun t' ->
    let (y, y0) = t' in
    (match y with
     | Coq_tSort s ->
       (match y0 with
        | [] -> ret (Obj.magic typing_monad) s
        | _ :: _ -> raise (Obj.magic monad_exc) (NotASort t0))
     | _ -> raise (Obj.magic monad_exc) (NotASort t0)))
| Coq_tEvar (_, _) ->
  bind (Obj.magic typing_monad) (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0)
    (fun t' ->
    let (y, y0) = t' in
    (match y with
     | Coq_tSort s ->
       (match y0 with
        | [] -> ret (Obj.magic typing_monad) s
        | _ :: _ -> raise (Obj.magic monad_exc) (NotASort t0))
     | _ -> raise (Obj.magic monad_exc) (NotASort t0)))
| Coq_tSort s -> ret (Obj.magic typing_monad) s
| Coq_tCast (_, _, _) ->
  bind (Obj.magic typing_monad) (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0)
    (fun t' ->
    let (y, y0) = t' in
    (match y with
     | Coq_tSort s ->
       (match y0 with
        | [] -> ret (Obj.magic typing_monad) s
        | _ :: _ -> raise (Obj.magic monad_exc) (NotASort t0))
     | _ -> raise (Obj.magic monad_exc) (NotASort t0)))
| Coq_tProd (_, _, _) ->
  bind (Obj.magic typing_monad) (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0)
    (fun t' ->
    let (y, y0) = t' in
    (match y with
     | Coq_tSort s ->
       (match y0 with
        | [] -> ret (Obj.magic typing_monad) s
        | _ :: _ -> raise (Obj.magic monad_exc) (NotASort t0))
     | _ -> raise (Obj.magic monad_exc) (NotASort t0)))
| Coq_tLambda (_, _, _) ->
  bind (Obj.magic typing_monad) (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0)
    (fun t' ->
    let (y, y0) = t' in
    (match y with
     | Coq_tSort s ->
       (match y0 with
        | [] -> ret (Obj.magic typing_monad) s
        | _ :: _ -> raise (Obj.magic monad_exc) (NotASort t0))
     | _ -> raise (Obj.magic monad_exc) (NotASort t0)))
| Coq_tLetIn (_, _, _, _) ->
  bind (Obj.magic typing_monad) (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0)
    (fun t' ->
    let (y, y0) = t' in
    (match y with
     | Coq_tSort s ->
       (match y0 with
        | [] -> ret (Obj.magic typing_monad) s
        | _ :: _ -> raise (Obj.magic monad_exc) (NotASort t0))
     | _ -> raise (Obj.magic monad_exc) (NotASort t0)))
| Coq_tApp (_, _) ->
  bind (Obj.magic typing_monad) (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0)
    (fun t' ->
    let (y, y0) = t' in
    (match y with
     | Coq_tSort s ->
       (match y0 with
        | [] -> ret (Obj.magic typing_monad) s
        | _ :: _ -> raise (Obj.magic monad_exc) (NotASort t0))
     | _ -> raise (Obj.magic monad_exc) (NotASort t0)))
| Coq_tConst (_, _) ->
  bind (Obj.magic typing_monad) (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0)
    (fun t' ->
    let (y, y0) = t' in
    (match y with
     | Coq_tSort s ->
       (match y0 with
        | [] -> ret (Obj.magic typing_monad) s
        | _ :: _ -> raise (Obj.magic monad_exc) (NotASort t0))
     | _ -> raise (Obj.magic monad_exc) (NotASort t0)))
| Coq_tInd (_, _) ->
  bind (Obj.magic typing_monad) (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0)
    (fun t' ->
    let (y, y0) = t' in
    (match y with
     | Coq_tSort s ->
       (match y0 with
        | [] -> ret (Obj.magic typing_monad) s
        | _ :: _ -> raise (Obj.magic monad_exc) (NotASort t0))
     | _ -> raise (Obj.magic monad_exc) (NotASort t0)))
| Coq_tConstruct (_, _, _) ->
  bind (Obj.magic typing_monad) (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0)
    (fun t' ->
    let (y, y0) = t' in
    (match y with
     | Coq_tSort s ->
       (match y0 with
        | [] -> ret (Obj.magic typing_monad) s
        | _ :: _ -> raise (Obj.magic monad_exc) (NotASort t0))
     | _ -> raise (Obj.magic monad_exc) (NotASort t0)))
| Coq_tCase (_, _, _, _) ->
  bind (Obj.magic typing_monad) (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0)
    (fun t' ->
    let (y, y0) = t' in
    (match y with
     | Coq_tSort s ->
       (match y0 with
        | [] -> ret (Obj.magic typing_monad) s
        | _ :: _ -> raise (Obj.magic monad_exc) (NotASort t0))
     | _ -> raise (Obj.magic monad_exc) (NotASort t0)))
| Coq_tProj (_, _) ->
  bind (Obj.magic typing_monad) (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0)
    (fun t' ->
    let (y, y0) = t' in
    (match y with
     | Coq_tSort s ->
       (match y0 with
        | [] -> ret (Obj.magic typing_monad) s
        | _ :: _ -> raise (Obj.magic monad_exc) (NotASort t0))
     | _ -> raise (Obj.magic monad_exc) (NotASort t0)))
| Coq_tFix (_, _) ->
  bind (Obj.magic typing_monad) (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0)
    (fun t' ->
    let (y, y0) = t' in
    (match y with
     | Coq_tSort s ->
       (match y0 with
        | [] -> ret (Obj.magic typing_monad) s
        | _ :: _ -> raise (Obj.magic monad_exc) (NotASort t0))
     | _ -> raise (Obj.magic monad_exc) (NotASort t0)))
| Coq_tCoFix (_, _) ->
  bind (Obj.magic typing_monad) (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0)
    (fun t' ->
    let (y, y0) = t' in
    (match y with
     | Coq_tSort s ->
       (match y0 with
        | [] -> ret (Obj.magic typing_monad) s
        | _ :: _ -> raise (Obj.magic monad_exc) (NotASort t0))
     | _ -> raise (Obj.magic monad_exc) (NotASort t0)))

(** val reduce_to_prod :
    coq_Fuel -> global_env -> context -> term -> (term * term) typing_result **)

let reduce_to_prod f _UU03a3_ _UU0393_ t0 = match t0 with
| Coq_tRel _ ->
  bind (Obj.magic typing_monad) (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0)
    (fun t' ->
    let (y, y0) = t' in
    (match y with
     | Coq_tProd (_, a, b) ->
       (match y0 with
        | [] -> ret (Obj.magic typing_monad) (a, b)
        | _ :: _ -> raise (Obj.magic monad_exc) (NotAProduct (t0, (zip t'))))
     | _ -> raise (Obj.magic monad_exc) (NotAProduct (t0, (zip t')))))
| Coq_tVar _ ->
  bind (Obj.magic typing_monad) (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0)
    (fun t' ->
    let (y, y0) = t' in
    (match y with
     | Coq_tProd (_, a, b) ->
       (match y0 with
        | [] -> ret (Obj.magic typing_monad) (a, b)
        | _ :: _ -> raise (Obj.magic monad_exc) (NotAProduct (t0, (zip t'))))
     | _ -> raise (Obj.magic monad_exc) (NotAProduct (t0, (zip t')))))
| Coq_tEvar (_, _) ->
  bind (Obj.magic typing_monad) (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0)
    (fun t' ->
    let (y, y0) = t' in
    (match y with
     | Coq_tProd (_, a, b) ->
       (match y0 with
        | [] -> ret (Obj.magic typing_monad) (a, b)
        | _ :: _ -> raise (Obj.magic monad_exc) (NotAProduct (t0, (zip t'))))
     | _ -> raise (Obj.magic monad_exc) (NotAProduct (t0, (zip t')))))
| Coq_tSort _ ->
  bind (Obj.magic typing_monad) (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0)
    (fun t' ->
    let (y, y0) = t' in
    (match y with
     | Coq_tProd (_, a, b) ->
       (match y0 with
        | [] -> ret (Obj.magic typing_monad) (a, b)
        | _ :: _ -> raise (Obj.magic monad_exc) (NotAProduct (t0, (zip t'))))
     | _ -> raise (Obj.magic monad_exc) (NotAProduct (t0, (zip t')))))
| Coq_tCast (_, _, _) ->
  bind (Obj.magic typing_monad) (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0)
    (fun t' ->
    let (y, y0) = t' in
    (match y with
     | Coq_tProd (_, a, b) ->
       (match y0 with
        | [] -> ret (Obj.magic typing_monad) (a, b)
        | _ :: _ -> raise (Obj.magic monad_exc) (NotAProduct (t0, (zip t'))))
     | _ -> raise (Obj.magic monad_exc) (NotAProduct (t0, (zip t')))))
| Coq_tProd (_, a, b) -> ret (Obj.magic typing_monad) (a, b)
| Coq_tLambda (_, _, _) ->
  bind (Obj.magic typing_monad) (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0)
    (fun t' ->
    let (y, y0) = t' in
    (match y with
     | Coq_tProd (_, a, b) ->
       (match y0 with
        | [] -> ret (Obj.magic typing_monad) (a, b)
        | _ :: _ -> raise (Obj.magic monad_exc) (NotAProduct (t0, (zip t'))))
     | _ -> raise (Obj.magic monad_exc) (NotAProduct (t0, (zip t')))))
| Coq_tLetIn (_, _, _, _) ->
  bind (Obj.magic typing_monad) (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0)
    (fun t' ->
    let (y, y0) = t' in
    (match y with
     | Coq_tProd (_, a, b) ->
       (match y0 with
        | [] -> ret (Obj.magic typing_monad) (a, b)
        | _ :: _ -> raise (Obj.magic monad_exc) (NotAProduct (t0, (zip t'))))
     | _ -> raise (Obj.magic monad_exc) (NotAProduct (t0, (zip t')))))
| Coq_tApp (_, _) ->
  bind (Obj.magic typing_monad) (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0)
    (fun t' ->
    let (y, y0) = t' in
    (match y with
     | Coq_tProd (_, a, b) ->
       (match y0 with
        | [] -> ret (Obj.magic typing_monad) (a, b)
        | _ :: _ -> raise (Obj.magic monad_exc) (NotAProduct (t0, (zip t'))))
     | _ -> raise (Obj.magic monad_exc) (NotAProduct (t0, (zip t')))))
| Coq_tConst (_, _) ->
  bind (Obj.magic typing_monad) (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0)
    (fun t' ->
    let (y, y0) = t' in
    (match y with
     | Coq_tProd (_, a, b) ->
       (match y0 with
        | [] -> ret (Obj.magic typing_monad) (a, b)
        | _ :: _ -> raise (Obj.magic monad_exc) (NotAProduct (t0, (zip t'))))
     | _ -> raise (Obj.magic monad_exc) (NotAProduct (t0, (zip t')))))
| Coq_tInd (_, _) ->
  bind (Obj.magic typing_monad) (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0)
    (fun t' ->
    let (y, y0) = t' in
    (match y with
     | Coq_tProd (_, a, b) ->
       (match y0 with
        | [] -> ret (Obj.magic typing_monad) (a, b)
        | _ :: _ -> raise (Obj.magic monad_exc) (NotAProduct (t0, (zip t'))))
     | _ -> raise (Obj.magic monad_exc) (NotAProduct (t0, (zip t')))))
| Coq_tConstruct (_, _, _) ->
  bind (Obj.magic typing_monad) (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0)
    (fun t' ->
    let (y, y0) = t' in
    (match y with
     | Coq_tProd (_, a, b) ->
       (match y0 with
        | [] -> ret (Obj.magic typing_monad) (a, b)
        | _ :: _ -> raise (Obj.magic monad_exc) (NotAProduct (t0, (zip t'))))
     | _ -> raise (Obj.magic monad_exc) (NotAProduct (t0, (zip t')))))
| Coq_tCase (_, _, _, _) ->
  bind (Obj.magic typing_monad) (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0)
    (fun t' ->
    let (y, y0) = t' in
    (match y with
     | Coq_tProd (_, a, b) ->
       (match y0 with
        | [] -> ret (Obj.magic typing_monad) (a, b)
        | _ :: _ -> raise (Obj.magic monad_exc) (NotAProduct (t0, (zip t'))))
     | _ -> raise (Obj.magic monad_exc) (NotAProduct (t0, (zip t')))))
| Coq_tProj (_, _) ->
  bind (Obj.magic typing_monad) (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0)
    (fun t' ->
    let (y, y0) = t' in
    (match y with
     | Coq_tProd (_, a, b) ->
       (match y0 with
        | [] -> ret (Obj.magic typing_monad) (a, b)
        | _ :: _ -> raise (Obj.magic monad_exc) (NotAProduct (t0, (zip t'))))
     | _ -> raise (Obj.magic monad_exc) (NotAProduct (t0, (zip t')))))
| Coq_tFix (_, _) ->
  bind (Obj.magic typing_monad) (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0)
    (fun t' ->
    let (y, y0) = t' in
    (match y with
     | Coq_tProd (_, a, b) ->
       (match y0 with
        | [] -> ret (Obj.magic typing_monad) (a, b)
        | _ :: _ -> raise (Obj.magic monad_exc) (NotAProduct (t0, (zip t'))))
     | _ -> raise (Obj.magic monad_exc) (NotAProduct (t0, (zip t')))))
| Coq_tCoFix (_, _) ->
  bind (Obj.magic typing_monad) (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0)
    (fun t' ->
    let (y, y0) = t' in
    (match y with
     | Coq_tProd (_, a, b) ->
       (match y0 with
        | [] -> ret (Obj.magic typing_monad) (a, b)
        | _ :: _ -> raise (Obj.magic monad_exc) (NotAProduct (t0, (zip t'))))
     | _ -> raise (Obj.magic monad_exc) (NotAProduct (t0, (zip t')))))

(** val reduce_to_ind :
    coq_Fuel -> global_env -> context -> term -> ((inductive * Level.t
    list) * term list) typing_result **)

let reduce_to_ind f _UU03a3_ _UU0393_ t0 =
  let (t1, l) = decompose_app t0 in
  (match t1 with
   | Coq_tRel _ ->
     bind (Obj.magic typing_monad)
       (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0) (fun t' ->
       let (y, l0) = t' in
       (match y with
        | Coq_tInd (i, u) -> ret (Obj.magic typing_monad) ((i, u), l0)
        | _ -> raise (Obj.magic monad_exc) (NotAnInductive t0)))
   | Coq_tVar _ ->
     bind (Obj.magic typing_monad)
       (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0) (fun t' ->
       let (y, l0) = t' in
       (match y with
        | Coq_tInd (i, u) -> ret (Obj.magic typing_monad) ((i, u), l0)
        | _ -> raise (Obj.magic monad_exc) (NotAnInductive t0)))
   | Coq_tEvar (_, _) ->
     bind (Obj.magic typing_monad)
       (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0) (fun t' ->
       let (y, l0) = t' in
       (match y with
        | Coq_tInd (i, u) -> ret (Obj.magic typing_monad) ((i, u), l0)
        | _ -> raise (Obj.magic monad_exc) (NotAnInductive t0)))
   | Coq_tSort _ ->
     bind (Obj.magic typing_monad)
       (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0) (fun t' ->
       let (y, l0) = t' in
       (match y with
        | Coq_tInd (i, u) -> ret (Obj.magic typing_monad) ((i, u), l0)
        | _ -> raise (Obj.magic monad_exc) (NotAnInductive t0)))
   | Coq_tCast (_, _, _) ->
     bind (Obj.magic typing_monad)
       (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0) (fun t' ->
       let (y, l0) = t' in
       (match y with
        | Coq_tInd (i, u) -> ret (Obj.magic typing_monad) ((i, u), l0)
        | _ -> raise (Obj.magic monad_exc) (NotAnInductive t0)))
   | Coq_tProd (_, _, _) ->
     bind (Obj.magic typing_monad)
       (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0) (fun t' ->
       let (y, l0) = t' in
       (match y with
        | Coq_tInd (i, u) -> ret (Obj.magic typing_monad) ((i, u), l0)
        | _ -> raise (Obj.magic monad_exc) (NotAnInductive t0)))
   | Coq_tLambda (_, _, _) ->
     bind (Obj.magic typing_monad)
       (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0) (fun t' ->
       let (y, l0) = t' in
       (match y with
        | Coq_tInd (i, u) -> ret (Obj.magic typing_monad) ((i, u), l0)
        | _ -> raise (Obj.magic monad_exc) (NotAnInductive t0)))
   | Coq_tLetIn (_, _, _, _) ->
     bind (Obj.magic typing_monad)
       (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0) (fun t' ->
       let (y, l0) = t' in
       (match y with
        | Coq_tInd (i, u) -> ret (Obj.magic typing_monad) ((i, u), l0)
        | _ -> raise (Obj.magic monad_exc) (NotAnInductive t0)))
   | Coq_tApp (_, _) ->
     bind (Obj.magic typing_monad)
       (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0) (fun t' ->
       let (y, l0) = t' in
       (match y with
        | Coq_tInd (i, u) -> ret (Obj.magic typing_monad) ((i, u), l0)
        | _ -> raise (Obj.magic monad_exc) (NotAnInductive t0)))
   | Coq_tConst (_, _) ->
     bind (Obj.magic typing_monad)
       (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0) (fun t' ->
       let (y, l0) = t' in
       (match y with
        | Coq_tInd (i, u) -> ret (Obj.magic typing_monad) ((i, u), l0)
        | _ -> raise (Obj.magic monad_exc) (NotAnInductive t0)))
   | Coq_tInd (i, u) -> ret (Obj.magic typing_monad) ((i, u), l)
   | Coq_tConstruct (_, _, _) ->
     bind (Obj.magic typing_monad)
       (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0) (fun t' ->
       let (y, l0) = t' in
       (match y with
        | Coq_tInd (i, u) -> ret (Obj.magic typing_monad) ((i, u), l0)
        | _ -> raise (Obj.magic monad_exc) (NotAnInductive t0)))
   | Coq_tCase (_, _, _, _) ->
     bind (Obj.magic typing_monad)
       (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0) (fun t' ->
       let (y, l0) = t' in
       (match y with
        | Coq_tInd (i, u) -> ret (Obj.magic typing_monad) ((i, u), l0)
        | _ -> raise (Obj.magic monad_exc) (NotAnInductive t0)))
   | Coq_tProj (_, _) ->
     bind (Obj.magic typing_monad)
       (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0) (fun t' ->
       let (y, l0) = t' in
       (match y with
        | Coq_tInd (i, u) -> ret (Obj.magic typing_monad) ((i, u), l0)
        | _ -> raise (Obj.magic monad_exc) (NotAnInductive t0)))
   | Coq_tFix (_, _) ->
     bind (Obj.magic typing_monad)
       (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0) (fun t' ->
       let (y, l0) = t' in
       (match y with
        | Coq_tInd (i, u) -> ret (Obj.magic typing_monad) ((i, u), l0)
        | _ -> raise (Obj.magic monad_exc) (NotAnInductive t0)))
   | Coq_tCoFix (_, _) ->
     bind (Obj.magic typing_monad)
       (Obj.magic hnf_stack f _UU03a3_ _UU0393_ t0) (fun t' ->
       let (y, l0) = t' in
       (match y with
        | Coq_tInd (i, u) -> ret (Obj.magic typing_monad) ((i, u), l0)
        | _ -> raise (Obj.magic monad_exc) (NotAnInductive t0))))

(** val convert_leq :
    checker_flags -> coq_Fuel -> global_env -> universes_graph -> context ->
    term -> term -> unit typing_result **)

let convert_leq cf f _UU03a3_ g _UU0393_ t0 u =
  if eq_term cf g t0 u
  then ret (Obj.magic typing_monad) ()
  else (match isconv cf _UU03a3_ g (fuel f) Cumul _UU0393_ t0 [] u [] with
        | Some b ->
          if b
          then ret (Obj.magic typing_monad) ()
          else raise (Obj.magic monad_exc) (NotConvertible (_UU0393_, t0, u,
                 t0, u))
        | None ->
          bind (Obj.magic typing_monad)
            (Obj.magic reduce f _UU03a3_ _UU0393_ t0) (fun t' ->
            bind (Obj.magic typing_monad)
              (Obj.magic reduce f _UU03a3_ _UU0393_ u) (fun u' ->
              if leq_term cf g t' u'
              then ret (Obj.magic typing_monad) ()
              else raise (Obj.magic monad_exc) (NotConvertible (_UU0393_, t0,
                     u, t', u')))))

(** val infer_spine :
    checker_flags -> coq_Fuel -> global_env -> universes_graph -> (context ->
    term -> term typing_result) -> context -> term -> term list -> term
    typing_result **)

let rec infer_spine cf f _UU03a3_ g infer0 _UU0393_ ty = function
| [] -> ret (Obj.magic typing_monad) ty
| x :: xs ->
  bind (Obj.magic typing_monad)
    (Obj.magic reduce_to_prod f _UU03a3_ _UU0393_ ty) (fun pi ->
    let (a1, b1) = pi in
    bind (Obj.magic typing_monad) (infer0 _UU0393_ x) (fun tx ->
      bind (Obj.magic typing_monad)
        (Obj.magic convert_leq cf f _UU03a3_ g _UU0393_ tx a1) (fun _ ->
        infer_spine cf f _UU03a3_ g infer0 _UU0393_ (subst1 x O b1) xs)))

(** val infer_type :
    coq_Fuel -> global_env -> (context -> term -> term typing_result) ->
    context -> term -> universe typing_result **)

let infer_type f _UU03a3_ infer0 _UU0393_ t0 =
  bind (Obj.magic typing_monad) (Obj.magic infer0 _UU0393_ t0) (fun tx ->
    reduce_to_sort f _UU03a3_ _UU0393_ tx)

(** val infer_cumul :
    checker_flags -> coq_Fuel -> global_env -> universes_graph -> (context ->
    term -> term typing_result) -> context -> term -> term -> unit
    typing_result **)

let infer_cumul cf f _UU03a3_ g infer0 _UU0393_ t0 t' =
  bind (Obj.magic typing_monad) (Obj.magic infer0 _UU0393_ t0) (fun tx ->
    convert_leq cf f _UU03a3_ g _UU0393_ tx t')

(** val polymorphic_constraints : universes_decl -> ConstraintSet.t **)

let polymorphic_constraints = function
| Monomorphic_ctx _ -> ConstraintSet.empty
| Polymorphic_ctx ctx -> snd (AUContext.repr ctx)
| Cumulative_ctx ctx0 -> let (ctx, _) = ctx0 in snd (AUContext.repr ctx)

(** val lookup_constant_type :
    global_env -> ident -> universe_instance -> term typing_result **)

let lookup_constant_type _UU03a3_ cst u =
  match lookup_env _UU03a3_ cst with
  | Some g ->
    (match g with
     | ConstantDecl (_, c) ->
       let { cst_type = ty; cst_body = _; cst_universes = _ } = c in
       ret (Obj.magic typing_monad) (subst_instance_constr u ty)
     | InductiveDecl (_, _) ->
       raise (Obj.magic monad_exc) (UndeclaredConstant cst))
  | None -> raise (Obj.magic monad_exc) (UndeclaredConstant cst)

(** val lookup_constant_type_cstrs :
    global_env -> ident -> universe_instance -> (term * constraints)
    typing_result **)

let lookup_constant_type_cstrs _UU03a3_ cst u =
  match lookup_env _UU03a3_ cst with
  | Some g ->
    (match g with
     | ConstantDecl (_, c) ->
       let { cst_type = ty; cst_body = _; cst_universes = uctx } = c in
       let cstrs = polymorphic_constraints uctx in
       ret (Obj.magic typing_monad) ((subst_instance_constr u ty),
         (subst_instance_cstrs u cstrs))
     | InductiveDecl (_, _) ->
       raise (Obj.magic monad_exc) (UndeclaredConstant cst))
  | None -> raise (Obj.magic monad_exc) (UndeclaredConstant cst)

(** val lookup_ind_decl :
    global_env -> ident -> nat -> ((one_inductive_body
    list * universes_decl) * one_inductive_body) typing_result **)

let lookup_ind_decl _UU03a3_ ind i =
  match lookup_env _UU03a3_ ind with
  | Some g ->
    (match g with
     | ConstantDecl (_, _) ->
       raise (Obj.magic monad_exc) (UndeclaredInductive { inductive_mind =
         ind; inductive_ind = i })
     | InductiveDecl (_, m) ->
       let { ind_finite = _; ind_npars = _; ind_params = _; ind_bodies = l;
         ind_universes = uctx } = m
       in
       (match nth_error l i with
        | Some body -> ret (Obj.magic typing_monad) ((l, uctx), body)
        | None ->
          raise (Obj.magic monad_exc) (UndeclaredInductive { inductive_mind =
            ind; inductive_ind = i })))
  | None ->
    raise (Obj.magic monad_exc) (UndeclaredInductive { inductive_mind = ind;
      inductive_ind = i })

(** val lookup_ind_type :
    global_env -> ident -> nat -> Level.t list -> term typing_result **)

let lookup_ind_type _UU03a3_ ind i u =
  bind (Obj.magic typing_monad) (Obj.magic lookup_ind_decl _UU03a3_ ind i)
    (fun res ->
    ret (Obj.magic typing_monad) (subst_instance_constr u (snd res).ind_type))

(** val lookup_ind_type_cstrs :
    global_env -> ident -> nat -> Level.t list -> (term * constraints)
    typing_result **)

let lookup_ind_type_cstrs _UU03a3_ ind i u =
  bind (Obj.magic typing_monad) (Obj.magic lookup_ind_decl _UU03a3_ ind i)
    (fun res ->
    let (y, body) = res in
    let (_, uctx) = y in
    let cstrs = polymorphic_constraints uctx in
    ret (Obj.magic typing_monad) ((subst_instance_constr u body.ind_type),
      (subst_instance_cstrs u cstrs)))

(** val lookup_constructor_decl :
    global_env -> ident -> nat -> nat -> ((one_inductive_body
    list * universes_decl) * term) typing_result **)

let lookup_constructor_decl _UU03a3_ ind i k =
  bind (Obj.magic typing_monad) (Obj.magic lookup_ind_decl _UU03a3_ ind i)
    (fun res ->
    let (y, body) = res in
    let (l, uctx) = y in
    (match nth_error body.ind_ctors k with
     | Some p ->
       let (p0, _) = p in
       let (_, ty) = p0 in ret (Obj.magic typing_monad) ((l, uctx), ty)
     | None ->
       raise (Obj.magic monad_exc) (UndeclaredConstructor ({ inductive_mind =
         ind; inductive_ind = i }, k))))

(** val lookup_constructor_type :
    global_env -> ident -> nat -> nat -> universe_instance -> term
    typing_result **)

let lookup_constructor_type _UU03a3_ ind i k u =
  bind (Obj.magic typing_monad)
    (Obj.magic lookup_constructor_decl _UU03a3_ ind i k) (fun res ->
    let (y, ty) = res in
    let (l, _) = y in
    ret (Obj.magic typing_monad)
      (subst (inds ind u l) O (subst_instance_constr u ty)))

(** val lookup_constructor_type_cstrs :
    global_env -> ident -> nat -> nat -> universe_instance ->
    (term * constraints) typing_result **)

let lookup_constructor_type_cstrs _UU03a3_ ind i k u =
  bind (Obj.magic typing_monad)
    (Obj.magic lookup_constructor_decl _UU03a3_ ind i k) (fun res ->
    let (y, ty) = res in
    let (l, uctx) = y in
    let cstrs = polymorphic_constraints uctx in
    ret (Obj.magic typing_monad)
      ((subst (inds ind u l) O (subst_instance_constr u ty)),
      (subst_instance_cstrs u cstrs)))

(** val check_consistent_constraints :
    checker_flags -> universes_graph -> constraints -> unit typing_result **)

let check_consistent_constraints cf g cstrs =
  if check_constraints cf g cstrs
  then ret (Obj.magic typing_monad) ()
  else raise (Obj.magic monad_exc) (UnsatisfiedConstraints cstrs)

(** val infer :
    checker_flags -> coq_Fuel -> global_env -> universes_graph -> context ->
    term -> term typing_result **)

let rec infer cf f _UU03a3_ g _UU0393_ = function
| Coq_tRel n ->
  (match nth_error _UU0393_ n with
   | Some d -> ret (Obj.magic typing_monad) (lift (S n) O d.decl_type)
   | None -> raise (Obj.magic monad_exc) (UnboundRel n))
| Coq_tVar n -> raise (Obj.magic monad_exc) (UnboundVar n)
| Coq_tEvar (ev, _) -> raise (Obj.magic monad_exc) (UnboundEvar ev)
| Coq_tSort s -> ret (Obj.magic typing_monad) (Coq_tSort (Universe.try_suc s))
| Coq_tCast (c, _, t1) ->
  bind (Obj.magic typing_monad)
    (Obj.magic infer_type f _UU03a3_ (infer cf f _UU03a3_ g) _UU0393_ t1)
    (fun _ ->
    bind (Obj.magic typing_monad)
      (Obj.magic infer_cumul cf f _UU03a3_ g (infer cf f _UU03a3_ g) _UU0393_
        c t1) (fun _ -> ret (Obj.magic typing_monad) t1))
| Coq_tProd (n, t1, b) ->
  bind (Obj.magic typing_monad)
    (Obj.magic infer_type f _UU03a3_ (infer cf f _UU03a3_ g) _UU0393_ t1)
    (fun s1 ->
    bind (Obj.magic typing_monad)
      (Obj.magic infer_type f _UU03a3_ (infer cf f _UU03a3_ g)
        (snoc _UU0393_ (vass n t1)) b) (fun s2 ->
      ret (Obj.magic typing_monad) (Coq_tSort
        (Universe.sort_of_product s1 s2))))
| Coq_tLambda (n, t1, b) ->
  bind (Obj.magic typing_monad)
    (Obj.magic infer_type f _UU03a3_ (infer cf f _UU03a3_ g) _UU0393_ t1)
    (fun _ ->
    bind (Obj.magic typing_monad)
      (infer cf f _UU03a3_ g (snoc _UU0393_ (vass n t1)) b) (fun t2 ->
      ret (Obj.magic typing_monad) (Coq_tProd (n, t1, t2))))
| Coq_tLetIn (n, b, b_ty, b') ->
  bind (Obj.magic typing_monad)
    (Obj.magic infer_type f _UU03a3_ (infer cf f _UU03a3_ g) _UU0393_ b_ty)
    (fun _ ->
    bind (Obj.magic typing_monad)
      (Obj.magic infer_cumul cf f _UU03a3_ g (infer cf f _UU03a3_ g) _UU0393_
        b b_ty) (fun _ ->
      bind (Obj.magic typing_monad)
        (infer cf f _UU03a3_ g (snoc _UU0393_ (vdef n b b_ty)) b')
        (fun b'_ty ->
        ret (Obj.magic typing_monad) (Coq_tLetIn (n, b, b_ty, b'_ty)))))
| Coq_tApp (t1, l) ->
  bind (Obj.magic typing_monad) (infer cf f _UU03a3_ g _UU0393_ t1)
    (fun t_ty ->
    infer_spine cf f _UU03a3_ g (infer cf f _UU03a3_ g) _UU0393_ t_ty l)
| Coq_tConst (cst, u) ->
  bind (Obj.magic typing_monad)
    (Obj.magic lookup_constant_type_cstrs _UU03a3_ cst u) (fun tycstrs ->
    let (ty, cstrs) = tycstrs in
    bind (Obj.magic typing_monad)
      (Obj.magic check_consistent_constraints cf g cstrs) (fun _ ->
      ret (Obj.magic typing_monad) ty))
| Coq_tInd (ind0, u) ->
  let { inductive_mind = ind; inductive_ind = i } = ind0 in
  bind (Obj.magic typing_monad)
    (Obj.magic lookup_ind_type_cstrs _UU03a3_ ind i u) (fun tycstrs ->
    let (ty, cstrs) = tycstrs in
    bind (Obj.magic typing_monad)
      (Obj.magic check_consistent_constraints cf g cstrs) (fun _ ->
      ret (Obj.magic typing_monad) ty))
| Coq_tConstruct (ind0, k, u) ->
  let { inductive_mind = ind; inductive_ind = i } = ind0 in
  bind (Obj.magic typing_monad)
    (Obj.magic lookup_constructor_type_cstrs _UU03a3_ ind i k u)
    (fun tycstrs ->
    let (ty, cstrs) = tycstrs in
    bind (Obj.magic typing_monad)
      (Obj.magic check_consistent_constraints cf g cstrs) (fun _ ->
      ret (Obj.magic typing_monad) ty))
| Coq_tCase (ind_and_nbparams, p, c, _) ->
  let (ind, par) = ind_and_nbparams in
  bind (Obj.magic typing_monad) (infer cf f _UU03a3_ g _UU0393_ c) (fun ty ->
    bind (Obj.magic typing_monad)
      (Obj.magic reduce_to_ind f _UU03a3_ _UU0393_ ty) (fun indargs ->
      let (y, args) = indargs in
      let (ind', u) = y in
      if eq_ind ind ind'
      then ret (Obj.magic typing_monad) (Coq_tApp (p,
             (app (skipn par args) (c :: []))))
      else let ind1 = Coq_tInd (ind, u) in
           let ind2 = Coq_tInd (ind', u) in
           raise (Obj.magic monad_exc) (NotConvertible (_UU0393_, ind1, ind2,
             ind1, ind2))))
| Coq_tProj (_, c) ->
  bind (Obj.magic typing_monad) (infer cf f _UU03a3_ g _UU0393_ c) (fun ty ->
    bind (Obj.magic typing_monad)
      (Obj.magic reduce_to_ind f _UU03a3_ _UU0393_ ty) (fun _ ->
      ret (Obj.magic typing_monad) ty))
| Coq_tFix (mfix, n) ->
  (match nth_error mfix n with
   | Some f0 -> ret (Obj.magic typing_monad) f0.dtype
   | None -> raise (Obj.magic monad_exc) (IllFormedFix (mfix, n)))
| Coq_tCoFix (mfix, n) ->
  (match nth_error mfix n with
   | Some f0 -> ret (Obj.magic typing_monad) f0.dtype
   | None -> raise (Obj.magic monad_exc) (IllFormedFix (mfix, n)))

(** val check :
    checker_flags -> coq_Fuel -> global_env -> universes_graph -> context ->
    term -> term -> unit typing_result **)

let check cf f _UU03a3_ g _UU0393_ t0 ty =
  bind (Obj.magic typing_monad) (Obj.magic infer cf f _UU03a3_ g _UU0393_ ty)
    (fun _ ->
    bind (Obj.magic typing_monad)
      (infer_cumul cf f _UU03a3_ g (infer cf f _UU03a3_ g) _UU0393_ t0 ty)
      (fun _ -> ret (Obj.magic typing_monad) ()))

(** val typechecking :
    checker_flags -> coq_Fuel -> global_env -> universes_graph -> context ->
    term -> term -> bool **)

let typechecking cf f _UU03a3_ g _UU0393_ t0 ty =
  match check cf f _UU03a3_ g _UU0393_ t0 ty with
  | Checked _ -> true
  | TypeError _ -> false

(** val cumul_convert_leq :
    checker_flags -> coq_Fuel -> global_env_ext -> Coq_wGraph.t -> context ->
    term -> term -> (cumul, __) isEquiv **)

let cumul_convert_leq _ _ _ _ _ _ _ =
  todo
    ('C'::('h'::('e'::('c'::('k'::('e'::('r'::('.'::('c'::('u'::('m'::('u'::('l'::('_'::('c'::('o'::('n'::('v'::('e'::('r'::('t'::('_'::('l'::('e'::('q'::[])))))))))))))))))))))))))

(** val cumul_reduce_to_sort :
    checker_flags -> coq_Fuel -> global_env_ext -> Coq_wGraph.t -> context ->
    term -> universe -> (cumul, (universe, __) sigT) isEquiv **)

let cumul_reduce_to_sort _ _ _ _ _ _ _ =
  todo
    ('C'::('h'::('e'::('c'::('k'::('e'::('r'::('.'::('c'::('u'::('m'::('u'::('l'::('_'::('r'::('e'::('d'::('u'::('c'::('e'::('_'::('t'::('o'::('_'::('s'::('o'::('r'::('t'::[]))))))))))))))))))))))))))))

(** val cumul_reduce_to_product :
    checker_flags -> coq_Fuel -> global_env_ext -> context -> term -> name ->
    term -> term -> cumul -> (term, (term, __ * cumul) sigT) sigT **)

let cumul_reduce_to_product _ _ _ _ _ _ _ _ _ =
  todo
    ('C'::('h'::('e'::('c'::('k'::('e'::('r'::('.'::('c'::('u'::('m'::('u'::('l'::('_'::('r'::('e'::('d'::('u'::('c'::('e'::('_'::('t'::('o'::('_'::('p'::('r'::('o'::('d'::('u'::('c'::('t'::[])))))))))))))))))))))))))))))))

(** val cumul_reduce_to_ind :
    checker_flags -> coq_Fuel -> global_env_ext -> context -> term ->
    inductive -> universe_instance -> term list -> (cumul, (term list,
    __ * cumul) sigT) isEquiv **)

let cumul_reduce_to_ind _ _ _ _ _ _ _ _ =
  todo
    ('C'::('h'::('e'::('c'::('k'::('e'::('r'::('.'::('c'::('u'::('m'::('u'::('l'::('_'::('r'::('e'::('d'::('u'::('c'::('e'::('_'::('t'::('o'::('_'::('i'::('n'::('d'::[])))))))))))))))))))))))))))

(** val lookup_constant_type_is_declared :
    global_env -> ident -> universe_instance -> (term * constraints) ->
    constant_body **)

let lookup_constant_type_is_declared _UU03a3_ cst _ _ =
  let o = Typing0.lookup_env _UU03a3_ cst in
  (match o with
   | Some g ->
     (match g with
      | ConstantDecl (_, c) -> c
      | InductiveDecl (_, _) -> assert false (* absurd case *))
   | None -> assert false (* absurd case *))

(** val infer_type_correct :
    checker_flags -> coq_Fuel -> global_env_ext -> universes_graph -> context
    -> term -> universe -> (context -> term -> __ -> typing) -> typing **)

let infer_type_correct = (fun f sigma ctx t x -> assert false)

(** val infer_cumul_correct :
    checker_flags -> coq_Fuel -> global_env_ext -> universes_graph -> context
    -> term -> term -> unit -> universe -> (context -> term -> __ -> typing)
    -> (context -> term -> __ -> typing) -> typing **)

let infer_cumul_correct cf f _UU03a3_ g _UU0393_ t0 u _ x' iH iH' =
  let t1 = infer cf f (fst _UU03a3_) g _UU0393_ t0 in
  (match t1 with
   | Checked a ->
     let t2 = convert_leq cf f (fst _UU03a3_) g _UU0393_ a u in
     (match t2 with
      | Checked _ ->
        let iH'0 = iH' _UU0393_ a __ in
        Coq_type_Conv (t0, a, u, iH'0, (Coq_inr
        (let h = infer_type_correct cf f _UU03a3_ g _UU0393_ u x' iH in
         Coq_existT (x', h))),
        ((cumul_convert_leq cf f _UU03a3_ g _UU0393_ a u).equiv_inv __))
      | TypeError _ -> assert false (* absurd case *))
   | TypeError _ -> assert false (* absurd case *))

(** val infer_correct :
    checker_flags -> coq_Fuel -> global_env_ext -> universes_graph -> context
    -> term -> term -> typing lift_typing coq_All_local_env -> typing **)

let infer_correct = (fun f sigma ctx t ty -> assert false)

(** val default_fuel : coq_Fuel **)

let default_fuel =
  Nat.pow (S (S O)) (S (S (S (S (S (S (S (S (S (S (S (S (S (S O))))))))))))))

(** val fresh : char list -> global_decl list -> bool **)

let rec fresh id = function
| [] -> true
| g :: env0 ->
  (&&) (negb (eq_constant (global_decl_ident g) id)) (fresh id env0)

type env_error =
| IllFormedDecl of char list * type_error
| AlreadyDeclared of char list

(** val env_error_rect :
    (char list -> type_error -> 'a1) -> (char list -> 'a1) -> env_error -> 'a1 **)

let env_error_rect f f0 = function
| IllFormedDecl (x, x0) -> f x x0
| AlreadyDeclared x -> f0 x

(** val env_error_rec :
    (char list -> type_error -> 'a1) -> (char list -> 'a1) -> env_error -> 'a1 **)

let env_error_rec f f0 = function
| IllFormedDecl (x, x0) -> f x x0
| AlreadyDeclared x -> f0 x

type 'a coq_EnvCheck =
| CorrectDecl of 'a
| EnvError of env_error

(** val coq_EnvCheck_rect :
    ('a1 -> 'a2) -> (env_error -> 'a2) -> 'a1 coq_EnvCheck -> 'a2 **)

let coq_EnvCheck_rect f f0 = function
| CorrectDecl x -> f x
| EnvError x -> f0 x

(** val coq_EnvCheck_rec :
    ('a1 -> 'a2) -> (env_error -> 'a2) -> 'a1 coq_EnvCheck -> 'a2 **)

let coq_EnvCheck_rec f f0 = function
| CorrectDecl x -> f x
| EnvError x -> f0 x

(** val envcheck_monad : __ coq_EnvCheck coq_Monad **)

let envcheck_monad =
  { ret = (fun _ a -> CorrectDecl a); bind = (fun _ _ m f ->
    match m with
    | CorrectDecl a -> f a
    | EnvError e -> EnvError e) }

(** val wrap_error : char list -> 'a1 typing_result -> 'a1 coq_EnvCheck **)

let wrap_error id = function
| Checked a -> CorrectDecl a
| TypeError e -> EnvError (IllFormedDecl (id, e))

(** val check_wf_type :
    checker_flags -> coq_Fuel -> char list -> global_env -> universes_graph
    -> term -> unit coq_EnvCheck **)

let check_wf_type cf f id _UU03a3_ g t0 =
  bind (Obj.magic envcheck_monad)
    (wrap_error id
      (Obj.magic infer_type f _UU03a3_ (infer cf f _UU03a3_ g) [] t0))
    (fun _ -> ret (Obj.magic envcheck_monad) ())

(** val check_wf_judgement :
    checker_flags -> coq_Fuel -> char list -> global_env -> universes_graph
    -> term -> term -> unit coq_EnvCheck **)

let check_wf_judgement cf f id _UU03a3_ g t0 ty =
  bind (Obj.magic envcheck_monad)
    (wrap_error id (check cf f _UU03a3_ g [] t0 ty)) (fun _ ->
    ret (Obj.magic envcheck_monad) ())

(** val infer_term :
    checker_flags -> coq_Fuel -> global_env -> universes_graph -> term ->
    term coq_EnvCheck **)

let infer_term cf f _UU03a3_ g t0 =
  wrap_error [] (infer cf f _UU03a3_ g [] t0)

(** val check_wf_decl :
    checker_flags -> coq_Fuel -> global_env -> universes_graph -> global_decl
    -> unit coq_EnvCheck **)

let check_wf_decl cf f _UU03a3_ g = function
| ConstantDecl (id, cst) ->
  (match cst.cst_body with
   | Some term0 -> check_wf_judgement cf f id _UU03a3_ g term0 cst.cst_type
   | None -> check_wf_type cf f id _UU03a3_ g cst.cst_type)
| InductiveDecl (_, inds0) ->
  fold_left (fun acc body ->
    bind (Obj.magic envcheck_monad) acc (fun _ ->
      check_wf_type cf f body.ind_name _UU03a3_ g body.ind_type))
    inds0.ind_bodies (ret (Obj.magic envcheck_monad) ())

(** val check_fresh : char list -> global_decl list -> unit coq_EnvCheck **)

let rec check_fresh id = function
| [] -> ret (Obj.magic envcheck_monad) ()
| g :: env0 ->
  bind (Obj.magic envcheck_monad) (check_fresh id env0) (fun _ ->
    if eq_constant id (global_decl_ident g)
    then EnvError (AlreadyDeclared id)
    else ret (Obj.magic envcheck_monad) ())

(** val monomorphic_constraints : universes_decl -> constraints **)

let monomorphic_constraints = function
| Monomorphic_ctx ctx -> snd ctx
| _ -> ConstraintSet.empty

(** val global_decl_univs : global_decl -> constraints **)

let global_decl_univs = function
| ConstantDecl (_, cb) -> monomorphic_constraints cb.cst_universes
| InductiveDecl (_, mb) -> monomorphic_constraints mb.ind_universes

(** val add_gc_constraints :
    GoodConstraintSet.t -> universes_graph -> universes_graph **)

let add_gc_constraints ctrs g =
  (((fst (fst g)),
    (GoodConstraintSet.fold (fun ctr ->
      Coq_wGraph.EdgeSet.add (edge_of_constraint ctr)) ctrs (snd (fst g)))),
    (snd g))

(** val check_wf_env :
    checker_flags -> coq_Fuel -> global_env -> universes_graph coq_EnvCheck **)

let rec check_wf_env cf f = function
| [] -> ret (Obj.magic envcheck_monad) init_graph
| g0 :: env ->
  bind (Obj.magic envcheck_monad) (check_wf_env cf f env) (fun g1 ->
    match gc_of_constraints cf (global_decl_univs g0) with
    | Some ctrs ->
      bind (Obj.magic envcheck_monad)
        (wrap_error []
          (Obj.magic check_consistent_constraints cf g1
            (global_decl_univs g0))) (fun _ ->
        let g' = add_gc_constraints ctrs g1 in
        bind (Obj.magic envcheck_monad)
          (Obj.magic check_wf_decl cf f env g' g0) (fun _ ->
          bind (Obj.magic envcheck_monad)
            (Obj.magic check_fresh (global_decl_ident g0) env) (fun _ ->
            ret (Obj.magic envcheck_monad) g')))
    | None ->
      EnvError (IllFormedDecl ((global_decl_ident g0),
        (UnsatisfiableConstraints (global_decl_univs g0)))))

(** val typecheck_program :
    checker_flags -> coq_Fuel -> program -> term coq_EnvCheck **)

let typecheck_program cf f p =
  let _UU03a3_ = List0.rev (fst p) in
  bind (Obj.magic envcheck_monad) (Obj.magic check_wf_env cf f _UU03a3_)
    (fun g -> infer_term cf f _UU03a3_ g (snd p))

(** val infer' :
    checker_flags -> coq_Fuel -> global_env_ext -> context -> term -> term
    typing_result **)

let infer' h h0 _UU03a3_ _UU0393_ t0 =
  let uctx = global_ext_uctx _UU03a3_ in
  (match gc_of_uctx h uctx with
   | Some uctx0 -> infer h h0 (fst _UU03a3_) (make_graph uctx0) _UU0393_ t0
   | None -> raise (Obj.magic monad_exc) (UnsatisfiableConstraints (snd uctx)))
