open Ast0
open AstUtils
open BasicAst
open Datatypes
open LiftSubst
open List0
open PeanoNat
open Specif
open String0
open Typing0
open UnivSubst0
open Universes0
open Config0
open Monad_utils
open UGraph0
open Utils

type __ = Obj.t

module RedFlags :
 sig
  type t = { beta : bool; iota : bool; zeta : bool; delta : bool;
             fix_ : bool; cofix_ : bool }

  val beta : t -> bool

  val iota : t -> bool

  val zeta : t -> bool

  val delta : t -> bool

  val fix_ : t -> bool

  val cofix_ : t -> bool

  val default : t
 end

val zip : (term * term list) -> term

val reduce_stack :
  RedFlags.t -> global_env -> context -> nat -> term -> term list ->
  (term * term list) option

val reduce_stack_term :
  RedFlags.t -> global_env -> context -> nat -> term -> term option

val fix_decls : term mfixpoint -> context_decl list

val map_constr_with_binders :
  (context -> term -> term) -> context -> term -> term

val reduce_opt :
  RedFlags.t -> global_env -> context -> nat -> term -> term option

val isConstruct : term -> bool

val isCoFix : term -> bool

type conv_pb =
| Conv
| Cumul

val conv_pb_rect : 'a1 -> 'a1 -> conv_pb -> 'a1

val conv_pb_rec : 'a1 -> 'a1 -> conv_pb -> 'a1

val eq_term : checker_flags -> universes_graph -> term -> term -> bool

val leq_term : checker_flags -> universes_graph -> term -> term -> bool

val nodelta_flags : RedFlags.t

val unfold_one_fix :
  global_env -> nat -> context -> term mfixpoint -> nat -> term list -> term
  option

val unfold_one_case : global_env -> nat -> context -> term -> term option

val reducible_head :
  global_env -> nat -> context -> term -> term list -> term option

val lookup_env : global_env -> ident -> global_decl option

val opt_bool_to_bool : bool option -> bool

val isconv :
  checker_flags -> global_env -> universes_graph -> nat -> conv_pb -> context
  -> term -> term list -> term -> term list -> bool option

val isconv_prog :
  checker_flags -> global_env -> universes_graph -> nat -> conv_pb -> context
  -> term -> term list -> term -> term list -> bool option

val try_reduce : global_env -> context -> nat -> term -> term

type type_error =
| UnboundRel of nat
| UnboundVar of char list
| UnboundMeta of nat
| UnboundEvar of nat
| UndeclaredConstant of char list
| UndeclaredInductive of inductive
| UndeclaredConstructor of inductive * nat
| NotConvertible of context * term * term * term * term
| NotASort of term
| NotAProduct of term * term
| NotAnInductive of term
| IllFormedFix of term mfixpoint * nat
| UnsatisfiedConstraints of ConstraintSet.t
| UnsatisfiableConstraints of ConstraintSet.t
| NotEnoughFuel of nat

val type_error_rect :
  (nat -> 'a1) -> (char list -> 'a1) -> (nat -> 'a1) -> (nat -> 'a1) ->
  (char list -> 'a1) -> (inductive -> 'a1) -> (inductive -> nat -> 'a1) ->
  (context -> term -> term -> term -> term -> 'a1) -> (term -> 'a1) -> (term
  -> term -> 'a1) -> (term -> 'a1) -> (term mfixpoint -> nat -> 'a1) ->
  (ConstraintSet.t -> 'a1) -> (ConstraintSet.t -> 'a1) -> (nat -> 'a1) ->
  type_error -> 'a1

val type_error_rec :
  (nat -> 'a1) -> (char list -> 'a1) -> (nat -> 'a1) -> (nat -> 'a1) ->
  (char list -> 'a1) -> (inductive -> 'a1) -> (inductive -> nat -> 'a1) ->
  (context -> term -> term -> term -> term -> 'a1) -> (term -> 'a1) -> (term
  -> term -> 'a1) -> (term -> 'a1) -> (term mfixpoint -> nat -> 'a1) ->
  (ConstraintSet.t -> 'a1) -> (ConstraintSet.t -> 'a1) -> (nat -> 'a1) ->
  type_error -> 'a1

val string_of_type_error : type_error -> char list

type 'a typing_result =
| Checked of 'a
| TypeError of type_error

val typing_result_rect :
  ('a1 -> 'a2) -> (type_error -> 'a2) -> 'a1 typing_result -> 'a2

val typing_result_rec :
  ('a1 -> 'a2) -> (type_error -> 'a2) -> 'a1 typing_result -> 'a2

val typing_monad : __ typing_result coq_Monad

val monad_exc : (type_error, __ typing_result) coq_MonadExc

val check_conv_gen :
  checker_flags -> coq_Fuel -> conv_pb -> global_env -> universes_graph ->
  context -> term -> term -> unit typing_result

val check_conv_leq :
  checker_flags -> coq_Fuel -> global_env -> universes_graph -> context ->
  term -> term -> unit typing_result

val check_conv :
  checker_flags -> coq_Fuel -> global_env -> universes_graph -> context ->
  term -> term -> unit typing_result

val conv_spec :
  checker_flags -> coq_Fuel -> global_env_ext -> Coq_wGraph.t -> context ->
  term -> term -> (conv, __) isEquiv

val cumul_spec :
  checker_flags -> coq_Fuel -> global_env_ext -> Coq_wGraph.t -> context ->
  term -> term -> (cumul, __) isEquiv

val reduce_cumul :
  checker_flags -> global_env_ext -> context -> nat -> term -> cumul

val hnf_stack :
  coq_Fuel -> global_env -> context -> term -> (term * term list)
  typing_result

val reduce : coq_Fuel -> global_env -> context -> term -> term typing_result

val reduce_to_sort :
  coq_Fuel -> global_env -> context -> term -> universe typing_result

val reduce_to_prod :
  coq_Fuel -> global_env -> context -> term -> (term * term) typing_result

val reduce_to_ind :
  coq_Fuel -> global_env -> context -> term -> ((inductive * Level.t
  list) * term list) typing_result

val convert_leq :
  checker_flags -> coq_Fuel -> global_env -> universes_graph -> context ->
  term -> term -> unit typing_result

val infer_spine :
  checker_flags -> coq_Fuel -> global_env -> universes_graph -> (context ->
  term -> term typing_result) -> context -> term -> term list -> term
  typing_result

val infer_type :
  coq_Fuel -> global_env -> (context -> term -> term typing_result) ->
  context -> term -> universe typing_result

val infer_cumul :
  checker_flags -> coq_Fuel -> global_env -> universes_graph -> (context ->
  term -> term typing_result) -> context -> term -> term -> unit typing_result

val polymorphic_constraints : universes_decl -> ConstraintSet.t

val lookup_constant_type :
  global_env -> ident -> universe_instance -> term typing_result

val lookup_constant_type_cstrs :
  global_env -> ident -> universe_instance -> (term * constraints)
  typing_result

val lookup_ind_decl :
  global_env -> ident -> nat -> ((one_inductive_body
  list * universes_decl) * one_inductive_body) typing_result

val lookup_ind_type :
  global_env -> ident -> nat -> Level.t list -> term typing_result

val lookup_ind_type_cstrs :
  global_env -> ident -> nat -> Level.t list -> (term * constraints)
  typing_result

val lookup_constructor_decl :
  global_env -> ident -> nat -> nat -> ((one_inductive_body
  list * universes_decl) * term) typing_result

val lookup_constructor_type :
  global_env -> ident -> nat -> nat -> universe_instance -> term typing_result

val lookup_constructor_type_cstrs :
  global_env -> ident -> nat -> nat -> universe_instance ->
  (term * constraints) typing_result

val check_consistent_constraints :
  checker_flags -> universes_graph -> constraints -> unit typing_result

val infer :
  checker_flags -> coq_Fuel -> global_env -> universes_graph -> context ->
  term -> term typing_result

val check :
  checker_flags -> coq_Fuel -> global_env -> universes_graph -> context ->
  term -> term -> unit typing_result

val typechecking :
  checker_flags -> coq_Fuel -> global_env -> universes_graph -> context ->
  term -> term -> bool

val cumul_convert_leq :
  checker_flags -> coq_Fuel -> global_env_ext -> Coq_wGraph.t -> context ->
  term -> term -> (cumul, __) isEquiv

val cumul_reduce_to_sort :
  checker_flags -> coq_Fuel -> global_env_ext -> Coq_wGraph.t -> context ->
  term -> universe -> (cumul, (universe, __) sigT) isEquiv

val cumul_reduce_to_product :
  checker_flags -> coq_Fuel -> global_env_ext -> context -> term -> name ->
  term -> term -> cumul -> (term, (term, __ * cumul) sigT) sigT

val cumul_reduce_to_ind :
  checker_flags -> coq_Fuel -> global_env_ext -> context -> term -> inductive
  -> universe_instance -> term list -> (cumul, (term list, __ * cumul) sigT)
  isEquiv

val lookup_constant_type_is_declared :
  global_env -> ident -> universe_instance -> (term * constraints) ->
  constant_body

val infer_type_correct :
  checker_flags -> coq_Fuel -> global_env_ext -> universes_graph -> context
  -> term -> universe -> (context -> term -> __ -> typing) -> typing

val infer_cumul_correct :
  checker_flags -> coq_Fuel -> global_env_ext -> universes_graph -> context
  -> term -> term -> unit -> universe -> (context -> term -> __ -> typing) ->
  (context -> term -> __ -> typing) -> typing

val infer_correct :
  checker_flags -> coq_Fuel -> global_env_ext -> universes_graph -> context
  -> term -> term -> typing lift_typing coq_All_local_env -> typing

val default_fuel : coq_Fuel

val fresh : char list -> global_decl list -> bool

type env_error =
| IllFormedDecl of char list * type_error
| AlreadyDeclared of char list

val env_error_rect :
  (char list -> type_error -> 'a1) -> (char list -> 'a1) -> env_error -> 'a1

val env_error_rec :
  (char list -> type_error -> 'a1) -> (char list -> 'a1) -> env_error -> 'a1

type 'a coq_EnvCheck =
| CorrectDecl of 'a
| EnvError of env_error

val coq_EnvCheck_rect :
  ('a1 -> 'a2) -> (env_error -> 'a2) -> 'a1 coq_EnvCheck -> 'a2

val coq_EnvCheck_rec :
  ('a1 -> 'a2) -> (env_error -> 'a2) -> 'a1 coq_EnvCheck -> 'a2

val envcheck_monad : __ coq_EnvCheck coq_Monad

val wrap_error : char list -> 'a1 typing_result -> 'a1 coq_EnvCheck

val check_wf_type :
  checker_flags -> coq_Fuel -> char list -> global_env -> universes_graph ->
  term -> unit coq_EnvCheck

val check_wf_judgement :
  checker_flags -> coq_Fuel -> char list -> global_env -> universes_graph ->
  term -> term -> unit coq_EnvCheck

val infer_term :
  checker_flags -> coq_Fuel -> global_env -> universes_graph -> term -> term
  coq_EnvCheck

val check_wf_decl :
  checker_flags -> coq_Fuel -> global_env -> universes_graph -> global_decl
  -> unit coq_EnvCheck

val check_fresh : char list -> global_decl list -> unit coq_EnvCheck

val monomorphic_constraints : universes_decl -> constraints

val global_decl_univs : global_decl -> constraints

val add_gc_constraints :
  GoodConstraintSet.t -> universes_graph -> universes_graph

val check_wf_env :
  checker_flags -> coq_Fuel -> global_env -> universes_graph coq_EnvCheck

val typecheck_program :
  checker_flags -> coq_Fuel -> program -> term coq_EnvCheck

val infer' :
  checker_flags -> coq_Fuel -> global_env_ext -> context -> term -> term
  typing_result
