open Ast0
open AstUtils
open BasicAst
open Datatypes
open Equality0
open LiftSubst
open List0
open Nat0
open PeanoNat
open Specif
open UnivSubst0
open Universes0
open Config0
open Utils

type __ = Obj.t
let __ = let rec f _ = Obj.repr f in Obj.repr f

(** val subst_context : term list -> nat -> context -> context **)

let subst_context s k _UU0393_ =
  fold_context (fun k' -> subst s (add k' k)) _UU0393_

(** val smash_context : context -> context -> context **)

let rec smash_context _UU0393_ = function
| [] -> _UU0393_
| d :: _UU0393_'0 ->
  let { decl_name = _; decl_body = decl_body0; decl_type = _ } = d in
  (match decl_body0 with
   | Some b -> smash_context (subst_context (b :: []) O _UU0393_) _UU0393_'0
   | None -> smash_context (app _UU0393_ (d :: [])) _UU0393_'0)

(** val global_decl_ident : global_decl -> kername **)

let global_decl_ident = function
| ConstantDecl (id, _) -> id
| InductiveDecl (id, _) -> id

(** val lookup_env : global_env -> ident -> global_decl option **)

let rec lookup_env _UU03a3_ id =
  match _UU03a3_ with
  | [] -> None
  | hd :: tl ->
    if ident_eq id (global_decl_ident hd) then Some hd else lookup_env tl id

(** val inds :
    kername -> universe_instance -> one_inductive_body list -> term list **)

let inds ind u l =
  let rec aux = function
  | O -> []
  | S n0 ->
    (Coq_tInd ({ inductive_mind = ind; inductive_ind = n0 }, u)) :: (aux n0)
  in aux (length l)

(** val type_of_constructor :
    mutual_inductive_body -> ((ident * term) * nat) -> (inductive * nat) ->
    Level.t list -> term **)

let type_of_constructor mdecl cdecl c u =
  let mind = (fst c).inductive_mind in
  subst (inds mind u mdecl.ind_bodies) O
    (subst_instance_constr u (snd (fst cdecl)))

(** val fix_subst : term mfixpoint -> term list **)

let fix_subst l =
  let rec aux = function
  | O -> []
  | S n0 -> (Coq_tFix (l, n0)) :: (aux n0)
  in aux (length l)

(** val unfold_fix : term mfixpoint -> nat -> (nat * term) option **)

let unfold_fix mfix idx =
  match nth_error mfix idx with
  | Some d -> Some (d.rarg, (subst (fix_subst mfix) O d.dbody))
  | None -> None

(** val cofix_subst : term mfixpoint -> term list **)

let cofix_subst l =
  let rec aux = function
  | O -> []
  | S n0 -> (Coq_tCoFix (l, n0)) :: (aux n0)
  in aux (length l)

(** val unfold_cofix : term mfixpoint -> nat -> (nat * term) option **)

let unfold_cofix mfix idx =
  match nth_error mfix idx with
  | Some d -> Some (d.rarg, (subst (cofix_subst mfix) O d.dbody))
  | None -> None

(** val isConstruct_app : term -> bool **)

let isConstruct_app t0 =
  match fst (decompose_app t0) with
  | Coq_tConstruct (_, _, _) -> true
  | _ -> false

(** val is_constructor : nat -> term list -> bool **)

let is_constructor n ts =
  match nth_error ts n with
  | Some a -> isConstruct_app a
  | None -> false

(** val fix_context : term mfixpoint -> context **)

let fix_context m =
  List0.rev (mapi (fun i d -> vass d.dname (lift i O d.dtype)) m)

(** val tDummy : term **)

let tDummy =
  Coq_tVar []

(** val iota_red : nat -> nat -> term list -> (nat * term) list -> term **)

let iota_red npar c args brs =
  mkApps (snd (nth c brs (O, tDummy))) (skipn npar args)

type red1 =
| Coq_red_beta of name * term * term * term * term list
| Coq_red_zeta of name * term * term * term
| Coq_red_rel of nat * term
| Coq_red_iota of inductive * nat * nat * universe_instance * term list
   * term * (nat * term) list
| Coq_red_fix of term mfixpoint * nat * term list * nat * term
| Coq_red_cofix_case of (inductive * nat) * term * term mfixpoint * nat
   * term list * nat * term * (nat * term) list
| Coq_red_cofix_proj of projection * term mfixpoint * nat * term list * 
   nat * term
| Coq_red_delta of ident * constant_body * term * universe_instance
| Coq_red_proj of inductive * nat * nat * term list * nat * universe_instance
   * term
| Coq_abs_red_l of name * term * term * term * red1
| Coq_abs_red_r of name * term * term * term * red1
| Coq_letin_red_def of name * term * term * term * term * red1
| Coq_letin_red_ty of name * term * term * term * term * red1
| Coq_letin_red_body of name * term * term * term * term * red1
| Coq_case_red_pred of (inductive * nat) * term * term * term
   * (nat * term) list * red1
| Coq_case_red_discr of (inductive * nat) * term * term * term
   * (nat * term) list * red1
| Coq_case_red_brs of (inductive * nat) * term * term * (nat * term) list
   * (nat * term) list * (nat * term, red1) coq_OnOne2
| Coq_proj_red of projection * term * term * red1
| Coq_app_red_l of term * term * term list * red1
| Coq_app_red_r of term list * term list * term * (term, red1) coq_OnOne2
| Coq_prod_red_l of name * term * term * term * red1
| Coq_prod_red_r of name * term * term * term * red1
| Coq_evar_red of nat * term list * term list * (term, red1) coq_OnOne2
| Coq_cast_red_l of term * cast_kind * term * term * red1
| Coq_cast_red_r of term * cast_kind * term * term * red1
| Coq_cast_red of term * cast_kind * term
| Coq_fix_red_ty of term def list * term def list * nat
   * (term def, red1 * __) coq_OnOne2
| Coq_fix_red_body of term mfixpoint * term def list * nat
   * (term def, red1 * __) coq_OnOne2
| Coq_cofix_red_ty of term def list * term def list * nat
   * (term def, red1 * __) coq_OnOne2
| Coq_cofix_red_body of term mfixpoint * term def list * nat
   * (term def, red1 * __) coq_OnOne2

(** val red1_rect :
    global_env -> (context -> name -> term -> term -> term -> term list ->
    'a1) -> (context -> name -> term -> term -> term -> 'a1) -> (context ->
    nat -> term -> __ -> 'a1) -> (context -> inductive -> nat -> nat ->
    universe_instance -> term list -> term -> (nat * term) list -> 'a1) ->
    (context -> term mfixpoint -> nat -> term list -> nat -> term -> __ -> __
    -> 'a1) -> (context -> (inductive * nat) -> term -> term mfixpoint -> nat
    -> term list -> nat -> term -> (nat * term) list -> __ -> 'a1) ->
    (context -> projection -> term mfixpoint -> nat -> term list -> nat ->
    term -> __ -> 'a1) -> (context -> ident -> constant_body -> term -> __ ->
    universe_instance -> __ -> 'a1) -> (context -> inductive -> nat -> nat ->
    term list -> nat -> universe_instance -> term -> __ -> 'a1) -> (context
    -> name -> term -> term -> term -> red1 -> 'a1 -> 'a1) -> (context ->
    name -> term -> term -> term -> red1 -> 'a1 -> 'a1) -> (context -> name
    -> term -> term -> term -> term -> red1 -> 'a1 -> 'a1) -> (context ->
    name -> term -> term -> term -> term -> red1 -> 'a1 -> 'a1) -> (context
    -> name -> term -> term -> term -> term -> red1 -> 'a1 -> 'a1) ->
    (context -> (inductive * nat) -> term -> term -> term -> (nat * term)
    list -> red1 -> 'a1 -> 'a1) -> (context -> (inductive * nat) -> term ->
    term -> term -> (nat * term) list -> red1 -> 'a1 -> 'a1) -> (context ->
    (inductive * nat) -> term -> term -> (nat * term) list -> (nat * term)
    list -> (nat * term, red1) coq_OnOne2 -> 'a1) -> (context -> projection
    -> term -> term -> red1 -> 'a1 -> 'a1) -> (context -> term -> term ->
    term list -> red1 -> 'a1 -> 'a1) -> (context -> term list -> term list ->
    term -> (term, red1) coq_OnOne2 -> 'a1) -> (context -> name -> term ->
    term -> term -> red1 -> 'a1 -> 'a1) -> (context -> name -> term -> term
    -> term -> red1 -> 'a1 -> 'a1) -> (context -> nat -> term list -> term
    list -> (term, red1) coq_OnOne2 -> 'a1) -> (context -> term -> cast_kind
    -> term -> term -> red1 -> 'a1 -> 'a1) -> (context -> term -> cast_kind
    -> term -> term -> red1 -> 'a1 -> 'a1) -> (context -> term -> cast_kind
    -> term -> 'a1) -> (context -> term def list -> term def list -> nat ->
    (term def, red1 * __) coq_OnOne2 -> 'a1) -> (context -> term mfixpoint ->
    term def list -> nat -> (term def, red1 * __) coq_OnOne2 -> 'a1) ->
    (context -> term def list -> term def list -> nat -> (term def,
    red1 * __) coq_OnOne2 -> 'a1) -> (context -> term mfixpoint -> term def
    list -> nat -> (term def, red1 * __) coq_OnOne2 -> 'a1) -> context ->
    term -> term -> red1 -> 'a1 **)

let rec red1_rect _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13 f14 f15 f16 f17 f18 f19 f20 f21 f22 f23 f24 f25 f26 f27 f28 _UU0393_ _ _ = function
| Coq_red_beta (na, t0, b, a, l) -> f _UU0393_ na t0 b a l
| Coq_red_zeta (na, b, t0, b') -> f0 _UU0393_ na b t0 b'
| Coq_red_rel (i, body) -> f1 _UU0393_ i body __
| Coq_red_iota (ind, pars, c, u, args, p, brs) ->
  f2 _UU0393_ ind pars c u args p brs
| Coq_red_fix (mfix, idx, args, narg, fn) ->
  f3 _UU0393_ mfix idx args narg fn __ __
| Coq_red_cofix_case (ip, p, mfix, idx, args, narg, fn, brs) ->
  f4 _UU0393_ ip p mfix idx args narg fn brs __
| Coq_red_cofix_proj (p, mfix, idx, args, narg, fn) ->
  f5 _UU0393_ p mfix idx args narg fn __
| Coq_red_delta (c, decl, body, u) -> f6 _UU0393_ c decl body __ u __
| Coq_red_proj (i, pars, narg, args, k, u, arg) ->
  f7 _UU0393_ i pars narg args k u arg __
| Coq_abs_red_l (na, m, m', n, r0) ->
  f8 _UU0393_ na m m' n r0
    (red1_rect _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13 f14
      f15 f16 f17 f18 f19 f20 f21 f22 f23 f24 f25 f26 f27 f28 _UU0393_ m m'
      r0)
| Coq_abs_red_r (na, m, m', n, r0) ->
  f9 _UU0393_ na m m' n r0
    (red1_rect _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13 f14
      f15 f16 f17 f18 f19 f20 f21 f22 f23 f24 f25 f26 f27 f28
      (snoc _UU0393_ (vass na n)) m m' r0)
| Coq_letin_red_def (na, b, t0, b', r0, r1) ->
  f10 _UU0393_ na b t0 b' r0 r1
    (red1_rect _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13 f14
      f15 f16 f17 f18 f19 f20 f21 f22 f23 f24 f25 f26 f27 f28 _UU0393_ b r0
      r1)
| Coq_letin_red_ty (na, b, t0, b', r0, r1) ->
  f11 _UU0393_ na b t0 b' r0 r1
    (red1_rect _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13 f14
      f15 f16 f17 f18 f19 f20 f21 f22 f23 f24 f25 f26 f27 f28 _UU0393_ t0 r0
      r1)
| Coq_letin_red_body (na, b, t0, b', r0, r1) ->
  f12 _UU0393_ na b t0 b' r0 r1
    (red1_rect _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13 f14
      f15 f16 f17 f18 f19 f20 f21 f22 f23 f24 f25 f26 f27 f28
      (snoc _UU0393_ (vdef na b t0)) b' r0 r1)
| Coq_case_red_pred (ind, p, p', c, brs, r0) ->
  f13 _UU0393_ ind p p' c brs r0
    (red1_rect _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13 f14
      f15 f16 f17 f18 f19 f20 f21 f22 f23 f24 f25 f26 f27 f28 _UU0393_ p p'
      r0)
| Coq_case_red_discr (ind, p, c, c', brs, r0) ->
  f14 _UU0393_ ind p c c' brs r0
    (red1_rect _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13 f14
      f15 f16 f17 f18 f19 f20 f21 f22 f23 f24 f25 f26 f27 f28 _UU0393_ c c'
      r0)
| Coq_case_red_brs (ind, p, c, brs, brs', o) ->
  f15 _UU0393_ ind p c brs brs' o
| Coq_proj_red (p, c, c', r0) ->
  f16 _UU0393_ p c c' r0
    (red1_rect _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13 f14
      f15 f16 f17 f18 f19 f20 f21 f22 f23 f24 f25 f26 f27 f28 _UU0393_ c c'
      r0)
| Coq_app_red_l (m1, n1, m2, r0) ->
  f17 _UU0393_ m1 n1 m2 r0
    (red1_rect _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13 f14
      f15 f16 f17 f18 f19 f20 f21 f22 f23 f24 f25 f26 f27 f28 _UU0393_ m1 n1
      r0)
| Coq_app_red_r (m2, n2, m1, o) -> f18 _UU0393_ m2 n2 m1 o
| Coq_prod_red_l (na, m1, m2, n1, r0) ->
  f19 _UU0393_ na m1 m2 n1 r0
    (red1_rect _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13 f14
      f15 f16 f17 f18 f19 f20 f21 f22 f23 f24 f25 f26 f27 f28 _UU0393_ m1 n1
      r0)
| Coq_prod_red_r (na, m2, n2, m1, r0) ->
  f20 _UU0393_ na m2 n2 m1 r0
    (red1_rect _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13 f14
      f15 f16 f17 f18 f19 f20 f21 f22 f23 f24 f25 f26 f27 f28
      (snoc _UU0393_ (vass na m1)) m2 n2 r0)
| Coq_evar_red (ev, l, l', o) -> f21 _UU0393_ ev l l' o
| Coq_cast_red_l (m1, k, m2, n1, r0) ->
  f22 _UU0393_ m1 k m2 n1 r0
    (red1_rect _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13 f14
      f15 f16 f17 f18 f19 f20 f21 f22 f23 f24 f25 f26 f27 f28 _UU0393_ m1 n1
      r0)
| Coq_cast_red_r (m2, k, n2, m1, r0) ->
  f23 _UU0393_ m2 k n2 m1 r0
    (red1_rect _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13 f14
      f15 f16 f17 f18 f19 f20 f21 f22 f23 f24 f25 f26 f27 f28 _UU0393_ m2 n2
      r0)
| Coq_cast_red (m1, k, m2) -> f24 _UU0393_ m1 k m2
| Coq_fix_red_ty (mfix0, mfix1, idx, o) -> f25 _UU0393_ mfix0 mfix1 idx o
| Coq_fix_red_body (mfix0, mfix1, idx, o) -> f26 _UU0393_ mfix0 mfix1 idx o
| Coq_cofix_red_ty (mfix0, mfix1, idx, o) -> f27 _UU0393_ mfix0 mfix1 idx o
| Coq_cofix_red_body (mfix0, mfix1, idx, o) -> f28 _UU0393_ mfix0 mfix1 idx o

(** val red1_rec :
    global_env -> (context -> name -> term -> term -> term -> term list ->
    'a1) -> (context -> name -> term -> term -> term -> 'a1) -> (context ->
    nat -> term -> __ -> 'a1) -> (context -> inductive -> nat -> nat ->
    universe_instance -> term list -> term -> (nat * term) list -> 'a1) ->
    (context -> term mfixpoint -> nat -> term list -> nat -> term -> __ -> __
    -> 'a1) -> (context -> (inductive * nat) -> term -> term mfixpoint -> nat
    -> term list -> nat -> term -> (nat * term) list -> __ -> 'a1) ->
    (context -> projection -> term mfixpoint -> nat -> term list -> nat ->
    term -> __ -> 'a1) -> (context -> ident -> constant_body -> term -> __ ->
    universe_instance -> __ -> 'a1) -> (context -> inductive -> nat -> nat ->
    term list -> nat -> universe_instance -> term -> __ -> 'a1) -> (context
    -> name -> term -> term -> term -> red1 -> 'a1 -> 'a1) -> (context ->
    name -> term -> term -> term -> red1 -> 'a1 -> 'a1) -> (context -> name
    -> term -> term -> term -> term -> red1 -> 'a1 -> 'a1) -> (context ->
    name -> term -> term -> term -> term -> red1 -> 'a1 -> 'a1) -> (context
    -> name -> term -> term -> term -> term -> red1 -> 'a1 -> 'a1) ->
    (context -> (inductive * nat) -> term -> term -> term -> (nat * term)
    list -> red1 -> 'a1 -> 'a1) -> (context -> (inductive * nat) -> term ->
    term -> term -> (nat * term) list -> red1 -> 'a1 -> 'a1) -> (context ->
    (inductive * nat) -> term -> term -> (nat * term) list -> (nat * term)
    list -> (nat * term, red1) coq_OnOne2 -> 'a1) -> (context -> projection
    -> term -> term -> red1 -> 'a1 -> 'a1) -> (context -> term -> term ->
    term list -> red1 -> 'a1 -> 'a1) -> (context -> term list -> term list ->
    term -> (term, red1) coq_OnOne2 -> 'a1) -> (context -> name -> term ->
    term -> term -> red1 -> 'a1 -> 'a1) -> (context -> name -> term -> term
    -> term -> red1 -> 'a1 -> 'a1) -> (context -> nat -> term list -> term
    list -> (term, red1) coq_OnOne2 -> 'a1) -> (context -> term -> cast_kind
    -> term -> term -> red1 -> 'a1 -> 'a1) -> (context -> term -> cast_kind
    -> term -> term -> red1 -> 'a1 -> 'a1) -> (context -> term -> cast_kind
    -> term -> 'a1) -> (context -> term def list -> term def list -> nat ->
    (term def, red1 * __) coq_OnOne2 -> 'a1) -> (context -> term mfixpoint ->
    term def list -> nat -> (term def, red1 * __) coq_OnOne2 -> 'a1) ->
    (context -> term def list -> term def list -> nat -> (term def,
    red1 * __) coq_OnOne2 -> 'a1) -> (context -> term mfixpoint -> term def
    list -> nat -> (term def, red1 * __) coq_OnOne2 -> 'a1) -> context ->
    term -> term -> red1 -> 'a1 **)

let rec red1_rec _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13 f14 f15 f16 f17 f18 f19 f20 f21 f22 f23 f24 f25 f26 f27 f28 _UU0393_ _ _ = function
| Coq_red_beta (na, t0, b, a, l) -> f _UU0393_ na t0 b a l
| Coq_red_zeta (na, b, t0, b') -> f0 _UU0393_ na b t0 b'
| Coq_red_rel (i, body) -> f1 _UU0393_ i body __
| Coq_red_iota (ind, pars, c, u, args, p, brs) ->
  f2 _UU0393_ ind pars c u args p brs
| Coq_red_fix (mfix, idx, args, narg, fn) ->
  f3 _UU0393_ mfix idx args narg fn __ __
| Coq_red_cofix_case (ip, p, mfix, idx, args, narg, fn, brs) ->
  f4 _UU0393_ ip p mfix idx args narg fn brs __
| Coq_red_cofix_proj (p, mfix, idx, args, narg, fn) ->
  f5 _UU0393_ p mfix idx args narg fn __
| Coq_red_delta (c, decl, body, u) -> f6 _UU0393_ c decl body __ u __
| Coq_red_proj (i, pars, narg, args, k, u, arg) ->
  f7 _UU0393_ i pars narg args k u arg __
| Coq_abs_red_l (na, m, m', n, r0) ->
  f8 _UU0393_ na m m' n r0
    (red1_rec _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13 f14
      f15 f16 f17 f18 f19 f20 f21 f22 f23 f24 f25 f26 f27 f28 _UU0393_ m m'
      r0)
| Coq_abs_red_r (na, m, m', n, r0) ->
  f9 _UU0393_ na m m' n r0
    (red1_rec _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13 f14
      f15 f16 f17 f18 f19 f20 f21 f22 f23 f24 f25 f26 f27 f28
      (snoc _UU0393_ (vass na n)) m m' r0)
| Coq_letin_red_def (na, b, t0, b', r0, r1) ->
  f10 _UU0393_ na b t0 b' r0 r1
    (red1_rec _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13 f14
      f15 f16 f17 f18 f19 f20 f21 f22 f23 f24 f25 f26 f27 f28 _UU0393_ b r0
      r1)
| Coq_letin_red_ty (na, b, t0, b', r0, r1) ->
  f11 _UU0393_ na b t0 b' r0 r1
    (red1_rec _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13 f14
      f15 f16 f17 f18 f19 f20 f21 f22 f23 f24 f25 f26 f27 f28 _UU0393_ t0 r0
      r1)
| Coq_letin_red_body (na, b, t0, b', r0, r1) ->
  f12 _UU0393_ na b t0 b' r0 r1
    (red1_rec _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13 f14
      f15 f16 f17 f18 f19 f20 f21 f22 f23 f24 f25 f26 f27 f28
      (snoc _UU0393_ (vdef na b t0)) b' r0 r1)
| Coq_case_red_pred (ind, p, p', c, brs, r0) ->
  f13 _UU0393_ ind p p' c brs r0
    (red1_rec _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13 f14
      f15 f16 f17 f18 f19 f20 f21 f22 f23 f24 f25 f26 f27 f28 _UU0393_ p p'
      r0)
| Coq_case_red_discr (ind, p, c, c', brs, r0) ->
  f14 _UU0393_ ind p c c' brs r0
    (red1_rec _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13 f14
      f15 f16 f17 f18 f19 f20 f21 f22 f23 f24 f25 f26 f27 f28 _UU0393_ c c'
      r0)
| Coq_case_red_brs (ind, p, c, brs, brs', o) ->
  f15 _UU0393_ ind p c brs brs' o
| Coq_proj_red (p, c, c', r0) ->
  f16 _UU0393_ p c c' r0
    (red1_rec _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13 f14
      f15 f16 f17 f18 f19 f20 f21 f22 f23 f24 f25 f26 f27 f28 _UU0393_ c c'
      r0)
| Coq_app_red_l (m1, n1, m2, r0) ->
  f17 _UU0393_ m1 n1 m2 r0
    (red1_rec _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13 f14
      f15 f16 f17 f18 f19 f20 f21 f22 f23 f24 f25 f26 f27 f28 _UU0393_ m1 n1
      r0)
| Coq_app_red_r (m2, n2, m1, o) -> f18 _UU0393_ m2 n2 m1 o
| Coq_prod_red_l (na, m1, m2, n1, r0) ->
  f19 _UU0393_ na m1 m2 n1 r0
    (red1_rec _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13 f14
      f15 f16 f17 f18 f19 f20 f21 f22 f23 f24 f25 f26 f27 f28 _UU0393_ m1 n1
      r0)
| Coq_prod_red_r (na, m2, n2, m1, r0) ->
  f20 _UU0393_ na m2 n2 m1 r0
    (red1_rec _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13 f14
      f15 f16 f17 f18 f19 f20 f21 f22 f23 f24 f25 f26 f27 f28
      (snoc _UU0393_ (vass na m1)) m2 n2 r0)
| Coq_evar_red (ev, l, l', o) -> f21 _UU0393_ ev l l' o
| Coq_cast_red_l (m1, k, m2, n1, r0) ->
  f22 _UU0393_ m1 k m2 n1 r0
    (red1_rec _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13 f14
      f15 f16 f17 f18 f19 f20 f21 f22 f23 f24 f25 f26 f27 f28 _UU0393_ m1 n1
      r0)
| Coq_cast_red_r (m2, k, n2, m1, r0) ->
  f23 _UU0393_ m2 k n2 m1 r0
    (red1_rec _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13 f14
      f15 f16 f17 f18 f19 f20 f21 f22 f23 f24 f25 f26 f27 f28 _UU0393_ m2 n2
      r0)
| Coq_cast_red (m1, k, m2) -> f24 _UU0393_ m1 k m2
| Coq_fix_red_ty (mfix0, mfix1, idx, o) -> f25 _UU0393_ mfix0 mfix1 idx o
| Coq_fix_red_body (mfix0, mfix1, idx, o) -> f26 _UU0393_ mfix0 mfix1 idx o
| Coq_cofix_red_ty (mfix0, mfix1, idx, o) -> f27 _UU0393_ mfix0 mfix1 idx o
| Coq_cofix_red_body (mfix0, mfix1, idx, o) -> f28 _UU0393_ mfix0 mfix1 idx o

(** val red1_ind_all :
    global_env -> (context -> name -> term -> term -> term -> term list ->
    'a1) -> (context -> name -> term -> term -> term -> 'a1) -> (context ->
    nat -> term -> __ -> 'a1) -> (context -> inductive -> nat -> nat ->
    universe_instance -> term list -> term -> (nat * term) list -> 'a1) ->
    (context -> term mfixpoint -> nat -> term list -> nat -> term -> __ -> __
    -> 'a1) -> (context -> (inductive * nat) -> term -> term mfixpoint -> nat
    -> term list -> nat -> term -> (nat * term) list -> __ -> 'a1) ->
    (context -> projection -> term mfixpoint -> nat -> term list -> nat ->
    term -> __ -> 'a1) -> (context -> ident -> constant_body -> term -> __ ->
    universe_instance -> __ -> 'a1) -> (context -> inductive -> nat -> nat ->
    term list -> nat -> universe_instance -> term -> __ -> 'a1) -> (context
    -> name -> term -> term -> term -> red1 -> 'a1 -> 'a1) -> (context ->
    name -> term -> term -> term -> red1 -> 'a1 -> 'a1) -> (context -> name
    -> term -> term -> term -> term -> red1 -> 'a1 -> 'a1) -> (context ->
    name -> term -> term -> term -> term -> red1 -> 'a1 -> 'a1) -> (context
    -> name -> term -> term -> term -> term -> red1 -> 'a1 -> 'a1) ->
    (context -> (inductive * nat) -> term -> term -> term -> (nat * term)
    list -> red1 -> 'a1 -> 'a1) -> (context -> (inductive * nat) -> term ->
    term -> term -> (nat * term) list -> red1 -> 'a1 -> 'a1) -> (context ->
    (inductive * nat) -> term -> term -> (nat * term) list -> (nat * term)
    list -> (nat * term, red1 * 'a1) coq_OnOne2 -> 'a1) -> (context ->
    projection -> term -> term -> red1 -> 'a1 -> 'a1) -> (context -> term ->
    term -> term list -> red1 -> 'a1 -> 'a1) -> (context -> term list -> term
    list -> term -> (term, red1 * 'a1) coq_OnOne2 -> 'a1) -> (context -> name
    -> term -> term -> term -> red1 -> 'a1 -> 'a1) -> (context -> name ->
    term -> term -> term -> red1 -> 'a1 -> 'a1) -> (context -> nat -> term
    list -> term list -> (term, red1 * 'a1) coq_OnOne2 -> 'a1) -> (context ->
    term -> cast_kind -> term -> term -> red1 -> 'a1 -> 'a1) -> (context ->
    term -> cast_kind -> term -> term -> red1 -> 'a1 -> 'a1) -> (context ->
    term -> cast_kind -> term -> 'a1) -> (context -> term def list -> term
    def list -> nat -> (term def, (red1 * __) * 'a1) coq_OnOne2 -> 'a1) ->
    (context -> term def list -> term def list -> nat -> (term def,
    (red1 * __) * 'a1) coq_OnOne2 -> 'a1) -> (context -> term def list ->
    term def list -> nat -> (term def, (red1 * __) * 'a1) coq_OnOne2 -> 'a1)
    -> (context -> term def list -> term def list -> nat -> (term def,
    (red1 * __) * 'a1) coq_OnOne2 -> 'a1) -> context -> term -> term -> red1
    -> 'a1 **)

let rec red1_ind_all _UU03a3_ x x0 x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12 x13 x14 x15 x16 x17 x18 x19 x20 x21 x22 x23 x24 x25 x26 x27 x28 _UU0393_ _ _ = function
| Coq_red_beta (na, t0, b, a, l) -> x _UU0393_ na t0 b a l
| Coq_red_zeta (na, b, t0, b') -> x0 _UU0393_ na b t0 b'
| Coq_red_rel (i, body) -> x1 _UU0393_ i body __
| Coq_red_iota (ind, pars, c, u, args, p, brs) ->
  x2 _UU0393_ ind pars c u args p brs
| Coq_red_fix (mfix, idx, args, narg, fn) ->
  x3 _UU0393_ mfix idx args narg fn __ __
| Coq_red_cofix_case (ip, p, mfix, idx, args, narg, fn, brs) ->
  x4 _UU0393_ ip p mfix idx args narg fn brs __
| Coq_red_cofix_proj (p, mfix, idx, args, narg, fn) ->
  x5 _UU0393_ p mfix idx args narg fn __
| Coq_red_delta (c, decl, body, u) -> x6 _UU0393_ c decl body __ u __
| Coq_red_proj (i, pars, narg, args, k, u, arg) ->
  x7 _UU0393_ i pars narg args k u arg __
| Coq_abs_red_l (na, m, m', n, xlast) ->
  x8 _UU0393_ na m m' n xlast
    (red1_ind_all _UU03a3_ x x0 x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12 x13
      x14 x15 x16 x17 x18 x19 x20 x21 x22 x23 x24 x25 x26 x27 x28 _UU0393_ m
      m' xlast)
| Coq_abs_red_r (na, m, m', n, xlast) ->
  x9 _UU0393_ na m m' n xlast
    (red1_ind_all _UU03a3_ x x0 x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12 x13
      x14 x15 x16 x17 x18 x19 x20 x21 x22 x23 x24 x25 x26 x27 x28
      (snoc _UU0393_ (vass na n)) m m' xlast)
| Coq_letin_red_def (na, b, t0, b', r, xlast) ->
  x10 _UU0393_ na b t0 b' r xlast
    (red1_ind_all _UU03a3_ x x0 x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12 x13
      x14 x15 x16 x17 x18 x19 x20 x21 x22 x23 x24 x25 x26 x27 x28 _UU0393_ b
      r xlast)
| Coq_letin_red_ty (na, b, t0, b', r, xlast) ->
  x11 _UU0393_ na b t0 b' r xlast
    (red1_ind_all _UU03a3_ x x0 x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12 x13
      x14 x15 x16 x17 x18 x19 x20 x21 x22 x23 x24 x25 x26 x27 x28 _UU0393_ t0
      r xlast)
| Coq_letin_red_body (na, b, t0, b', r, xlast) ->
  x12 _UU0393_ na b t0 b' r xlast
    (red1_ind_all _UU03a3_ x x0 x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12 x13
      x14 x15 x16 x17 x18 x19 x20 x21 x22 x23 x24 x25 x26 x27 x28
      (snoc _UU0393_ (vdef na b t0)) b' r xlast)
| Coq_case_red_pred (ind, p, p', c, brs, xlast) ->
  x13 _UU0393_ ind p p' c brs xlast
    (red1_ind_all _UU03a3_ x x0 x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12 x13
      x14 x15 x16 x17 x18 x19 x20 x21 x22 x23 x24 x25 x26 x27 x28 _UU0393_ p
      p' xlast)
| Coq_case_red_discr (ind, p, c, c', brs, xlast) ->
  x14 _UU0393_ ind p c c' brs xlast
    (red1_ind_all _UU03a3_ x x0 x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12 x13
      x14 x15 x16 x17 x18 x19 x20 x21 x22 x23 x24 x25 x26 x27 x28 _UU0393_ c
      c' xlast)
| Coq_case_red_brs (ind, p, c, brs, brs', o) ->
  x15 _UU0393_ ind p c brs brs'
    (let rec auxl _ _ = function
     | OnOne2_hd (hd, hd', tl, r) ->
       OnOne2_hd (hd, hd', tl, (r,
         (red1_ind_all _UU03a3_ x x0 x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12
           x13 x14 x15 x16 x17 x18 x19 x20 x21 x22 x23 x24 x25 x26 x27 x28
           _UU0393_ (snd hd) (snd hd') r)))
     | OnOne2_tl (hd, tl, tl', hl0) ->
       OnOne2_tl (hd, tl, tl', (auxl tl tl' hl0))
     in auxl brs brs' o)
| Coq_proj_red (p, c, c', xlast) ->
  x16 _UU0393_ p c c' xlast
    (red1_ind_all _UU03a3_ x x0 x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12 x13
      x14 x15 x16 x17 x18 x19 x20 x21 x22 x23 x24 x25 x26 x27 x28 _UU0393_ c
      c' xlast)
| Coq_app_red_l (m1, n1, m2, xlast) ->
  x17 _UU0393_ m1 n1 m2 xlast
    (red1_ind_all _UU03a3_ x x0 x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12 x13
      x14 x15 x16 x17 x18 x19 x20 x21 x22 x23 x24 x25 x26 x27 x28 _UU0393_ m1
      n1 xlast)
| Coq_app_red_r (m2, n2, m1, o) ->
  x18 _UU0393_ m2 n2 m1
    (let rec auxl _ _ = function
     | OnOne2_hd (hd, hd', tl, r) ->
       OnOne2_hd (hd, hd', tl, (r,
         (red1_ind_all _UU03a3_ x x0 x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12
           x13 x14 x15 x16 x17 x18 x19 x20 x21 x22 x23 x24 x25 x26 x27 x28
           _UU0393_ hd hd' r)))
     | OnOne2_tl (hd, tl, tl', hl0) ->
       OnOne2_tl (hd, tl, tl', (auxl tl tl' hl0))
     in auxl m2 n2 o)
| Coq_prod_red_l (na, m1, m2, n1, xlast) ->
  x19 _UU0393_ na m1 m2 n1 xlast
    (red1_ind_all _UU03a3_ x x0 x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12 x13
      x14 x15 x16 x17 x18 x19 x20 x21 x22 x23 x24 x25 x26 x27 x28 _UU0393_ m1
      n1 xlast)
| Coq_prod_red_r (na, m2, n2, m1, xlast) ->
  x20 _UU0393_ na m2 n2 m1 xlast
    (red1_ind_all _UU03a3_ x x0 x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12 x13
      x14 x15 x16 x17 x18 x19 x20 x21 x22 x23 x24 x25 x26 x27 x28
      (snoc _UU0393_ (vass na m1)) m2 n2 xlast)
| Coq_evar_red (ev, l, l', o) ->
  x21 _UU0393_ ev l l'
    (let rec auxl _ _ = function
     | OnOne2_hd (hd, hd', tl, r) ->
       OnOne2_hd (hd, hd', tl, (r,
         (red1_ind_all _UU03a3_ x x0 x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12
           x13 x14 x15 x16 x17 x18 x19 x20 x21 x22 x23 x24 x25 x26 x27 x28
           _UU0393_ hd hd' r)))
     | OnOne2_tl (hd, tl, tl', hl0) ->
       OnOne2_tl (hd, tl, tl', (auxl tl tl' hl0))
     in auxl l l' o)
| Coq_cast_red_l (m1, k, m2, n1, xlast) ->
  x22 _UU0393_ m1 k m2 n1 xlast
    (red1_ind_all _UU03a3_ x x0 x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12 x13
      x14 x15 x16 x17 x18 x19 x20 x21 x22 x23 x24 x25 x26 x27 x28 _UU0393_ m1
      n1 xlast)
| Coq_cast_red_r (m2, k, n2, m1, xlast) ->
  x23 _UU0393_ m2 k n2 m1 xlast
    (red1_ind_all _UU03a3_ x x0 x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12 x13
      x14 x15 x16 x17 x18 x19 x20 x21 x22 x23 x24 x25 x26 x27 x28 _UU0393_ m2
      n2 xlast)
| Coq_cast_red (m1, k, m2) -> x24 _UU0393_ m1 k m2
| Coq_fix_red_ty (mfix0, mfix1, idx, o) ->
  x25 _UU0393_ mfix0 mfix1 idx
    (let rec auxl _ _ = function
     | OnOne2_hd (hd, hd', tl, p) ->
       OnOne2_hd (hd, hd', tl, (p,
         (let (x30, _) = p in
          red1_ind_all _UU03a3_ x x0 x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12
            x13 x14 x15 x16 x17 x18 x19 x20 x21 x22 x23 x24 x25 x26 x27 x28
            _UU0393_ hd.dtype hd'.dtype x30)))
     | OnOne2_tl (hd, tl, tl', hl0) ->
       OnOne2_tl (hd, tl, tl', (auxl tl tl' hl0))
     in auxl mfix0 mfix1 o)
| Coq_fix_red_body (mfix0, mfix1, idx, o) ->
  x26 _UU0393_ mfix0 mfix1 idx
    (let c = fix_context mfix0 in
     let rec auxl _ _ = function
     | OnOne2_hd (hd, hd', tl, p) ->
       OnOne2_hd (hd, hd', tl, (p,
         (let (x30, _) = p in
          red1_ind_all _UU03a3_ x x0 x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12
            x13 x14 x15 x16 x17 x18 x19 x20 x21 x22 x23 x24 x25 x26 x27 x28
            (app_context _UU0393_ c) hd.dbody hd'.dbody x30)))
     | OnOne2_tl (hd, tl, tl', hl0) ->
       OnOne2_tl (hd, tl, tl', (auxl tl tl' hl0))
     in auxl mfix0 mfix1 o)
| Coq_cofix_red_ty (mfix0, mfix1, idx, o) ->
  x27 _UU0393_ mfix0 mfix1 idx
    (let rec auxl _ _ = function
     | OnOne2_hd (hd, hd', tl, p) ->
       OnOne2_hd (hd, hd', tl, (p,
         (let (x30, _) = p in
          red1_ind_all _UU03a3_ x x0 x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12
            x13 x14 x15 x16 x17 x18 x19 x20 x21 x22 x23 x24 x25 x26 x27 x28
            _UU0393_ hd.dtype hd'.dtype x30)))
     | OnOne2_tl (hd, tl, tl', hl0) ->
       OnOne2_tl (hd, tl, tl', (auxl tl tl' hl0))
     in auxl mfix0 mfix1 o)
| Coq_cofix_red_body (mfix0, mfix1, idx, o) ->
  x28 _UU0393_ mfix0 mfix1 idx
    (let c = fix_context mfix0 in
     let rec auxl _ _ = function
     | OnOne2_hd (hd, hd', tl, p) ->
       OnOne2_hd (hd, hd', tl, (p,
         (let (x30, _) = p in
          red1_ind_all _UU03a3_ x x0 x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12
            x13 x14 x15 x16 x17 x18 x19 x20 x21 x22 x23 x24 x25 x26 x27 x28
            (app_context _UU0393_ c) hd.dbody hd'.dbody x30)))
     | OnOne2_tl (hd, tl, tl', hl0) ->
       OnOne2_tl (hd, tl, tl', (auxl tl tl' hl0))
     in auxl mfix0 mfix1 o)

type red =
| Coq_refl_red
| Coq_trans_red of term * term * red * red1

(** val red_rect :
    global_env -> context -> term -> 'a1 -> (term -> term -> red -> 'a1 ->
    red1 -> 'a1) -> term -> red -> 'a1 **)

let rec red_rect _UU03a3_ _UU0393_ m f f0 _ = function
| Coq_refl_red -> f
| Coq_trans_red (p, n, r0, r1) ->
  f0 p n r0 (red_rect _UU03a3_ _UU0393_ m f f0 p r0) r1

(** val red_rec :
    global_env -> context -> term -> 'a1 -> (term -> term -> red -> 'a1 ->
    red1 -> 'a1) -> term -> red -> 'a1 **)

let rec red_rec _UU03a3_ _UU0393_ m f f0 _ = function
| Coq_refl_red -> f
| Coq_trans_red (p, n, r0, r1) ->
  f0 p n r0 (red_rec _UU03a3_ _UU0393_ m f f0 p r0) r1

(** val strip_casts : term -> term **)

let rec strip_casts t0 = match t0 with
| Coq_tEvar (ev, args) -> Coq_tEvar (ev, (map strip_casts args))
| Coq_tCast (c, _, _) -> strip_casts c
| Coq_tProd (na, a, b) -> Coq_tProd (na, (strip_casts a), (strip_casts b))
| Coq_tLambda (na, t1, m) ->
  Coq_tLambda (na, (strip_casts t1), (strip_casts m))
| Coq_tLetIn (na, b, t1, b') ->
  Coq_tLetIn (na, (strip_casts b), (strip_casts t1), (strip_casts b'))
| Coq_tApp (u, v) -> mkApps (strip_casts u) (map strip_casts v)
| Coq_tCase (ind, p, c, brs) ->
  let brs' = map (on_snd strip_casts) brs in
  Coq_tCase (ind, (strip_casts p), (strip_casts c), brs')
| Coq_tProj (p, c) -> Coq_tProj (p, (strip_casts c))
| Coq_tFix (mfix, idx) ->
  let mfix' = map (map_def strip_casts strip_casts) mfix in
  Coq_tFix (mfix', idx)
| Coq_tCoFix (mfix, idx) ->
  let mfix' = map (map_def strip_casts strip_casts) mfix in
  Coq_tCoFix (mfix', idx)
| _ -> t0

(** val destArity :
    context_decl list -> term -> (context_decl list * universe) option **)

let rec destArity _UU0393_ = function
| Coq_tSort s -> Some (_UU0393_, s)
| Coq_tProd (na, t1, b) -> destArity (snoc _UU0393_ (vass na t1)) b
| Coq_tLetIn (na, b, b_ty, b') ->
  destArity (snoc _UU0393_ (vdef na b b_ty)) b'
| _ -> None

(** val instantiate_params_subst :
    context_decl list -> term list -> term list -> term -> (term list * term)
    option **)

let rec instantiate_params_subst params pars s ty =
  match params with
  | [] -> (match pars with
           | [] -> Some (s, ty)
           | _ :: _ -> None)
  | d :: params0 ->
    (match d.decl_body with
     | Some b ->
       (match ty with
        | Coq_tLetIn (_, _, _, b') ->
          instantiate_params_subst params0 pars ((subst s O b) :: s) b'
        | _ -> None)
     | None ->
       (match ty with
        | Coq_tProd (_, _, b) ->
          (match pars with
           | [] -> None
           | hd :: tl -> instantiate_params_subst params0 tl (hd :: s) b)
        | _ -> None))

(** val instantiate_params :
    context_decl list -> term list -> term -> term option **)

let instantiate_params params pars ty =
  match instantiate_params_subst (List0.rev params) pars [] ty with
  | Some p -> let (s, ty0) = p in Some (subst s O ty0)
  | None -> None

(** val build_branches_type :
    inductive -> mutual_inductive_body -> one_inductive_body -> term list ->
    universe_instance -> term -> (nat * term) option list **)

let build_branches_type ind mdecl idecl params u p =
  let inds0 = inds ind.inductive_mind u mdecl.ind_bodies in
  let branch_type = fun i pat ->
    let (y, ar) = pat in
    let (_, t0) = y in
    let ty = subst inds0 O (subst_instance_constr u t0) in
    (match instantiate_params mdecl.ind_params params ty with
     | Some ty0 ->
       let (sign, ccl) = decompose_prod_assum [] ty0 in
       let nargs = length sign in
       let allargs = snd (decompose_app ccl) in
       let (paramrels, args) = chop mdecl.ind_npars allargs in
       let cstr = Coq_tConstruct (ind, i, u) in
       let args0 =
         app args
           ((mkApps cstr (app paramrels (to_extended_list sign))) :: [])
       in
       Some (ar, (it_mkProd_or_LetIn sign (mkApps (lift nargs O p) args0)))
     | None -> None)
  in
  mapi branch_type idecl.ind_ctors

(** val types_of_case :
    inductive -> mutual_inductive_body -> one_inductive_body -> term list ->
    universe_instance -> term -> term -> (((context_decl list * context_decl
    list) * universe) * (nat * term) list) option **)

let types_of_case ind mdecl idecl params u p pty =
  let brtys = build_branches_type ind mdecl idecl params u p in
  (match instantiate_params mdecl.ind_params params idecl.ind_type with
   | Some ity ->
     (match destArity [] ity with
      | Some p0 ->
        let (args, _) = p0 in
        (match destArity [] pty with
         | Some p1 ->
           let (args', s') = p1 in
           (match map_option_out brtys with
            | Some brtys0 -> Some (((args, args'), s'), brtys0)
            | None -> None)
         | None -> None)
      | None -> None)
   | None -> None)

(** val on_udecl_decl : (universes_decl -> 'a1) -> global_decl -> 'a1 **)

let on_udecl_decl f = function
| ConstantDecl (_, cb) -> f cb.cst_universes
| InductiveDecl (_, mb) -> f mb.ind_universes

(** val monomorphic_udecl_decl : global_decl -> ContextSet.t **)

let monomorphic_udecl_decl =
  on_udecl_decl monomorphic_udecl

(** val monomorphic_levels_decl : global_decl -> LevelSet.t **)

let monomorphic_levels_decl x =
  fst (monomorphic_udecl_decl x)

(** val monomorphic_constraints_decl : global_decl -> constraints **)

let monomorphic_constraints_decl x =
  snd (monomorphic_udecl_decl x)

(** val universes_decl_of_decl : global_decl -> universes_decl **)

let universes_decl_of_decl =
  on_udecl_decl (fun x -> x)

(** val coq_LevelSet_pair : LevelSet.elt -> LevelSet.elt -> LevelSet.t **)

let coq_LevelSet_pair x y =
  LevelSet.add y (LevelSet.singleton x)

(** val global_levels : global_env -> LevelSet.t **)

let global_levels _UU03a3_ =
  fold_right (fun decl lvls ->
    LevelSet.union (monomorphic_levels_decl decl) lvls)
    (coq_LevelSet_pair Level.Coq_lSet Level.Coq_lProp) _UU03a3_

(** val global_constraints : global_env -> constraints **)

let global_constraints _UU03a3_ =
  fold_right (fun decl ctrs ->
    ConstraintSet.union (monomorphic_constraints_decl decl) ctrs)
    ConstraintSet.empty _UU03a3_

(** val global_ext_levels : global_env_ext -> LevelSet.t **)

let global_ext_levels _UU03a3_ =
  LevelSet.union (levels_of_udecl (snd _UU03a3_))
    (global_levels (fst _UU03a3_))

(** val global_ext_constraints : global_env_ext -> constraints **)

let global_ext_constraints _UU03a3_ =
  ConstraintSet.union (constraints_of_udecl (snd _UU03a3_))
    (global_constraints (fst _UU03a3_))

(** val global_ext_uctx : global_env_ext -> ContextSet.t **)

let global_ext_uctx _UU03a3_ =
  ((global_ext_levels _UU03a3_), (global_ext_constraints _UU03a3_))

type cumul =
| Coq_cumul_refl of term * term
| Coq_cumul_red_l of term * term * term * red1 * cumul
| Coq_cumul_red_r of term * term * term * cumul * red1

(** val cumul_rect :
    checker_flags -> global_env_ext -> context -> (term -> term -> __ -> 'a1)
    -> (term -> term -> term -> red1 -> cumul -> 'a1 -> 'a1) -> (term -> term
    -> term -> cumul -> 'a1 -> red1 -> 'a1) -> term -> term -> cumul -> 'a1 **)

let rec cumul_rect h _UU03a3_ _UU0393_ f f0 f1 _ _ = function
| Coq_cumul_refl (t0, u) -> f t0 u __
| Coq_cumul_red_l (t0, u, v, r, c0) ->
  f0 t0 u v r c0 (cumul_rect h _UU03a3_ _UU0393_ f f0 f1 v u c0)
| Coq_cumul_red_r (t0, u, v, c0, r) ->
  f1 t0 u v c0 (cumul_rect h _UU03a3_ _UU0393_ f f0 f1 t0 v c0) r

(** val cumul_rec :
    checker_flags -> global_env_ext -> context -> (term -> term -> __ -> 'a1)
    -> (term -> term -> term -> red1 -> cumul -> 'a1 -> 'a1) -> (term -> term
    -> term -> cumul -> 'a1 -> red1 -> 'a1) -> term -> term -> cumul -> 'a1 **)

let rec cumul_rec h _UU03a3_ _UU0393_ f f0 f1 _ _ = function
| Coq_cumul_refl (t0, u) -> f t0 u __
| Coq_cumul_red_l (t0, u, v, r, c0) ->
  f0 t0 u v r c0 (cumul_rec h _UU03a3_ _UU0393_ f f0 f1 v u c0)
| Coq_cumul_red_r (t0, u, v, c0, r) ->
  f1 t0 u v c0 (cumul_rec h _UU03a3_ _UU0393_ f f0 f1 t0 v c0) r

type conv = cumul * cumul

(** val conv_refl :
    checker_flags -> global_env_ext -> context -> term -> conv **)

let conv_refl _ _ _ _ =
  todo ('c'::('o'::('n'::('v'::('_'::('r'::('e'::('f'::('l'::[])))))))))

(** val cumul_refl' :
    checker_flags -> global_env_ext -> context -> term -> cumul **)

let cumul_refl' _ _ _ _ =
  todo
    ('c'::('u'::('m'::('u'::('l'::('_'::('r'::('e'::('f'::('l'::('\''::[])))))))))))

(** val cumul_trans :
    checker_flags -> global_env_ext -> context -> term -> term -> term ->
    cumul -> cumul -> cumul **)

let cumul_trans _ _ _ _ _ _ _ _ =
  todo
    ('c'::('u'::('m'::('u'::('l'::('_'::('t'::('r'::('a'::('n'::('s'::[])))))))))))

(** val congr_cumul_prod :
    checker_flags -> global_env_ext -> context -> name -> name -> term ->
    term -> term -> term -> cumul -> cumul -> cumul **)

let congr_cumul_prod _ _ _ _ _ _ _ _ _ _ _ =
  todo
    ('c'::('o'::('n'::('g'::('r'::('_'::('c'::('u'::('m'::('u'::('l'::('_'::('p'::('r'::('o'::('d'::[]))))))))))))))))

type eq_context = (context_decl, context_decl, __) coq_All2

type check_correct_arity = eq_context

type 'typing coq_All_local_env =
| Coq_localenv_nil
| Coq_localenv_cons_abs of context * name * term * 'typing coq_All_local_env
   * 'typing
| Coq_localenv_cons_def of context * name * term * term
   * 'typing coq_All_local_env * 'typing * 'typing

(** val coq_All_local_env_rect :
    'a2 -> (context -> name -> term -> 'a1 coq_All_local_env -> 'a2 -> 'a1 ->
    'a2) -> (context -> name -> term -> term -> 'a1 coq_All_local_env -> 'a2
    -> 'a1 -> 'a1 -> 'a2) -> context -> 'a1 coq_All_local_env -> 'a2 **)

let rec coq_All_local_env_rect f f0 f1 _ = function
| Coq_localenv_nil -> f
| Coq_localenv_cons_abs (_UU0393_, na, t0, a0, t1) ->
  f0 _UU0393_ na t0 a0 (coq_All_local_env_rect f f0 f1 _UU0393_ a0) t1
| Coq_localenv_cons_def (_UU0393_, na, b, t0, a0, t1, t2) ->
  f1 _UU0393_ na b t0 a0 (coq_All_local_env_rect f f0 f1 _UU0393_ a0) t1 t2

(** val coq_All_local_env_rec :
    'a2 -> (context -> name -> term -> 'a1 coq_All_local_env -> 'a2 -> 'a1 ->
    'a2) -> (context -> name -> term -> term -> 'a1 coq_All_local_env -> 'a2
    -> 'a1 -> 'a1 -> 'a2) -> context -> 'a1 coq_All_local_env -> 'a2 **)

let rec coq_All_local_env_rec f f0 f1 _ = function
| Coq_localenv_nil -> f
| Coq_localenv_cons_abs (_UU0393_, na, t0, a0, t1) ->
  f0 _UU0393_ na t0 a0 (coq_All_local_env_rec f f0 f1 _UU0393_ a0) t1
| Coq_localenv_cons_def (_UU0393_, na, b, t0, a0, t1, t2) ->
  f1 _UU0393_ na b t0 a0 (coq_All_local_env_rec f f0 f1 _UU0393_ a0) t1 t2

type 'p lift_typing = __

type 'p on_local_decl = __

type ('typing, 'property) coq_All_local_env_over =
| Coq_localenv_over_nil
| Coq_localenv_over_cons_abs of context * name * term
   * 'typing lift_typing coq_All_local_env
   * ('typing, 'property) coq_All_local_env_over * (universe, 'typing) sigT
   * 'property
| Coq_localenv_over_cons_def of context * name * term * term
   * 'typing lift_typing coq_All_local_env * 'typing
   * ('typing, 'property) coq_All_local_env_over * 'property
   * (universe, 'typing) sigT * 'property

(** val coq_All_local_env_over_rect :
    global_env_ext -> 'a3 -> (context -> name -> term -> 'a1 lift_typing
    coq_All_local_env -> ('a1, 'a2) coq_All_local_env_over -> 'a3 ->
    (universe, 'a1) sigT -> 'a2 -> 'a3) -> (context -> name -> term -> term
    -> 'a1 lift_typing coq_All_local_env -> 'a1 -> ('a1, 'a2)
    coq_All_local_env_over -> 'a3 -> 'a2 -> (universe, 'a1) sigT -> 'a2 ->
    'a3) -> context -> 'a1 lift_typing coq_All_local_env -> ('a1, 'a2)
    coq_All_local_env_over -> 'a3 **)

let rec coq_All_local_env_over_rect _UU03a3_ f f0 f1 _ _ = function
| Coq_localenv_over_nil -> f
| Coq_localenv_over_cons_abs (_UU0393_, na, t0, all, a0, tu, p) ->
  f0 _UU0393_ na t0 all a0
    (coq_All_local_env_over_rect _UU03a3_ f f0 f1 _UU0393_ all a0) tu p
| Coq_localenv_over_cons_def (_UU0393_, na, b, t0, all, tb, a0, p, tu, p0) ->
  f1 _UU0393_ na b t0 all tb a0
    (coq_All_local_env_over_rect _UU03a3_ f f0 f1 _UU0393_ all a0) p tu p0

(** val coq_All_local_env_over_rec :
    global_env_ext -> 'a3 -> (context -> name -> term -> 'a1 lift_typing
    coq_All_local_env -> ('a1, 'a2) coq_All_local_env_over -> 'a3 ->
    (universe, 'a1) sigT -> 'a2 -> 'a3) -> (context -> name -> term -> term
    -> 'a1 lift_typing coq_All_local_env -> 'a1 -> ('a1, 'a2)
    coq_All_local_env_over -> 'a3 -> 'a2 -> (universe, 'a1) sigT -> 'a2 ->
    'a3) -> context -> 'a1 lift_typing coq_All_local_env -> ('a1, 'a2)
    coq_All_local_env_over -> 'a3 **)

let rec coq_All_local_env_over_rec _UU03a3_ f f0 f1 _ _ = function
| Coq_localenv_over_nil -> f
| Coq_localenv_over_cons_abs (_UU0393_, na, t0, all, a0, tu, p) ->
  f0 _UU0393_ na t0 all a0
    (coq_All_local_env_over_rec _UU03a3_ f f0 f1 _UU0393_ all a0) tu p
| Coq_localenv_over_cons_def (_UU0393_, na, b, t0, all, tb, a0, p, tu, p0) ->
  f1 _UU0393_ na b t0 all tb a0
    (coq_All_local_env_over_rec _UU03a3_ f f0 f1 _UU0393_ all a0) p tu p0

type 'typing isWfArity =
  (context_decl list, (universe, __ * 'typing lift_typing coq_All_local_env)
  sigT) sigT

type ('typing, 'property) isWfArity_prop =
  ('typing isWfArity, ('typing, 'property) coq_All_local_env_over) sigT

(** val fix_guard : term mfixpoint -> bool **)

let fix_guard = fun m -> assert false

(** val ind_guard : mutual_inductive_body -> bool **)

let ind_guard = fun m -> assert false

type typing =
| Coq_type_Rel of nat * context_decl * typing lift_typing coq_All_local_env
| Coq_type_Sort of LevelSet.elt * typing lift_typing coq_All_local_env
| Coq_type_Cast of term * cast_kind * term * universe * typing * typing
| Coq_type_Prod of name * term * term * universe * universe * typing * typing
| Coq_type_Lambda of name * term * term * universe * term * typing * typing
| Coq_type_LetIn of name * term * term * term * universe * term * typing
   * typing * typing
| Coq_type_App of term * term list * term * term * typing * typing_spine
| Coq_type_Const of ident * universe_instance
   * typing lift_typing coq_All_local_env * constant_body
| Coq_type_Ind of inductive * universe_instance
   * typing lift_typing coq_All_local_env * mutual_inductive_body
   * one_inductive_body
| Coq_type_Construct of inductive * nat * universe_instance
   * typing lift_typing coq_All_local_env * mutual_inductive_body
   * one_inductive_body * ((ident * term) * nat)
| Coq_type_Case of inductive * universe_instance * nat * term * term
   * (nat * term) list * term list * mutual_inductive_body
   * one_inductive_body * term * typing * context_decl list
   * context_decl list * universe * (nat * term) list * check_correct_arity
   * typing * (nat * term, nat * term, __ * typing) coq_All2
| Coq_type_Proj of projection * term * universe_instance
   * mutual_inductive_body * one_inductive_body * (ident * term) * term list
   * typing
| Coq_type_Fix of term mfixpoint * nat * term def
   * typing lift_typing coq_All_local_env * (term def, typing * __) coq_All
| Coq_type_CoFix of term mfixpoint * nat * term def
   * typing lift_typing coq_All_local_env * (term def, typing) coq_All
| Coq_type_Conv of term * term * term * typing
   * (typing isWfArity, (universe, typing) sigT) sum * cumul
and typing_spine =
| Coq_type_spine_nil of term
| Coq_type_spine_cons of term * term list * name * term * term * universe
   * term * term * typing * cumul * typing * typing_spine

(** val typing_rect :
    checker_flags -> global_env_ext -> (context -> nat -> context_decl ->
    typing lift_typing coq_All_local_env -> __ -> 'a1) -> (context ->
    LevelSet.elt -> typing lift_typing coq_All_local_env -> __ -> 'a1) ->
    (context -> term -> cast_kind -> term -> universe -> typing -> 'a1 ->
    typing -> 'a1 -> 'a1) -> (context -> name -> term -> term -> universe ->
    universe -> typing -> 'a1 -> typing -> 'a1 -> 'a1) -> (context -> name ->
    term -> term -> universe -> term -> typing -> 'a1 -> typing -> 'a1 ->
    'a1) -> (context -> name -> term -> term -> term -> universe -> term ->
    typing -> 'a1 -> typing -> 'a1 -> typing -> 'a1 -> 'a1) -> (context ->
    term -> term list -> term -> term -> typing -> 'a1 -> __ -> __ ->
    typing_spine -> 'a1) -> (context -> ident -> universe_instance -> typing
    lift_typing coq_All_local_env -> constant_body -> __ -> __ -> 'a1) ->
    (context -> inductive -> universe_instance -> typing lift_typing
    coq_All_local_env -> mutual_inductive_body -> one_inductive_body -> __ ->
    __ -> 'a1) -> (context -> inductive -> nat -> universe_instance -> typing
    lift_typing coq_All_local_env -> mutual_inductive_body ->
    one_inductive_body -> ((ident * term) * nat) -> __ -> __ -> 'a1) ->
    (context -> inductive -> universe_instance -> nat -> term -> term ->
    (nat * term) list -> term list -> mutual_inductive_body ->
    one_inductive_body -> __ -> __ -> term -> typing -> 'a1 -> context_decl
    list -> context_decl list -> universe -> (nat * term) list -> __ ->
    check_correct_arity -> __ -> typing -> 'a1 -> (nat * term, nat * term,
    __ * typing) coq_All2 -> 'a1) -> (context -> projection -> term ->
    universe_instance -> mutual_inductive_body -> one_inductive_body ->
    (ident * term) -> __ -> term list -> typing -> 'a1 -> __ -> 'a1) ->
    (context -> term mfixpoint -> nat -> term def -> __ -> __ -> typing
    lift_typing coq_All_local_env -> (term def, typing * __) coq_All -> 'a1)
    -> (context -> term mfixpoint -> nat -> term def -> __ -> __ -> typing
    lift_typing coq_All_local_env -> (term def, typing) coq_All -> 'a1) ->
    (context -> term -> term -> term -> typing -> 'a1 -> (typing isWfArity,
    (universe, typing) sigT) sum -> cumul -> 'a1) -> context -> term -> term
    -> typing -> 'a1 **)

let rec typing_rect h _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13 _UU0393_ _ _ = function
| Coq_type_Rel (n, decl, a) -> f _UU0393_ n decl a __
| Coq_type_Sort (l, a) -> f0 _UU0393_ l a __
| Coq_type_Cast (c, k, t1, s, t2, t3) ->
  f1 _UU0393_ c k t1 s t2
    (typing_rect h _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13
      _UU0393_ t1 (Coq_tSort s) t2) t3
    (typing_rect h _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13
      _UU0393_ c t1 t3)
| Coq_type_Prod (n, t1, b, s1, s2, t2, t3) ->
  f2 _UU0393_ n t1 b s1 s2 t2
    (typing_rect h _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13
      _UU0393_ t1 (Coq_tSort s1) t2) t3
    (typing_rect h _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13
      (snoc _UU0393_ (vass n t1)) b (Coq_tSort s2) t3)
| Coq_type_Lambda (n, t1, b, s1, bty, t2, t3) ->
  f3 _UU0393_ n t1 b s1 bty t2
    (typing_rect h _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13
      _UU0393_ t1 (Coq_tSort s1) t2) t3
    (typing_rect h _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13
      (snoc _UU0393_ (vass n t1)) b bty t3)
| Coq_type_LetIn (n, b, b_ty, b', s1, b'_ty, t1, t2, t3) ->
  f4 _UU0393_ n b b_ty b' s1 b'_ty t1
    (typing_rect h _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13
      _UU0393_ b_ty (Coq_tSort s1) t1) t2
    (typing_rect h _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13
      _UU0393_ b b_ty t2) t3
    (typing_rect h _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13
      (snoc _UU0393_ (vdef n b b_ty)) b' b'_ty t3)
| Coq_type_App (t1, l, t_ty, t', t2, t3) ->
  f5 _UU0393_ t1 l t_ty t' t2
    (typing_rect h _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13
      _UU0393_ t1 t_ty t2) __ __ t3
| Coq_type_Const (cst, u, a, decl) -> f6 _UU0393_ cst u a decl __ __
| Coq_type_Ind (ind, u, a, mdecl, idecl) ->
  f7 _UU0393_ ind u a mdecl idecl __ __
| Coq_type_Construct (ind, i, u, a, mdecl, idecl, cdecl) ->
  f8 _UU0393_ ind i u a mdecl idecl cdecl __ __
| Coq_type_Case (ind, u, npar, p, c, brs, args, mdecl, idecl, x, x0, x1, x2,
                 x3, x4, x5, x6, x7) ->
  f9 _UU0393_ ind u npar p c brs args mdecl idecl __ __ x x0
    (typing_rect h _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13
      _UU0393_ p x x0) x1 x2 x3 x4 __ x5 __ x6
    (typing_rect h _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13
      _UU0393_ c (mkApps (Coq_tInd (ind, u)) args) x6) x7
| Coq_type_Proj (p, c, u, mdecl, idecl, pdecl, args, t1) ->
  f10 _UU0393_ p c u mdecl idecl pdecl __ args t1
    (typing_rect h _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13
      _UU0393_ c (mkApps (Coq_tInd ((fst (fst p)), u)) args) t1) __
| Coq_type_Fix (mfix, n, decl, x, x0) -> f11 _UU0393_ mfix n decl __ __ x x0
| Coq_type_CoFix (mfix, n, decl, x, x0) -> f12 _UU0393_ mfix n decl __ __ x x0
| Coq_type_Conv (t1, a, b, t2, s, c) ->
  f13 _UU0393_ t1 a b t2
    (typing_rect h _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13
      _UU0393_ t1 a t2) s c

(** val typing_rec :
    checker_flags -> global_env_ext -> (context -> nat -> context_decl ->
    typing lift_typing coq_All_local_env -> __ -> 'a1) -> (context ->
    LevelSet.elt -> typing lift_typing coq_All_local_env -> __ -> 'a1) ->
    (context -> term -> cast_kind -> term -> universe -> typing -> 'a1 ->
    typing -> 'a1 -> 'a1) -> (context -> name -> term -> term -> universe ->
    universe -> typing -> 'a1 -> typing -> 'a1 -> 'a1) -> (context -> name ->
    term -> term -> universe -> term -> typing -> 'a1 -> typing -> 'a1 ->
    'a1) -> (context -> name -> term -> term -> term -> universe -> term ->
    typing -> 'a1 -> typing -> 'a1 -> typing -> 'a1 -> 'a1) -> (context ->
    term -> term list -> term -> term -> typing -> 'a1 -> __ -> __ ->
    typing_spine -> 'a1) -> (context -> ident -> universe_instance -> typing
    lift_typing coq_All_local_env -> constant_body -> __ -> __ -> 'a1) ->
    (context -> inductive -> universe_instance -> typing lift_typing
    coq_All_local_env -> mutual_inductive_body -> one_inductive_body -> __ ->
    __ -> 'a1) -> (context -> inductive -> nat -> universe_instance -> typing
    lift_typing coq_All_local_env -> mutual_inductive_body ->
    one_inductive_body -> ((ident * term) * nat) -> __ -> __ -> 'a1) ->
    (context -> inductive -> universe_instance -> nat -> term -> term ->
    (nat * term) list -> term list -> mutual_inductive_body ->
    one_inductive_body -> __ -> __ -> term -> typing -> 'a1 -> context_decl
    list -> context_decl list -> universe -> (nat * term) list -> __ ->
    check_correct_arity -> __ -> typing -> 'a1 -> (nat * term, nat * term,
    __ * typing) coq_All2 -> 'a1) -> (context -> projection -> term ->
    universe_instance -> mutual_inductive_body -> one_inductive_body ->
    (ident * term) -> __ -> term list -> typing -> 'a1 -> __ -> 'a1) ->
    (context -> term mfixpoint -> nat -> term def -> __ -> __ -> typing
    lift_typing coq_All_local_env -> (term def, typing * __) coq_All -> 'a1)
    -> (context -> term mfixpoint -> nat -> term def -> __ -> __ -> typing
    lift_typing coq_All_local_env -> (term def, typing) coq_All -> 'a1) ->
    (context -> term -> term -> term -> typing -> 'a1 -> (typing isWfArity,
    (universe, typing) sigT) sum -> cumul -> 'a1) -> context -> term -> term
    -> typing -> 'a1 **)

let rec typing_rec h _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13 _UU0393_ _ _ = function
| Coq_type_Rel (n, decl, a) -> f _UU0393_ n decl a __
| Coq_type_Sort (l, a) -> f0 _UU0393_ l a __
| Coq_type_Cast (c, k, t1, s, t2, t3) ->
  f1 _UU0393_ c k t1 s t2
    (typing_rec h _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13
      _UU0393_ t1 (Coq_tSort s) t2) t3
    (typing_rec h _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13
      _UU0393_ c t1 t3)
| Coq_type_Prod (n, t1, b, s1, s2, t2, t3) ->
  f2 _UU0393_ n t1 b s1 s2 t2
    (typing_rec h _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13
      _UU0393_ t1 (Coq_tSort s1) t2) t3
    (typing_rec h _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13
      (snoc _UU0393_ (vass n t1)) b (Coq_tSort s2) t3)
| Coq_type_Lambda (n, t1, b, s1, bty, t2, t3) ->
  f3 _UU0393_ n t1 b s1 bty t2
    (typing_rec h _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13
      _UU0393_ t1 (Coq_tSort s1) t2) t3
    (typing_rec h _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13
      (snoc _UU0393_ (vass n t1)) b bty t3)
| Coq_type_LetIn (n, b, b_ty, b', s1, b'_ty, t1, t2, t3) ->
  f4 _UU0393_ n b b_ty b' s1 b'_ty t1
    (typing_rec h _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13
      _UU0393_ b_ty (Coq_tSort s1) t1) t2
    (typing_rec h _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13
      _UU0393_ b b_ty t2) t3
    (typing_rec h _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13
      (snoc _UU0393_ (vdef n b b_ty)) b' b'_ty t3)
| Coq_type_App (t1, l, t_ty, t', t2, t3) ->
  f5 _UU0393_ t1 l t_ty t' t2
    (typing_rec h _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13
      _UU0393_ t1 t_ty t2) __ __ t3
| Coq_type_Const (cst, u, a, decl) -> f6 _UU0393_ cst u a decl __ __
| Coq_type_Ind (ind, u, a, mdecl, idecl) ->
  f7 _UU0393_ ind u a mdecl idecl __ __
| Coq_type_Construct (ind, i, u, a, mdecl, idecl, cdecl) ->
  f8 _UU0393_ ind i u a mdecl idecl cdecl __ __
| Coq_type_Case (ind, u, npar, p, c, brs, args, mdecl, idecl, x, x0, x1, x2,
                 x3, x4, x5, x6, x7) ->
  f9 _UU0393_ ind u npar p c brs args mdecl idecl __ __ x x0
    (typing_rec h _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13
      _UU0393_ p x x0) x1 x2 x3 x4 __ x5 __ x6
    (typing_rec h _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13
      _UU0393_ c (mkApps (Coq_tInd (ind, u)) args) x6) x7
| Coq_type_Proj (p, c, u, mdecl, idecl, pdecl, args, t1) ->
  f10 _UU0393_ p c u mdecl idecl pdecl __ args t1
    (typing_rec h _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13
      _UU0393_ c (mkApps (Coq_tInd ((fst (fst p)), u)) args) t1) __
| Coq_type_Fix (mfix, n, decl, x, x0) -> f11 _UU0393_ mfix n decl __ __ x x0
| Coq_type_CoFix (mfix, n, decl, x, x0) -> f12 _UU0393_ mfix n decl __ __ x x0
| Coq_type_Conv (t1, a, b, t2, s, c) ->
  f13 _UU0393_ t1 a b t2
    (typing_rec h _UU03a3_ f f0 f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 f13
      _UU0393_ t1 a t2) s c

(** val typing_spine_rect :
    checker_flags -> global_env_ext -> (context -> term -> 'a1) -> (context
    -> term -> term list -> name -> term -> term -> universe -> term -> term
    -> typing -> cumul -> typing -> typing_spine -> 'a1 -> 'a1) -> context ->
    term -> term list -> term -> typing_spine -> 'a1 **)

let rec typing_spine_rect h _UU03a3_ f f0 _UU0393_ _ _ _ = function
| Coq_type_spine_nil ty -> f _UU0393_ ty
| Coq_type_spine_cons (hd, tl, na, a, b, s, t1, b', t2, c, t3, t4) ->
  f0 _UU0393_ hd tl na a b s t1 b' t2 c t3 t4
    (typing_spine_rect h _UU03a3_ f f0 _UU0393_ (subst1 hd O b) tl b' t4)

(** val typing_spine_rec :
    checker_flags -> global_env_ext -> (context -> term -> 'a1) -> (context
    -> term -> term list -> name -> term -> term -> universe -> term -> term
    -> typing -> cumul -> typing -> typing_spine -> 'a1 -> 'a1) -> context ->
    term -> term list -> term -> typing_spine -> 'a1 **)

let rec typing_spine_rec h _UU03a3_ f f0 _UU0393_ _ _ _ = function
| Coq_type_spine_nil ty -> f _UU0393_ ty
| Coq_type_spine_cons (hd, tl, na, a, b, s, t1, b', t2, c, t3, t4) ->
  f0 _UU0393_ hd tl na a b s t1 b' t2 c t3 t4
    (typing_spine_rec h _UU03a3_ f f0 _UU0393_ (subst1 hd O b) tl b' t4)

type isType = (universe, typing) sigT

type 'p unlift_opt_pred = 'p

type 'p type_local_ctx = __

type 'p on_type = 'p

type constructor_shape = { cshape_args : context;
                           cshape_args_univ : universe;
                           cshape_indices : term list }

(** val cshape_args :
    mutual_inductive_body -> nat -> one_inductive_body ->
    ((ident * term) * nat) -> constructor_shape -> context **)

let cshape_args _ _ _ _ x = x.cshape_args

(** val cshape_args_univ :
    mutual_inductive_body -> nat -> one_inductive_body ->
    ((ident * term) * nat) -> constructor_shape -> universe **)

let cshape_args_univ _ _ _ _ x = x.cshape_args_univ

(** val cshape_concl_head :
    mutual_inductive_body -> nat -> one_inductive_body ->
    ((ident * term) * nat) -> constructor_shape -> term **)

let cshape_concl_head mdecl i _ _ c =
  Coq_tRel
    (add
      (add (sub (length mdecl.ind_bodies) (S i)) (length mdecl.ind_params))
      (length c.cshape_args))

(** val cshape_indices :
    mutual_inductive_body -> nat -> one_inductive_body ->
    ((ident * term) * nat) -> constructor_shape -> term list **)

let cshape_indices _ _ _ _ x = x.cshape_indices

type 'p on_constructor =
  'p on_type * (constructor_shape, 'p type_local_ctx) sigT

type 'p on_constructors = ((ident * term) * nat, 'p on_constructor) coq_Alli

type 'p on_projection = 'p on_type

type 'p on_projections =
  (ident * term, 'p on_projection) coq_Alli
  (* singleton inductive, whose constructor was Build_on_projections *)

(** val on_projs :
    global_env_ext -> kername -> mutual_inductive_body -> nat ->
    one_inductive_body -> context -> (ident * term) list -> 'a1
    on_projections -> (ident * term, 'a1 on_projection) coq_Alli **)

let on_projs _ _ _ _ _ _ _ o =
  o

(** val is_prop_elim : sort_family list -> bool **)

let is_prop_elim = function
| [] -> false
| y :: l0 ->
  (match y with
   | InProp -> (match l0 with
                | [] -> true
                | _ :: _ -> false)
   | _ -> false)

(** val is_all_elim : sort_family list -> bool **)

let is_all_elim = function
| [] -> false
| y :: l0 ->
  (match y with
   | InProp ->
     (match l0 with
      | [] -> false
      | s :: l1 ->
        (match s with
         | InSet ->
           (match l1 with
            | [] -> false
            | s0 :: l2 ->
              (match s0 with
               | InType -> (match l2 with
                            | [] -> true
                            | _ :: _ -> false)
               | _ -> false))
         | _ -> false))
   | _ -> false)

(** val leb_sort_family : sort_family -> sort_family -> bool **)

let leb_sort_family x y =
  match x with
  | InProp -> true
  | InSet -> (match y with
              | InProp -> false
              | _ -> true)
  | InType -> (match y with
               | InType -> true
               | _ -> false)

type 'p on_ind_body = { ind_indices : context; ind_sort : universe;
                        onArity : 'p on_type;
                        onConstructors : 'p on_constructors;
                        onProjections : (__ -> 'p on_projections) }

(** val ind_indices :
    checker_flags -> global_env_ext -> kername -> mutual_inductive_body ->
    nat -> one_inductive_body -> 'a1 on_ind_body -> context **)

let ind_indices _ _ _ _ _ _ x = x.ind_indices

(** val ind_sort :
    checker_flags -> global_env_ext -> kername -> mutual_inductive_body ->
    nat -> one_inductive_body -> 'a1 on_ind_body -> universe **)

let ind_sort _ _ _ _ _ _ x = x.ind_sort

(** val onArity :
    checker_flags -> global_env_ext -> kername -> mutual_inductive_body ->
    nat -> one_inductive_body -> 'a1 on_ind_body -> 'a1 on_type **)

let onArity _ _ _ _ _ _ x = x.onArity

(** val onConstructors :
    checker_flags -> global_env_ext -> kername -> mutual_inductive_body ->
    nat -> one_inductive_body -> 'a1 on_ind_body -> 'a1 on_constructors **)

let onConstructors _ _ _ _ _ _ x = x.onConstructors

(** val onProjections :
    checker_flags -> global_env_ext -> kername -> mutual_inductive_body ->
    nat -> one_inductive_body -> 'a1 on_ind_body -> 'a1 on_projections **)

let onProjections _ _ _ _ _ _ o =
  o.onProjections __

type 'p on_context = 'p coq_All_local_env

type 'p on_inductive = { onInductives : (one_inductive_body, 'p on_ind_body)
                                        coq_Alli; onParams : 'p on_context }

(** val onInductives :
    checker_flags -> global_env_ext -> kername -> mutual_inductive_body ->
    'a1 on_inductive -> (one_inductive_body, 'a1 on_ind_body) coq_Alli **)

let onInductives _ _ _ _ x = x.onInductives

(** val onParams :
    checker_flags -> global_env_ext -> kername -> mutual_inductive_body ->
    'a1 on_inductive -> 'a1 on_context **)

let onParams _ _ _ _ x = x.onParams

type 'p on_constant_decl = __

type 'p on_global_decl = __

type 'p on_global_env =
| Coq_globenv_nil
| Coq_globenv_decl of global_env * global_decl * 'p on_global_env
   * 'p on_global_decl

(** val on_global_env_rect :
    checker_flags -> checker_flags -> 'a2 -> (global_env -> global_decl ->
    'a1 on_global_env -> 'a2 -> __ -> __ -> 'a1 on_global_decl -> 'a2) ->
    global_env -> 'a1 on_global_env -> 'a2 **)

let rec on_global_env_rect cf h f f0 _ = function
| Coq_globenv_nil -> f
| Coq_globenv_decl (_UU03a3_, d, o0, x) ->
  f0 _UU03a3_ d o0 (on_global_env_rect cf h f f0 _UU03a3_ o0) __ __ x

(** val on_global_env_rec :
    checker_flags -> checker_flags -> 'a2 -> (global_env -> global_decl ->
    'a1 on_global_env -> 'a2 -> __ -> __ -> 'a1 on_global_decl -> 'a2) ->
    global_env -> 'a1 on_global_env -> 'a2 **)

let rec on_global_env_rec cf h f f0 _ = function
| Coq_globenv_nil -> f
| Coq_globenv_decl (_UU03a3_, d, o0, x) ->
  f0 _UU03a3_ d o0 (on_global_env_rec cf h f f0 _UU03a3_ o0) __ __ x

type 'p on_global_env_ext = 'p on_global_env * __

type 'p sort_irrelevant =
  global_env_ext -> context -> term option -> universe -> universe -> 'p -> 'p

(** val coq_All_local_env_impl :
    context -> 'a1 coq_All_local_env -> (context -> term -> term option ->
    'a1 -> 'a2) -> 'a2 coq_All_local_env **)

let rec coq_All_local_env_impl _ a x =
  match a with
  | Coq_localenv_nil -> Coq_localenv_nil
  | Coq_localenv_cons_abs (_UU0393_, na, t0, a0, t1) ->
    Coq_localenv_cons_abs (_UU0393_, na, t0,
      (coq_All_local_env_impl _UU0393_ a0 x), (x _UU0393_ t0 None t1))
  | Coq_localenv_cons_def (_UU0393_, na, b, t0, a0, t1, t2) ->
    Coq_localenv_cons_def (_UU0393_, na, b, t0,
      (coq_All_local_env_impl _UU0393_ a0 x), (x _UU0393_ t0 None t1),
      (x _UU0393_ b (Some t0) t2))

(** val on_global_env_mix :
    checker_flags -> global_env -> 'a2 sort_irrelevant -> 'a1 on_global_env
    -> 'a2 on_global_env -> ('a1 * 'a2) on_global_env **)

let on_global_env_mix _ _ =
  todo
    ('s'::('i'::('m'::('p'::('l'::('i'::('f'::('y'::(' '::('o'::('n'::('C'::('o'::('n'::('s'::('t'::('r'::('u'::('c'::('t'::('o'::('r'::('s'::[])))))))))))))))))))))))

(** val coq_Alli_impl_trans :
    'a1 list -> nat -> ('a1, 'a2) coq_Alli -> (nat -> 'a1 -> 'a2 -> 'a3) ->
    ('a1, 'a3) coq_Alli **)

let rec coq_Alli_impl_trans _ n x x0 =
  match x with
  | Alli_nil -> Alli_nil
  | Alli_cons (hd, tl, y, a) ->
    Alli_cons (hd, tl, (x0 n hd y), (coq_Alli_impl_trans tl (S n) a x0))

(** val on_global_env_impl :
    checker_flags -> global_env -> ((global_env * universes_decl) -> context
    -> term -> term option -> 'a1 on_global_env -> 'a1 -> 'a2) -> 'a1
    on_global_env -> 'a2 on_global_env **)

let on_global_env_impl h _UU03a3_ x x0 =
  let on_global_env_impl_subproof = fun _ x1 _UU03a3_0 m x2 _ _ _ x3 c ->
    let rec f = function
    | [] -> Obj.magic __
    | y :: l0 ->
      let { decl_name = _; decl_body = decl_body0; decl_type = x4 } = y in
      (match decl_body0 with
       | Some _ -> f l0
       | None ->
         (fun t0 -> ((Obj.magic f l0 (let (x5, _) = t0 in x5)),
           (x1 (_UU03a3_0, m.ind_universes)
             (app_context
               (app_context (arities_context m.ind_bodies) m.ind_params) l0)
             x4 (Some (Coq_tSort x3.cshape_args_univ)) x2
             (let (_, x5) = t0 in x5)))))
    in f c
  in
  let rec f _ = function
  | Coq_globenv_nil -> Coq_globenv_nil
  | Coq_globenv_decl (_UU03a3_0, d, o0, x1) ->
    Coq_globenv_decl (_UU03a3_0, d, (f _UU03a3_0 o0),
      (match d with
       | ConstantDecl (_, c) ->
         let { cst_type = cst_type0; cst_body = cst_body0; cst_universes =
           cst_universes0 } = c
         in
         (match cst_body0 with
          | Some t0 ->
            Obj.magic x (_UU03a3_0, cst_universes0) [] t0 (Some cst_type0) o0
              x1
          | None ->
            Obj.magic x (_UU03a3_0, cst_universes0) [] cst_type0 None o0 x1)
       | InductiveDecl (k, m) ->
         let udecl = universes_decl_of_decl (InductiveDecl (k, m)) in
         let { onInductives = onI; onParams = onP } = Obj.magic x1 in
         Obj.magic { onInductives =
           (coq_Alli_impl m.ind_bodies O onI (fun n x2 x3 -> { ind_indices =
             x3.ind_indices; ind_sort = x3.ind_sort; onArity =
             (let x4 = x3.onArity in
              x (_UU03a3_0, m.ind_universes) [] x2.ind_type None o0 x4);
             onConstructors =
             (let x4 = x3.onConstructors in
              coq_Alli_impl_trans x2.ind_ctors O (Obj.magic x4)
                (fun _ x5 x6 ->
                let (p, s) = x6 in
                let Coq_existT (x7, t0) = s in
                ((x (_UU03a3_0, m.ind_universes)
                   (arities_context m.ind_bodies) (snd (fst x5)) None o0 p),
                (Coq_existT (x7,
                (Obj.magic on_global_env_impl_subproof h x _UU03a3_0 m o0 n
                  x2 x5 x7 x7.cshape_args t0)))))); onProjections = (fun _ ->
             let o1 = onProjections h (_UU03a3_0, udecl) k m n x2 x3 in
             coq_Alli_impl x2.ind_projs O o1 (fun _ x4 x5 ->
               x (_UU03a3_0, m.ind_universes)
                 (snoc (smash_context [] m.ind_params)
                   (vass (Coq_nNamed x2.ind_name)
                     (mkApps (Coq_tInd ({ inductive_mind = k; inductive_ind =
                       n }, (polymorphic_instance m.ind_universes)))
                       (to_extended_list (smash_context [] m.ind_params)))))
                 (snd x4) None o0 x5)) })); onParams =
           (coq_All_local_env_impl m.ind_params onP (fun _UU0393_ t0 t1 x2 ->
             x (_UU03a3_0, m.ind_universes) _UU0393_ t0 t1 o0 x2)) }))
  in f _UU03a3_ x0

(** val on_global_env_proj :
    checker_flags -> global_env -> ('a1 * 'a2) on_global_env -> 'a1
    on_global_env **)

let on_global_env_proj h _UU03a3_ x =
  on_global_env_impl h _UU03a3_ (fun _ _ _ _ _ x1 -> let (x0, _) = x1 in x0) x

type 'p coq_Forall_decls_typing = 'p lift_typing on_global_env

type type_local_decl = __

(** val refine_type :
    checker_flags -> global_env_ext -> context -> term -> term -> term ->
    typing -> typing **)

let refine_type _ _ _ _ _ _ ht =
  ht

type 'p coq_Forall_typing_spine =
| Forall_type_spine_nil of term
| Forall_type_spine_cons of term * term list * name * term * term * universe
   * term * term * typing_spine * typing * cumul * typing * 'p * 'p
   * 'p coq_Forall_typing_spine

(** val coq_Forall_typing_spine_rect :
    checker_flags -> global_env_ext -> context -> (term -> 'a2) -> (term ->
    term list -> name -> term -> term -> universe -> term -> term ->
    typing_spine -> typing -> cumul -> typing -> 'a1 -> 'a1 -> 'a1
    coq_Forall_typing_spine -> 'a2 -> 'a2) -> term -> term list -> term ->
    typing_spine -> 'a1 coq_Forall_typing_spine -> 'a2 **)

let rec coq_Forall_typing_spine_rect h _UU03a3_ _UU0393_ f f0 _ _ _ _ = function
| Forall_type_spine_nil t0 -> f t0
| Forall_type_spine_cons (hd, tl, na, a, b, s, t0, b', tls, typrod, cumul0,
                          ty, y, y0, f2) ->
  f0 hd tl na a b s t0 b' tls typrod cumul0 ty y y0 f2
    (coq_Forall_typing_spine_rect h _UU03a3_ _UU0393_ f f0 (subst1 hd O b) tl
      b' tls f2)

(** val coq_Forall_typing_spine_rec :
    checker_flags -> global_env_ext -> context -> (term -> 'a2) -> (term ->
    term list -> name -> term -> term -> universe -> term -> term ->
    typing_spine -> typing -> cumul -> typing -> 'a1 -> 'a1 -> 'a1
    coq_Forall_typing_spine -> 'a2 -> 'a2) -> term -> term list -> term ->
    typing_spine -> 'a1 coq_Forall_typing_spine -> 'a2 **)

let rec coq_Forall_typing_spine_rec h _UU03a3_ _UU0393_ f f0 _ _ _ _ = function
| Forall_type_spine_nil t0 -> f t0
| Forall_type_spine_cons (hd, tl, na, a, b, s, t0, b', tls, typrod, cumul0,
                          ty, y, y0, f2) ->
  f0 hd tl na a b s t0 b' tls typrod cumul0 ty y y0 f2
    (coq_Forall_typing_spine_rec h _UU03a3_ _UU0393_ f f0 (subst1 hd O b) tl
      b' tls f2)

(** val wf_local_size :
    checker_flags -> global_env_ext -> (global_env_ext -> context -> term ->
    term -> typing -> size) -> context -> typing lift_typing
    coq_All_local_env -> size **)

let rec wf_local_size h _UU03a3_ fn _ = function
| Coq_localenv_nil -> O
| Coq_localenv_cons_abs (_UU0393_, _, t0, wf_UU0393_, tty) ->
  add
    (fn _UU03a3_ _UU0393_ t0 (Coq_tSort (projT1 (Obj.magic tty)))
      (projT2 (Obj.magic tty)))
    (wf_local_size h _UU03a3_ fn _UU0393_ wf_UU0393_)
| Coq_localenv_cons_def (_UU0393_, _, b, t0, wf_UU0393_, tty, tty') ->
  add
    (add
      (fn _UU03a3_ _UU0393_ t0 (Coq_tSort (projT1 (Obj.magic tty)))
        (projT2 (Obj.magic tty))) (Obj.magic fn _UU03a3_ _UU0393_ b t0 tty'))
    (wf_local_size h _UU03a3_ fn _UU0393_ wf_UU0393_)

(** val typing_spine_size :
    checker_flags -> (global_env_ext -> context -> term -> term -> typing ->
    size) -> global_env_ext -> context -> term -> term list -> term ->
    typing_spine -> size **)

let rec typing_spine_size h fn _UU03a3_ _UU0393_ _ _ _ = function
| Coq_type_spine_nil _ -> O
| Coq_type_spine_cons (hd, tl, na, a, b, s0, _, b', typrod, _, ty, s') ->
  add
    (add (fn _UU03a3_ _UU0393_ hd a ty)
      (fn _UU03a3_ _UU0393_ (Coq_tProd (na, a, b)) (Coq_tSort s0) typrod))
    (typing_spine_size h fn _UU03a3_ _UU0393_ (subst1 hd O b) tl b' s')

(** val typing_size :
    checker_flags -> global_env_ext -> context -> term -> term -> typing ->
    size **)

let rec typing_size h _UU03a3_ _UU0393_ _ _ = function
| Coq_type_Rel (_, _, a) ->
  S
    (wf_local_size h _UU03a3_ (fun x x0 x1 x2 x3 ->
      typing_size h x x0 x1 x2 x3) _UU0393_ a)
| Coq_type_Sort (_, a) ->
  S
    (wf_local_size h _UU03a3_ (fun x x0 x1 x2 x3 ->
      typing_size h x x0 x1 x2 x3) _UU0393_ a)
| Coq_type_Cast (c, _, t0, s, d1, d2) ->
  let d3 = typing_size h _UU03a3_ _UU0393_ c t0 d2 in
  let d4 = typing_size h _UU03a3_ _UU0393_ t0 (Coq_tSort s) d1 in
  S (Nat.max d4 d3)
| Coq_type_Prod (n, t0, b, s1, s2, d1, d2) ->
  let d3 =
    typing_size h _UU03a3_ (snoc _UU0393_ (vass n t0)) b (Coq_tSort s2) d2
  in
  let d4 = typing_size h _UU03a3_ _UU0393_ t0 (Coq_tSort s1) d1 in
  S (Nat.max d4 d3)
| Coq_type_Lambda (n, t0, b, s1, bty, d1, d2) ->
  let d3 = typing_size h _UU03a3_ (snoc _UU0393_ (vass n t0)) b bty d2 in
  let d4 = typing_size h _UU03a3_ _UU0393_ t0 (Coq_tSort s1) d1 in
  S (Nat.max d4 d3)
| Coq_type_LetIn (n, b, b_ty, b', s1, b'_ty, d1, d2, d3) ->
  let d4 = typing_size h _UU03a3_ (snoc _UU0393_ (vdef n b b_ty)) b' b'_ty d3
  in
  let d5 = typing_size h _UU03a3_ _UU0393_ b b_ty d2 in
  let d6 = typing_size h _UU03a3_ _UU0393_ b_ty (Coq_tSort s1) d1 in
  S (Nat.max d6 (Nat.max d5 d4))
| Coq_type_App (t0, l, t_ty, t', d0, t1) ->
  let d1 = typing_size h _UU03a3_ _UU0393_ t0 t_ty d0 in
  S
  (Nat.max d1
    (typing_spine_size h (fun x x0 x1 x2 x3 -> typing_size h x x0 x1 x2 x3)
      _UU03a3_ _UU0393_ t_ty l t' t1))
| Coq_type_Const (_, _, a, _) ->
  S (S
    (wf_local_size h _UU03a3_ (fun x x0 x1 x2 x3 ->
      typing_size h x x0 x1 x2 x3) _UU0393_ a))
| Coq_type_Ind (_, _, a, _, _) ->
  S (S
    (wf_local_size h _UU03a3_ (fun x x0 x1 x2 x3 ->
      typing_size h x x0 x1 x2 x3) _UU0393_ a))
| Coq_type_Construct (_, _, _, a, _, _, _) ->
  S (S
    (wf_local_size h _UU03a3_ (fun x x0 x1 x2 x3 ->
      typing_size h x x0 x1 x2 x3) _UU0393_ a))
| Coq_type_Case (ind, u, _, p, c, brs, args, _, _, x, x0, _, _, _, x1, _, x2,
                 x3) ->
  let d2 =
    typing_size h _UU03a3_ _UU0393_ c (mkApps (Coq_tInd (ind, u)) args) x2
  in
  let d1 = typing_size h _UU03a3_ _UU0393_ p x x0 in
  S
  (Nat.max d1
    (Nat.max d2
      (all2_size (fun x4 y p0 ->
        typing_size h _UU03a3_ _UU0393_ (snd x4) (snd y) (snd p0)) brs x1 x3)))
| Coq_type_Proj (p, c, u, _, _, _, args, d0) ->
  let d1 =
    typing_size h _UU03a3_ _UU0393_ c
      (mkApps (Coq_tInd ((fst (fst p)), u)) args) d0
  in
  S d1
| Coq_type_Fix (mfix, _, _, x, x0) ->
  let types = fix_context mfix in
  S
  (Nat.max
    (wf_local_size h _UU03a3_ (fun x1 x2 x3 x4 x5 ->
      typing_size h x1 x2 x3 x4 x5) (app_context _UU0393_ types) x)
    (all_size (fun x1 p ->
      typing_size h _UU03a3_ (app_context _UU0393_ types) x1.dbody
        (lift (length types) O x1.dtype) (fst p)) mfix x0))
| Coq_type_CoFix (mfix, _, _, x, x0) ->
  let types = fix_context mfix in
  S
  (Nat.max
    (wf_local_size h _UU03a3_ (fun x1 x2 x3 x4 x5 ->
      typing_size h x1 x2 x3 x4 x5) (app_context _UU0393_ types) x)
    (all_size (fun x1 p ->
      typing_size h _UU03a3_ (app_context _UU0393_ types) x1.dbody
        (lift (length types) O x1.dtype) p) mfix x0))
| Coq_type_Conv (t0, a, b, d0, s, _) ->
  let d1 = typing_size h _UU03a3_ _UU0393_ t0 a d0 in
  (match s with
   | Coq_inl i ->
     S
       (Nat.max
         (wf_local_size h _UU03a3_ (fun x x0 x1 x2 x3 ->
           typing_size h x x0 x1 x2 x3) (app_context _UU0393_ (projT1 i))
           (snd (projT2 (projT2 i)))) d1)
   | Coq_inr s0 ->
     let Coq_existT (u, hu) = s0 in
     let hu0 = typing_size h _UU03a3_ _UU0393_ b (Coq_tSort u) hu in
     S (Nat.max d1 hu0))

(** val globenv_size : global_env -> size **)

let rec globenv_size = function
| [] -> S O
| _ :: _UU03a3_0 -> S (globenv_size _UU03a3_0)

type wf = typing coq_Forall_decls_typing

type wf_ext = typing lift_typing on_global_env_ext

type 'p env_prop =
  (global_env * universes_decl) -> wf -> context -> typing lift_typing
  coq_All_local_env -> term -> term -> typing -> 'p
  coq_Forall_decls_typing * 'p

(** val env_prop_typing :
    checker_flags -> 'a1 env_prop -> (global_env * universes_decl) -> wf ->
    context -> typing lift_typing coq_All_local_env -> term -> term -> typing
    -> 'a1 **)

let env_prop_typing _ x _UU03a3_ wf_UU03a3_ _UU0393_ wf_UU0393_ t0 t1 x0 =
  let (_, x1) = x _UU03a3_ wf_UU03a3_ _UU0393_ wf_UU0393_ t0 t1 x0 in x1

(** val type_Prop : checker_flags -> global_env_ext -> typing **)

let type_Prop _ _ =
  Coq_type_Sort (Level.prop, Coq_localenv_nil)

(** val env_prop_sigma :
    checker_flags -> 'a1 env_prop -> global_env -> wf -> 'a1
    coq_Forall_decls_typing **)

let env_prop_sigma h x _UU03a3_ wf_UU03a3_ =
  let _UU0393_ = [] in
  let wf_UU0393_ = Coq_localenv_nil in
  let t0 = Coq_tSort Universe.type0m in
  let t1 = Coq_tSort Universe.type1 in
  let x0 = type_Prop h (empty_ext _UU03a3_) in
  let (x1, _) = x (empty_ext _UU03a3_) wf_UU03a3_ _UU0393_ wf_UU0393_ t0 t1 x0
  in
  x1

(** val wf_local_app :
    checker_flags -> global_env_ext -> context -> context -> typing
    lift_typing coq_All_local_env -> typing lift_typing coq_All_local_env **)

let rec wf_local_app h _UU03a3_ _UU0393_ _UU0393_' x =
  match _UU0393_' with
  | [] -> x
  | _ :: l ->
    (match x with
     | Coq_localenv_nil -> assert false (* absurd case *)
     | Coq_localenv_cons_abs (_, _, _, x0, _) ->
       wf_local_app h _UU03a3_ _UU0393_ l x0
     | Coq_localenv_cons_def (_, _, _, _, x0, _, _) ->
       wf_local_app h _UU03a3_ _UU0393_ l x0)

(** val typing_wf_local :
    checker_flags -> global_env_ext -> context -> term -> term -> typing ->
    typing lift_typing coq_All_local_env **)

let rec typing_wf_local h _UU03a3_ _UU0393_ _ _ = function
| Coq_type_Rel (_, _, a) -> a
| Coq_type_Sort (_, a) -> a
| Coq_type_Cast (c, _, t0, _, _, t1) ->
  typing_wf_local h _UU03a3_ _UU0393_ c t0 t1
| Coq_type_Prod (_, t0, _, s1, _, t1, _) ->
  typing_wf_local h _UU03a3_ _UU0393_ t0 (Coq_tSort s1) t1
| Coq_type_Lambda (_, t0, _, s1, _, t1, _) ->
  typing_wf_local h _UU03a3_ _UU0393_ t0 (Coq_tSort s1) t1
| Coq_type_LetIn (_, b, b_ty, _, _, _, _, t0, _) ->
  typing_wf_local h _UU03a3_ _UU0393_ b b_ty t0
| Coq_type_App (t0, _, t_ty, _, t1, _) ->
  typing_wf_local h _UU03a3_ _UU0393_ t0 t_ty t1
| Coq_type_Const (_, _, a, _) -> a
| Coq_type_Ind (_, _, a, _, _) -> a
| Coq_type_Construct (_, _, _, a, _, _, _) -> a
| Coq_type_Case (ind, u, _, _, c, _, args, _, _, _, _, _, _, _, _, _, x0, _) ->
  typing_wf_local h _UU03a3_ _UU0393_ c (mkApps (Coq_tInd (ind, u)) args) x0
| Coq_type_Proj (p, c, u, _, _, _, args, t0) ->
  typing_wf_local h _UU03a3_ _UU0393_ c
    (mkApps (Coq_tInd ((fst (fst p)), u)) args) t0
| Coq_type_Fix (mfix, _, _, x0, _) ->
  let types = fix_context mfix in wf_local_app h _UU03a3_ _UU0393_ types x0
| Coq_type_CoFix (mfix, _, _, x0, _) ->
  let types = fix_context mfix in wf_local_app h _UU03a3_ _UU0393_ types x0
| Coq_type_Conv (t0, a, _, t1, _, _) ->
  typing_wf_local h _UU03a3_ _UU0393_ t0 a t1

(** val wf_local_inv :
    checker_flags -> global_env_ext -> context -> typing lift_typing
    coq_All_local_env -> context_decl -> context_decl list -> (typing
    lift_typing coq_All_local_env, __) sigT **)

let wf_local_inv _ _ _ w _ _ =
  match w with
  | Coq_localenv_nil -> assert false (* absurd case *)
  | Coq_localenv_cons_abs (_, _, _, w0, l) ->
    Coq_existT (w0,
      (let Coq_existT (x, t0) = Obj.magic l in Obj.magic (Coq_existT (x, t0))))
  | Coq_localenv_cons_def (_, _, _, _, w0, l, l0) ->
    Coq_existT (w0,
      (let Coq_existT (u, h) = Obj.magic l in
       Obj.magic (Coq_existT (u, (Coq_existT (l0, h))))))

(** val type_local_ctx_impl :
    global_env_ext -> context -> context -> universe -> 'a1 type_local_ctx ->
    (context -> term -> term option -> 'a1 -> 'a2) -> 'a2 type_local_ctx **)

let rec type_local_ctx_impl _UU03a3_ _UU0393_ _UU0394_ u hP hPQ =
  match _UU0394_ with
  | [] -> Obj.magic __ _UU0393_ hPQ hP
  | y :: l ->
    let { decl_name = _; decl_body = decl_body0; decl_type = x } = y in
    (match decl_body0 with
     | Some _ -> type_local_ctx_impl _UU03a3_ _UU0393_ l u hP hPQ
     | None ->
       let (x0, x1) = Obj.magic hP in
       Obj.magic ((type_local_ctx_impl _UU03a3_ _UU0393_ l u x0 hPQ),
         (hPQ (app_context _UU0393_ l) x (Some (Coq_tSort u)) x1)))

(** val typing_ind_env :
    checker_flags -> ((global_env * universes_decl) -> wf -> context ->
    typing lift_typing coq_All_local_env -> nat -> context_decl -> __ ->
    (typing, 'a1) coq_All_local_env_over -> 'a1) ->
    ((global_env * universes_decl) -> wf -> context -> typing lift_typing
    coq_All_local_env -> Level.t -> (typing, 'a1) coq_All_local_env_over ->
    __ -> 'a1) -> ((global_env * universes_decl) -> wf -> context -> typing
    lift_typing coq_All_local_env -> term -> cast_kind -> term -> universe ->
    typing -> 'a1 -> typing -> 'a1 -> 'a1) -> ((global_env * universes_decl)
    -> wf -> context -> typing lift_typing coq_All_local_env -> name -> term
    -> term -> universe -> universe -> (typing, 'a1) coq_All_local_env_over
    -> typing -> 'a1 -> typing -> 'a1 -> 'a1) ->
    ((global_env * universes_decl) -> wf -> context -> typing lift_typing
    coq_All_local_env -> name -> term -> term -> universe -> term -> (typing,
    'a1) coq_All_local_env_over -> typing -> 'a1 -> typing -> 'a1 -> 'a1) ->
    ((global_env * universes_decl) -> wf -> context -> typing lift_typing
    coq_All_local_env -> name -> term -> term -> term -> universe -> term ->
    (typing, 'a1) coq_All_local_env_over -> typing -> 'a1 -> typing -> 'a1 ->
    typing -> 'a1 -> 'a1) -> ((global_env * universes_decl) -> wf -> context
    -> typing lift_typing coq_All_local_env -> term -> term list -> term ->
    term -> typing -> 'a1 -> __ -> __ -> typing_spine -> 'a1
    coq_Forall_typing_spine -> 'a1) -> ((global_env * universes_decl) -> wf
    -> context -> typing lift_typing coq_All_local_env -> ident ->
    universe_instance -> constant_body -> 'a1 coq_Forall_decls_typing ->
    (typing, 'a1) coq_All_local_env_over -> __ -> __ -> 'a1) ->
    ((global_env * universes_decl) -> wf -> context -> typing lift_typing
    coq_All_local_env -> inductive -> universe_instance ->
    mutual_inductive_body -> one_inductive_body -> __ -> 'a1
    coq_Forall_decls_typing -> (typing, 'a1) coq_All_local_env_over -> __ ->
    'a1) -> ((global_env * universes_decl) -> wf -> context -> typing
    lift_typing coq_All_local_env -> inductive -> nat -> universe_instance ->
    mutual_inductive_body -> one_inductive_body -> ((ident * term) * nat) ->
    __ -> 'a1 coq_Forall_decls_typing -> (typing, 'a1) coq_All_local_env_over
    -> __ -> 'a1) -> ((global_env * universes_decl) -> wf -> context ->
    typing lift_typing coq_All_local_env -> inductive -> universe_instance ->
    nat -> term -> term -> (nat * term) list -> term list ->
    mutual_inductive_body -> one_inductive_body -> __ -> 'a1
    coq_Forall_decls_typing -> (typing, 'a1) coq_All_local_env_over -> __ ->
    term -> typing -> context_decl list -> context_decl list -> universe ->
    (nat * term) list -> __ -> check_correct_arity -> __ -> 'a1 -> typing ->
    'a1 -> (nat * term, nat * term, (__ * typing) * 'a1) coq_All2 -> 'a1) ->
    ((global_env * universes_decl) -> wf -> context -> typing lift_typing
    coq_All_local_env -> projection -> term -> universe_instance ->
    mutual_inductive_body -> one_inductive_body -> (ident * term) -> __ ->
    term list -> 'a1 coq_Forall_decls_typing -> (typing, 'a1)
    coq_All_local_env_over -> typing -> 'a1 -> __ -> 'a1) ->
    ((global_env * universes_decl) -> wf -> context -> typing lift_typing
    coq_All_local_env -> term def list -> nat -> term def -> __ -> __ ->
    (typing * 'a1) lift_typing coq_All_local_env -> (term def,
    (typing * __) * 'a1) coq_All -> 'a1) -> ((global_env * universes_decl) ->
    wf -> context -> typing lift_typing coq_All_local_env -> term def list ->
    nat -> term def -> __ -> (typing * 'a1) lift_typing coq_All_local_env ->
    (term def, typing * 'a1) coq_All -> __ -> 'a1) ->
    ((global_env * universes_decl) -> wf -> context -> typing lift_typing
    coq_All_local_env -> term -> term -> term -> (typing, 'a1)
    coq_All_local_env_over -> typing -> 'a1 -> ((typing, 'a1) isWfArity_prop,
    (universe, typing * 'a1) sigT) sum -> cumul -> 'a1) -> 'a1 env_prop **)

let typing_ind_env cf x x0 xcast x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12 _UU03a3_ wf_UU03a3_ _UU0393_ wf_UU0393_ t0 t1 h =
  let p0 = fun x13 x14 ->
    let rec fix_F x15 =
      x13 x15 (fun y _ -> fix_F y)
    in fix_F x14
  in
  let h0 = fun x13 iH ->
    let Coq_existT (_UU03a3_0, s) = x13 in
    let Coq_existT (wf_UU03a3_0, s0) = s in
    let Coq_existT (_UU0393_0, s1) = s0 in
    let Coq_existT (wf_UU0393_0, s2) = s1 in
    let Coq_existT (t2, s3) = s2 in
    let Coq_existT (t3, h0) = s3 in
    ((let (_UU03a3_1, _) = _UU03a3_0 in
      (match _UU03a3_1 with
       | [] -> Coq_globenv_nil
       | g :: _UU03a3_2 ->
         (match wf_UU03a3_0 with
          | Coq_globenv_nil -> assert false (* absurd case *)
          | Coq_globenv_decl (_, _, x14, x15) ->
            let udecl = universes_decl_of_decl g in
            (match wf_UU03a3_0 with
             | Coq_globenv_nil -> assert false (* absurd case *)
             | Coq_globenv_decl (_, _, _, _) ->
               Coq_globenv_decl (_UU03a3_2, g,
                 (let iH' =
                    Obj.magic iH (Coq_existT ((_UU03a3_2, udecl), (Coq_existT
                      (x14, (Coq_existT ([], (Coq_existT (Coq_localenv_nil,
                      (Coq_existT ((Coq_tSort Universe.type0m), (Coq_existT
                      ((Coq_tSort Universe.type1),
                      (type_Prop cf (_UU03a3_2, udecl))))))))))))))
                  in
                  let iH'0 = iH' __ in let (x16, _) = iH'0 in x16),
                 (match g with
                  | ConstantDecl (k, c) ->
                    let { cst_type = cst_type0; cst_body = cst_body0;
                      cst_universes = cst_universes0 } = c
                    in
                    let udecl0 =
                      universes_decl_of_decl (ConstantDecl (k, { cst_type =
                        cst_type0; cst_body = cst_body0; cst_universes =
                        cst_universes0 }))
                    in
                    (match cst_body0 with
                     | Some t4 ->
                       let iH0 =
                         iH (Coq_existT ((_UU03a3_2, udecl0), (Coq_existT
                           (x14, (Coq_existT ([], (Coq_existT
                           (Coq_localenv_nil, (Coq_existT (t4, (Coq_existT
                           (cst_type0, x15))))))))))))
                       in
                       let iH1 = iH0 __ in let (_, x16) = iH1 in x16
                     | None ->
                       let Coq_existT (s4, hs) = Obj.magic x15 in
                       let iH0 =
                         iH (Coq_existT ((_UU03a3_2, udecl0), (Coq_existT
                           (x14, (Coq_existT ([], (Coq_existT
                           (Coq_localenv_nil, (Coq_existT (cst_type0,
                           (Coq_existT ((Coq_tSort s4), hs))))))))))))
                       in
                       let iH1 = iH0 __ in
                       Obj.magic (Coq_existT (s4,
                         (let (_, x16) = iH1 in x16))))
                  | InductiveDecl (k, m) ->
                    let udecl0 = universes_decl_of_decl (InductiveDecl (k, m))
                    in
                    let { onInductives = onI; onParams = onP } = Obj.magic x15
                    in
                    Obj.magic { onInductives =
                      (coq_Alli_impl m.ind_bodies O (Obj.magic onI)
                        (fun n x16 xg -> { ind_indices = xg.ind_indices;
                        ind_sort = xg.ind_sort; onArity =
                        (let xg0 = xg.onArity in
                         let Coq_existT (s4, hs) = xg0 in
                         Coq_existT (s4,
                         (let iH0 =
                            Obj.magic iH (Coq_existT ((_UU03a3_2, udecl0),
                              (Coq_existT (x14, (Coq_existT ([], (Coq_existT
                              (Coq_localenv_nil, (Coq_existT (x16.ind_type,
                              (Coq_existT ((Coq_tSort s4), hs))))))))))))
                          in
                          let (_, x17) = iH0 __ in x17))); onConstructors =
                        (let xg0 = xg.onConstructors in
                         coq_Alli_impl_trans x16.ind_ctors O xg0
                           (fun _ x17 x18 ->
                           let (o, x19) = x18 in
                           let Coq_existT (s4, hs) = o in
                           let Coq_existT (cs, hargsu) = x19 in
                           ((let x20 =
                               typing_wf_local cf (_UU03a3_2, udecl0)
                                 (arities_context m.ind_bodies)
                                 (snd (fst x17)) (Coq_tSort s4) hs
                             in
                             let iH0 =
                               Obj.magic iH (Coq_existT ((_UU03a3_2, udecl0),
                                 (Coq_existT (x14, (Coq_existT
                                 ((arities_context m.ind_bodies), (Coq_existT
                                 (x20, (Coq_existT ((snd (fst x17)),
                                 (Coq_existT ((Coq_tSort s4), hs))))))))))))
                             in
                             Coq_existT (s4, (let (_, x21) = iH0 __ in x21))),
                           (Coq_existT (cs,
                           (type_local_ctx_impl (_UU03a3_2, m.ind_universes)
                             (app_context (arities_context m.ind_bodies)
                               m.ind_params) cs.cshape_args
                             cs.cshape_args_univ hargsu
                             (fun _UU0393_1 t4 t5 x20 ->
                             match t5 with
                             | Some t6 ->
                               let x21 =
                                 typing_wf_local cf (_UU03a3_2,
                                   m.ind_universes) _UU0393_1 t4 t6 x20
                               in
                               let iH0 =
                                 Obj.magic iH (Coq_existT ((_UU03a3_2,
                                   udecl0), (Coq_existT (x14, (Coq_existT
                                   (_UU0393_1, (Coq_existT (x21, (Coq_existT
                                   (t4, (Coq_existT (t6, x20))))))))))))
                               in
                               let (_, x22) = iH0 __ in x22
                             | None ->
                               let Coq_existT (u, hu) = Obj.magic x20 in
                               Coq_existT (u,
                               (let x21 =
                                  typing_wf_local cf (_UU03a3_2,
                                    m.ind_universes) _UU0393_1 t4 (Coq_tSort
                                    u) hu
                                in
                                let iH0 =
                                  Obj.magic iH (Coq_existT ((_UU03a3_2,
                                    udecl0), (Coq_existT (x14, (Coq_existT
                                    (_UU0393_1, (Coq_existT (x21, (Coq_existT
                                    (t4, (Coq_existT ((Coq_tSort u),
                                    hu))))))))))))
                                in
                                let (_, x22) = iH0 __ in x22)))))))));
                        onProjections = (fun _ ->
                        let x17 =
                          onProjections cf (_UU03a3_2, udecl0) k m n x16 xg
                        in
                        coq_Alli_impl x16.ind_projs O x17 (fun _ x18 x19 ->
                          let Coq_existT (s4, hs) = x19 in
                          Coq_existT (s4,
                          (let x20 =
                             typing_wf_local cf (_UU03a3_2, udecl0)
                               (snoc (smash_context [] m.ind_params)
                                 (vass (Coq_nNamed x16.ind_name)
                                   (mkApps (Coq_tInd ({ inductive_mind = k;
                                     inductive_ind = n },
                                     (polymorphic_instance m.ind_universes)))
                                     (to_extended_list
                                       (smash_context [] m.ind_params)))))
                               (snd x18) (Coq_tSort s4) hs
                           in
                           let iH0 =
                             Obj.magic iH (Coq_existT ((_UU03a3_2, udecl0),
                               (Coq_existT (x14, (Coq_existT
                               ((snoc (smash_context [] m.ind_params)
                                  (vass (Coq_nNamed x16.ind_name)
                                    (mkApps (Coq_tInd ({ inductive_mind = k;
                                      inductive_ind = n },
                                      (polymorphic_instance m.ind_universes)))
                                      (to_extended_list
                                        (smash_context [] m.ind_params))))),
                               (Coq_existT (x20, (Coq_existT ((snd x18),
                               (Coq_existT ((Coq_tSort s4), hs))))))))))))
                           in
                           let (_, x21) = iH0 __ in x21)))) })); onParams =
                      (coq_All_local_env_impl m.ind_params onP
                        (fun _UU0393_1 t4 t5 x16 ->
                        match t5 with
                        | Some t6 ->
                          let iH0 =
                            Obj.magic iH (Coq_existT ((_UU03a3_2, udecl0),
                              (Coq_existT (x14, (Coq_existT (_UU0393_1,
                              (Coq_existT
                              ((typing_wf_local cf (_UU03a3_2, udecl0)
                                 _UU0393_1 t4 t6 x16), (Coq_existT (t4,
                              (Coq_existT (t6, x16))))))))))))
                          in
                          let (_, x17) = iH0 __ in x17
                        | None ->
                          let Coq_existT (u, hu) = Obj.magic x16 in
                          let iH0 =
                            Obj.magic iH (Coq_existT ((_UU03a3_2, udecl0),
                              (Coq_existT (x14, (Coq_existT (_UU0393_1,
                              (Coq_existT
                              ((typing_wf_local cf (_UU03a3_2, udecl0)
                                 _UU0393_1 t4 (Coq_tSort u) hu), (Coq_existT
                              (t4, (Coq_existT ((Coq_tSort u), hu))))))))))))
                          in
                          Coq_existT (u, (let (_, x17) = iH0 __ in x17)))) })))))),
    (let x14 = fun _UU0393_1 wf_UU0393_1 t4 t5 hty ->
       iH (Coq_existT (_UU03a3_0, (Coq_existT (wf_UU03a3_0, (Coq_existT
         (_UU0393_1, (Coq_existT (wf_UU0393_1, (Coq_existT (t4, (Coq_existT
         (t5, hty)))))))))))) __
     in
     let x15 =
       let foo = typing_wf_local cf _UU03a3_0 _UU0393_0 t2 t3 h0 in
       let rec f _ = function
       | Coq_localenv_nil -> Coq_localenv_over_nil
       | Coq_localenv_cons_abs (_UU0393_1, na, t4, a0, t5) ->
         let Coq_existT (u, hu) = Obj.magic t5 in
         Coq_localenv_over_cons_abs (_UU0393_1, na, t4, a0, (f _UU0393_1 a0),
         (Coq_existT (u, hu)),
         (let (_, x15) = x14 _UU0393_1 a0 t4 (Coq_tSort u) hu in x15))
       | Coq_localenv_cons_def (_UU0393_1, na, b, t4, a0, t5, t6) ->
         let Coq_existT (u, hu) = Obj.magic t5 in
         Coq_localenv_over_cons_def (_UU0393_1, na, b, t4, a0, t6,
         (f _UU0393_1 a0), (let (_, x15) = x14 _UU0393_1 a0 b t4 t6 in x15),
         (Coq_existT (u, hu)),
         (let (_, x15) = x14 _UU0393_1 a0 t4 (Coq_tSort u) hu in x15))
       in f _UU0393_0 foo
     in
     (match h0 with
      | Coq_type_Rel (n, decl, a) ->
        Obj.magic x _UU03a3_0 wf_UU03a3_0 _UU0393_0
          (typing_wf_local cf _UU03a3_0 _UU0393_0 (Coq_tRel n)
            (lift (S n) O decl.decl_type) (Coq_type_Rel (n, decl, a))) n decl
          __ x15
      | Coq_type_Sort (l, a) ->
        Obj.magic x0 _UU03a3_0 wf_UU03a3_0 _UU0393_0
          (typing_wf_local cf _UU03a3_0 _UU0393_0 (Coq_tSort
            (Universe.make l)) (Coq_tSort (Universe.super l)) (Coq_type_Sort
            (l, a))) l x15 __
      | Coq_type_Cast (c, k, t4, s4, h1, h2) ->
        xcast _UU03a3_0 wf_UU03a3_0 _UU0393_0 wf_UU0393_0 c k t4 s4 h1
          (let x16 = fun _UU0393_1 wf_UU0393_1 t5 t6 hty ->
             let (_, x16) = x14 _UU0393_1 wf_UU0393_1 t5 t6 hty in x16
           in
           Obj.magic x16 _UU0393_0 wf_UU0393_0 t4 (Coq_tSort s4) h1) h2
          (let x16 = fun _UU0393_1 wf_UU0393_1 t5 t6 hty ->
             let (_, x16) = x14 _UU0393_1 wf_UU0393_1 t5 t6 hty in x16
           in
           Obj.magic x16 _UU0393_0 wf_UU0393_0 c t4 h2)
      | Coq_type_Prod (n, t4, b, s4, s5, h1, h2) ->
        Obj.magic x1 _UU03a3_0 wf_UU03a3_0 _UU0393_0
          (typing_wf_local cf _UU03a3_0 _UU0393_0 (Coq_tProd (n, t4, b))
            (Coq_tSort (Universe.sort_of_product s4 s5)) (Coq_type_Prod (n,
            t4, b, s4, s5, h1, h2))) n t4 b s4 s5 x15 h1
          (let x16 = fun _UU0393_1 wf_UU0393_1 t5 t6 hty ->
             let (_, x16) = x14 _UU0393_1 wf_UU0393_1 t5 t6 hty in x16
           in
           Obj.magic x16 _UU0393_0 wf_UU0393_0 t4 (Coq_tSort s4) h1) h2
          (let x16 = fun _UU0393_1 wf_UU0393_1 t5 t6 hty ->
             let (_, x16) = x14 _UU0393_1 wf_UU0393_1 t5 t6 hty in x16
           in
           Obj.magic x16 (snoc _UU0393_0 (vass n t4))
             (typing_wf_local cf _UU03a3_0 (snoc _UU0393_0 (vass n t4)) b
               (Coq_tSort s5) h2) b (Coq_tSort s5) h2)
      | Coq_type_Lambda (n, t4, b, s4, bty, h1, h2) ->
        Obj.magic x2 _UU03a3_0 wf_UU03a3_0 _UU0393_0
          (typing_wf_local cf _UU03a3_0 _UU0393_0 (Coq_tLambda (n, t4, b))
            (Coq_tProd (n, t4, bty)) (Coq_type_Lambda (n, t4, b, s4, bty, h1,
            h2))) n t4 b s4 bty x15 h1
          (let x16 = fun _UU0393_1 wf_UU0393_1 t5 t6 hty ->
             let (_, x16) = x14 _UU0393_1 wf_UU0393_1 t5 t6 hty in x16
           in
           Obj.magic x16 _UU0393_0 wf_UU0393_0 t4 (Coq_tSort s4) h1) h2
          (let x16 = fun _UU0393_1 wf_UU0393_1 t5 t6 hty ->
             let (_, x16) = x14 _UU0393_1 wf_UU0393_1 t5 t6 hty in x16
           in
           Obj.magic x16 (snoc _UU0393_0 (vass n t4)) (Coq_localenv_cons_abs
             (_UU0393_0, n, t4, (Obj.magic wf_UU0393_0), (Coq_existT (s4,
             h1)))) b bty h2)
      | Coq_type_LetIn (n, b, b_ty, b', s4, b'_ty, h1, h2, h3) ->
        Obj.magic x3 _UU03a3_0 wf_UU03a3_0 _UU0393_0
          (typing_wf_local cf _UU03a3_0 _UU0393_0 (Coq_tLetIn (n, b, b_ty,
            b')) (Coq_tLetIn (n, b, b_ty, b'_ty)) (Coq_type_LetIn (n, b,
            b_ty, b', s4, b'_ty, h1, h2, h3))) n b b_ty b' s4 b'_ty x15 h1
          (let x16 = fun _UU0393_1 wf_UU0393_1 t4 t5 hty ->
             let (_, x16) = x14 _UU0393_1 wf_UU0393_1 t4 t5 hty in x16
           in
           Obj.magic x16 _UU0393_0 wf_UU0393_0 b_ty (Coq_tSort s4) h1) h2
          (let x16 = fun _UU0393_1 wf_UU0393_1 t4 t5 hty ->
             let (_, x16) = x14 _UU0393_1 wf_UU0393_1 t4 t5 hty in x16
           in
           Obj.magic x16 _UU0393_0 wf_UU0393_0 b b_ty h2) h3
          (let x16 = fun _UU0393_1 wf_UU0393_1 t4 t5 hty ->
             let (_, x16) = x14 _UU0393_1 wf_UU0393_1 t4 t5 hty in x16
           in
           Obj.magic x16 (snoc _UU0393_0 (vdef n b b_ty))
             (Coq_localenv_cons_def (_UU0393_0, n, b, b_ty,
             (Obj.magic wf_UU0393_0), (Obj.magic (Coq_existT (s4, h1))), h2))
             b' b'_ty h3)
      | Coq_type_App (t4, l, t_ty, t', h1, t5) ->
        x4 _UU03a3_0 wf_UU03a3_0 _UU0393_0 wf_UU0393_0 t4 l t_ty t' h1
          (let x16 = fun _UU0393_1 wf_UU0393_1 t6 t7 hty ->
             let (_, x16) = x14 _UU0393_1 wf_UU0393_1 t6 t7 hty in x16
           in
           Obj.magic x16 _UU0393_0 wf_UU0393_0 t4 t_ty h1) __ __ t5
          (let rec f _UU0393_1 _ _ _ t6 wf_UU0393_1 x16 =
             match t6 with
             | Coq_type_spine_nil ty -> Forall_type_spine_nil ty
             | Coq_type_spine_cons (hd, tl, na, a, b, s4, t7, b', t8, c, t9,
                                    t10) ->
               Forall_type_spine_cons (hd, tl, na, a, b, s4, t7, b', t10, t8,
                 c, t9,
                 (let x17 = fun _UU0393_2 x17 t11 t12 hty ->
                    let (_, x18) = x16 _UU0393_2 x17 t11 t12 hty __ in x18
                  in
                  Obj.magic x17 _UU0393_1 wf_UU0393_1 (Coq_tProd (na, a, b))
                    (Coq_tSort s4) t8),
                 (let x17 = fun _UU0393_2 x17 t11 t12 hty ->
                    let (_, x18) = x16 _UU0393_2 x17 t11 t12 hty __ in x18
                  in
                  Obj.magic x17 _UU0393_1 wf_UU0393_1 hd a t9),
                 (f _UU0393_1 (subst1 hd O b) tl b' t10 wf_UU0393_1
                   (fun _UU0393_2 x17 t11 t12 hty _ ->
                   x16 _UU0393_2 x17 t11 t12 hty __)))
           in f _UU0393_0 t_ty l t' t5 wf_UU0393_0
                (fun _UU0393_1 x16 t6 t7 hty _ -> x14 _UU0393_1 x16 t6 t7 hty))
      | Coq_type_Const (cst, u, a, decl) ->
        Obj.magic x5 _UU03a3_0 wf_UU03a3_0 _UU0393_0
          (typing_wf_local cf _UU03a3_0 _UU0393_0 (Coq_tConst (cst, u))
            (subst_instance_constr u decl.cst_type) (Coq_type_Const (cst, u,
            a, decl))) cst u decl
          (let x16 = fun _ ->
             Obj.magic x14 [] Coq_localenv_nil (Coq_tSort Universe.type0m)
               (Coq_tSort Universe.type1) (type_Prop cf _UU03a3_0)
           in
           let x17 = x16 __ in let (x18, _) = x17 in x18) x15 __ __
      | Coq_type_Ind (ind, u, a, mdecl, idecl) ->
        Obj.magic x6 _UU03a3_0 wf_UU03a3_0 _UU0393_0
          (typing_wf_local cf _UU03a3_0 _UU0393_0 (Coq_tInd (ind, u))
            (subst_instance_constr u idecl.ind_type) (Coq_type_Ind (ind, u,
            a, mdecl, idecl))) ind u mdecl idecl __
          (let x16 = fun _ ->
             Obj.magic x14 [] Coq_localenv_nil (Coq_tSort Universe.type0m)
               (Coq_tSort Universe.type1) (type_Prop cf _UU03a3_0)
           in
           let x17 = x16 __ in let (x18, _) = x17 in x18) x15 __
      | Coq_type_Construct (ind, i, u, a, mdecl, idecl, cdecl) ->
        Obj.magic x7 _UU03a3_0 wf_UU03a3_0 _UU0393_0
          (typing_wf_local cf _UU03a3_0 _UU0393_0 (Coq_tConstruct (ind, i,
            u)) (type_of_constructor mdecl cdecl (ind, i) u)
            (Coq_type_Construct (ind, i, u, a, mdecl, idecl, cdecl))) ind i u
          mdecl idecl cdecl __
          (let x16 = fun _ ->
             Obj.magic x14 [] Coq_localenv_nil (Coq_tSort Universe.type0m)
               (Coq_tSort Universe.type1) (type_Prop cf _UU03a3_0)
           in
           let x17 = x16 __ in let (x18, _) = x17 in x18) x15 __
      | Coq_type_Case (ind, u, npar, p, c, brs, args, mdecl, idecl, x16, x17,
                       x18, x19, x20, x21, x22, x23, x24) ->
        Obj.magic x8 _UU03a3_0 wf_UU03a3_0 _UU0393_0
          (typing_wf_local cf _UU03a3_0 _UU0393_0 (Coq_tCase ((ind, npar), p,
            c, brs)) (mkApps p (app (skipn npar args) (c :: [])))
            (Coq_type_Case (ind, u, npar, p, c, brs, args, mdecl, idecl, x16,
            x17, x18, x19, x20, x21, x22, x23, x24))) ind u npar p c brs args
          mdecl idecl __
          (let (x25, _) = Obj.magic x14 _UU0393_0 wf_UU0393_0 p x16 x17 in x25)
          x15 __ x16 x17 x18 x19 x20 x21 __ x22 __
          (let (_, x25) = Obj.magic x14 _UU0393_0 wf_UU0393_0 p x16 x17 in x25)
          x23
          (let (_, x25) =
             Obj.magic x14 _UU0393_0 wf_UU0393_0 c
               (mkApps (Coq_tInd (ind, u)) args) x23
           in
           x25)
          (let rec f _ _ a x25 =
             match a with
             | All2_nil -> All2_nil
             | All2_cons (x26, y, l, l', y0, a0) ->
               let (_, t4) = y0 in
               All2_cons (x26, y, l, l', ((__, t4),
               (let (_, x27) =
                  Obj.magic x25 _UU0393_0 wf_UU0393_0 (snd x26) (snd y) t4 __
                in
                x27)),
               (f l l' a0 (fun _UU0393_1 wf_UU0393_1 t5 t6 hty _ ->
                 x25 _UU0393_1 wf_UU0393_1 t5 t6 hty __)))
           in f brs x21 x24 (fun _UU0393_1 wf_UU0393_1 t4 t5 hty _ ->
                x14 _UU0393_1 wf_UU0393_1 t4 t5 hty))
      | Coq_type_Proj (p, c, u, mdecl, idecl, pdecl, args, h1) ->
        let ty = snd pdecl in
        Obj.magic x9 _UU03a3_0 wf_UU03a3_0 _UU0393_0
          (typing_wf_local cf _UU03a3_0 _UU0393_0 (Coq_tProj (p, c))
            (subst (c :: (List0.rev args)) O (subst_instance_constr u ty))
            (Coq_type_Proj (p, c, u, mdecl, idecl, pdecl, args, h1))) p c u
          mdecl idecl pdecl __ args
          (let x16 = fun _ ->
             Obj.magic x14 [] Coq_localenv_nil (Coq_tSort Universe.type0m)
               (Coq_tSort Universe.type1) (type_Prop cf _UU03a3_0)
           in
           let x17 = x16 __ in let (x18, _) = x17 in x18) x15 h1
          (let x16 = fun _UU0393_1 wf_UU0393_1 t4 t5 hty ->
             let (_, x16) = x14 _UU0393_1 wf_UU0393_1 t4 t5 hty in x16
           in
           Obj.magic x16 _UU0393_0 wf_UU0393_0 c
             (mkApps (Coq_tInd ((fst (fst p)), u)) args) h1) __
      | Coq_type_Fix (mfix, n, decl, x16, x17) ->
        Obj.magic x10 _UU03a3_0 wf_UU03a3_0 _UU0393_0 wf_UU0393_0 mfix n decl
          __ __
          (let _UU0393_' = app_context _UU0393_0 (fix_context mfix) in
           let rec f _ a x18 _UU0393_1 wf_UU0393_1 =
             match a with
             | Coq_localenv_nil -> Coq_localenv_nil
             | Coq_localenv_cons_abs (_UU0393_2, na, t4, a0, t5) ->
               Coq_localenv_cons_abs (_UU0393_2, na, t4,
                 (f _UU0393_2 a0 (fun _UU0393_3 x19 t6 t7 hty _ ->
                   x18 _UU0393_3 x19 t6 t7 hty __) _UU0393_1 wf_UU0393_1),
                 (let Coq_existT (x19, t6) = Obj.magic t5 in
                  Obj.magic (Coq_existT (x19, (t6,
                    (let (_, x20) = x18 _UU0393_2 a0 t4 (Coq_tSort x19) t6 __
                     in
                     x20))))))
             | Coq_localenv_cons_def (_UU0393_2, na, b, t4, a0, t5, t6) ->
               Coq_localenv_cons_def (_UU0393_2, na, b, t4,
                 (f _UU0393_2 a0 (fun _UU0393_3 x19 t7 t8 hty _ ->
                   x18 _UU0393_3 x19 t7 t8 hty __) _UU0393_2 a0),
                 (let Coq_existT (u, t7) = Obj.magic t5 in
                  Obj.magic (Coq_existT (u, (t7,
                    (let t8 = Coq_tSort u in
                     let (_, x19) = x18 _UU0393_2 a0 t4 t8 t7 __ in x19))))),
                 (t6, (let (_, x19) = x18 _UU0393_2 a0 b t4 t6 __ in x19)))
           in f _UU0393_' x16 (fun _UU0393_1 x18 t4 t5 hty _ ->
                x14 _UU0393_1 x18 t4 t5 hty) _UU0393_0 wf_UU0393_0)
          (let mfixcontext = fix_context mfix in
           let rec f _ a x18 =
             match a with
             | All_nil -> All_nil
             | All_cons (x19, l, y, a0) ->
               All_cons (x19, l, (y,
                 (let (_, x20) =
                    Obj.magic x18 (app_context _UU0393_0 mfixcontext) x16
                      x19.dbody (lift (length mfixcontext) O x19.dtype)
                      (fst y) __
                  in
                  x20)),
                 (f l a0 (fun _UU0393_1 x20 t4 t5 hty _ ->
                   x18 _UU0393_1 x20 t4 t5 hty __)))
           in f mfix x17 (fun _UU0393_1 x18 t4 t5 hty _ ->
                x14 _UU0393_1 x18 t4 t5 hty))
      | Coq_type_CoFix (mfix, n, decl, x16, x17) ->
        Obj.magic x11 _UU03a3_0 wf_UU03a3_0 _UU0393_0 wf_UU0393_0 mfix n decl
          __
          (let _UU0393_' = app_context _UU0393_0 (fix_context mfix) in
           let rec f _ a x18 _UU0393_1 wf_UU0393_1 =
             match a with
             | Coq_localenv_nil -> Coq_localenv_nil
             | Coq_localenv_cons_abs (_UU0393_2, na, t4, a0, t5) ->
               Coq_localenv_cons_abs (_UU0393_2, na, t4,
                 (f _UU0393_2 a0 (fun _UU0393_3 x19 t6 t7 hty _ ->
                   x18 _UU0393_3 x19 t6 t7 hty __) _UU0393_1 wf_UU0393_1),
                 (let Coq_existT (x19, t6) = Obj.magic t5 in
                  Obj.magic (Coq_existT (x19, (t6,
                    (let (_, x20) = x18 _UU0393_2 a0 t4 (Coq_tSort x19) t6 __
                     in
                     x20))))))
             | Coq_localenv_cons_def (_UU0393_2, na, b, t4, a0, t5, t6) ->
               Coq_localenv_cons_def (_UU0393_2, na, b, t4,
                 (f _UU0393_2 a0 (fun _UU0393_3 x19 t7 t8 hty _ ->
                   x18 _UU0393_3 x19 t7 t8 hty __) _UU0393_2 a0),
                 (let Coq_existT (u, t7) = Obj.magic t5 in
                  Obj.magic (Coq_existT (u, (t7,
                    (let t8 = Coq_tSort u in
                     let (_, x19) = x18 _UU0393_2 a0 t4 t8 t7 __ in x19))))),
                 (t6, (let (_, x19) = x18 _UU0393_2 a0 b t4 t6 __ in x19)))
           in f _UU0393_' x16 (fun _UU0393_1 x18 t4 t5 hty _ ->
                x14 _UU0393_1 x18 t4 t5 hty) _UU0393_0 wf_UU0393_0)
          (let mfixcontext = fix_context mfix in
           let rec f _ a x18 =
             match a with
             | All_nil -> All_nil
             | All_cons (x19, l, y, a0) ->
               All_cons (x19, l, (y,
                 (let (_, x20) =
                    Obj.magic x18 (app_context _UU0393_0 mfixcontext) x16
                      x19.dbody (lift (length mfixcontext) O x19.dtype) y __
                  in
                  x20)),
                 (f l a0 (fun _UU0393_1 x20 t4 t5 hty _ ->
                   x18 _UU0393_1 x20 t4 t5 hty __)))
           in f mfix x17 (fun _UU0393_1 x18 t4 t5 hty _ ->
                x14 _UU0393_1 x18 t4 t5 hty)) __
      | Coq_type_Conv (t4, a, b, h1, s4, c) ->
        Obj.magic x12 _UU03a3_0 wf_UU03a3_0 _UU0393_0
          (typing_wf_local cf _UU03a3_0 _UU0393_0 t4 b (Coq_type_Conv (t4, a,
            b, h1, s4, c))) t4 a b x15 h1
          (let (_, x16) = Obj.magic x14 _UU0393_0 wf_UU0393_0 t4 a h1 in x16)
          (match s4 with
           | Coq_inl s5 ->
             Coq_inl (Coq_existT (s5,
               (let a0 = snd (projT2 (projT2 s5)) in
                let c0 = app_context _UU0393_0 (projT1 s5) in
                let rec f _ a1 x16 =
                  match a1 with
                  | Coq_localenv_nil -> Coq_localenv_over_nil
                  | Coq_localenv_cons_abs (_UU0393_1, na, t5, a2, t6) ->
                    Coq_localenv_over_cons_abs (_UU0393_1, na, t5, a2,
                      (f _UU0393_1 a2
                        (fun _UU0393_2 wf_UU0393_1 t7 t8 hty _ ->
                        x16 _UU0393_2 wf_UU0393_1 t7 t8 hty __)),
                      (Obj.magic t6),
                      (let (_, x17) =
                         x16 _UU0393_1 a2 t5 (Coq_tSort
                           (projT1 (Obj.magic t6))) (projT2 (Obj.magic t6)) __
                       in
                       x17))
                  | Coq_localenv_cons_def (_UU0393_1, na, b0, t5, a2, t6, t7) ->
                    Coq_localenv_over_cons_def (_UU0393_1, na, b0, t5, a2,
                      t7,
                      (f _UU0393_1 a2
                        (fun _UU0393_2 wf_UU0393_1 t8 t9 hty _ ->
                        x16 _UU0393_2 wf_UU0393_1 t8 t9 hty __)),
                      (let (_, x17) = x16 _UU0393_1 a2 b0 t5 t7 __ in x17),
                      (Obj.magic t6),
                      (let Coq_existT (x17, t8) = Obj.magic t6 in
                       let t9 = Coq_tSort (projT1 (Coq_existT (x17, t8))) in
                       let (_, x18) = x16 _UU0393_1 a2 t5 t9 t8 __ in x18))
                in f c0 a0 (fun _UU0393_1 wf_UU0393_1 t5 t6 hty _ ->
                     x14 _UU0393_1 wf_UU0393_1 t5 t6 hty))))
           | Coq_inr s5 ->
             let Coq_existT (u, hu) = s5 in
             Coq_inr (Coq_existT (u, (hu,
             (let x16 = fun _UU0393_1 wf_UU0393_1 t5 t6 hty ->
                let (_, x16) = x14 _UU0393_1 wf_UU0393_1 t5 t6 hty in x16
              in
              Obj.magic x16 _UU0393_0 wf_UU0393_0 b (Coq_tSort u) hu))))) c)))
  in
  Obj.magic p0 h0 (Coq_existT (_UU03a3_, (Coq_existT (wf_UU03a3_, (Coq_existT
    (_UU0393_, (Coq_existT (wf_UU0393_, (Coq_existT (t0, (Coq_existT (t1,
    h))))))))))))

(** val nth_error_All_local_env :
    context_decl list -> nat -> 'a1 coq_All_local_env -> (context_decl, 'a1
    on_local_decl) on_some **)

let rec nth_error_All_local_env _ n = function
| Coq_localenv_nil ->
  (match n with
   | O -> Obj.magic __ __
   | S x0 -> Obj.magic __ x0 __)
| Coq_localenv_cons_abs (_UU0393_, _, _, a, t0) ->
  (match n with
   | O -> Obj.magic t0
   | S n0 -> nth_error_All_local_env _UU0393_ n0 a)
| Coq_localenv_cons_def (_UU0393_, _, _, _, a, _, t0) ->
  (match n with
   | O -> Obj.magic t0
   | S n0 -> nth_error_All_local_env _UU0393_ n0 a)

(** val lookup_on_global_env :
    checker_flags -> global_env -> ident -> global_decl -> 'a1 on_global_env
    -> (global_env * universes_decl, ('a1 on_global_env, 'a1 on_global_decl)
    sigT) sigT **)

let lookup_on_global_env _ _UU03a3_ c decl x =
  let rec f _ = function
  | Coq_globenv_nil -> (fun _ -> assert false (* absurd case *))
  | Coq_globenv_decl (_UU03a3_0, d, o0, x0) ->
    let b = ident_eq c (global_decl_ident d) in
    if b
    then (fun _ ->
           let udecl = universes_decl_of_decl decl in
           Coq_existT ((_UU03a3_0, udecl), (Coq_existT (o0, x0))))
    else f _UU03a3_0 o0
  in f _UU03a3_ x __

(** val coq_All_local_env_app :
    context -> context -> 'a1 coq_All_local_env -> 'a1
    coq_All_local_env * 'a1 coq_All_local_env **)

let rec coq_All_local_env_app l = function
| [] -> (fun x -> (x, Coq_localenv_nil))
| _ :: l0 ->
  let iHl' = coq_All_local_env_app l l0 in
  (fun x ->
  ((match x with
    | Coq_localenv_nil -> assert false (* absurd case *)
    | Coq_localenv_cons_abs (_, _, _, x0, _) ->
      let x1 = iHl' x0 in let (x2, _) = x1 in x2
    | Coq_localenv_cons_def (_, _, _, _, x0, _, _) ->
      let (x1, _) = iHl' x0 in x1),
  (match x with
   | Coq_localenv_nil -> assert false (* absurd case *)
   | Coq_localenv_cons_abs (_, na, t0, x0, x1) ->
     Coq_localenv_cons_abs (l0, na, t0, (let (_, x2) = iHl' x0 in x2), x1)
   | Coq_localenv_cons_def (_, na, b, t0, x0, x1, x2) ->
     Coq_localenv_cons_def (l0, na, b, t0, (let (_, x3) = iHl' x0 in x3), x1,
       x2))))

(** val coq_All_local_env_lookup :
    context -> nat -> context_decl -> 'a1 coq_All_local_env -> 'a1
    on_local_decl **)

let coq_All_local_env_lookup _UU0393_ n decl x =
  let rec f _ a n0 decl0 =
    match a with
    | Coq_localenv_nil -> (fun _ -> assert false (* absurd case *))
    | Coq_localenv_cons_abs (_UU0393_0, _, _, a0, t0) ->
      (match n0 with
       | O -> (fun _ -> Obj.magic t0)
       | S n1 -> f _UU0393_0 a0 n1 decl0)
    | Coq_localenv_cons_def (_UU0393_0, _, _, _, a0, _, t0) ->
      (match n0 with
       | O -> (fun _ -> Obj.magic t0)
       | S n1 -> f _UU0393_0 a0 n1 decl0)
  in f _UU0393_ x n decl __

(** val coq_All_local_env_app_inv :
    checker_flags -> context -> context -> ('a1 coq_All_local_env * 'a1
    coq_All_local_env) -> 'a1 coq_All_local_env **)

let rec coq_All_local_env_app_inv h l l' x =
  match l' with
  | [] -> let (x0, _) = x in x0
  | y :: l0 ->
    let (x0, x1) = x in
    let x2 = fun x2 x3 -> coq_All_local_env_app_inv h l l0 (x2, x3) in
    let x3 = x2 x0 in
    let { decl_name = decl_name0; decl_body = decl_body0; decl_type =
      decl_type0 } = y
    in
    (match decl_body0 with
     | Some t0 ->
       (match x1 with
        | Coq_localenv_cons_def (_, _, _, _, x4, x5, x6) ->
          Coq_localenv_cons_def ((app_context l l0), decl_name0, t0,
            decl_type0, (x3 x4), x5, x6)
        | _ -> assert false (* absurd case *))
     | None ->
       (match x1 with
        | Coq_localenv_cons_abs (_, _, _, x4, x5) ->
          Coq_localenv_cons_abs ((app_context l l0), decl_name0, decl_type0,
            (x3 x4), x5)
        | _ -> assert false (* absurd case *)))

(** val coq_All_local_env_map :
    checker_flags -> (term -> term) -> context -> 'a1 coq_All_local_env ->
    'a1 coq_All_local_env **)

let rec coq_All_local_env_map h f _ = function
| Coq_localenv_nil -> Coq_localenv_nil
| Coq_localenv_cons_abs (_UU0393_, na, t0, a, t1) ->
  Coq_localenv_cons_abs
    ((let rec map0 = function
      | [] -> []
      | a0 :: t2 -> (map_decl f a0) :: (map0 t2)
      in map0 _UU0393_), (vass na t0).decl_name, (f (vass na t0).decl_type),
    (coq_All_local_env_map h f _UU0393_ a), t1)
| Coq_localenv_cons_def (_UU0393_, na, b, t0, a, t1, t2) ->
  Coq_localenv_cons_def
    ((let rec map0 = function
      | [] -> []
      | a0 :: t3 -> (map_decl f a0) :: (map0 t3)
      in map0 _UU0393_), (vdef na b t0).decl_name, (f b),
    (f (vdef na b t0).decl_type), (coq_All_local_env_map h f _UU0393_ a), t1,
    t2)

(** val lookup_wf_local :
    context -> 'a1 coq_All_local_env -> nat -> 'a1 coq_All_local_env **)

let rec lookup_wf_local _ a n =
  match a with
  | Coq_localenv_nil -> Coq_localenv_nil
  | Coq_localenv_cons_abs (_UU0393_, _, _, a0, _) ->
    (match n with
     | O -> a0
     | S n0 -> lookup_wf_local _UU0393_ a0 n0)
  | Coq_localenv_cons_def (_UU0393_, _, _, _, a0, _, _) ->
    (match n with
     | O -> a0
     | S n0 -> lookup_wf_local _UU0393_ a0 n0)

(** val lookup_wf_local_decl :
    context -> 'a1 coq_All_local_env -> nat -> context_decl -> ('a1
    coq_All_local_env, 'a1 on_local_decl) sigT **)

let rec lookup_wf_local_decl _ a n decl =
  match a with
  | Coq_localenv_nil -> assert false (* absurd case *)
  | Coq_localenv_cons_abs (_UU0393_, _, _, a0, t0) ->
    (match n with
     | O -> Coq_existT (a0, (Obj.magic t0))
     | S n0 -> lookup_wf_local_decl _UU0393_ a0 n0 decl)
  | Coq_localenv_cons_def (_UU0393_, _, _, _, a0, _, t0) ->
    (match n with
     | O -> Coq_existT (a0, (Obj.magic t0))
     | S n0 -> lookup_wf_local_decl _UU0393_ a0 n0 decl)

type 'p on_wf_local_decl = __

(** val nth_error_All_local_env_over :
    checker_flags -> global_env_ext -> context_decl list -> nat ->
    context_decl -> typing lift_typing coq_All_local_env -> (typing, 'a1)
    coq_All_local_env_over -> (typing, 'a1) coq_All_local_env_over * 'a1
    on_wf_local_decl **)

let rec nth_error_All_local_env_over h _UU03a3_ _ n decl _ = function
| Coq_localenv_over_nil -> assert false (* absurd case *)
| Coq_localenv_over_cons_abs (_UU0393_, na, t0, all, a, tu, p) ->
  (match n with
   | O ->
     solution_right (vass na t0) (fun _ _ x0 _ _ _ ->
       simplification_K (Some (vass na t0)) (fun _ ->
         simplification_heq __ __ (fun _ p0 -> (x0, (Obj.magic p0))))) decl
       _UU0393_ all a tu (fun n0 decl0 _ ->
       nth_error_All_local_env_over h _UU03a3_ _UU0393_ n0 decl0 all a) __ __
       p
   | S n0 -> nth_error_All_local_env_over h _UU03a3_ _UU0393_ n0 decl all a)
| Coq_localenv_over_cons_def (_UU0393_, na, b, t0, all, tb, a, p, tu, p0) ->
  (match n with
   | O ->
     solution_right (vdef na b t0) (fun _ _ _ ->
       simplification_heq __ __ (fun _ ->
         solution_right __ (fun _ _ x0 p1 _ _ _ -> (x0, (Obj.magic p1))) __))
       decl _UU0393_ __ __ all tb a p tu p0 (fun n0 decl0 _ ->
       nth_error_All_local_env_over h _UU03a3_ _UU0393_ n0 decl0 all a)
   | S n0 -> nth_error_All_local_env_over h _UU03a3_ _UU0393_ n0 decl all a)

(** val wf_ext_wf : checker_flags -> global_env_ext -> wf_ext -> wf **)

let wf_ext_wf _ _ =
  fst
