open Ast0
open AstUtils
open BasicAst
open Datatypes
open Equality0
open LiftSubst
open List0
open Nat0
open PeanoNat
open Specif
open UnivSubst0
open Universes0
open Config0
open Utils

type __ = Obj.t

val subst_context : term list -> nat -> context -> context

val smash_context : context -> context -> context

val global_decl_ident : global_decl -> kername

val lookup_env : global_env -> ident -> global_decl option

val inds :
  kername -> universe_instance -> one_inductive_body list -> term list

val type_of_constructor :
  mutual_inductive_body -> ((ident * term) * nat) -> (inductive * nat) ->
  Level.t list -> term

val fix_subst : term mfixpoint -> term list

val unfold_fix : term mfixpoint -> nat -> (nat * term) option

val cofix_subst : term mfixpoint -> term list

val unfold_cofix : term mfixpoint -> nat -> (nat * term) option

val isConstruct_app : term -> bool

val is_constructor : nat -> term list -> bool

val fix_context : term mfixpoint -> context

val tDummy : term

val iota_red : nat -> nat -> term list -> (nat * term) list -> term

type red1 =
| Coq_red_beta of name * term * term * term * term list
| Coq_red_zeta of name * term * term * term
| Coq_red_rel of nat * term
| Coq_red_iota of inductive * nat * nat * universe_instance * term list
   * term * (nat * term) list
| Coq_red_fix of term mfixpoint * nat * term list * nat * term
| Coq_red_cofix_case of (inductive * nat) * term * term mfixpoint * nat
   * term list * nat * term * (nat * term) list
| Coq_red_cofix_proj of projection * term mfixpoint * nat * term list * 
   nat * term
| Coq_red_delta of ident * constant_body * term * universe_instance
| Coq_red_proj of inductive * nat * nat * term list * nat * universe_instance
   * term
| Coq_abs_red_l of name * term * term * term * red1
| Coq_abs_red_r of name * term * term * term * red1
| Coq_letin_red_def of name * term * term * term * term * red1
| Coq_letin_red_ty of name * term * term * term * term * red1
| Coq_letin_red_body of name * term * term * term * term * red1
| Coq_case_red_pred of (inductive * nat) * term * term * term
   * (nat * term) list * red1
| Coq_case_red_discr of (inductive * nat) * term * term * term
   * (nat * term) list * red1
| Coq_case_red_brs of (inductive * nat) * term * term * (nat * term) list
   * (nat * term) list * (nat * term, red1) coq_OnOne2
| Coq_proj_red of projection * term * term * red1
| Coq_app_red_l of term * term * term list * red1
| Coq_app_red_r of term list * term list * term * (term, red1) coq_OnOne2
| Coq_prod_red_l of name * term * term * term * red1
| Coq_prod_red_r of name * term * term * term * red1
| Coq_evar_red of nat * term list * term list * (term, red1) coq_OnOne2
| Coq_cast_red_l of term * cast_kind * term * term * red1
| Coq_cast_red_r of term * cast_kind * term * term * red1
| Coq_cast_red of term * cast_kind * term
| Coq_fix_red_ty of term def list * term def list * nat
   * (term def, red1 * __) coq_OnOne2
| Coq_fix_red_body of term mfixpoint * term def list * nat
   * (term def, red1 * __) coq_OnOne2
| Coq_cofix_red_ty of term def list * term def list * nat
   * (term def, red1 * __) coq_OnOne2
| Coq_cofix_red_body of term mfixpoint * term def list * nat
   * (term def, red1 * __) coq_OnOne2

val red1_rect :
  global_env -> (context -> name -> term -> term -> term -> term list -> 'a1)
  -> (context -> name -> term -> term -> term -> 'a1) -> (context -> nat ->
  term -> __ -> 'a1) -> (context -> inductive -> nat -> nat ->
  universe_instance -> term list -> term -> (nat * term) list -> 'a1) ->
  (context -> term mfixpoint -> nat -> term list -> nat -> term -> __ -> __
  -> 'a1) -> (context -> (inductive * nat) -> term -> term mfixpoint -> nat
  -> term list -> nat -> term -> (nat * term) list -> __ -> 'a1) -> (context
  -> projection -> term mfixpoint -> nat -> term list -> nat -> term -> __ ->
  'a1) -> (context -> ident -> constant_body -> term -> __ ->
  universe_instance -> __ -> 'a1) -> (context -> inductive -> nat -> nat ->
  term list -> nat -> universe_instance -> term -> __ -> 'a1) -> (context ->
  name -> term -> term -> term -> red1 -> 'a1 -> 'a1) -> (context -> name ->
  term -> term -> term -> red1 -> 'a1 -> 'a1) -> (context -> name -> term ->
  term -> term -> term -> red1 -> 'a1 -> 'a1) -> (context -> name -> term ->
  term -> term -> term -> red1 -> 'a1 -> 'a1) -> (context -> name -> term ->
  term -> term -> term -> red1 -> 'a1 -> 'a1) -> (context ->
  (inductive * nat) -> term -> term -> term -> (nat * term) list -> red1 ->
  'a1 -> 'a1) -> (context -> (inductive * nat) -> term -> term -> term ->
  (nat * term) list -> red1 -> 'a1 -> 'a1) -> (context -> (inductive * nat)
  -> term -> term -> (nat * term) list -> (nat * term) list -> (nat * term,
  red1) coq_OnOne2 -> 'a1) -> (context -> projection -> term -> term -> red1
  -> 'a1 -> 'a1) -> (context -> term -> term -> term list -> red1 -> 'a1 ->
  'a1) -> (context -> term list -> term list -> term -> (term, red1)
  coq_OnOne2 -> 'a1) -> (context -> name -> term -> term -> term -> red1 ->
  'a1 -> 'a1) -> (context -> name -> term -> term -> term -> red1 -> 'a1 ->
  'a1) -> (context -> nat -> term list -> term list -> (term, red1)
  coq_OnOne2 -> 'a1) -> (context -> term -> cast_kind -> term -> term -> red1
  -> 'a1 -> 'a1) -> (context -> term -> cast_kind -> term -> term -> red1 ->
  'a1 -> 'a1) -> (context -> term -> cast_kind -> term -> 'a1) -> (context ->
  term def list -> term def list -> nat -> (term def, red1 * __) coq_OnOne2
  -> 'a1) -> (context -> term mfixpoint -> term def list -> nat -> (term def,
  red1 * __) coq_OnOne2 -> 'a1) -> (context -> term def list -> term def list
  -> nat -> (term def, red1 * __) coq_OnOne2 -> 'a1) -> (context -> term
  mfixpoint -> term def list -> nat -> (term def, red1 * __) coq_OnOne2 ->
  'a1) -> context -> term -> term -> red1 -> 'a1

val red1_rec :
  global_env -> (context -> name -> term -> term -> term -> term list -> 'a1)
  -> (context -> name -> term -> term -> term -> 'a1) -> (context -> nat ->
  term -> __ -> 'a1) -> (context -> inductive -> nat -> nat ->
  universe_instance -> term list -> term -> (nat * term) list -> 'a1) ->
  (context -> term mfixpoint -> nat -> term list -> nat -> term -> __ -> __
  -> 'a1) -> (context -> (inductive * nat) -> term -> term mfixpoint -> nat
  -> term list -> nat -> term -> (nat * term) list -> __ -> 'a1) -> (context
  -> projection -> term mfixpoint -> nat -> term list -> nat -> term -> __ ->
  'a1) -> (context -> ident -> constant_body -> term -> __ ->
  universe_instance -> __ -> 'a1) -> (context -> inductive -> nat -> nat ->
  term list -> nat -> universe_instance -> term -> __ -> 'a1) -> (context ->
  name -> term -> term -> term -> red1 -> 'a1 -> 'a1) -> (context -> name ->
  term -> term -> term -> red1 -> 'a1 -> 'a1) -> (context -> name -> term ->
  term -> term -> term -> red1 -> 'a1 -> 'a1) -> (context -> name -> term ->
  term -> term -> term -> red1 -> 'a1 -> 'a1) -> (context -> name -> term ->
  term -> term -> term -> red1 -> 'a1 -> 'a1) -> (context ->
  (inductive * nat) -> term -> term -> term -> (nat * term) list -> red1 ->
  'a1 -> 'a1) -> (context -> (inductive * nat) -> term -> term -> term ->
  (nat * term) list -> red1 -> 'a1 -> 'a1) -> (context -> (inductive * nat)
  -> term -> term -> (nat * term) list -> (nat * term) list -> (nat * term,
  red1) coq_OnOne2 -> 'a1) -> (context -> projection -> term -> term -> red1
  -> 'a1 -> 'a1) -> (context -> term -> term -> term list -> red1 -> 'a1 ->
  'a1) -> (context -> term list -> term list -> term -> (term, red1)
  coq_OnOne2 -> 'a1) -> (context -> name -> term -> term -> term -> red1 ->
  'a1 -> 'a1) -> (context -> name -> term -> term -> term -> red1 -> 'a1 ->
  'a1) -> (context -> nat -> term list -> term list -> (term, red1)
  coq_OnOne2 -> 'a1) -> (context -> term -> cast_kind -> term -> term -> red1
  -> 'a1 -> 'a1) -> (context -> term -> cast_kind -> term -> term -> red1 ->
  'a1 -> 'a1) -> (context -> term -> cast_kind -> term -> 'a1) -> (context ->
  term def list -> term def list -> nat -> (term def, red1 * __) coq_OnOne2
  -> 'a1) -> (context -> term mfixpoint -> term def list -> nat -> (term def,
  red1 * __) coq_OnOne2 -> 'a1) -> (context -> term def list -> term def list
  -> nat -> (term def, red1 * __) coq_OnOne2 -> 'a1) -> (context -> term
  mfixpoint -> term def list -> nat -> (term def, red1 * __) coq_OnOne2 ->
  'a1) -> context -> term -> term -> red1 -> 'a1

val red1_ind_all :
  global_env -> (context -> name -> term -> term -> term -> term list -> 'a1)
  -> (context -> name -> term -> term -> term -> 'a1) -> (context -> nat ->
  term -> __ -> 'a1) -> (context -> inductive -> nat -> nat ->
  universe_instance -> term list -> term -> (nat * term) list -> 'a1) ->
  (context -> term mfixpoint -> nat -> term list -> nat -> term -> __ -> __
  -> 'a1) -> (context -> (inductive * nat) -> term -> term mfixpoint -> nat
  -> term list -> nat -> term -> (nat * term) list -> __ -> 'a1) -> (context
  -> projection -> term mfixpoint -> nat -> term list -> nat -> term -> __ ->
  'a1) -> (context -> ident -> constant_body -> term -> __ ->
  universe_instance -> __ -> 'a1) -> (context -> inductive -> nat -> nat ->
  term list -> nat -> universe_instance -> term -> __ -> 'a1) -> (context ->
  name -> term -> term -> term -> red1 -> 'a1 -> 'a1) -> (context -> name ->
  term -> term -> term -> red1 -> 'a1 -> 'a1) -> (context -> name -> term ->
  term -> term -> term -> red1 -> 'a1 -> 'a1) -> (context -> name -> term ->
  term -> term -> term -> red1 -> 'a1 -> 'a1) -> (context -> name -> term ->
  term -> term -> term -> red1 -> 'a1 -> 'a1) -> (context ->
  (inductive * nat) -> term -> term -> term -> (nat * term) list -> red1 ->
  'a1 -> 'a1) -> (context -> (inductive * nat) -> term -> term -> term ->
  (nat * term) list -> red1 -> 'a1 -> 'a1) -> (context -> (inductive * nat)
  -> term -> term -> (nat * term) list -> (nat * term) list -> (nat * term,
  red1 * 'a1) coq_OnOne2 -> 'a1) -> (context -> projection -> term -> term ->
  red1 -> 'a1 -> 'a1) -> (context -> term -> term -> term list -> red1 -> 'a1
  -> 'a1) -> (context -> term list -> term list -> term -> (term, red1 * 'a1)
  coq_OnOne2 -> 'a1) -> (context -> name -> term -> term -> term -> red1 ->
  'a1 -> 'a1) -> (context -> name -> term -> term -> term -> red1 -> 'a1 ->
  'a1) -> (context -> nat -> term list -> term list -> (term, red1 * 'a1)
  coq_OnOne2 -> 'a1) -> (context -> term -> cast_kind -> term -> term -> red1
  -> 'a1 -> 'a1) -> (context -> term -> cast_kind -> term -> term -> red1 ->
  'a1 -> 'a1) -> (context -> term -> cast_kind -> term -> 'a1) -> (context ->
  term def list -> term def list -> nat -> (term def, (red1 * __) * 'a1)
  coq_OnOne2 -> 'a1) -> (context -> term def list -> term def list -> nat ->
  (term def, (red1 * __) * 'a1) coq_OnOne2 -> 'a1) -> (context -> term def
  list -> term def list -> nat -> (term def, (red1 * __) * 'a1) coq_OnOne2 ->
  'a1) -> (context -> term def list -> term def list -> nat -> (term def,
  (red1 * __) * 'a1) coq_OnOne2 -> 'a1) -> context -> term -> term -> red1 ->
  'a1

type red =
| Coq_refl_red
| Coq_trans_red of term * term * red * red1

val red_rect :
  global_env -> context -> term -> 'a1 -> (term -> term -> red -> 'a1 -> red1
  -> 'a1) -> term -> red -> 'a1

val red_rec :
  global_env -> context -> term -> 'a1 -> (term -> term -> red -> 'a1 -> red1
  -> 'a1) -> term -> red -> 'a1

val strip_casts : term -> term

val destArity :
  context_decl list -> term -> (context_decl list * universe) option

val instantiate_params_subst :
  context_decl list -> term list -> term list -> term -> (term list * term)
  option

val instantiate_params : context_decl list -> term list -> term -> term option

val build_branches_type :
  inductive -> mutual_inductive_body -> one_inductive_body -> term list ->
  universe_instance -> term -> (nat * term) option list

val types_of_case :
  inductive -> mutual_inductive_body -> one_inductive_body -> term list ->
  universe_instance -> term -> term -> (((context_decl list * context_decl
  list) * universe) * (nat * term) list) option

val on_udecl_decl : (universes_decl -> 'a1) -> global_decl -> 'a1

val monomorphic_udecl_decl : global_decl -> ContextSet.t

val monomorphic_levels_decl : global_decl -> LevelSet.t

val monomorphic_constraints_decl : global_decl -> constraints

val universes_decl_of_decl : global_decl -> universes_decl

val coq_LevelSet_pair : LevelSet.elt -> LevelSet.elt -> LevelSet.t

val global_levels : global_env -> LevelSet.t

val global_constraints : global_env -> constraints

val global_ext_levels : global_env_ext -> LevelSet.t

val global_ext_constraints : global_env_ext -> constraints

val global_ext_uctx : global_env_ext -> ContextSet.t

type cumul =
| Coq_cumul_refl of term * term
| Coq_cumul_red_l of term * term * term * red1 * cumul
| Coq_cumul_red_r of term * term * term * cumul * red1

val cumul_rect :
  checker_flags -> global_env_ext -> context -> (term -> term -> __ -> 'a1)
  -> (term -> term -> term -> red1 -> cumul -> 'a1 -> 'a1) -> (term -> term
  -> term -> cumul -> 'a1 -> red1 -> 'a1) -> term -> term -> cumul -> 'a1

val cumul_rec :
  checker_flags -> global_env_ext -> context -> (term -> term -> __ -> 'a1)
  -> (term -> term -> term -> red1 -> cumul -> 'a1 -> 'a1) -> (term -> term
  -> term -> cumul -> 'a1 -> red1 -> 'a1) -> term -> term -> cumul -> 'a1

type conv = cumul * cumul

val conv_refl : checker_flags -> global_env_ext -> context -> term -> conv

val cumul_refl' : checker_flags -> global_env_ext -> context -> term -> cumul

val cumul_trans :
  checker_flags -> global_env_ext -> context -> term -> term -> term -> cumul
  -> cumul -> cumul

val congr_cumul_prod :
  checker_flags -> global_env_ext -> context -> name -> name -> term -> term
  -> term -> term -> cumul -> cumul -> cumul

type eq_context = (context_decl, context_decl, __) coq_All2

type check_correct_arity = eq_context

type 'typing coq_All_local_env =
| Coq_localenv_nil
| Coq_localenv_cons_abs of context * name * term * 'typing coq_All_local_env
   * 'typing
| Coq_localenv_cons_def of context * name * term * term
   * 'typing coq_All_local_env * 'typing * 'typing

val coq_All_local_env_rect :
  'a2 -> (context -> name -> term -> 'a1 coq_All_local_env -> 'a2 -> 'a1 ->
  'a2) -> (context -> name -> term -> term -> 'a1 coq_All_local_env -> 'a2 ->
  'a1 -> 'a1 -> 'a2) -> context -> 'a1 coq_All_local_env -> 'a2

val coq_All_local_env_rec :
  'a2 -> (context -> name -> term -> 'a1 coq_All_local_env -> 'a2 -> 'a1 ->
  'a2) -> (context -> name -> term -> term -> 'a1 coq_All_local_env -> 'a2 ->
  'a1 -> 'a1 -> 'a2) -> context -> 'a1 coq_All_local_env -> 'a2

type 'p lift_typing = __

type 'p on_local_decl = __

type ('typing, 'property) coq_All_local_env_over =
| Coq_localenv_over_nil
| Coq_localenv_over_cons_abs of context * name * term
   * 'typing lift_typing coq_All_local_env
   * ('typing, 'property) coq_All_local_env_over * (universe, 'typing) sigT
   * 'property
| Coq_localenv_over_cons_def of context * name * term * term
   * 'typing lift_typing coq_All_local_env * 'typing
   * ('typing, 'property) coq_All_local_env_over * 'property
   * (universe, 'typing) sigT * 'property

val coq_All_local_env_over_rect :
  global_env_ext -> 'a3 -> (context -> name -> term -> 'a1 lift_typing
  coq_All_local_env -> ('a1, 'a2) coq_All_local_env_over -> 'a3 -> (universe,
  'a1) sigT -> 'a2 -> 'a3) -> (context -> name -> term -> term -> 'a1
  lift_typing coq_All_local_env -> 'a1 -> ('a1, 'a2) coq_All_local_env_over
  -> 'a3 -> 'a2 -> (universe, 'a1) sigT -> 'a2 -> 'a3) -> context -> 'a1
  lift_typing coq_All_local_env -> ('a1, 'a2) coq_All_local_env_over -> 'a3

val coq_All_local_env_over_rec :
  global_env_ext -> 'a3 -> (context -> name -> term -> 'a1 lift_typing
  coq_All_local_env -> ('a1, 'a2) coq_All_local_env_over -> 'a3 -> (universe,
  'a1) sigT -> 'a2 -> 'a3) -> (context -> name -> term -> term -> 'a1
  lift_typing coq_All_local_env -> 'a1 -> ('a1, 'a2) coq_All_local_env_over
  -> 'a3 -> 'a2 -> (universe, 'a1) sigT -> 'a2 -> 'a3) -> context -> 'a1
  lift_typing coq_All_local_env -> ('a1, 'a2) coq_All_local_env_over -> 'a3

type 'typing isWfArity =
  (context_decl list, (universe, __ * 'typing lift_typing coq_All_local_env)
  sigT) sigT

type ('typing, 'property) isWfArity_prop =
  ('typing isWfArity, ('typing, 'property) coq_All_local_env_over) sigT

val fix_guard : term mfixpoint -> bool

val ind_guard : mutual_inductive_body -> bool

type typing =
| Coq_type_Rel of nat * context_decl * typing lift_typing coq_All_local_env
| Coq_type_Sort of LevelSet.elt * typing lift_typing coq_All_local_env
| Coq_type_Cast of term * cast_kind * term * universe * typing * typing
| Coq_type_Prod of name * term * term * universe * universe * typing * typing
| Coq_type_Lambda of name * term * term * universe * term * typing * typing
| Coq_type_LetIn of name * term * term * term * universe * term * typing
   * typing * typing
| Coq_type_App of term * term list * term * term * typing * typing_spine
| Coq_type_Const of ident * universe_instance
   * typing lift_typing coq_All_local_env * constant_body
| Coq_type_Ind of inductive * universe_instance
   * typing lift_typing coq_All_local_env * mutual_inductive_body
   * one_inductive_body
| Coq_type_Construct of inductive * nat * universe_instance
   * typing lift_typing coq_All_local_env * mutual_inductive_body
   * one_inductive_body * ((ident * term) * nat)
| Coq_type_Case of inductive * universe_instance * nat * term * term
   * (nat * term) list * term list * mutual_inductive_body
   * one_inductive_body * term * typing * context_decl list
   * context_decl list * universe * (nat * term) list * check_correct_arity
   * typing * (nat * term, nat * term, __ * typing) coq_All2
| Coq_type_Proj of projection * term * universe_instance
   * mutual_inductive_body * one_inductive_body * (ident * term) * term list
   * typing
| Coq_type_Fix of term mfixpoint * nat * term def
   * typing lift_typing coq_All_local_env * (term def, typing * __) coq_All
| Coq_type_CoFix of term mfixpoint * nat * term def
   * typing lift_typing coq_All_local_env * (term def, typing) coq_All
| Coq_type_Conv of term * term * term * typing
   * (typing isWfArity, (universe, typing) sigT) sum * cumul
and typing_spine =
| Coq_type_spine_nil of term
| Coq_type_spine_cons of term * term list * name * term * term * universe
   * term * term * typing * cumul * typing * typing_spine

val typing_rect :
  checker_flags -> global_env_ext -> (context -> nat -> context_decl ->
  typing lift_typing coq_All_local_env -> __ -> 'a1) -> (context ->
  LevelSet.elt -> typing lift_typing coq_All_local_env -> __ -> 'a1) ->
  (context -> term -> cast_kind -> term -> universe -> typing -> 'a1 ->
  typing -> 'a1 -> 'a1) -> (context -> name -> term -> term -> universe ->
  universe -> typing -> 'a1 -> typing -> 'a1 -> 'a1) -> (context -> name ->
  term -> term -> universe -> term -> typing -> 'a1 -> typing -> 'a1 -> 'a1)
  -> (context -> name -> term -> term -> term -> universe -> term -> typing
  -> 'a1 -> typing -> 'a1 -> typing -> 'a1 -> 'a1) -> (context -> term ->
  term list -> term -> term -> typing -> 'a1 -> __ -> __ -> typing_spine ->
  'a1) -> (context -> ident -> universe_instance -> typing lift_typing
  coq_All_local_env -> constant_body -> __ -> __ -> 'a1) -> (context ->
  inductive -> universe_instance -> typing lift_typing coq_All_local_env ->
  mutual_inductive_body -> one_inductive_body -> __ -> __ -> 'a1) -> (context
  -> inductive -> nat -> universe_instance -> typing lift_typing
  coq_All_local_env -> mutual_inductive_body -> one_inductive_body ->
  ((ident * term) * nat) -> __ -> __ -> 'a1) -> (context -> inductive ->
  universe_instance -> nat -> term -> term -> (nat * term) list -> term list
  -> mutual_inductive_body -> one_inductive_body -> __ -> __ -> term ->
  typing -> 'a1 -> context_decl list -> context_decl list -> universe ->
  (nat * term) list -> __ -> check_correct_arity -> __ -> typing -> 'a1 ->
  (nat * term, nat * term, __ * typing) coq_All2 -> 'a1) -> (context ->
  projection -> term -> universe_instance -> mutual_inductive_body ->
  one_inductive_body -> (ident * term) -> __ -> term list -> typing -> 'a1 ->
  __ -> 'a1) -> (context -> term mfixpoint -> nat -> term def -> __ -> __ ->
  typing lift_typing coq_All_local_env -> (term def, typing * __) coq_All ->
  'a1) -> (context -> term mfixpoint -> nat -> term def -> __ -> __ -> typing
  lift_typing coq_All_local_env -> (term def, typing) coq_All -> 'a1) ->
  (context -> term -> term -> term -> typing -> 'a1 -> (typing isWfArity,
  (universe, typing) sigT) sum -> cumul -> 'a1) -> context -> term -> term ->
  typing -> 'a1

val typing_rec :
  checker_flags -> global_env_ext -> (context -> nat -> context_decl ->
  typing lift_typing coq_All_local_env -> __ -> 'a1) -> (context ->
  LevelSet.elt -> typing lift_typing coq_All_local_env -> __ -> 'a1) ->
  (context -> term -> cast_kind -> term -> universe -> typing -> 'a1 ->
  typing -> 'a1 -> 'a1) -> (context -> name -> term -> term -> universe ->
  universe -> typing -> 'a1 -> typing -> 'a1 -> 'a1) -> (context -> name ->
  term -> term -> universe -> term -> typing -> 'a1 -> typing -> 'a1 -> 'a1)
  -> (context -> name -> term -> term -> term -> universe -> term -> typing
  -> 'a1 -> typing -> 'a1 -> typing -> 'a1 -> 'a1) -> (context -> term ->
  term list -> term -> term -> typing -> 'a1 -> __ -> __ -> typing_spine ->
  'a1) -> (context -> ident -> universe_instance -> typing lift_typing
  coq_All_local_env -> constant_body -> __ -> __ -> 'a1) -> (context ->
  inductive -> universe_instance -> typing lift_typing coq_All_local_env ->
  mutual_inductive_body -> one_inductive_body -> __ -> __ -> 'a1) -> (context
  -> inductive -> nat -> universe_instance -> typing lift_typing
  coq_All_local_env -> mutual_inductive_body -> one_inductive_body ->
  ((ident * term) * nat) -> __ -> __ -> 'a1) -> (context -> inductive ->
  universe_instance -> nat -> term -> term -> (nat * term) list -> term list
  -> mutual_inductive_body -> one_inductive_body -> __ -> __ -> term ->
  typing -> 'a1 -> context_decl list -> context_decl list -> universe ->
  (nat * term) list -> __ -> check_correct_arity -> __ -> typing -> 'a1 ->
  (nat * term, nat * term, __ * typing) coq_All2 -> 'a1) -> (context ->
  projection -> term -> universe_instance -> mutual_inductive_body ->
  one_inductive_body -> (ident * term) -> __ -> term list -> typing -> 'a1 ->
  __ -> 'a1) -> (context -> term mfixpoint -> nat -> term def -> __ -> __ ->
  typing lift_typing coq_All_local_env -> (term def, typing * __) coq_All ->
  'a1) -> (context -> term mfixpoint -> nat -> term def -> __ -> __ -> typing
  lift_typing coq_All_local_env -> (term def, typing) coq_All -> 'a1) ->
  (context -> term -> term -> term -> typing -> 'a1 -> (typing isWfArity,
  (universe, typing) sigT) sum -> cumul -> 'a1) -> context -> term -> term ->
  typing -> 'a1

val typing_spine_rect :
  checker_flags -> global_env_ext -> (context -> term -> 'a1) -> (context ->
  term -> term list -> name -> term -> term -> universe -> term -> term ->
  typing -> cumul -> typing -> typing_spine -> 'a1 -> 'a1) -> context -> term
  -> term list -> term -> typing_spine -> 'a1

val typing_spine_rec :
  checker_flags -> global_env_ext -> (context -> term -> 'a1) -> (context ->
  term -> term list -> name -> term -> term -> universe -> term -> term ->
  typing -> cumul -> typing -> typing_spine -> 'a1 -> 'a1) -> context -> term
  -> term list -> term -> typing_spine -> 'a1

type isType = (universe, typing) sigT

type 'p unlift_opt_pred = 'p

type 'p type_local_ctx = __

type 'p on_type = 'p

type constructor_shape = { cshape_args : context;
                           cshape_args_univ : universe;
                           cshape_indices : term list }

val cshape_args :
  mutual_inductive_body -> nat -> one_inductive_body ->
  ((ident * term) * nat) -> constructor_shape -> context

val cshape_args_univ :
  mutual_inductive_body -> nat -> one_inductive_body ->
  ((ident * term) * nat) -> constructor_shape -> universe

val cshape_concl_head :
  mutual_inductive_body -> nat -> one_inductive_body ->
  ((ident * term) * nat) -> constructor_shape -> term

val cshape_indices :
  mutual_inductive_body -> nat -> one_inductive_body ->
  ((ident * term) * nat) -> constructor_shape -> term list

type 'p on_constructor =
  'p on_type * (constructor_shape, 'p type_local_ctx) sigT

type 'p on_constructors = ((ident * term) * nat, 'p on_constructor) coq_Alli

type 'p on_projection = 'p on_type

type 'p on_projections =
  (ident * term, 'p on_projection) coq_Alli
  (* singleton inductive, whose constructor was Build_on_projections *)

val on_projs :
  global_env_ext -> kername -> mutual_inductive_body -> nat ->
  one_inductive_body -> context -> (ident * term) list -> 'a1 on_projections
  -> (ident * term, 'a1 on_projection) coq_Alli

val is_prop_elim : sort_family list -> bool

val is_all_elim : sort_family list -> bool

val leb_sort_family : sort_family -> sort_family -> bool

type 'p on_ind_body = { ind_indices : context; ind_sort : universe;
                        onArity : 'p on_type;
                        onConstructors : 'p on_constructors;
                        onProjections : (__ -> 'p on_projections) }

val ind_indices :
  checker_flags -> global_env_ext -> kername -> mutual_inductive_body -> nat
  -> one_inductive_body -> 'a1 on_ind_body -> context

val ind_sort :
  checker_flags -> global_env_ext -> kername -> mutual_inductive_body -> nat
  -> one_inductive_body -> 'a1 on_ind_body -> universe

val onArity :
  checker_flags -> global_env_ext -> kername -> mutual_inductive_body -> nat
  -> one_inductive_body -> 'a1 on_ind_body -> 'a1 on_type

val onConstructors :
  checker_flags -> global_env_ext -> kername -> mutual_inductive_body -> nat
  -> one_inductive_body -> 'a1 on_ind_body -> 'a1 on_constructors

val onProjections :
  checker_flags -> global_env_ext -> kername -> mutual_inductive_body -> nat
  -> one_inductive_body -> 'a1 on_ind_body -> 'a1 on_projections

type 'p on_context = 'p coq_All_local_env

type 'p on_inductive = { onInductives : (one_inductive_body, 'p on_ind_body)
                                        coq_Alli; onParams : 'p on_context }

val onInductives :
  checker_flags -> global_env_ext -> kername -> mutual_inductive_body -> 'a1
  on_inductive -> (one_inductive_body, 'a1 on_ind_body) coq_Alli

val onParams :
  checker_flags -> global_env_ext -> kername -> mutual_inductive_body -> 'a1
  on_inductive -> 'a1 on_context

type 'p on_constant_decl = __

type 'p on_global_decl = __

type 'p on_global_env =
| Coq_globenv_nil
| Coq_globenv_decl of global_env * global_decl * 'p on_global_env
   * 'p on_global_decl

val on_global_env_rect :
  checker_flags -> checker_flags -> 'a2 -> (global_env -> global_decl -> 'a1
  on_global_env -> 'a2 -> __ -> __ -> 'a1 on_global_decl -> 'a2) ->
  global_env -> 'a1 on_global_env -> 'a2

val on_global_env_rec :
  checker_flags -> checker_flags -> 'a2 -> (global_env -> global_decl -> 'a1
  on_global_env -> 'a2 -> __ -> __ -> 'a1 on_global_decl -> 'a2) ->
  global_env -> 'a1 on_global_env -> 'a2

type 'p on_global_env_ext = 'p on_global_env * __

type 'p sort_irrelevant =
  global_env_ext -> context -> term option -> universe -> universe -> 'p -> 'p

val coq_All_local_env_impl :
  context -> 'a1 coq_All_local_env -> (context -> term -> term option -> 'a1
  -> 'a2) -> 'a2 coq_All_local_env

val on_global_env_mix :
  checker_flags -> global_env -> 'a2 sort_irrelevant -> 'a1 on_global_env ->
  'a2 on_global_env -> ('a1 * 'a2) on_global_env

val coq_Alli_impl_trans :
  'a1 list -> nat -> ('a1, 'a2) coq_Alli -> (nat -> 'a1 -> 'a2 -> 'a3) ->
  ('a1, 'a3) coq_Alli

val on_global_env_impl :
  checker_flags -> global_env -> ((global_env * universes_decl) -> context ->
  term -> term option -> 'a1 on_global_env -> 'a1 -> 'a2) -> 'a1
  on_global_env -> 'a2 on_global_env

val on_global_env_proj :
  checker_flags -> global_env -> ('a1 * 'a2) on_global_env -> 'a1
  on_global_env

type 'p coq_Forall_decls_typing = 'p lift_typing on_global_env

type type_local_decl = __

val refine_type :
  checker_flags -> global_env_ext -> context -> term -> term -> term ->
  typing -> typing

type 'p coq_Forall_typing_spine =
| Forall_type_spine_nil of term
| Forall_type_spine_cons of term * term list * name * term * term * universe
   * term * term * typing_spine * typing * cumul * typing * 'p * 'p
   * 'p coq_Forall_typing_spine

val coq_Forall_typing_spine_rect :
  checker_flags -> global_env_ext -> context -> (term -> 'a2) -> (term ->
  term list -> name -> term -> term -> universe -> term -> term ->
  typing_spine -> typing -> cumul -> typing -> 'a1 -> 'a1 -> 'a1
  coq_Forall_typing_spine -> 'a2 -> 'a2) -> term -> term list -> term ->
  typing_spine -> 'a1 coq_Forall_typing_spine -> 'a2

val coq_Forall_typing_spine_rec :
  checker_flags -> global_env_ext -> context -> (term -> 'a2) -> (term ->
  term list -> name -> term -> term -> universe -> term -> term ->
  typing_spine -> typing -> cumul -> typing -> 'a1 -> 'a1 -> 'a1
  coq_Forall_typing_spine -> 'a2 -> 'a2) -> term -> term list -> term ->
  typing_spine -> 'a1 coq_Forall_typing_spine -> 'a2

val wf_local_size :
  checker_flags -> global_env_ext -> (global_env_ext -> context -> term ->
  term -> typing -> size) -> context -> typing lift_typing coq_All_local_env
  -> size

val typing_spine_size :
  checker_flags -> (global_env_ext -> context -> term -> term -> typing ->
  size) -> global_env_ext -> context -> term -> term list -> term ->
  typing_spine -> size

val typing_size :
  checker_flags -> global_env_ext -> context -> term -> term -> typing -> size

val globenv_size : global_env -> size

type wf = typing coq_Forall_decls_typing

type wf_ext = typing lift_typing on_global_env_ext

type 'p env_prop =
  (global_env * universes_decl) -> wf -> context -> typing lift_typing
  coq_All_local_env -> term -> term -> typing -> 'p
  coq_Forall_decls_typing * 'p

val env_prop_typing :
  checker_flags -> 'a1 env_prop -> (global_env * universes_decl) -> wf ->
  context -> typing lift_typing coq_All_local_env -> term -> term -> typing
  -> 'a1

val type_Prop : checker_flags -> global_env_ext -> typing

val env_prop_sigma :
  checker_flags -> 'a1 env_prop -> global_env -> wf -> 'a1
  coq_Forall_decls_typing

val wf_local_app :
  checker_flags -> global_env_ext -> context -> context -> typing lift_typing
  coq_All_local_env -> typing lift_typing coq_All_local_env

val typing_wf_local :
  checker_flags -> global_env_ext -> context -> term -> term -> typing ->
  typing lift_typing coq_All_local_env

val wf_local_inv :
  checker_flags -> global_env_ext -> context -> typing lift_typing
  coq_All_local_env -> context_decl -> context_decl list -> (typing
  lift_typing coq_All_local_env, __) sigT

val type_local_ctx_impl :
  global_env_ext -> context -> context -> universe -> 'a1 type_local_ctx ->
  (context -> term -> term option -> 'a1 -> 'a2) -> 'a2 type_local_ctx

val typing_ind_env :
  checker_flags -> ((global_env * universes_decl) -> wf -> context -> typing
  lift_typing coq_All_local_env -> nat -> context_decl -> __ -> (typing, 'a1)
  coq_All_local_env_over -> 'a1) -> ((global_env * universes_decl) -> wf ->
  context -> typing lift_typing coq_All_local_env -> Level.t -> (typing, 'a1)
  coq_All_local_env_over -> __ -> 'a1) -> ((global_env * universes_decl) ->
  wf -> context -> typing lift_typing coq_All_local_env -> term -> cast_kind
  -> term -> universe -> typing -> 'a1 -> typing -> 'a1 -> 'a1) ->
  ((global_env * universes_decl) -> wf -> context -> typing lift_typing
  coq_All_local_env -> name -> term -> term -> universe -> universe ->
  (typing, 'a1) coq_All_local_env_over -> typing -> 'a1 -> typing -> 'a1 ->
  'a1) -> ((global_env * universes_decl) -> wf -> context -> typing
  lift_typing coq_All_local_env -> name -> term -> term -> universe -> term
  -> (typing, 'a1) coq_All_local_env_over -> typing -> 'a1 -> typing -> 'a1
  -> 'a1) -> ((global_env * universes_decl) -> wf -> context -> typing
  lift_typing coq_All_local_env -> name -> term -> term -> term -> universe
  -> term -> (typing, 'a1) coq_All_local_env_over -> typing -> 'a1 -> typing
  -> 'a1 -> typing -> 'a1 -> 'a1) -> ((global_env * universes_decl) -> wf ->
  context -> typing lift_typing coq_All_local_env -> term -> term list ->
  term -> term -> typing -> 'a1 -> __ -> __ -> typing_spine -> 'a1
  coq_Forall_typing_spine -> 'a1) -> ((global_env * universes_decl) -> wf ->
  context -> typing lift_typing coq_All_local_env -> ident ->
  universe_instance -> constant_body -> 'a1 coq_Forall_decls_typing ->
  (typing, 'a1) coq_All_local_env_over -> __ -> __ -> 'a1) ->
  ((global_env * universes_decl) -> wf -> context -> typing lift_typing
  coq_All_local_env -> inductive -> universe_instance ->
  mutual_inductive_body -> one_inductive_body -> __ -> 'a1
  coq_Forall_decls_typing -> (typing, 'a1) coq_All_local_env_over -> __ ->
  'a1) -> ((global_env * universes_decl) -> wf -> context -> typing
  lift_typing coq_All_local_env -> inductive -> nat -> universe_instance ->
  mutual_inductive_body -> one_inductive_body -> ((ident * term) * nat) -> __
  -> 'a1 coq_Forall_decls_typing -> (typing, 'a1) coq_All_local_env_over ->
  __ -> 'a1) -> ((global_env * universes_decl) -> wf -> context -> typing
  lift_typing coq_All_local_env -> inductive -> universe_instance -> nat ->
  term -> term -> (nat * term) list -> term list -> mutual_inductive_body ->
  one_inductive_body -> __ -> 'a1 coq_Forall_decls_typing -> (typing, 'a1)
  coq_All_local_env_over -> __ -> term -> typing -> context_decl list ->
  context_decl list -> universe -> (nat * term) list -> __ ->
  check_correct_arity -> __ -> 'a1 -> typing -> 'a1 -> (nat * term,
  nat * term, (__ * typing) * 'a1) coq_All2 -> 'a1) ->
  ((global_env * universes_decl) -> wf -> context -> typing lift_typing
  coq_All_local_env -> projection -> term -> universe_instance ->
  mutual_inductive_body -> one_inductive_body -> (ident * term) -> __ -> term
  list -> 'a1 coq_Forall_decls_typing -> (typing, 'a1) coq_All_local_env_over
  -> typing -> 'a1 -> __ -> 'a1) -> ((global_env * universes_decl) -> wf ->
  context -> typing lift_typing coq_All_local_env -> term def list -> nat ->
  term def -> __ -> __ -> (typing * 'a1) lift_typing coq_All_local_env ->
  (term def, (typing * __) * 'a1) coq_All -> 'a1) ->
  ((global_env * universes_decl) -> wf -> context -> typing lift_typing
  coq_All_local_env -> term def list -> nat -> term def -> __ ->
  (typing * 'a1) lift_typing coq_All_local_env -> (term def, typing * 'a1)
  coq_All -> __ -> 'a1) -> ((global_env * universes_decl) -> wf -> context ->
  typing lift_typing coq_All_local_env -> term -> term -> term -> (typing,
  'a1) coq_All_local_env_over -> typing -> 'a1 -> ((typing, 'a1)
  isWfArity_prop, (universe, typing * 'a1) sigT) sum -> cumul -> 'a1) -> 'a1
  env_prop

val nth_error_All_local_env :
  context_decl list -> nat -> 'a1 coq_All_local_env -> (context_decl, 'a1
  on_local_decl) on_some

val lookup_on_global_env :
  checker_flags -> global_env -> ident -> global_decl -> 'a1 on_global_env ->
  (global_env * universes_decl, ('a1 on_global_env, 'a1 on_global_decl) sigT)
  sigT

val coq_All_local_env_app :
  context -> context -> 'a1 coq_All_local_env -> 'a1 coq_All_local_env * 'a1
  coq_All_local_env

val coq_All_local_env_lookup :
  context -> nat -> context_decl -> 'a1 coq_All_local_env -> 'a1 on_local_decl

val coq_All_local_env_app_inv :
  checker_flags -> context -> context -> ('a1 coq_All_local_env * 'a1
  coq_All_local_env) -> 'a1 coq_All_local_env

val coq_All_local_env_map :
  checker_flags -> (term -> term) -> context -> 'a1 coq_All_local_env -> 'a1
  coq_All_local_env

val lookup_wf_local :
  context -> 'a1 coq_All_local_env -> nat -> 'a1 coq_All_local_env

val lookup_wf_local_decl :
  context -> 'a1 coq_All_local_env -> nat -> context_decl -> ('a1
  coq_All_local_env, 'a1 on_local_decl) sigT

type 'p on_wf_local_decl = __

val nth_error_All_local_env_over :
  checker_flags -> global_env_ext -> context_decl list -> nat -> context_decl
  -> typing lift_typing coq_All_local_env -> (typing, 'a1)
  coq_All_local_env_over -> (typing, 'a1) coq_All_local_env_over * 'a1
  on_wf_local_decl

val wf_ext_wf : checker_flags -> global_env_ext -> wf_ext -> wf
