open BinPos
open Bool
open Datatypes
open MSetFacts
open MSetProperties
open MSetWeakList
open Nat0
open PeanoNat
open Specif
open String0
open Universes0
open Config0
open Monad_utils
open Utils
open WGraph

type __ = Obj.t

module ConstraintSetFact = WFactsOn(UnivConstraintDec)(ConstraintSet)

module ConstraintSetProp = WPropertiesOn(UnivConstraintDec)(ConstraintSet)

type variable_level =
| Coq_mLevel of char list
| Coq_mVar of nat

(** val variable_level_rect :
    (char list -> 'a1) -> (nat -> 'a1) -> variable_level -> 'a1 **)

let variable_level_rect f f0 = function
| Coq_mLevel x -> f x
| Coq_mVar x -> f0 x

(** val variable_level_rec :
    (char list -> 'a1) -> (nat -> 'a1) -> variable_level -> 'a1 **)

let variable_level_rec f f0 = function
| Coq_mLevel x -> f x
| Coq_mVar x -> f0 x

type good_constraint =
| Coq_gc_le of variable_level * variable_level
| Coq_gc_lt of variable_level * variable_level
| Coq_gc_lt_set of nat
| Coq_gc_eq_set of nat

(** val good_constraint_rect :
    (variable_level -> variable_level -> 'a1) -> (variable_level ->
    variable_level -> 'a1) -> (nat -> 'a1) -> (nat -> 'a1) -> good_constraint
    -> 'a1 **)

let good_constraint_rect f f0 f1 f2 = function
| Coq_gc_le (x, x0) -> f x x0
| Coq_gc_lt (x, x0) -> f0 x x0
| Coq_gc_lt_set x -> f1 x
| Coq_gc_eq_set x -> f2 x

(** val good_constraint_rec :
    (variable_level -> variable_level -> 'a1) -> (variable_level ->
    variable_level -> 'a1) -> (nat -> 'a1) -> (nat -> 'a1) -> good_constraint
    -> 'a1 **)

let good_constraint_rec f f0 f1 f2 = function
| Coq_gc_le (x, x0) -> f x x0
| Coq_gc_lt (x, x0) -> f0 x x0
| Coq_gc_lt_set x -> f1 x
| Coq_gc_eq_set x -> f2 x

module GoodConstraintDec =
 struct
  type t = good_constraint

  (** val variable_level_dec : variable_level -> variable_level -> bool **)

  let variable_level_dec x y =
    match x with
    | Coq_mLevel x0 ->
      (match y with
       | Coq_mLevel s0 -> string_dec x0 s0
       | Coq_mVar _ -> false)
    | Coq_mVar x0 ->
      (match y with
       | Coq_mLevel _ -> false
       | Coq_mVar n0 -> Nat.eq_dec x0 n0)

  (** val eq_dec : t -> t -> bool **)

  let eq_dec x y =
    match x with
    | Coq_gc_le (x0, x1) ->
      (match y with
       | Coq_gc_le (v1, v2) ->
         if variable_level_dec x0 v1 then variable_level_dec x1 v2 else false
       | _ -> false)
    | Coq_gc_lt (x0, x1) ->
      (match y with
       | Coq_gc_lt (v1, v2) ->
         if variable_level_dec x0 v1 then variable_level_dec x1 v2 else false
       | _ -> false)
    | Coq_gc_lt_set x0 ->
      (match y with
       | Coq_gc_lt_set n0 -> Nat.eq_dec x0 n0
       | _ -> false)
    | Coq_gc_eq_set x0 ->
      (match y with
       | Coq_gc_eq_set n0 -> Nat.eq_dec x0 n0
       | _ -> false)
 end

module GoodConstraintSet = Make(GoodConstraintDec)

module GoodConstraintSetFact = WFactsOn(GoodConstraintDec)(GoodConstraintSet)

module GoodConstraintSetProp =
 WPropertiesOn(GoodConstraintDec)(GoodConstraintSet)

(** val coq_GoodConstraintSet_pair :
    GoodConstraintSet.elt -> GoodConstraintSet.elt -> GoodConstraintSet.t **)

let coq_GoodConstraintSet_pair x y =
  GoodConstraintSet.add y (GoodConstraintSet.singleton x)

(** val gc_val0 : valuation -> variable_level -> nat **)

let gc_val0 v = function
| Coq_mLevel s0 -> Pos.to_nat (v.valuation_mono s0)
| Coq_mVar x -> v.valuation_poly x

(** val gc_satisfies0 : valuation -> good_constraint -> bool **)

let gc_satisfies0 v = function
| Coq_gc_le (l, l') -> leb (gc_val0 v l) (gc_val0 v l')
| Coq_gc_lt (l, l') -> ltb (gc_val0 v l) (gc_val0 v l')
| Coq_gc_lt_set l -> ltb O (v.valuation_poly l)
| Coq_gc_eq_set l -> Nat0.eqb O (v.valuation_poly l)

(** val gc_satisfies : valuation -> GoodConstraintSet.t -> bool **)

let gc_satisfies v =
  GoodConstraintSet.for_all (gc_satisfies0 v)

(** val gc_of_constraint :
    checker_flags -> univ_constraint -> GoodConstraintSet.t option **)

let gc_of_constraint h uc =
  let empty0 = Some GoodConstraintSet.empty in
  let singleton0 = fun x -> Some (GoodConstraintSet.singleton x) in
  let pair = fun x y -> Some (coq_GoodConstraintSet_pair x y) in
  let (p, u) = uc in
  let (u0, t0) = p in
  (match u0 with
   | Level.Coq_lProp ->
     (match t0 with
      | ConstraintType.Lt ->
        (match u with
         | Level.Coq_lProp -> None
         | _ -> if h.prop_sub_type then empty0 else None)
      | ConstraintType.Le ->
        (match u with
         | Level.Coq_lProp -> empty0
         | _ -> if h.prop_sub_type then empty0 else None)
      | ConstraintType.Eq ->
        (match u with
         | Level.Coq_lProp -> empty0
         | _ -> None))
   | Level.Coq_lSet ->
     (match t0 with
      | ConstraintType.Lt ->
        (match u with
         | Level.Level _ -> empty0
         | Level.Var n -> singleton0 (Coq_gc_lt_set n)
         | _ -> None)
      | ConstraintType.Le ->
        (match u with
         | Level.Coq_lProp -> None
         | _ -> empty0)
      | ConstraintType.Eq ->
        (match u with
         | Level.Coq_lSet -> empty0
         | Level.Var n -> singleton0 (Coq_gc_eq_set n)
         | _ -> None))
   | Level.Level l ->
     (match t0 with
      | ConstraintType.Lt ->
        (match u with
         | Level.Level l' ->
           singleton0 (Coq_gc_lt ((Coq_mLevel l), (Coq_mLevel l')))
         | Level.Var n ->
           singleton0 (Coq_gc_lt ((Coq_mLevel l), (Coq_mVar n)))
         | _ -> None)
      | ConstraintType.Le ->
        (match u with
         | Level.Level l' ->
           singleton0 (Coq_gc_le ((Coq_mLevel l), (Coq_mLevel l')))
         | Level.Var n ->
           singleton0 (Coq_gc_le ((Coq_mLevel l), (Coq_mVar n)))
         | _ -> None)
      | ConstraintType.Eq ->
        (match u with
         | Level.Level l' ->
           pair (Coq_gc_le ((Coq_mLevel l), (Coq_mLevel l'))) (Coq_gc_le
             ((Coq_mLevel l'), (Coq_mLevel l)))
         | Level.Var n ->
           pair (Coq_gc_le ((Coq_mLevel l), (Coq_mVar n))) (Coq_gc_le
             ((Coq_mVar n), (Coq_mLevel l)))
         | _ -> None))
   | Level.Var n ->
     (match t0 with
      | ConstraintType.Lt ->
        (match u with
         | Level.Level l ->
           singleton0 (Coq_gc_lt ((Coq_mVar n), (Coq_mLevel l)))
         | Level.Var n' ->
           singleton0 (Coq_gc_lt ((Coq_mVar n), (Coq_mVar n')))
         | _ -> None)
      | ConstraintType.Le ->
        (match u with
         | Level.Coq_lProp -> None
         | Level.Coq_lSet -> singleton0 (Coq_gc_eq_set n)
         | Level.Level l ->
           singleton0 (Coq_gc_le ((Coq_mVar n), (Coq_mLevel l)))
         | Level.Var n' ->
           singleton0 (Coq_gc_le ((Coq_mVar n), (Coq_mVar n'))))
      | ConstraintType.Eq ->
        (match u with
         | Level.Coq_lProp -> None
         | Level.Coq_lSet -> singleton0 (Coq_gc_eq_set n)
         | Level.Level l ->
           pair (Coq_gc_le ((Coq_mVar n), (Coq_mLevel l))) (Coq_gc_le
             ((Coq_mLevel l), (Coq_mVar n)))
         | Level.Var n' ->
           pair (Coq_gc_le ((Coq_mVar n), (Coq_mVar n'))) (Coq_gc_le
             ((Coq_mVar n'), (Coq_mVar n))))))

(** val add_gc_of_constraint :
    checker_flags -> univ_constraint -> GoodConstraintSet.t option ->
    GoodConstraintSet.t option **)

let add_gc_of_constraint cf uc s0 =
  bind (Obj.magic option_monad) s0 (fun s1 ->
    bind (Obj.magic option_monad) (gc_of_constraint cf uc) (fun s2 ->
      ret (Obj.magic option_monad) (GoodConstraintSet.union s1 s2)))

(** val gc_of_constraints :
    checker_flags -> constraints -> GoodConstraintSet.t option **)

let gc_of_constraints cf ctrs =
  ConstraintSet.fold (add_gc_of_constraint cf) ctrs (Some
    GoodConstraintSet.empty)

type no_prop_level =
| Coq_lSet
| Coq_vtn of variable_level

(** val no_prop_level_rect :
    'a1 -> (variable_level -> 'a1) -> no_prop_level -> 'a1 **)

let no_prop_level_rect f f0 = function
| Coq_lSet -> f
| Coq_vtn x -> f0 x

(** val no_prop_level_rec :
    'a1 -> (variable_level -> 'a1) -> no_prop_level -> 'a1 **)

let no_prop_level_rec f f0 = function
| Coq_lSet -> f
| Coq_vtn x -> f0 x

module VariableLevel =
 struct
  type t = variable_level

  (** val compare : t -> t -> comparison **)

  let compare x y =
    match x with
    | Coq_mLevel s0 ->
      (match y with
       | Coq_mLevel s' -> string_compare s0 s'
       | Coq_mVar _ -> Lt)
    | Coq_mVar n ->
      (match y with
       | Coq_mLevel _ -> Gt
       | Coq_mVar n' -> Nat0.compare n n')

  (** val eq_dec : t -> t -> bool **)

  let eq_dec x y =
    match x with
    | Coq_mLevel s0 ->
      (match y with
       | Coq_mLevel s' -> string_dec s0 s'
       | Coq_mVar _ -> false)
    | Coq_mVar n ->
      (match y with
       | Coq_mLevel _ -> false
       | Coq_mVar n' -> Nat.eq_dec n n')
 end

module NoPropLevel =
 struct
  type t = no_prop_level

  (** val compare : t -> t -> comparison **)

  let compare x y =
    match x with
    | Coq_lSet -> (match y with
                   | Coq_lSet -> Eq
                   | Coq_vtn _ -> Lt)
    | Coq_vtn v ->
      (match y with
       | Coq_lSet -> Gt
       | Coq_vtn v' -> VariableLevel.compare v v')

  (** val eq_dec : t -> t -> bool **)

  let eq_dec x y =
    match x with
    | Coq_lSet -> (match y with
                   | Coq_lSet -> true
                   | Coq_vtn _ -> false)
    | Coq_vtn x0 ->
      (match y with
       | Coq_lSet -> false
       | Coq_vtn l0 -> VariableLevel.eq_dec x0 l0)
 end

module Coq_wGraph = WeightedGraph(NoPropLevel)

type universes_graph = Coq_wGraph.t

(** val init_graph : universes_graph **)

let init_graph =
  (((Coq_wGraph.VSet.singleton Coq_lSet), Coq_wGraph.EdgeSet.empty), Coq_lSet)

(** val no_prop_of_level : Level.t -> no_prop_level option **)

let no_prop_of_level = function
| Level.Coq_lProp -> None
| Level.Coq_lSet -> Some Coq_lSet
| Level.Level s0 -> Some (Coq_vtn (Coq_mLevel s0))
| Level.Var n -> Some (Coq_vtn (Coq_mVar n))

(** val no_prop_levels : LevelSet.t -> Coq_wGraph.VSet.t **)

let no_prop_levels x =
  LevelSet.fold (fun l x0 ->
    match no_prop_of_level l with
    | Some l0 -> Coq_wGraph.VSet.add l0 x0
    | None -> x0) x Coq_wGraph.VSet.empty

(** val gc_of_uctx :
    checker_flags -> ContextSet.t ->
    (Coq_wGraph.VSet.t * GoodConstraintSet.t) option **)

let gc_of_uctx h uctx =
  bind (Obj.magic option_monad) (Obj.magic gc_of_constraints h (snd uctx))
    (fun ctrs ->
    ret (Obj.magic option_monad) ((no_prop_levels (fst uctx)), ctrs))

(** val level_of_variable : variable_level -> Level.t **)

let level_of_variable = function
| Coq_mLevel s0 -> Level.Level s0
| Coq_mVar n -> Level.Var n

(** val level_of_no_prop : no_prop_level -> Level.t **)

let level_of_no_prop = function
| Coq_lSet -> Level.Coq_lSet
| Coq_vtn l0 -> level_of_variable l0

(** val edge_of_level : variable_level -> Coq_wGraph.EdgeSet.elt **)

let edge_of_level = function
| Coq_mLevel l0 -> ((Coq_lSet, (S O)), (Coq_vtn (Coq_mLevel l0)))
| Coq_mVar n -> ((Coq_lSet, O), (Coq_vtn (Coq_mVar n)))

(** val coq_EdgeSet_pair :
    Coq_wGraph.EdgeSet.elt -> Coq_wGraph.EdgeSet.elt -> Coq_wGraph.EdgeSet.t **)

let coq_EdgeSet_pair x y =
  Coq_wGraph.EdgeSet.add y (Coq_wGraph.EdgeSet.singleton x)

(** val coq_EdgeSet_triple :
    Coq_wGraph.EdgeSet.elt -> Coq_wGraph.EdgeSet.elt ->
    Coq_wGraph.EdgeSet.elt -> Coq_wGraph.EdgeSet.t **)

let coq_EdgeSet_triple x y z =
  Coq_wGraph.EdgeSet.add z (coq_EdgeSet_pair x y)

(** val edge_of_constraint : good_constraint -> Coq_wGraph.EdgeSet.elt **)

let edge_of_constraint = function
| Coq_gc_le (l, l') -> (((Coq_vtn l), O), (Coq_vtn l'))
| Coq_gc_lt (l, l') -> (((Coq_vtn l), (S O)), (Coq_vtn l'))
| Coq_gc_lt_set n -> ((Coq_lSet, (S O)), (Coq_vtn (Coq_mVar n)))
| Coq_gc_eq_set n -> (((Coq_vtn (Coq_mVar n)), O), Coq_lSet)

(** val make_graph :
    (Coq_wGraph.VSet.t * GoodConstraintSet.t) -> Coq_wGraph.t **)

let make_graph uctx =
  let init_edges =
    Coq_wGraph.VSet.fold (fun l e ->
      match l with
      | Coq_lSet -> e
      | Coq_vtn l0 -> Coq_wGraph.EdgeSet.add (edge_of_level l0) e) (fst uctx)
      Coq_wGraph.EdgeSet.empty
  in
  let edges =
    GoodConstraintSet.fold (fun ctr ->
      Coq_wGraph.EdgeSet.add (edge_of_constraint ctr)) (snd uctx) init_edges
  in
  (((fst uctx), edges), Coq_lSet)

(** val labelling_of_valuation : valuation -> Coq_wGraph.labelling **)

let labelling_of_valuation v = function
| Coq_lSet -> O
| Coq_vtn l0 ->
  (match l0 with
   | Coq_mLevel l -> Pos.to_nat (v.valuation_mono l)
   | Coq_mVar n -> v.valuation_poly n)

(** val valuation_of_labelling : Coq_wGraph.labelling -> valuation **)

let valuation_of_labelling l =
  { valuation_mono = (fun s0 -> Pos.of_nat (l (Coq_vtn (Coq_mLevel s0))));
    valuation_poly = (fun n -> l (Coq_vtn (Coq_mVar n))) }

(** val is_consistent : checker_flags -> ContextSet.t -> bool **)

let is_consistent h uctx =
  match gc_of_uctx h uctx with
  | Some uctx0 -> Coq_wGraph.is_acyclic (make_graph uctx0)
  | None -> false

(** val leqb_no_prop_n :
    universes_graph -> nat -> no_prop_level -> no_prop_level -> bool **)

let leqb_no_prop_n =
  Coq_wGraph.leqb_vertices

(** val leqb_level_n :
    universes_graph -> nat -> Level.t -> Level.t -> bool **)

let leqb_level_n g n l l' =
  match no_prop_of_level l with
  | Some l0 ->
    (match no_prop_of_level l' with
     | Some l'0 -> leqb_no_prop_n g n l0 l'0
     | None -> false)
  | None ->
    (match no_prop_of_level l' with
     | Some l'0 ->
       (match n with
        | O -> true
        | S n0 -> leqb_no_prop_n g n0 Coq_lSet l'0)
     | None -> Nat0.eqb n O)

(** val leqb_level :
    checker_flags -> universes_graph -> Level.t -> Level.t -> bool **)

let leqb_level cf g l l' =
  (||) (negb cf.check_univs) (leqb_level_n g O l l')

(** val eqb_level :
    checker_flags -> universes_graph -> Level.t -> Level.t -> bool **)

let eqb_level cf g l l' =
  (&&) (leqb_level cf g l l') (leqb_level cf g l' l)

(** val eqb_univ_instance :
    checker_flags -> universes_graph -> universe_instance ->
    universe_instance -> bool **)

let eqb_univ_instance cf g u1 u2 =
  forallb2 (eqb_level cf g) u1 u2

(** val leqb_expr_n :
    universes_graph -> nat -> Universe.Expr.t -> Universe.Expr.t -> bool **)

let leqb_expr_n g n e1 e2 =
  if (&&) (snd e1) (negb (Level.is_prop (fst e1)))
  then if (&&) (snd e2) (negb (Level.is_prop (fst e2)))
       then leqb_level_n g n (fst e1) (fst e2)
       else leqb_level_n g (S n) (fst e1) (fst e2)
  else if (&&) (snd e2) (negb (Level.is_prop (fst e2)))
       then leqb_level_n g (pred n) (fst e1) (fst e2)
       else leqb_level_n g n (fst e1) (fst e2)

(** val leqb_univ_expr_n :
    universes_graph -> nat -> universe -> Universe.Expr.t -> bool **)

let rec leqb_univ_expr_n g n u e2 =
  match u with
  | NEL.Coq_sing e1 -> leqb_expr_n g n e1 e2
  | NEL.Coq_cons (e1, u0) ->
    (&&) (leqb_expr_n g n e1 e2) (leqb_univ_expr_n g n u0 e2)

(** val leqb_univ_expr :
    checker_flags -> universes_graph -> universe -> Universe.Expr.t -> bool **)

let leqb_univ_expr cf g u e2 =
  (||) (negb cf.check_univs) (leqb_univ_expr_n g O u e2)

(** val try_leqb_universe_n :
    universes_graph -> nat -> universe -> universe -> bool **)

let rec try_leqb_universe_n g n u1 = function
| NEL.Coq_sing e -> leqb_univ_expr_n g n u1 e
| NEL.Coq_cons (e, u3) ->
  (||) (leqb_univ_expr_n g n u1 e) (try_leqb_universe_n g n u1 u3)

(** val try_leqb_universe :
    checker_flags -> universes_graph -> universe -> universe -> bool **)

let try_leqb_universe cf g u1 u2 =
  (||) (negb cf.check_univs) (try_leqb_universe_n g O u1 u2)

(** val try_eqb_universe :
    checker_flags -> universes_graph -> universe -> universe -> bool **)

let try_eqb_universe cf g u1 u2 =
  (||) (negb cf.check_univs)
    ((&&) (try_leqb_universe_n g O u1 u2) (try_leqb_universe_n g O u2 u1))

(** val check_gc_constraint :
    checker_flags -> universes_graph -> good_constraint -> bool **)

let check_gc_constraint cf g gc =
  (||) (negb cf.check_univs)
    (match gc with
     | Coq_gc_le (l, l') -> leqb_no_prop_n g O (Coq_vtn l) (Coq_vtn l')
     | Coq_gc_lt (l, l') -> leqb_no_prop_n g (S O) (Coq_vtn l) (Coq_vtn l')
     | Coq_gc_lt_set n ->
       leqb_no_prop_n g (S O) Coq_lSet (Coq_vtn (Coq_mVar n))
     | Coq_gc_eq_set n -> leqb_no_prop_n g O (Coq_vtn (Coq_mVar n)) Coq_lSet)

(** val check_gc_constraints :
    checker_flags -> universes_graph -> GoodConstraintSet.t -> bool **)

let check_gc_constraints cf g =
  GoodConstraintSet.for_all (check_gc_constraint cf g)

(** val check_constraints :
    checker_flags -> universes_graph -> constraints -> bool **)

let check_constraints cf g ctrs =
  match gc_of_constraints cf ctrs with
  | Some ctrs0 -> check_gc_constraints cf g ctrs0
  | None -> false
