open BinPos
open Bool
open Datatypes
open MSetFacts
open MSetProperties
open MSetWeakList
open Nat0
open PeanoNat
open Specif
open String0
open Universes0
open Config0
open Monad_utils
open Utils
open WGraph

type __ = Obj.t

module ConstraintSetFact :
 sig
  val eqb : univ_constraint -> univ_constraint -> bool
 end

module ConstraintSetProp :
 sig
  module Dec :
   sig
    module F :
     sig
      val eqb : univ_constraint -> univ_constraint -> bool
     end

    module MSetLogicalFacts :
     sig
     end

    module MSetDecideAuxiliary :
     sig
     end

    module MSetDecideTestCases :
     sig
     end
   end

  module FM :
   sig
    val eqb : univ_constraint -> univ_constraint -> bool
   end

  val coq_In_dec : ConstraintSet.elt -> ConstraintSet.t -> bool

  val of_list : ConstraintSet.elt list -> ConstraintSet.t

  val to_list : ConstraintSet.t -> ConstraintSet.elt list

  val fold_rec :
    (ConstraintSet.elt -> 'a1 -> 'a1) -> 'a1 -> ConstraintSet.t ->
    (ConstraintSet.t -> __ -> 'a2) -> (ConstraintSet.elt -> 'a1 ->
    ConstraintSet.t -> ConstraintSet.t -> __ -> __ -> __ -> 'a2 -> 'a2) -> 'a2

  val fold_rec_bis :
    (ConstraintSet.elt -> 'a1 -> 'a1) -> 'a1 -> ConstraintSet.t ->
    (ConstraintSet.t -> ConstraintSet.t -> 'a1 -> __ -> 'a2 -> 'a2) -> 'a2 ->
    (ConstraintSet.elt -> 'a1 -> ConstraintSet.t -> __ -> __ -> 'a2 -> 'a2)
    -> 'a2

  val fold_rec_nodep :
    (ConstraintSet.elt -> 'a1 -> 'a1) -> 'a1 -> ConstraintSet.t -> 'a2 ->
    (ConstraintSet.elt -> 'a1 -> __ -> 'a2 -> 'a2) -> 'a2

  val fold_rec_weak :
    (ConstraintSet.elt -> 'a1 -> 'a1) -> 'a1 -> (ConstraintSet.t ->
    ConstraintSet.t -> 'a1 -> __ -> 'a2 -> 'a2) -> 'a2 -> (ConstraintSet.elt
    -> 'a1 -> ConstraintSet.t -> __ -> 'a2 -> 'a2) -> ConstraintSet.t -> 'a2

  val fold_rel :
    (ConstraintSet.elt -> 'a1 -> 'a1) -> (ConstraintSet.elt -> 'a2 -> 'a2) ->
    'a1 -> 'a2 -> ConstraintSet.t -> 'a3 -> (ConstraintSet.elt -> 'a1 -> 'a2
    -> __ -> 'a3 -> 'a3) -> 'a3

  val set_induction :
    (ConstraintSet.t -> __ -> 'a1) -> (ConstraintSet.t -> ConstraintSet.t ->
    'a1 -> ConstraintSet.elt -> __ -> __ -> 'a1) -> ConstraintSet.t -> 'a1

  val set_induction_bis :
    (ConstraintSet.t -> ConstraintSet.t -> __ -> 'a1 -> 'a1) -> 'a1 ->
    (ConstraintSet.elt -> ConstraintSet.t -> __ -> 'a1 -> 'a1) ->
    ConstraintSet.t -> 'a1

  val cardinal_inv_2 : ConstraintSet.t -> nat -> ConstraintSet.elt

  val cardinal_inv_2b : ConstraintSet.t -> ConstraintSet.elt
 end

type variable_level =
| Coq_mLevel of char list
| Coq_mVar of nat

val variable_level_rect :
  (char list -> 'a1) -> (nat -> 'a1) -> variable_level -> 'a1

val variable_level_rec :
  (char list -> 'a1) -> (nat -> 'a1) -> variable_level -> 'a1

type good_constraint =
| Coq_gc_le of variable_level * variable_level
| Coq_gc_lt of variable_level * variable_level
| Coq_gc_lt_set of nat
| Coq_gc_eq_set of nat

val good_constraint_rect :
  (variable_level -> variable_level -> 'a1) -> (variable_level ->
  variable_level -> 'a1) -> (nat -> 'a1) -> (nat -> 'a1) -> good_constraint
  -> 'a1

val good_constraint_rec :
  (variable_level -> variable_level -> 'a1) -> (variable_level ->
  variable_level -> 'a1) -> (nat -> 'a1) -> (nat -> 'a1) -> good_constraint
  -> 'a1

module GoodConstraintDec :
 sig
  type t = good_constraint

  val variable_level_dec : variable_level -> variable_level -> bool

  val eq_dec : t -> t -> bool
 end

module GoodConstraintSet :
 sig
  module Raw :
   sig
    type elt = good_constraint

    type t = elt list

    val empty : t

    val is_empty : t -> bool

    val mem : elt -> t -> bool

    val add : elt -> t -> t

    val singleton : elt -> t

    val remove : elt -> t -> t

    val fold : (elt -> 'a1 -> 'a1) -> t -> 'a1 -> 'a1

    val union : t -> t -> t

    val diff : t -> t -> t

    val inter : t -> t -> t

    val subset : t -> t -> bool

    val equal : t -> t -> bool

    val filter : (elt -> bool) -> t -> t

    val for_all : (elt -> bool) -> t -> bool

    val exists_ : (elt -> bool) -> t -> bool

    val partition : (elt -> bool) -> t -> t * t

    val cardinal : t -> nat

    val elements : t -> elt list

    val choose : t -> elt option

    val isok : elt list -> bool
   end

  module E :
   sig
    type t = good_constraint

    val eq_dec : good_constraint -> good_constraint -> bool
   end

  type elt = good_constraint

  type t_ = Raw.t
    (* singleton inductive, whose constructor was Mkt *)

  val this : t_ -> Raw.t

  type t = t_

  val mem : elt -> t -> bool

  val add : elt -> t -> t

  val remove : elt -> t -> t

  val singleton : elt -> t

  val union : t -> t -> t

  val inter : t -> t -> t

  val diff : t -> t -> t

  val equal : t -> t -> bool

  val subset : t -> t -> bool

  val empty : t

  val is_empty : t -> bool

  val elements : t -> elt list

  val choose : t -> elt option

  val fold : (elt -> 'a1 -> 'a1) -> t -> 'a1 -> 'a1

  val cardinal : t -> nat

  val filter : (elt -> bool) -> t -> t

  val for_all : (elt -> bool) -> t -> bool

  val exists_ : (elt -> bool) -> t -> bool

  val partition : (elt -> bool) -> t -> t * t

  val eq_dec : t -> t -> bool
 end

module GoodConstraintSetFact :
 sig
  val eqb : good_constraint -> good_constraint -> bool
 end

module GoodConstraintSetProp :
 sig
  module Dec :
   sig
    module F :
     sig
      val eqb : good_constraint -> good_constraint -> bool
     end

    module MSetLogicalFacts :
     sig
     end

    module MSetDecideAuxiliary :
     sig
     end

    module MSetDecideTestCases :
     sig
     end
   end

  module FM :
   sig
    val eqb : good_constraint -> good_constraint -> bool
   end

  val coq_In_dec : GoodConstraintSet.elt -> GoodConstraintSet.t -> bool

  val of_list : GoodConstraintSet.elt list -> GoodConstraintSet.t

  val to_list : GoodConstraintSet.t -> GoodConstraintSet.elt list

  val fold_rec :
    (GoodConstraintSet.elt -> 'a1 -> 'a1) -> 'a1 -> GoodConstraintSet.t ->
    (GoodConstraintSet.t -> __ -> 'a2) -> (GoodConstraintSet.elt -> 'a1 ->
    GoodConstraintSet.t -> GoodConstraintSet.t -> __ -> __ -> __ -> 'a2 ->
    'a2) -> 'a2

  val fold_rec_bis :
    (GoodConstraintSet.elt -> 'a1 -> 'a1) -> 'a1 -> GoodConstraintSet.t ->
    (GoodConstraintSet.t -> GoodConstraintSet.t -> 'a1 -> __ -> 'a2 -> 'a2)
    -> 'a2 -> (GoodConstraintSet.elt -> 'a1 -> GoodConstraintSet.t -> __ ->
    __ -> 'a2 -> 'a2) -> 'a2

  val fold_rec_nodep :
    (GoodConstraintSet.elt -> 'a1 -> 'a1) -> 'a1 -> GoodConstraintSet.t ->
    'a2 -> (GoodConstraintSet.elt -> 'a1 -> __ -> 'a2 -> 'a2) -> 'a2

  val fold_rec_weak :
    (GoodConstraintSet.elt -> 'a1 -> 'a1) -> 'a1 -> (GoodConstraintSet.t ->
    GoodConstraintSet.t -> 'a1 -> __ -> 'a2 -> 'a2) -> 'a2 ->
    (GoodConstraintSet.elt -> 'a1 -> GoodConstraintSet.t -> __ -> 'a2 -> 'a2)
    -> GoodConstraintSet.t -> 'a2

  val fold_rel :
    (GoodConstraintSet.elt -> 'a1 -> 'a1) -> (GoodConstraintSet.elt -> 'a2 ->
    'a2) -> 'a1 -> 'a2 -> GoodConstraintSet.t -> 'a3 ->
    (GoodConstraintSet.elt -> 'a1 -> 'a2 -> __ -> 'a3 -> 'a3) -> 'a3

  val set_induction :
    (GoodConstraintSet.t -> __ -> 'a1) -> (GoodConstraintSet.t ->
    GoodConstraintSet.t -> 'a1 -> GoodConstraintSet.elt -> __ -> __ -> 'a1)
    -> GoodConstraintSet.t -> 'a1

  val set_induction_bis :
    (GoodConstraintSet.t -> GoodConstraintSet.t -> __ -> 'a1 -> 'a1) -> 'a1
    -> (GoodConstraintSet.elt -> GoodConstraintSet.t -> __ -> 'a1 -> 'a1) ->
    GoodConstraintSet.t -> 'a1

  val cardinal_inv_2 : GoodConstraintSet.t -> nat -> GoodConstraintSet.elt

  val cardinal_inv_2b : GoodConstraintSet.t -> GoodConstraintSet.elt
 end

val coq_GoodConstraintSet_pair :
  GoodConstraintSet.elt -> GoodConstraintSet.elt -> GoodConstraintSet.t

val gc_val0 : valuation -> variable_level -> nat

val gc_satisfies0 : valuation -> good_constraint -> bool

val gc_satisfies : valuation -> GoodConstraintSet.t -> bool

val gc_of_constraint :
  checker_flags -> univ_constraint -> GoodConstraintSet.t option

val add_gc_of_constraint :
  checker_flags -> univ_constraint -> GoodConstraintSet.t option ->
  GoodConstraintSet.t option

val gc_of_constraints :
  checker_flags -> constraints -> GoodConstraintSet.t option

type no_prop_level =
| Coq_lSet
| Coq_vtn of variable_level

val no_prop_level_rect :
  'a1 -> (variable_level -> 'a1) -> no_prop_level -> 'a1

val no_prop_level_rec : 'a1 -> (variable_level -> 'a1) -> no_prop_level -> 'a1

module VariableLevel :
 sig
  type t = variable_level

  val compare : t -> t -> comparison

  val eq_dec : t -> t -> bool
 end

module NoPropLevel :
 sig
  type t = no_prop_level

  val compare : t -> t -> comparison

  val eq_dec : t -> t -> bool
 end

module Coq_wGraph :
 sig
  module VSet :
   sig
    module Raw :
     sig
      module MX :
       sig
        module OrderTac :
         sig
          module OTF :
           sig
            type t = no_prop_level

            val compare : no_prop_level -> no_prop_level -> comparison

            val eq_dec : no_prop_level -> no_prop_level -> bool
           end

          module TO :
           sig
            type t = no_prop_level

            val compare : no_prop_level -> no_prop_level -> comparison

            val eq_dec : no_prop_level -> no_prop_level -> bool
           end
         end

        val eq_dec : no_prop_level -> no_prop_level -> bool

        val lt_dec : no_prop_level -> no_prop_level -> bool

        val eqb : no_prop_level -> no_prop_level -> bool
       end

      module ML :
       sig
       end

      type elt = no_prop_level

      type t = elt list

      val empty : t

      val is_empty : t -> bool

      val mem : no_prop_level -> no_prop_level list -> bool

      val add : no_prop_level -> no_prop_level list -> no_prop_level list

      val singleton : elt -> elt list

      val remove : no_prop_level -> no_prop_level list -> t

      val union : t -> t -> t

      val inter : t -> t -> t

      val diff : t -> t -> t

      val equal : t -> t -> bool

      val subset : no_prop_level list -> no_prop_level list -> bool

      val fold : (elt -> 'a1 -> 'a1) -> t -> 'a1 -> 'a1

      val filter : (elt -> bool) -> t -> t

      val for_all : (elt -> bool) -> t -> bool

      val exists_ : (elt -> bool) -> t -> bool

      val partition : (elt -> bool) -> t -> t * t

      val cardinal : t -> nat

      val elements : t -> elt list

      val min_elt : t -> elt option

      val max_elt : t -> elt option

      val choose : t -> elt option

      val compare : no_prop_level list -> no_prop_level list -> comparison

      val inf : no_prop_level -> no_prop_level list -> bool

      val isok : no_prop_level list -> bool

      module L :
       sig
        module MO :
         sig
          module OrderTac :
           sig
            module OTF :
             sig
              type t = no_prop_level

              val compare : no_prop_level -> no_prop_level -> comparison

              val eq_dec : no_prop_level -> no_prop_level -> bool
             end

            module TO :
             sig
              type t = no_prop_level

              val compare : no_prop_level -> no_prop_level -> comparison

              val eq_dec : no_prop_level -> no_prop_level -> bool
             end
           end

          val eq_dec : no_prop_level -> no_prop_level -> bool

          val lt_dec : no_prop_level -> no_prop_level -> bool

          val eqb : no_prop_level -> no_prop_level -> bool
         end
       end
     end

    module E :
     sig
      type t = no_prop_level

      val compare : no_prop_level -> no_prop_level -> comparison

      val eq_dec : no_prop_level -> no_prop_level -> bool
     end

    type elt = no_prop_level

    type t_ = Raw.t
      (* singleton inductive, whose constructor was Mkt *)

    val this : t_ -> Raw.t

    type t = t_

    val mem : elt -> t -> bool

    val add : elt -> t -> t

    val remove : elt -> t -> t

    val singleton : elt -> t

    val union : t -> t -> t

    val inter : t -> t -> t

    val diff : t -> t -> t

    val equal : t -> t -> bool

    val subset : t -> t -> bool

    val empty : t

    val is_empty : t -> bool

    val elements : t -> elt list

    val choose : t -> elt option

    val fold : (elt -> 'a1 -> 'a1) -> t -> 'a1 -> 'a1

    val cardinal : t -> nat

    val filter : (elt -> bool) -> t -> t

    val for_all : (elt -> bool) -> t -> bool

    val exists_ : (elt -> bool) -> t -> bool

    val partition : (elt -> bool) -> t -> t * t

    val eq_dec : t -> t -> bool

    val compare : t -> t -> comparison

    val min_elt : t -> elt option

    val max_elt : t -> elt option
   end

  module VSetFact :
   sig
    val eqb : no_prop_level -> no_prop_level -> bool
   end

  module VSetProp :
   sig
    module Dec :
     sig
      module F :
       sig
        val eqb : no_prop_level -> no_prop_level -> bool
       end

      module MSetLogicalFacts :
       sig
       end

      module MSetDecideAuxiliary :
       sig
       end

      module MSetDecideTestCases :
       sig
       end
     end

    module FM :
     sig
      val eqb : no_prop_level -> no_prop_level -> bool
     end

    val coq_In_dec : VSet.elt -> VSet.t -> bool

    val of_list : VSet.elt list -> VSet.t

    val to_list : VSet.t -> VSet.elt list

    val fold_rec :
      (VSet.elt -> 'a1 -> 'a1) -> 'a1 -> VSet.t -> (VSet.t -> __ -> 'a2) ->
      (VSet.elt -> 'a1 -> VSet.t -> VSet.t -> __ -> __ -> __ -> 'a2 -> 'a2)
      -> 'a2

    val fold_rec_bis :
      (VSet.elt -> 'a1 -> 'a1) -> 'a1 -> VSet.t -> (VSet.t -> VSet.t -> 'a1
      -> __ -> 'a2 -> 'a2) -> 'a2 -> (VSet.elt -> 'a1 -> VSet.t -> __ -> __
      -> 'a2 -> 'a2) -> 'a2

    val fold_rec_nodep :
      (VSet.elt -> 'a1 -> 'a1) -> 'a1 -> VSet.t -> 'a2 -> (VSet.elt -> 'a1 ->
      __ -> 'a2 -> 'a2) -> 'a2

    val fold_rec_weak :
      (VSet.elt -> 'a1 -> 'a1) -> 'a1 -> (VSet.t -> VSet.t -> 'a1 -> __ ->
      'a2 -> 'a2) -> 'a2 -> (VSet.elt -> 'a1 -> VSet.t -> __ -> 'a2 -> 'a2)
      -> VSet.t -> 'a2

    val fold_rel :
      (VSet.elt -> 'a1 -> 'a1) -> (VSet.elt -> 'a2 -> 'a2) -> 'a1 -> 'a2 ->
      VSet.t -> 'a3 -> (VSet.elt -> 'a1 -> 'a2 -> __ -> 'a3 -> 'a3) -> 'a3

    val set_induction :
      (VSet.t -> __ -> 'a1) -> (VSet.t -> VSet.t -> 'a1 -> VSet.elt -> __ ->
      __ -> 'a1) -> VSet.t -> 'a1

    val set_induction_bis :
      (VSet.t -> VSet.t -> __ -> 'a1 -> 'a1) -> 'a1 -> (VSet.elt -> VSet.t ->
      __ -> 'a1 -> 'a1) -> VSet.t -> 'a1

    val cardinal_inv_2 : VSet.t -> nat -> VSet.elt

    val cardinal_inv_2b : VSet.t -> VSet.elt
   end

  module Edge :
   sig
    type t = (no_prop_level * nat) * no_prop_level

    val compare : t -> t -> comparison

    val eq_dec : t -> t -> bool

    val eqb : t -> t -> bool
   end

  module EdgeSet :
   sig
    module Raw :
     sig
      module MX :
       sig
        module OrderTac :
         sig
          module OTF :
           sig
            type t = (no_prop_level * nat) * no_prop_level

            val compare :
              ((no_prop_level * nat) * no_prop_level) ->
              ((no_prop_level * nat) * no_prop_level) -> comparison

            val eq_dec :
              ((no_prop_level * nat) * no_prop_level) ->
              ((no_prop_level * nat) * no_prop_level) -> bool
           end

          module TO :
           sig
            type t = (no_prop_level * nat) * no_prop_level

            val compare :
              ((no_prop_level * nat) * no_prop_level) ->
              ((no_prop_level * nat) * no_prop_level) -> comparison

            val eq_dec :
              ((no_prop_level * nat) * no_prop_level) ->
              ((no_prop_level * nat) * no_prop_level) -> bool
           end
         end

        val eq_dec :
          ((no_prop_level * nat) * no_prop_level) ->
          ((no_prop_level * nat) * no_prop_level) -> bool

        val lt_dec :
          ((no_prop_level * nat) * no_prop_level) ->
          ((no_prop_level * nat) * no_prop_level) -> bool

        val eqb :
          ((no_prop_level * nat) * no_prop_level) ->
          ((no_prop_level * nat) * no_prop_level) -> bool
       end

      module ML :
       sig
       end

      type elt = (no_prop_level * nat) * no_prop_level

      type t = elt list

      val empty : t

      val is_empty : t -> bool

      val mem :
        ((no_prop_level * nat) * no_prop_level) ->
        ((no_prop_level * nat) * no_prop_level) list -> bool

      val add :
        ((no_prop_level * nat) * no_prop_level) ->
        ((no_prop_level * nat) * no_prop_level) list ->
        ((no_prop_level * nat) * no_prop_level) list

      val singleton : elt -> elt list

      val remove :
        ((no_prop_level * nat) * no_prop_level) ->
        ((no_prop_level * nat) * no_prop_level) list -> t

      val union : t -> t -> t

      val inter : t -> t -> t

      val diff : t -> t -> t

      val equal : t -> t -> bool

      val subset :
        ((no_prop_level * nat) * no_prop_level) list ->
        ((no_prop_level * nat) * no_prop_level) list -> bool

      val fold : (elt -> 'a1 -> 'a1) -> t -> 'a1 -> 'a1

      val filter : (elt -> bool) -> t -> t

      val for_all : (elt -> bool) -> t -> bool

      val exists_ : (elt -> bool) -> t -> bool

      val partition : (elt -> bool) -> t -> t * t

      val cardinal : t -> nat

      val elements : t -> elt list

      val min_elt : t -> elt option

      val max_elt : t -> elt option

      val choose : t -> elt option

      val compare :
        ((no_prop_level * nat) * no_prop_level) list ->
        ((no_prop_level * nat) * no_prop_level) list -> comparison

      val inf :
        ((no_prop_level * nat) * no_prop_level) ->
        ((no_prop_level * nat) * no_prop_level) list -> bool

      val isok : ((no_prop_level * nat) * no_prop_level) list -> bool

      module L :
       sig
        module MO :
         sig
          module OrderTac :
           sig
            module OTF :
             sig
              type t = (no_prop_level * nat) * no_prop_level

              val compare :
                ((no_prop_level * nat) * no_prop_level) ->
                ((no_prop_level * nat) * no_prop_level) -> comparison

              val eq_dec :
                ((no_prop_level * nat) * no_prop_level) ->
                ((no_prop_level * nat) * no_prop_level) -> bool
             end

            module TO :
             sig
              type t = (no_prop_level * nat) * no_prop_level

              val compare :
                ((no_prop_level * nat) * no_prop_level) ->
                ((no_prop_level * nat) * no_prop_level) -> comparison

              val eq_dec :
                ((no_prop_level * nat) * no_prop_level) ->
                ((no_prop_level * nat) * no_prop_level) -> bool
             end
           end

          val eq_dec :
            ((no_prop_level * nat) * no_prop_level) ->
            ((no_prop_level * nat) * no_prop_level) -> bool

          val lt_dec :
            ((no_prop_level * nat) * no_prop_level) ->
            ((no_prop_level * nat) * no_prop_level) -> bool

          val eqb :
            ((no_prop_level * nat) * no_prop_level) ->
            ((no_prop_level * nat) * no_prop_level) -> bool
         end
       end
     end

    module E :
     sig
      type t = (no_prop_level * nat) * no_prop_level

      val compare :
        ((no_prop_level * nat) * no_prop_level) ->
        ((no_prop_level * nat) * no_prop_level) -> comparison

      val eq_dec :
        ((no_prop_level * nat) * no_prop_level) ->
        ((no_prop_level * nat) * no_prop_level) -> bool
     end

    type elt = (no_prop_level * nat) * no_prop_level

    type t_ = Raw.t
      (* singleton inductive, whose constructor was Mkt *)

    val this : t_ -> Raw.t

    type t = t_

    val mem : elt -> t -> bool

    val add : elt -> t -> t

    val remove : elt -> t -> t

    val singleton : elt -> t

    val union : t -> t -> t

    val inter : t -> t -> t

    val diff : t -> t -> t

    val equal : t -> t -> bool

    val subset : t -> t -> bool

    val empty : t

    val is_empty : t -> bool

    val elements : t -> elt list

    val choose : t -> elt option

    val fold : (elt -> 'a1 -> 'a1) -> t -> 'a1 -> 'a1

    val cardinal : t -> nat

    val filter : (elt -> bool) -> t -> t

    val for_all : (elt -> bool) -> t -> bool

    val exists_ : (elt -> bool) -> t -> bool

    val partition : (elt -> bool) -> t -> t * t

    val eq_dec : t -> t -> bool

    val compare : t -> t -> comparison

    val min_elt : t -> elt option

    val max_elt : t -> elt option
   end

  module EdgeSetFact :
   sig
    val eqb :
      ((no_prop_level * nat) * no_prop_level) ->
      ((no_prop_level * nat) * no_prop_level) -> bool
   end

  module EdgeSetProp :
   sig
    module Dec :
     sig
      module F :
       sig
        val eqb :
          ((no_prop_level * nat) * no_prop_level) ->
          ((no_prop_level * nat) * no_prop_level) -> bool
       end

      module MSetLogicalFacts :
       sig
       end

      module MSetDecideAuxiliary :
       sig
       end

      module MSetDecideTestCases :
       sig
       end
     end

    module FM :
     sig
      val eqb :
        ((no_prop_level * nat) * no_prop_level) ->
        ((no_prop_level * nat) * no_prop_level) -> bool
     end

    val coq_In_dec : EdgeSet.elt -> EdgeSet.t -> bool

    val of_list : EdgeSet.elt list -> EdgeSet.t

    val to_list : EdgeSet.t -> EdgeSet.elt list

    val fold_rec :
      (EdgeSet.elt -> 'a1 -> 'a1) -> 'a1 -> EdgeSet.t -> (EdgeSet.t -> __ ->
      'a2) -> (EdgeSet.elt -> 'a1 -> EdgeSet.t -> EdgeSet.t -> __ -> __ -> __
      -> 'a2 -> 'a2) -> 'a2

    val fold_rec_bis :
      (EdgeSet.elt -> 'a1 -> 'a1) -> 'a1 -> EdgeSet.t -> (EdgeSet.t ->
      EdgeSet.t -> 'a1 -> __ -> 'a2 -> 'a2) -> 'a2 -> (EdgeSet.elt -> 'a1 ->
      EdgeSet.t -> __ -> __ -> 'a2 -> 'a2) -> 'a2

    val fold_rec_nodep :
      (EdgeSet.elt -> 'a1 -> 'a1) -> 'a1 -> EdgeSet.t -> 'a2 -> (EdgeSet.elt
      -> 'a1 -> __ -> 'a2 -> 'a2) -> 'a2

    val fold_rec_weak :
      (EdgeSet.elt -> 'a1 -> 'a1) -> 'a1 -> (EdgeSet.t -> EdgeSet.t -> 'a1 ->
      __ -> 'a2 -> 'a2) -> 'a2 -> (EdgeSet.elt -> 'a1 -> EdgeSet.t -> __ ->
      'a2 -> 'a2) -> EdgeSet.t -> 'a2

    val fold_rel :
      (EdgeSet.elt -> 'a1 -> 'a1) -> (EdgeSet.elt -> 'a2 -> 'a2) -> 'a1 ->
      'a2 -> EdgeSet.t -> 'a3 -> (EdgeSet.elt -> 'a1 -> 'a2 -> __ -> 'a3 ->
      'a3) -> 'a3

    val set_induction :
      (EdgeSet.t -> __ -> 'a1) -> (EdgeSet.t -> EdgeSet.t -> 'a1 ->
      EdgeSet.elt -> __ -> __ -> 'a1) -> EdgeSet.t -> 'a1

    val set_induction_bis :
      (EdgeSet.t -> EdgeSet.t -> __ -> 'a1 -> 'a1) -> 'a1 -> (EdgeSet.elt ->
      EdgeSet.t -> __ -> 'a1 -> 'a1) -> EdgeSet.t -> 'a1

    val cardinal_inv_2 : EdgeSet.t -> nat -> EdgeSet.elt

    val cardinal_inv_2b : EdgeSet.t -> EdgeSet.elt
   end

  type t = (VSet.t * EdgeSet.t) * no_prop_level

  val coq_V : t -> VSet.t

  val coq_E : t -> EdgeSet.t

  val s : t -> no_prop_level

  val e_source : Edge.t -> no_prop_level

  val e_target : Edge.t -> no_prop_level

  val e_weight : Edge.t -> nat

  type labelling = no_prop_level -> nat

  val add_node : t -> VSet.elt -> t

  val add_edge : t -> Edge.t -> t

  type coq_Edges = (nat, __) sigT

  type coq_Paths =
  | Coq_paths_refl of no_prop_level
  | Coq_paths_step of no_prop_level * no_prop_level * no_prop_level
     * coq_Edges * coq_Paths

  val coq_Paths_rect :
    t -> (no_prop_level -> 'a1) -> (no_prop_level -> no_prop_level ->
    no_prop_level -> coq_Edges -> coq_Paths -> 'a1 -> 'a1) -> no_prop_level
    -> no_prop_level -> coq_Paths -> 'a1

  val coq_Paths_rec :
    t -> (no_prop_level -> 'a1) -> (no_prop_level -> no_prop_level ->
    no_prop_level -> coq_Edges -> coq_Paths -> 'a1 -> 'a1) -> no_prop_level
    -> no_prop_level -> coq_Paths -> 'a1

  val weight : t -> no_prop_level -> no_prop_level -> coq_Paths -> nat

  val nodes : t -> no_prop_level -> no_prop_level -> coq_Paths -> VSet.t

  val concat :
    t -> no_prop_level -> no_prop_level -> no_prop_level -> coq_Paths ->
    coq_Paths -> coq_Paths

  val length : t -> no_prop_level -> no_prop_level -> coq_Paths -> nat

  type coq_SimplePaths =
  | Coq_spaths_refl of VSet.t * no_prop_level
  | Coq_spaths_step of VSet.t * VSet.t * no_prop_level * no_prop_level
     * no_prop_level * coq_Edges * coq_SimplePaths

  val coq_SimplePaths_rect :
    t -> (VSet.t -> no_prop_level -> 'a1) -> (VSet.t -> VSet.t ->
    no_prop_level -> no_prop_level -> no_prop_level -> __ -> coq_Edges ->
    coq_SimplePaths -> 'a1 -> 'a1) -> VSet.t -> no_prop_level ->
    no_prop_level -> coq_SimplePaths -> 'a1

  val coq_SimplePaths_rec :
    t -> (VSet.t -> no_prop_level -> 'a1) -> (VSet.t -> VSet.t ->
    no_prop_level -> no_prop_level -> no_prop_level -> __ -> coq_Edges ->
    coq_SimplePaths -> 'a1 -> 'a1) -> VSet.t -> no_prop_level ->
    no_prop_level -> coq_SimplePaths -> 'a1

  val to_paths :
    t -> VSet.t -> no_prop_level -> no_prop_level -> coq_SimplePaths ->
    coq_Paths

  val sweight :
    t -> VSet.t -> no_prop_level -> no_prop_level -> coq_SimplePaths -> nat

  val is_simple : t -> no_prop_level -> no_prop_level -> coq_Paths -> bool

  val to_simple :
    t -> no_prop_level -> no_prop_level -> coq_Paths -> coq_SimplePaths

  val add_end :
    t -> VSet.t -> no_prop_level -> no_prop_level -> coq_SimplePaths ->
    no_prop_level -> coq_Edges -> VSet.t -> coq_SimplePaths

  val coq_SimplePaths_sub :
    t -> VSet.t -> VSet.t -> no_prop_level -> no_prop_level ->
    coq_SimplePaths -> coq_SimplePaths

  val snodes :
    t -> VSet.t -> no_prop_level -> no_prop_level -> coq_SimplePaths -> VSet.t

  val split :
    t -> VSet.t -> no_prop_level -> no_prop_level -> coq_SimplePaths ->
    VSet.elt -> bool -> coq_SimplePaths * coq_SimplePaths

  val split' :
    t -> VSet.t -> no_prop_level -> no_prop_level -> coq_SimplePaths ->
    coq_SimplePaths * coq_SimplePaths

  val simplify :
    t -> VSet.t -> no_prop_level -> no_prop_level -> coq_Paths ->
    coq_SimplePaths -> VSet.t -> (no_prop_level, coq_SimplePaths) sigT

  val succs : t -> no_prop_level -> (nat * no_prop_level) list

  val lsp00 : t -> nat -> VSet.t -> no_prop_level -> no_prop_level -> Nbar.t

  val lsp0 : t -> VSet.t -> no_prop_level -> no_prop_level -> Nbar.t

  val lsp : t -> no_prop_level -> no_prop_level -> Nbar.t

  val reduce :
    t -> VSet.t -> no_prop_level -> no_prop_level -> coq_SimplePaths ->
    coq_SimplePaths

  val simplify2 :
    t -> no_prop_level -> no_prop_level -> coq_Paths -> coq_SimplePaths

  val simplify2' :
    t -> no_prop_level -> no_prop_level -> coq_Paths -> coq_SimplePaths

  val coq_VSet_Forall_reflect :
    (VSet.elt -> bool) -> (VSet.elt -> reflect) -> VSet.t -> reflect

  val reflect_logically_equiv : bool -> reflect -> reflect

  val is_acyclic : t -> bool

  val coq_K : nat -> nat -> nat

  val to_G' :
    t -> no_prop_level -> nat -> nat -> no_prop_level -> no_prop_level ->
    coq_Paths -> coq_Paths

  val from_G' :
    t -> no_prop_level -> nat -> nat -> VSet.t -> no_prop_level ->
    no_prop_level -> coq_SimplePaths -> (coq_SimplePaths,
    coq_SimplePaths * coq_SimplePaths) sum

  val leqb_vertices : t -> nat -> no_prop_level -> VSet.elt -> bool
 end

type universes_graph = Coq_wGraph.t

val init_graph : universes_graph

val no_prop_of_level : Level.t -> no_prop_level option

val no_prop_levels : LevelSet.t -> Coq_wGraph.VSet.t

val gc_of_uctx :
  checker_flags -> ContextSet.t -> (Coq_wGraph.VSet.t * GoodConstraintSet.t)
  option

val level_of_variable : variable_level -> Level.t

val level_of_no_prop : no_prop_level -> Level.t

val edge_of_level : variable_level -> Coq_wGraph.EdgeSet.elt

val coq_EdgeSet_pair :
  Coq_wGraph.EdgeSet.elt -> Coq_wGraph.EdgeSet.elt -> Coq_wGraph.EdgeSet.t

val coq_EdgeSet_triple :
  Coq_wGraph.EdgeSet.elt -> Coq_wGraph.EdgeSet.elt -> Coq_wGraph.EdgeSet.elt
  -> Coq_wGraph.EdgeSet.t

val edge_of_constraint : good_constraint -> Coq_wGraph.EdgeSet.elt

val make_graph : (Coq_wGraph.VSet.t * GoodConstraintSet.t) -> Coq_wGraph.t

val labelling_of_valuation : valuation -> Coq_wGraph.labelling

val valuation_of_labelling : Coq_wGraph.labelling -> valuation

val is_consistent : checker_flags -> ContextSet.t -> bool

val leqb_no_prop_n :
  universes_graph -> nat -> no_prop_level -> no_prop_level -> bool

val leqb_level_n : universes_graph -> nat -> Level.t -> Level.t -> bool

val leqb_level :
  checker_flags -> universes_graph -> Level.t -> Level.t -> bool

val eqb_level : checker_flags -> universes_graph -> Level.t -> Level.t -> bool

val eqb_univ_instance :
  checker_flags -> universes_graph -> universe_instance -> universe_instance
  -> bool

val leqb_expr_n :
  universes_graph -> nat -> Universe.Expr.t -> Universe.Expr.t -> bool

val leqb_univ_expr_n :
  universes_graph -> nat -> universe -> Universe.Expr.t -> bool

val leqb_univ_expr :
  checker_flags -> universes_graph -> universe -> Universe.Expr.t -> bool

val try_leqb_universe_n :
  universes_graph -> nat -> universe -> universe -> bool

val try_leqb_universe :
  checker_flags -> universes_graph -> universe -> universe -> bool

val try_eqb_universe :
  checker_flags -> universes_graph -> universe -> universe -> bool

val check_gc_constraint :
  checker_flags -> universes_graph -> good_constraint -> bool

val check_gc_constraints :
  checker_flags -> universes_graph -> GoodConstraintSet.t -> bool

val check_constraints :
  checker_flags -> universes_graph -> constraints -> bool
