open Ast0
open BasicAst
open Bool
open Datatypes
open List0
open Nat0
open PeanoNat
open String0
open Universes0
open Utils

(** val string_of_list_aux :
    ('a1 -> char list) -> char list -> 'a1 list -> char list **)

let rec string_of_list_aux f sep = function
| [] -> []
| a :: l0 ->
  (match l0 with
   | [] -> f a
   | _ :: _ -> append (f a) (append sep (string_of_list_aux f sep l0)))

(** val string_of_list : ('a1 -> char list) -> 'a1 list -> char list **)

let string_of_list f l =
  append ('['::[]) (append (string_of_list_aux f (','::[]) l) (']'::[]))

(** val string_of_level : Level.t -> char list **)

let string_of_level = function
| Level.Coq_lProp -> 'P'::('r'::('o'::('p'::[])))
| Level.Coq_lSet -> 'S'::('e'::('t'::[]))
| Level.Level s -> s
| Level.Var n -> append ('V'::('a'::('r'::[]))) (string_of_nat n)

(** val string_of_level_expr : (Level.t * bool) -> char list **)

let string_of_level_expr = function
| (l0, b) -> append (string_of_level l0) (if b then '+'::('1'::[]) else [])

(** val string_of_sort : universe -> char list **)

let string_of_sort u =
  string_of_list string_of_level_expr (universe_coercion u)

(** val string_of_name : name -> char list **)

let string_of_name = function
| Coq_nAnon -> 'A'::('n'::('o'::('n'::('y'::('m'::('o'::('u'::('s'::[]))))))))
| Coq_nNamed n -> n

(** val string_of_universe_instance : Level.t list -> char list **)

let string_of_universe_instance u =
  string_of_list string_of_level u

(** val string_of_def : ('a1 -> char list) -> 'a1 def -> char list **)

let string_of_def f def0 =
  append ('('::[])
    (append (string_of_name def0.dname)
      (append (','::[])
        (append (f def0.dtype)
          (append (','::[])
            (append (f def0.dbody)
              (append (','::[]) (append (string_of_nat def0.rarg) (')'::[]))))))))

(** val string_of_inductive : inductive -> char list **)

let string_of_inductive i =
  append i.inductive_mind (append (','::[]) (string_of_nat i.inductive_ind))

(** val string_of_term : term -> char list **)

let rec string_of_term = function
| Coq_tRel n ->
  append ('R'::('e'::('l'::('('::[])))) (append (string_of_nat n) (')'::[]))
| Coq_tVar n -> append ('V'::('a'::('r'::('('::[])))) (append n (')'::[]))
| Coq_tEvar (ev, args) ->
  append ('E'::('v'::('a'::('r'::('('::[])))))
    (append (string_of_nat ev)
      (append (','::[])
        (append (string_of_list string_of_term args) (')'::[]))))
| Coq_tSort s ->
  append ('S'::('o'::('r'::('t'::('('::[])))))
    (append (string_of_sort s) (')'::[]))
| Coq_tCast (c, _, t1) ->
  append ('C'::('a'::('s'::('t'::('('::[])))))
    (append (string_of_term c)
      (append (','::[]) (append (string_of_term t1) (')'::[]))))
| Coq_tProd (na, b, t1) ->
  append ('P'::('r'::('o'::('d'::('('::[])))))
    (append (string_of_name na)
      (append (','::[])
        (append (string_of_term b)
          (append (','::[]) (append (string_of_term t1) (')'::[]))))))
| Coq_tLambda (na, b, t1) ->
  append ('L'::('a'::('m'::('b'::('d'::('a'::('('::[])))))))
    (append (string_of_name na)
      (append (','::[])
        (append (string_of_term b)
          (append (','::[]) (append (string_of_term t1) (')'::[]))))))
| Coq_tLetIn (na, b, t', t1) ->
  append ('L'::('e'::('t'::('I'::('n'::('('::[]))))))
    (append (string_of_name na)
      (append (','::[])
        (append (string_of_term b)
          (append (','::[])
            (append (string_of_term t')
              (append (','::[]) (append (string_of_term t1) (')'::[]))))))))
| Coq_tApp (f, l) ->
  append ('A'::('p'::('p'::('('::[]))))
    (append (string_of_term f)
      (append (','::[]) (append (string_of_list string_of_term l) (')'::[]))))
| Coq_tConst (c, u) ->
  append ('C'::('o'::('n'::('s'::('t'::('('::[]))))))
    (append c
      (append (','::[]) (append (string_of_universe_instance u) (')'::[]))))
| Coq_tInd (i, u) ->
  append ('I'::('n'::('d'::('('::[]))))
    (append (string_of_inductive i)
      (append (','::[]) (append (string_of_universe_instance u) (')'::[]))))
| Coq_tConstruct (i, n, u) ->
  append
    ('C'::('o'::('n'::('s'::('t'::('r'::('u'::('c'::('t'::('('::[]))))))))))
    (append (string_of_inductive i)
      (append (','::[])
        (append (string_of_nat n)
          (append (','::[])
            (append (string_of_universe_instance u) (')'::[]))))))
| Coq_tCase (ind_and_nbparams, t1, p, brs) ->
  let (ind, i) = ind_and_nbparams in
  append ('C'::('a'::('s'::('e'::('('::[])))))
    (append (string_of_inductive ind)
      (append (','::[])
        (append (string_of_nat i)
          (append (','::[])
            (append (string_of_term t1)
              (append (','::[])
                (append (string_of_term p)
                  (append (','::[])
                    (append
                      (string_of_list (fun b -> string_of_term (snd b)) brs)
                      (')'::[]))))))))))
| Coq_tProj (proj, c) ->
  let (p, k) = proj in
  let (ind, i) = p in
  append ('P'::('r'::('o'::('j'::('('::[])))))
    (append (string_of_inductive ind)
      (append (','::[])
        (append (string_of_nat i)
          (append (','::[])
            (append (string_of_nat k)
              (append (','::[]) (append (string_of_term c) (')'::[]))))))))
| Coq_tFix (l, n) ->
  append ('F'::('i'::('x'::('('::[]))))
    (append (string_of_list (string_of_def string_of_term) l)
      (append (','::[]) (append (string_of_nat n) (')'::[]))))
| Coq_tCoFix (l, n) ->
  append ('C'::('o'::('F'::('i'::('x'::('('::[]))))))
    (append (string_of_list (string_of_def string_of_term) l)
      (append (','::[]) (append (string_of_nat n) (')'::[]))))

(** val mkLambda_or_LetIn : context_decl -> term -> term **)

let mkLambda_or_LetIn d t0 =
  match d.decl_body with
  | Some b -> Coq_tLetIn (d.decl_name, b, d.decl_type, t0)
  | None -> Coq_tLambda (d.decl_name, d.decl_type, t0)

(** val it_mkLambda_or_LetIn : context -> term -> term **)

let it_mkLambda_or_LetIn l t0 =
  fold_left (fun acc d -> mkLambda_or_LetIn d acc) l t0

(** val mkProd_or_LetIn : context_decl -> term -> term **)

let mkProd_or_LetIn d t0 =
  match d.decl_body with
  | Some b -> Coq_tLetIn (d.decl_name, b, d.decl_type, t0)
  | None -> Coq_tProd (d.decl_name, d.decl_type, t0)

(** val it_mkProd_or_LetIn : context -> term -> term **)

let it_mkProd_or_LetIn l t0 =
  fold_left (fun acc d -> mkProd_or_LetIn d acc) l t0

(** val map_decl : (term -> term) -> context_decl -> context_decl **)

let map_decl f d =
  { decl_name = d.decl_name; decl_body = (option_map f d.decl_body);
    decl_type = (f d.decl_type) }

(** val map_context :
    (term -> term) -> context_decl list -> context_decl list **)

let map_context f c =
  map (map_decl f) c

(** val map_constant_body :
    (term -> term) -> constant_body -> constant_body **)

let map_constant_body f decl =
  { cst_type = (f decl.cst_type); cst_body = (option_map f decl.cst_body);
    cst_universes = decl.cst_universes }

(** val map_def : ('a1 -> 'a2) -> ('a1 -> 'a2) -> 'a1 def -> 'a2 def **)

let map_def tyf bodyf d =
  { dname = d.dname; dtype = (tyf d.dtype); dbody = (bodyf d.dbody); rarg =
    d.rarg }

(** val app_context : context -> context -> context **)

let app_context _UU0393_ _UU0393_' =
  app _UU0393_' _UU0393_

(** val string_of_gref : global_reference -> char list **)

let string_of_gref = function
| VarRef s ->
  append ('V'::('a'::('r'::('i'::('a'::('b'::('l'::('e'::(' '::[]))))))))) s
| ConstRef s -> s
| IndRef i ->
  let { inductive_mind = s; inductive_ind = n } = i in
  append
    ('I'::('n'::('d'::('u'::('c'::('t'::('i'::('v'::('e'::(' '::[]))))))))))
    (append s (append (' '::[]) (string_of_nat n)))
| ConstructRef (i, k) ->
  let { inductive_mind = s; inductive_ind = n } = i in
  append
    ('C'::('o'::('n'::('s'::('t'::('r'::('u'::('c'::('t'::('o'::('r'::(' '::[]))))))))))))
    (append s
      (append (' '::[])
        (append (string_of_nat n) (append (' '::[]) (string_of_nat k)))))

(** val gref_eq_dec : global_reference -> global_reference -> bool **)

let gref_eq_dec gr gr' =
  match gr with
  | VarRef x -> (match gr' with
                 | VarRef i0 -> string_dec x i0
                 | _ -> false)
  | ConstRef x -> (match gr' with
                   | ConstRef k0 -> string_dec x k0
                   | _ -> false)
  | IndRef x ->
    (match gr' with
     | IndRef i0 ->
       let { inductive_mind = inductive_mind0; inductive_ind =
         inductive_ind0 } = x
       in
       let { inductive_mind = inductive_mind2; inductive_ind =
         inductive_ind2 } = i0
       in
       if string_dec inductive_mind0 inductive_mind2
       then Nat.eq_dec inductive_ind0 inductive_ind2
       else false
     | _ -> false)
  | ConstructRef (x, x0) ->
    (match gr' with
     | ConstructRef (i0, n0) ->
       if let { inductive_mind = inductive_mind0; inductive_ind =
            inductive_ind0 } = x
          in
          let { inductive_mind = inductive_mind2; inductive_ind =
            inductive_ind2 } = i0
          in
          if string_dec inductive_mind0 inductive_mind2
          then Nat.eq_dec inductive_ind0 inductive_ind2
          else false
       then Nat.eq_dec x0 n0
       else false
     | _ -> false)

(** val ident_eq : ident -> ident -> bool **)

let ident_eq x y =
  match string_compare x y with
  | Eq -> true
  | _ -> false

(** val ident_eq_spec : ident -> ident -> reflect **)

let ident_eq_spec x y =
  let c = string_compare x y in (match c with
                                 | Eq -> ReflectT
                                 | _ -> ReflectF)

(** val eq_string : char list -> char list -> bool **)

let eq_string s s' =
  match string_compare s s' with
  | Eq -> true
  | _ -> false

(** val eq_ind : inductive -> inductive -> bool **)

let eq_ind i i' =
  let { inductive_mind = i0; inductive_ind = n } = i in
  let { inductive_mind = i'0; inductive_ind = n' } = i' in
  (&&) (eq_string i0 i'0) (Nat.eqb n n')

(** val eq_constant : char list -> char list -> bool **)

let eq_constant =
  eq_string

(** val eq_nat : nat -> nat -> bool **)

let eq_nat =
  Nat.eqb

(** val eq_evar : nat -> nat -> bool **)

let eq_evar =
  eq_nat

(** val eq_projection :
    ((inductive * nat) * nat) -> ((inductive * nat) * nat) -> bool **)

let eq_projection p p' =
  let (y, arg) = p in
  let (ind, pars) = y in
  let (y0, arg') = p' in
  let (ind', pars') = y0 in
  (&&) ((&&) (eq_ind ind ind') (eq_nat pars pars')) (eq_nat arg arg')

(** val decompose_app : term -> term * term list **)

let decompose_app t0 = match t0 with
| Coq_tApp (f, l) -> (f, l)
| _ -> (t0, [])

(** val decompose_prod : term -> (name list * term list) * term **)

let rec decompose_prod t0 = match t0 with
| Coq_tProd (n, a, b) ->
  let (nAs, b0) = decompose_prod b in
  let (ns, as0) = nAs in (((n :: ns), (a :: as0)), b0)
| _ -> (([], []), t0)

(** val get_ident : name -> char list **)

let get_ident = function
| Coq_nAnon -> 'X'::('X'::[])
| Coq_nNamed i -> i

(** val remove_arity : nat -> term -> term **)

let rec remove_arity n t0 =
  match n with
  | O -> t0
  | S n0 -> (match t0 with
             | Coq_tProd (_, _, b) -> remove_arity n0 b
             | _ -> t0)

(** val lookup_mind_decl :
    ident -> global_env -> mutual_inductive_body option **)

let rec lookup_mind_decl id = function
| [] -> None
| g :: tl ->
  (match g with
   | ConstantDecl (_, _) -> lookup_mind_decl id tl
   | InductiveDecl (kn, d) ->
     (match string_compare kn id with
      | Eq -> Some d
      | _ -> lookup_mind_decl id tl))

(** val mind_body_to_entry :
    mutual_inductive_body -> mutual_inductive_entry **)

let mind_body_to_entry decl =
  { mind_entry_record = None; mind_entry_finite = Finite; mind_entry_params =
    (match hd_error decl.ind_bodies with
     | Some i0 ->
       List0.rev
         (let typ = decompose_prod i0.ind_type in
          let (p, _) = typ in
          let (names, types) = p in
          let names0 = firstn decl.ind_npars names in
          let types0 = firstn decl.ind_npars types in
          combine (map get_ident names0) (map (fun x -> LocalAssum x) types0))
     | None -> []); mind_entry_inds =
    (map (fun h ->
      let { ind_name = ind_name0; ind_type = ind_type0; ind_kelim = _;
        ind_ctors = ind_ctors0; ind_projs = _ } = h
      in
      { mind_entry_typename = ind_name0; mind_entry_arity =
      (remove_arity decl.ind_npars ind_type0); mind_entry_template = false;
      mind_entry_consnames = (map (fun x -> fst (fst x)) ind_ctors0);
      mind_entry_lc =
      (map (fun x -> remove_arity decl.ind_npars (snd (fst x))) ind_ctors0) })
      decl.ind_bodies); mind_entry_universes = decl.ind_universes;
    mind_entry_private = None }

(** val arities_context : one_inductive_body list -> context_decl list **)

let arities_context l =
  rev_map (fun ind -> vass (Coq_nNamed ind.ind_name) ind.ind_type) l

(** val decompose_prod_assum : context -> term -> context * term **)

let rec decompose_prod_assum _UU0393_ t0 = match t0 with
| Coq_tCast (t1, _, _) -> decompose_prod_assum _UU0393_ t1
| Coq_tProd (n, a, b) -> decompose_prod_assum (snoc _UU0393_ (vass n a)) b
| Coq_tLetIn (na, b, bty, b') ->
  decompose_prod_assum (snoc _UU0393_ (vdef na b bty)) b'
| _ -> (_UU0393_, t0)

(** val strip_outer_cast : term -> term **)

let rec strip_outer_cast t0 = match t0 with
| Coq_tCast (t1, _, _) -> strip_outer_cast t1
| _ -> t0

(** val decompose_prod_n_assum :
    context -> nat -> term -> (context * term) option **)

let rec decompose_prod_n_assum _UU0393_ n t0 =
  match n with
  | O -> Some (_UU0393_, t0)
  | S n0 ->
    (match strip_outer_cast t0 with
     | Coq_tProd (na, a, b) ->
       decompose_prod_n_assum (snoc _UU0393_ (vass na a)) n0 b
     | Coq_tLetIn (na, b, bty, b') ->
       decompose_prod_n_assum (snoc _UU0393_ (vdef na b bty)) n0 b'
     | _ -> None)

(** val is_ind_app_head : term -> bool **)

let is_ind_app_head = function
| Coq_tApp (f, _) -> (match f with
                      | Coq_tInd (_, _) -> true
                      | _ -> false)
| Coq_tInd (_, _) -> true
| _ -> false

(** val reln : term list -> nat -> context_decl list -> term list **)

let rec reln l p = function
| [] -> l
| c :: hyps ->
  let { decl_name = _; decl_body = decl_body0; decl_type = _ } = c in
  (match decl_body0 with
   | Some _ -> reln l (add p (S O)) hyps
   | None -> reln ((Coq_tRel p) :: l) (add p (S O)) hyps)

(** val to_extended_list_k : context_decl list -> nat -> term list **)

let to_extended_list_k _UU0393_ k =
  reln [] k _UU0393_

(** val to_extended_list : context_decl list -> term list **)

let to_extended_list _UU0393_ =
  to_extended_list_k _UU0393_ O

(** val reln_alt : nat -> context_decl list -> term list **)

let rec reln_alt p = function
| [] -> []
| y :: _UU0393_0 ->
  let { decl_name = _; decl_body = decl_body0; decl_type = _ } = y in
  (match decl_body0 with
   | Some _ -> reln_alt (add p (S O)) _UU0393_0
   | None -> (Coq_tRel p) :: (reln_alt (add p (S O)) _UU0393_0))

(** val polymorphic_instance : universes_decl -> Instance.t **)

let polymorphic_instance = function
| Monomorphic_ctx _ -> Instance.empty
| Polymorphic_ctx c -> fst (AUContext.repr c)
| Cumulative_ctx ctx -> let (c, _) = ctx in fst (AUContext.repr c)

(** val context_assumptions : context -> nat **)

let rec context_assumptions = function
| [] -> O
| d :: _UU0393_0 ->
  (match d.decl_body with
   | Some _ -> context_assumptions _UU0393_0
   | None -> S (context_assumptions _UU0393_0))

(** val map_one_inductive_body :
    nat -> nat -> (nat -> term -> term) -> nat -> one_inductive_body ->
    one_inductive_body **)

let map_one_inductive_body npars arities f _ m =
  let { ind_name = ind_name0; ind_type = ind_type0; ind_kelim = ind_kelim0;
    ind_ctors = ind_ctors0; ind_projs = ind_projs0 } = m
  in
  { ind_name = ind_name0; ind_type = (f O ind_type0); ind_kelim = ind_kelim0;
  ind_ctors = (map (on_pi2 (f arities)) ind_ctors0); ind_projs =
  (map (on_snd (f (S npars))) ind_projs0) }

(** val fold_context : (nat -> term -> term) -> context -> context **)

let fold_context f _UU0393_ =
  List0.rev (mapi (fun k' decl -> map_decl (f k') decl) (List0.rev _UU0393_))

(** val map_mutual_inductive_body :
    (nat -> term -> term) -> mutual_inductive_body -> mutual_inductive_body **)

let map_mutual_inductive_body f m =
  let { ind_finite = finite; ind_npars = ind_npars0; ind_params = ind_pars;
    ind_bodies = ind_bodies0; ind_universes = ind_universes0 } = m
  in
  let arities = arities_context ind_bodies0 in
  let pars = fold_context f ind_pars in
  { ind_finite = finite; ind_npars = ind_npars0; ind_params = pars;
  ind_bodies =
  (mapi
    (map_one_inductive_body (context_assumptions pars)
      (Datatypes.length arities) f) ind_bodies0); ind_universes =
  ind_universes0 }

(** val test_def : ('a1 -> bool) -> ('a1 -> bool) -> 'a1 def -> bool **)

let test_def tyf bodyf d =
  (&&) (tyf d.dtype) (bodyf d.dbody)

(** val fst_ctx : global_env_ext -> global_env **)

let fst_ctx =
  fst

(** val empty_ext : global_env -> global_env_ext **)

let empty_ext _UU03a3_ =
  (_UU03a3_, (Monomorphic_ctx ContextSet.empty))
