open Datatypes

type ident = char list

type qualid = char list

type kername = char list

type name =
| Coq_nAnon
| Coq_nNamed of ident

(** val name_rect : 'a1 -> (ident -> 'a1) -> name -> 'a1 **)

let name_rect f f0 = function
| Coq_nAnon -> f
| Coq_nNamed x -> f0 x

(** val name_rec : 'a1 -> (ident -> 'a1) -> name -> 'a1 **)

let name_rec f f0 = function
| Coq_nAnon -> f
| Coq_nNamed x -> f0 x

type inductive = { inductive_mind : kername; inductive_ind : nat }

(** val inductive_mind : inductive -> kername **)

let inductive_mind x = x.inductive_mind

(** val inductive_ind : inductive -> nat **)

let inductive_ind x = x.inductive_ind

type projection = (inductive * nat) * nat

type 'term def = { dname : name; dtype : 'term; dbody : 'term; rarg : nat }

(** val dname : 'a1 def -> name **)

let dname x = x.dname

(** val dtype : 'a1 def -> 'a1 **)

let dtype x = x.dtype

(** val dbody : 'a1 def -> 'a1 **)

let dbody x = x.dbody

(** val rarg : 'a1 def -> nat **)

let rarg x = x.rarg

type 'term mfixpoint = 'term def list

type cast_kind =
| VmCast
| NativeCast
| Cast
| RevertCast

(** val cast_kind_rect : 'a1 -> 'a1 -> 'a1 -> 'a1 -> cast_kind -> 'a1 **)

let cast_kind_rect f f0 f1 f2 = function
| VmCast -> f
| NativeCast -> f0
| Cast -> f1
| RevertCast -> f2

(** val cast_kind_rec : 'a1 -> 'a1 -> 'a1 -> 'a1 -> cast_kind -> 'a1 **)

let cast_kind_rec f f0 f1 f2 = function
| VmCast -> f
| NativeCast -> f0
| Cast -> f1
| RevertCast -> f2

type recursivity_kind =
| Finite
| CoFinite
| BiFinite

(** val recursivity_kind_rect :
    'a1 -> 'a1 -> 'a1 -> recursivity_kind -> 'a1 **)

let recursivity_kind_rect f f0 f1 = function
| Finite -> f
| CoFinite -> f0
| BiFinite -> f1

(** val recursivity_kind_rec :
    'a1 -> 'a1 -> 'a1 -> recursivity_kind -> 'a1 **)

let recursivity_kind_rec f f0 f1 = function
| Finite -> f
| CoFinite -> f0
| BiFinite -> f1

type global_reference =
| VarRef of ident
| ConstRef of kername
| IndRef of inductive
| ConstructRef of inductive * nat

(** val global_reference_rect :
    (ident -> 'a1) -> (kername -> 'a1) -> (inductive -> 'a1) -> (inductive ->
    nat -> 'a1) -> global_reference -> 'a1 **)

let global_reference_rect f f0 f1 f2 = function
| VarRef x -> f x
| ConstRef x -> f0 x
| IndRef x -> f1 x
| ConstructRef (x, x0) -> f2 x x0

(** val global_reference_rec :
    (ident -> 'a1) -> (kername -> 'a1) -> (inductive -> 'a1) -> (inductive ->
    nat -> 'a1) -> global_reference -> 'a1 **)

let global_reference_rec f f0 f1 f2 = function
| VarRef x -> f x
| ConstRef x -> f0 x
| IndRef x -> f1 x
| ConstructRef (x, x0) -> f2 x x0
