open BasicAst
open BinInt
open Bool
open Datatypes
open MSetFacts
open MSetProperties
open MSetWeakList
open PeanoNat
open String0
open Utils

type __ = Obj.t

module Level =
 struct
  type t =
  | Coq_lProp
  | Coq_lSet
  | Level of char list
  | Var of nat

  (** val t_rect :
      'a1 -> 'a1 -> (char list -> 'a1) -> (nat -> 'a1) -> t -> 'a1 **)

  let t_rect f f0 f1 f2 = function
  | Coq_lProp -> f
  | Coq_lSet -> f0
  | Level x -> f1 x
  | Var x -> f2 x

  (** val t_rec :
      'a1 -> 'a1 -> (char list -> 'a1) -> (nat -> 'a1) -> t -> 'a1 **)

  let t_rec f f0 f1 f2 = function
  | Coq_lProp -> f
  | Coq_lSet -> f0
  | Level x -> f1 x
  | Var x -> f2 x

  (** val set : t **)

  let set =
    Coq_lSet

  (** val prop : t **)

  let prop =
    Coq_lProp

  (** val is_small : t -> bool **)

  let is_small = function
  | Coq_lProp -> true
  | Coq_lSet -> true
  | _ -> false

  (** val is_prop : t -> bool **)

  let is_prop = function
  | Coq_lProp -> true
  | _ -> false

  (** val is_set : t -> bool **)

  let is_set = function
  | Coq_lSet -> true
  | _ -> false

  (** val is_var : t -> bool **)

  let is_var = function
  | Var _ -> true
  | _ -> false

  (** val compare : t -> t -> comparison **)

  let compare l1 l2 =
    match l1 with
    | Coq_lProp -> (match l2 with
                    | Coq_lProp -> Eq
                    | _ -> Lt)
    | Coq_lSet -> (match l2 with
                   | Coq_lProp -> Gt
                   | Coq_lSet -> Eq
                   | _ -> Lt)
    | Level s1 ->
      (match l2 with
       | Level s2 -> string_compare s1 s2
       | Var _ -> Lt
       | _ -> Gt)
    | Var n -> (match l2 with
                | Var m -> Nat.compare n m
                | _ -> Gt)

  (** val eq_dec : t -> t -> bool **)

  let eq_dec l1 l2 =
    match l1 with
    | Coq_lProp -> (match l2 with
                    | Coq_lProp -> true
                    | _ -> false)
    | Coq_lSet -> (match l2 with
                   | Coq_lSet -> true
                   | _ -> false)
    | Level x -> (match l2 with
                  | Level s0 -> string_dec x s0
                  | _ -> false)
    | Var x -> (match l2 with
                | Var n0 -> Nat.eq_dec x n0
                | _ -> false)

  (** val equal : t -> t -> bool **)

  let equal l1 l2 =
    match compare l1 l2 with
    | Eq -> true
    | _ -> false
 end

module LevelDecidableType =
 struct
  type t = Level.t

  (** val eq_dec : t -> t -> bool **)

  let eq_dec =
    Level.eq_dec
 end

module LevelSet = Make(LevelDecidableType)

module LevelSetFact = WFactsOn(LevelDecidableType)(LevelSet)

module LevelSetProp = WPropertiesOn(LevelDecidableType)(LevelSet)

(** val coq_LevelSet_pair : LevelSet.elt -> LevelSet.elt -> LevelSet.t **)

let coq_LevelSet_pair x y =
  LevelSet.add y (LevelSet.singleton x)

type universe_level = Level.t

module Universe =
 struct
  module Expr =
   struct
    type t = Level.t * bool

    (** val is_small : t -> bool **)

    let is_small = function
    | (t0, b) ->
      (match t0 with
       | Level.Coq_lProp -> true
       | Level.Coq_lSet -> if b then false else true
       | _ -> false)

    (** val is_level : t -> bool **)

    let is_level = function
    | (t0, b) ->
      (match t0 with
       | Level.Coq_lProp -> true
       | _ -> if b then false else true)

    (** val is_prop : t -> bool **)

    let is_prop = function
    | (t0, _) -> (match t0 with
                  | Level.Coq_lProp -> true
                  | _ -> false)

    (** val equal : t -> t -> bool **)

    let equal e1 e2 =
      let (l1, b1) = e1 in
      (match l1 with
       | Level.Coq_lProp ->
         let (l2, b2) = e2 in
         (match l2 with
          | Level.Coq_lProp -> true
          | _ -> (&&) (Level.equal l1 l2) (Bool.eqb b1 b2))
       | _ -> let (l2, b2) = e2 in (&&) (Level.equal l1 l2) (Bool.eqb b1 b2))

    (** val get_level : t -> Level.t **)

    let get_level =
      fst

    (** val prop : t **)

    let prop =
      (Level.prop, false)

    (** val set : t **)

    let set =
      (Level.set, false)

    (** val type1 : t **)

    let type1 =
      (Level.set, true)
   end

  type t = Expr.t non_empty_list

  (** val equal : t -> t -> bool **)

  let equal u1 u2 =
    NEL.forallb2 Expr.equal u1 u2

  (** val make : Level.t -> t **)

  let make l =
    NEL.Coq_sing (l, false)

  (** val make' : Expr.t -> t **)

  let make' e =
    NEL.Coq_sing e

  (** val make'' : Expr.t -> Expr.t list -> t **)

  let make'' =
    NEL.cons'

  (** val is_level : t -> bool **)

  let is_level = function
  | NEL.Coq_sing e -> Expr.is_level e
  | NEL.Coq_cons (_, _) -> false

  (** val is_levels : t -> bool **)

  let is_levels u =
    NEL.forallb Expr.is_level u

  (** val level : t -> Level.t option **)

  let level = function
  | NEL.Coq_sing t0 ->
    let (l, b) = t0 in
    (match l with
     | Level.Coq_lProp -> Some Level.Coq_lProp
     | _ -> if b then None else Some l)
  | NEL.Coq_cons (_, _) -> None

  (** val is_small : t -> bool **)

  let is_small u =
    NEL.forallb Expr.is_small u

  (** val is_prop : t -> bool **)

  let is_prop u =
    NEL.forallb Expr.is_prop u

  (** val type0m : t **)

  let type0m =
    make Level.prop

  (** val type0 : t **)

  let type0 =
    make Level.set

  (** val type1 : t **)

  let type1 =
    make' Expr.type1

  (** val super : Level.t -> t **)

  let super l =
    if Level.is_small l then type1 else make' (l, true)

  (** val try_suc : t -> t **)

  let try_suc u =
    NEL.map (fun pat -> let (l, _) = pat in (l, true)) u

  (** val sup : t -> t -> t **)

  let sup =
    NEL.app

  (** val for_all : (Expr.t -> bool) -> t -> bool **)

  let for_all =
    NEL.forallb

  (** val sort_of_product : t -> t -> t **)

  let sort_of_product domsort rangsort =
    if is_prop rangsort then rangsort else sup domsort rangsort
 end

type universe = Universe.t

(** val universe_coercion : universe -> Universe.Expr.t list **)

let universe_coercion =
  NEL.to_list

type sort_family =
| InProp
| InSet
| InType

(** val sort_family_rect : 'a1 -> 'a1 -> 'a1 -> sort_family -> 'a1 **)

let sort_family_rect f f0 f1 = function
| InProp -> f
| InSet -> f0
| InType -> f1

(** val sort_family_rec : 'a1 -> 'a1 -> 'a1 -> sort_family -> 'a1 **)

let sort_family_rec f f0 f1 = function
| InProp -> f
| InSet -> f0
| InType -> f1

(** val leb_sort_family : sort_family -> sort_family -> bool **)

let leb_sort_family x y =
  match x with
  | InProp -> true
  | InSet -> (match y with
              | InProp -> false
              | _ -> true)
  | InType -> (match y with
               | InType -> true
               | _ -> false)

(** val universe_family : universe -> sort_family **)

let universe_family u =
  if Universe.is_prop u
  then InProp
  else if Universe.is_small u then InSet else InType

module ConstraintType =
 struct
  type t =
  | Lt
  | Le
  | Eq

  (** val t_rect : 'a1 -> 'a1 -> 'a1 -> t -> 'a1 **)

  let t_rect f f0 f1 = function
  | Lt -> f
  | Le -> f0
  | Eq -> f1

  (** val t_rec : 'a1 -> 'a1 -> 'a1 -> t -> 'a1 **)

  let t_rec f f0 f1 = function
  | Lt -> f
  | Le -> f0
  | Eq -> f1
 end

type constraint_type = ConstraintType.t

type univ_constraint = (universe_level * ConstraintType.t) * universe_level

module UnivConstraintDec =
 struct
  type t = univ_constraint

  (** val eq_dec : t -> t -> bool **)

  let eq_dec x y =
    let (x0, x1) = x in
    let (p, u) = y in
    let (x2, x3) = x0 in
    let (u0, t0) = p in
    if LevelDecidableType.eq_dec x2 u0
    then (match x3 with
          | ConstraintType.Lt ->
            (match t0 with
             | ConstraintType.Lt ->
               (match x1 with
                | Level.Coq_lProp ->
                  (match u with
                   | Level.Coq_lProp -> true
                   | _ -> false)
                | Level.Coq_lSet ->
                  (match u with
                   | Level.Coq_lSet -> true
                   | _ -> false)
                | Level.Level x4 ->
                  (match u with
                   | Level.Level s0 -> string_dec x4 s0
                   | _ -> false)
                | Level.Var x4 ->
                  (match u with
                   | Level.Var n0 -> Nat.eq_dec x4 n0
                   | _ -> false))
             | _ -> false)
          | ConstraintType.Le ->
            (match t0 with
             | ConstraintType.Le ->
               (match x1 with
                | Level.Coq_lProp ->
                  (match u with
                   | Level.Coq_lProp -> true
                   | _ -> false)
                | Level.Coq_lSet ->
                  (match u with
                   | Level.Coq_lSet -> true
                   | _ -> false)
                | Level.Level x4 ->
                  (match u with
                   | Level.Level s0 -> string_dec x4 s0
                   | _ -> false)
                | Level.Var x4 ->
                  (match u with
                   | Level.Var n0 -> Nat.eq_dec x4 n0
                   | _ -> false))
             | _ -> false)
          | ConstraintType.Eq ->
            (match t0 with
             | ConstraintType.Eq ->
               (match x1 with
                | Level.Coq_lProp ->
                  (match u with
                   | Level.Coq_lProp -> true
                   | _ -> false)
                | Level.Coq_lSet ->
                  (match u with
                   | Level.Coq_lSet -> true
                   | _ -> false)
                | Level.Level x4 ->
                  (match u with
                   | Level.Level s0 -> string_dec x4 s0
                   | _ -> false)
                | Level.Var x4 ->
                  (match u with
                   | Level.Var n0 -> Nat.eq_dec x4 n0
                   | _ -> false))
             | _ -> false))
    else false
 end

module ConstraintSet = Make(UnivConstraintDec)

module ConstraintSetFact = WFactsOn(UnivConstraintDec)(ConstraintSet)

module ConstraintSetProp = WPropertiesOn(UnivConstraintDec)(ConstraintSet)

(** val make_univ_constraint :
    universe_level -> constraint_type -> universe_level -> univ_constraint **)

let make_univ_constraint x y z =
  ((x, y), z)

type constraints = ConstraintSet.t

(** val empty_constraint : constraints **)

let empty_constraint =
  ConstraintSet.empty

module Instance =
 struct
  type t = Level.t list

  (** val empty : t **)

  let empty =
    []

  (** val is_empty : t -> bool **)

  let is_empty = function
  | [] -> true
  | _ :: _ -> false

  (** val equal : t -> t -> bool **)

  let equal i j =
    forallb2 Level.equal i j

  (** val equal_upto : (Level.t -> Level.t -> bool) -> t -> t -> bool **)

  let equal_upto =
    forallb2
 end

type universe_instance = Instance.t

module UContext =
 struct
  type t = universe_instance * constraints

  (** val make : Instance.t -> ConstraintSet.t -> t **)

  let make x x0 =
    (x, x0)

  (** val empty : t **)

  let empty =
    (Instance.empty, ConstraintSet.empty)

  (** val instance : t -> Instance.t **)

  let instance =
    fst

  (** val constraints : t -> constraints **)

  let constraints =
    snd

  (** val dest : t -> Instance.t * ConstraintSet.t **)

  let dest x =
    x
 end

module AUContext =
 struct
  type t = ident list * constraints

  (** val make : ident list -> constraints -> t **)

  let make ids ctrs =
    (ids, ctrs)

  (** val repr : t -> UContext.t **)

  let repr = function
  | (u, cst) -> ((mapi (fun i _ -> Level.Var i) u), cst)

  (** val levels : t -> LevelSet.t **)

  let levels uctx =
    LevelSetProp.of_list (fst (repr uctx))
 end

module ContextSet =
 struct
  type t = LevelSet.t * constraints

  (** val empty : t **)

  let empty =
    (LevelSet.empty, ConstraintSet.empty)

  (** val is_empty : t -> bool **)

  let is_empty uctx =
    (&&) (LevelSet.is_empty (fst uctx)) (ConstraintSet.is_empty (snd uctx))
 end

module Variance =
 struct
  type t =
  | Irrelevant
  | Covariant
  | Invariant

  (** val t_rect : 'a1 -> 'a1 -> 'a1 -> t -> 'a1 **)

  let t_rect f f0 f1 = function
  | Irrelevant -> f
  | Covariant -> f0
  | Invariant -> f1

  (** val t_rec : 'a1 -> 'a1 -> 'a1 -> t -> 'a1 **)

  let t_rec f f0 f1 = function
  | Irrelevant -> f
  | Covariant -> f0
  | Invariant -> f1
 end

module ACumulativityInfo =
 struct
  type t = AUContext.t * Variance.t list

  (** val make : AUContext.t -> Variance.t list -> t **)

  let make ctx var =
    (ctx, var)

  (** val univ_context : t -> AUContext.t **)

  let univ_context =
    fst

  (** val variance : t -> Variance.t list **)

  let variance =
    snd
 end

type universes_decl =
| Monomorphic_ctx of ContextSet.t
| Polymorphic_ctx of AUContext.t
| Cumulative_ctx of ACumulativityInfo.t

(** val universes_decl_rect :
    (ContextSet.t -> 'a1) -> (AUContext.t -> 'a1) -> (ACumulativityInfo.t ->
    'a1) -> universes_decl -> 'a1 **)

let universes_decl_rect f f0 f1 = function
| Monomorphic_ctx x -> f x
| Polymorphic_ctx x -> f0 x
| Cumulative_ctx x -> f1 x

(** val universes_decl_rec :
    (ContextSet.t -> 'a1) -> (AUContext.t -> 'a1) -> (ACumulativityInfo.t ->
    'a1) -> universes_decl -> 'a1 **)

let universes_decl_rec f f0 f1 = function
| Monomorphic_ctx x -> f x
| Polymorphic_ctx x -> f0 x
| Cumulative_ctx x -> f1 x

(** val monomorphic_udecl : universes_decl -> ContextSet.t **)

let monomorphic_udecl = function
| Monomorphic_ctx ctx -> ctx
| _ -> ContextSet.empty

(** val monomorphic_levels : universes_decl -> LevelSet.t **)

let monomorphic_levels _UU03c6_ =
  fst (monomorphic_udecl _UU03c6_)

(** val monomorphic_constraints : universes_decl -> constraints **)

let monomorphic_constraints _UU03c6_ =
  snd (monomorphic_udecl _UU03c6_)

(** val levels_of_udecl : universes_decl -> LevelSet.t **)

let levels_of_udecl = function
| Monomorphic_ctx ctx -> fst ctx
| Polymorphic_ctx ctx -> AUContext.levels ctx
| Cumulative_ctx ctx0 -> let (ctx, _) = ctx0 in AUContext.levels ctx

(** val constraints_of_udecl : universes_decl -> constraints **)

let constraints_of_udecl = function
| Monomorphic_ctx ctx -> snd ctx
| Polymorphic_ctx ctx -> snd (AUContext.repr ctx)
| Cumulative_ctx ctx0 -> let (ctx, _) = ctx0 in snd (AUContext.repr ctx)

type valuation = { valuation_mono : (char list -> int);
                   valuation_poly : (nat -> nat) }

(** val valuation_mono : valuation -> char list -> int **)

let valuation_mono x = x.valuation_mono

(** val valuation_poly : valuation -> nat -> nat **)

let valuation_poly x = x.valuation_poly

(** val val0 : valuation -> Level.t -> int **)

let val0 v = function
| Level.Coq_lProp -> (~-) 1
| Level.Coq_lSet -> 0
| Level.Level s -> (v.valuation_mono s)
| Level.Var x -> Z.of_nat (v.valuation_poly x)

(** val val1 : valuation -> Universe.Expr.t -> int **)

let val1 v e =
  let n = val0 v (fst e) in
  if (&&) (snd e) (negb (Level.is_prop (fst e))) then Z.add n 1 else n

(** val coq_val : valuation -> universe -> int **)

let coq_val v = function
| NEL.Coq_sing e -> val1 v e
| NEL.Coq_cons (e, u0) ->
  NEL.fold_left (fun n e0 -> Z.max (val1 v e0) n) u0 (val1 v e)

(** val fresh_universe : universe **)

let fresh_universe =
  Universe.type0

(** val is_prop_sort : Universe.t -> bool **)

let is_prop_sort s =
  match Universe.level s with
  | Some l -> Level.is_prop l
  | None -> false
