open BasicAst
open BinInt
open Bool
open Datatypes
open MSetFacts
open MSetProperties
open MSetWeakList
open PeanoNat
open String0
open Utils

type __ = Obj.t

module Level :
 sig
  type t =
  | Coq_lProp
  | Coq_lSet
  | Level of char list
  | Var of nat

  val t_rect : 'a1 -> 'a1 -> (char list -> 'a1) -> (nat -> 'a1) -> t -> 'a1

  val t_rec : 'a1 -> 'a1 -> (char list -> 'a1) -> (nat -> 'a1) -> t -> 'a1

  val set : t

  val prop : t

  val is_small : t -> bool

  val is_prop : t -> bool

  val is_set : t -> bool

  val is_var : t -> bool

  val compare : t -> t -> comparison

  val eq_dec : t -> t -> bool

  val equal : t -> t -> bool
 end

module LevelDecidableType :
 sig
  type t = Level.t

  val eq_dec : t -> t -> bool
 end

module LevelSet :
 sig
  module Raw :
   sig
    type elt = Level.t

    type t = elt list

    val empty : t

    val is_empty : t -> bool

    val mem : elt -> t -> bool

    val add : elt -> t -> t

    val singleton : elt -> t

    val remove : elt -> t -> t

    val fold : (elt -> 'a1 -> 'a1) -> t -> 'a1 -> 'a1

    val union : t -> t -> t

    val diff : t -> t -> t

    val inter : t -> t -> t

    val subset : t -> t -> bool

    val equal : t -> t -> bool

    val filter : (elt -> bool) -> t -> t

    val for_all : (elt -> bool) -> t -> bool

    val exists_ : (elt -> bool) -> t -> bool

    val partition : (elt -> bool) -> t -> t * t

    val cardinal : t -> nat

    val elements : t -> elt list

    val choose : t -> elt option

    val isok : elt list -> bool
   end

  module E :
   sig
    type t = Level.t

    val eq_dec : Level.t -> Level.t -> bool
   end

  type elt = Level.t

  type t_ = Raw.t
    (* singleton inductive, whose constructor was Mkt *)

  val this : t_ -> Raw.t

  type t = t_

  val mem : elt -> t -> bool

  val add : elt -> t -> t

  val remove : elt -> t -> t

  val singleton : elt -> t

  val union : t -> t -> t

  val inter : t -> t -> t

  val diff : t -> t -> t

  val equal : t -> t -> bool

  val subset : t -> t -> bool

  val empty : t

  val is_empty : t -> bool

  val elements : t -> elt list

  val choose : t -> elt option

  val fold : (elt -> 'a1 -> 'a1) -> t -> 'a1 -> 'a1

  val cardinal : t -> nat

  val filter : (elt -> bool) -> t -> t

  val for_all : (elt -> bool) -> t -> bool

  val exists_ : (elt -> bool) -> t -> bool

  val partition : (elt -> bool) -> t -> t * t

  val eq_dec : t -> t -> bool
 end

module LevelSetFact :
 sig
  val eqb : Level.t -> Level.t -> bool
 end

module LevelSetProp :
 sig
  module Dec :
   sig
    module F :
     sig
      val eqb : Level.t -> Level.t -> bool
     end

    module MSetLogicalFacts :
     sig
     end

    module MSetDecideAuxiliary :
     sig
     end

    module MSetDecideTestCases :
     sig
     end
   end

  module FM :
   sig
    val eqb : Level.t -> Level.t -> bool
   end

  val coq_In_dec : LevelSet.elt -> LevelSet.t -> bool

  val of_list : LevelSet.elt list -> LevelSet.t

  val to_list : LevelSet.t -> LevelSet.elt list

  val fold_rec :
    (LevelSet.elt -> 'a1 -> 'a1) -> 'a1 -> LevelSet.t -> (LevelSet.t -> __ ->
    'a2) -> (LevelSet.elt -> 'a1 -> LevelSet.t -> LevelSet.t -> __ -> __ ->
    __ -> 'a2 -> 'a2) -> 'a2

  val fold_rec_bis :
    (LevelSet.elt -> 'a1 -> 'a1) -> 'a1 -> LevelSet.t -> (LevelSet.t ->
    LevelSet.t -> 'a1 -> __ -> 'a2 -> 'a2) -> 'a2 -> (LevelSet.elt -> 'a1 ->
    LevelSet.t -> __ -> __ -> 'a2 -> 'a2) -> 'a2

  val fold_rec_nodep :
    (LevelSet.elt -> 'a1 -> 'a1) -> 'a1 -> LevelSet.t -> 'a2 -> (LevelSet.elt
    -> 'a1 -> __ -> 'a2 -> 'a2) -> 'a2

  val fold_rec_weak :
    (LevelSet.elt -> 'a1 -> 'a1) -> 'a1 -> (LevelSet.t -> LevelSet.t -> 'a1
    -> __ -> 'a2 -> 'a2) -> 'a2 -> (LevelSet.elt -> 'a1 -> LevelSet.t -> __
    -> 'a2 -> 'a2) -> LevelSet.t -> 'a2

  val fold_rel :
    (LevelSet.elt -> 'a1 -> 'a1) -> (LevelSet.elt -> 'a2 -> 'a2) -> 'a1 ->
    'a2 -> LevelSet.t -> 'a3 -> (LevelSet.elt -> 'a1 -> 'a2 -> __ -> 'a3 ->
    'a3) -> 'a3

  val set_induction :
    (LevelSet.t -> __ -> 'a1) -> (LevelSet.t -> LevelSet.t -> 'a1 ->
    LevelSet.elt -> __ -> __ -> 'a1) -> LevelSet.t -> 'a1

  val set_induction_bis :
    (LevelSet.t -> LevelSet.t -> __ -> 'a1 -> 'a1) -> 'a1 -> (LevelSet.elt ->
    LevelSet.t -> __ -> 'a1 -> 'a1) -> LevelSet.t -> 'a1

  val cardinal_inv_2 : LevelSet.t -> nat -> LevelSet.elt

  val cardinal_inv_2b : LevelSet.t -> LevelSet.elt
 end

val coq_LevelSet_pair : LevelSet.elt -> LevelSet.elt -> LevelSet.t

type universe_level = Level.t

module Universe :
 sig
  module Expr :
   sig
    type t = Level.t * bool

    val is_small : t -> bool

    val is_level : t -> bool

    val is_prop : t -> bool

    val equal : t -> t -> bool

    val get_level : t -> Level.t

    val prop : t

    val set : t

    val type1 : t
   end

  type t = Expr.t non_empty_list

  val equal : t -> t -> bool

  val make : Level.t -> t

  val make' : Expr.t -> t

  val make'' : Expr.t -> Expr.t list -> t

  val is_level : t -> bool

  val is_levels : t -> bool

  val level : t -> Level.t option

  val is_small : t -> bool

  val is_prop : t -> bool

  val type0m : t

  val type0 : t

  val type1 : t

  val super : Level.t -> t

  val try_suc : t -> t

  val sup : t -> t -> t

  val for_all : (Expr.t -> bool) -> t -> bool

  val sort_of_product : t -> t -> t
 end

type universe = Universe.t

val universe_coercion : universe -> Universe.Expr.t list

type sort_family =
| InProp
| InSet
| InType

val sort_family_rect : 'a1 -> 'a1 -> 'a1 -> sort_family -> 'a1

val sort_family_rec : 'a1 -> 'a1 -> 'a1 -> sort_family -> 'a1

val leb_sort_family : sort_family -> sort_family -> bool

val universe_family : universe -> sort_family

module ConstraintType :
 sig
  type t =
  | Lt
  | Le
  | Eq

  val t_rect : 'a1 -> 'a1 -> 'a1 -> t -> 'a1

  val t_rec : 'a1 -> 'a1 -> 'a1 -> t -> 'a1
 end

type constraint_type = ConstraintType.t

type univ_constraint = (universe_level * ConstraintType.t) * universe_level

module UnivConstraintDec :
 sig
  type t = univ_constraint

  val eq_dec : t -> t -> bool
 end

module ConstraintSet :
 sig
  module Raw :
   sig
    type elt = univ_constraint

    type t = elt list

    val empty : t

    val is_empty : t -> bool

    val mem : elt -> t -> bool

    val add : elt -> t -> t

    val singleton : elt -> t

    val remove : elt -> t -> t

    val fold : (elt -> 'a1 -> 'a1) -> t -> 'a1 -> 'a1

    val union : t -> t -> t

    val diff : t -> t -> t

    val inter : t -> t -> t

    val subset : t -> t -> bool

    val equal : t -> t -> bool

    val filter : (elt -> bool) -> t -> t

    val for_all : (elt -> bool) -> t -> bool

    val exists_ : (elt -> bool) -> t -> bool

    val partition : (elt -> bool) -> t -> t * t

    val cardinal : t -> nat

    val elements : t -> elt list

    val choose : t -> elt option

    val isok : elt list -> bool
   end

  module E :
   sig
    type t = univ_constraint

    val eq_dec : univ_constraint -> univ_constraint -> bool
   end

  type elt = univ_constraint

  type t_ = Raw.t
    (* singleton inductive, whose constructor was Mkt *)

  val this : t_ -> Raw.t

  type t = t_

  val mem : elt -> t -> bool

  val add : elt -> t -> t

  val remove : elt -> t -> t

  val singleton : elt -> t

  val union : t -> t -> t

  val inter : t -> t -> t

  val diff : t -> t -> t

  val equal : t -> t -> bool

  val subset : t -> t -> bool

  val empty : t

  val is_empty : t -> bool

  val elements : t -> elt list

  val choose : t -> elt option

  val fold : (elt -> 'a1 -> 'a1) -> t -> 'a1 -> 'a1

  val cardinal : t -> nat

  val filter : (elt -> bool) -> t -> t

  val for_all : (elt -> bool) -> t -> bool

  val exists_ : (elt -> bool) -> t -> bool

  val partition : (elt -> bool) -> t -> t * t

  val eq_dec : t -> t -> bool
 end

module ConstraintSetFact :
 sig
  val eqb : univ_constraint -> univ_constraint -> bool
 end

module ConstraintSetProp :
 sig
  module Dec :
   sig
    module F :
     sig
      val eqb : univ_constraint -> univ_constraint -> bool
     end

    module MSetLogicalFacts :
     sig
     end

    module MSetDecideAuxiliary :
     sig
     end

    module MSetDecideTestCases :
     sig
     end
   end

  module FM :
   sig
    val eqb : univ_constraint -> univ_constraint -> bool
   end

  val coq_In_dec : ConstraintSet.elt -> ConstraintSet.t -> bool

  val of_list : ConstraintSet.elt list -> ConstraintSet.t

  val to_list : ConstraintSet.t -> ConstraintSet.elt list

  val fold_rec :
    (ConstraintSet.elt -> 'a1 -> 'a1) -> 'a1 -> ConstraintSet.t ->
    (ConstraintSet.t -> __ -> 'a2) -> (ConstraintSet.elt -> 'a1 ->
    ConstraintSet.t -> ConstraintSet.t -> __ -> __ -> __ -> 'a2 -> 'a2) -> 'a2

  val fold_rec_bis :
    (ConstraintSet.elt -> 'a1 -> 'a1) -> 'a1 -> ConstraintSet.t ->
    (ConstraintSet.t -> ConstraintSet.t -> 'a1 -> __ -> 'a2 -> 'a2) -> 'a2 ->
    (ConstraintSet.elt -> 'a1 -> ConstraintSet.t -> __ -> __ -> 'a2 -> 'a2)
    -> 'a2

  val fold_rec_nodep :
    (ConstraintSet.elt -> 'a1 -> 'a1) -> 'a1 -> ConstraintSet.t -> 'a2 ->
    (ConstraintSet.elt -> 'a1 -> __ -> 'a2 -> 'a2) -> 'a2

  val fold_rec_weak :
    (ConstraintSet.elt -> 'a1 -> 'a1) -> 'a1 -> (ConstraintSet.t ->
    ConstraintSet.t -> 'a1 -> __ -> 'a2 -> 'a2) -> 'a2 -> (ConstraintSet.elt
    -> 'a1 -> ConstraintSet.t -> __ -> 'a2 -> 'a2) -> ConstraintSet.t -> 'a2

  val fold_rel :
    (ConstraintSet.elt -> 'a1 -> 'a1) -> (ConstraintSet.elt -> 'a2 -> 'a2) ->
    'a1 -> 'a2 -> ConstraintSet.t -> 'a3 -> (ConstraintSet.elt -> 'a1 -> 'a2
    -> __ -> 'a3 -> 'a3) -> 'a3

  val set_induction :
    (ConstraintSet.t -> __ -> 'a1) -> (ConstraintSet.t -> ConstraintSet.t ->
    'a1 -> ConstraintSet.elt -> __ -> __ -> 'a1) -> ConstraintSet.t -> 'a1

  val set_induction_bis :
    (ConstraintSet.t -> ConstraintSet.t -> __ -> 'a1 -> 'a1) -> 'a1 ->
    (ConstraintSet.elt -> ConstraintSet.t -> __ -> 'a1 -> 'a1) ->
    ConstraintSet.t -> 'a1

  val cardinal_inv_2 : ConstraintSet.t -> nat -> ConstraintSet.elt

  val cardinal_inv_2b : ConstraintSet.t -> ConstraintSet.elt
 end

val make_univ_constraint :
  universe_level -> constraint_type -> universe_level -> univ_constraint

type constraints = ConstraintSet.t

val empty_constraint : constraints

module Instance :
 sig
  type t = Level.t list

  val empty : t

  val is_empty : t -> bool

  val equal : t -> t -> bool

  val equal_upto : (Level.t -> Level.t -> bool) -> t -> t -> bool
 end

type universe_instance = Instance.t

module UContext :
 sig
  type t = universe_instance * constraints

  val make : Instance.t -> ConstraintSet.t -> t

  val empty : t

  val instance : t -> Instance.t

  val constraints : t -> constraints

  val dest : t -> Instance.t * ConstraintSet.t
 end

module AUContext :
 sig
  type t = ident list * constraints

  val make : ident list -> constraints -> t

  val repr : t -> UContext.t

  val levels : t -> LevelSet.t
 end

module ContextSet :
 sig
  type t = LevelSet.t * constraints

  val empty : t

  val is_empty : t -> bool
 end

module Variance :
 sig
  type t =
  | Irrelevant
  | Covariant
  | Invariant

  val t_rect : 'a1 -> 'a1 -> 'a1 -> t -> 'a1

  val t_rec : 'a1 -> 'a1 -> 'a1 -> t -> 'a1
 end

module ACumulativityInfo :
 sig
  type t = AUContext.t * Variance.t list

  val make : AUContext.t -> Variance.t list -> t

  val univ_context : t -> AUContext.t

  val variance : t -> Variance.t list
 end

type universes_decl =
| Monomorphic_ctx of ContextSet.t
| Polymorphic_ctx of AUContext.t
| Cumulative_ctx of ACumulativityInfo.t

val universes_decl_rect :
  (ContextSet.t -> 'a1) -> (AUContext.t -> 'a1) -> (ACumulativityInfo.t ->
  'a1) -> universes_decl -> 'a1

val universes_decl_rec :
  (ContextSet.t -> 'a1) -> (AUContext.t -> 'a1) -> (ACumulativityInfo.t ->
  'a1) -> universes_decl -> 'a1

val monomorphic_udecl : universes_decl -> ContextSet.t

val monomorphic_levels : universes_decl -> LevelSet.t

val monomorphic_constraints : universes_decl -> constraints

val levels_of_udecl : universes_decl -> LevelSet.t

val constraints_of_udecl : universes_decl -> constraints

type valuation = { valuation_mono : (char list -> int);
                   valuation_poly : (nat -> nat) }

val valuation_mono : valuation -> char list -> int

val valuation_poly : valuation -> nat -> nat

val val0 : valuation -> Level.t -> int

val val1 : valuation -> Universe.Expr.t -> int

val coq_val : valuation -> universe -> int

val fresh_universe : universe

val is_prop_sort : Universe.t -> bool
